; *******************************************************
; cette routine calibre une image NOAA au format NOAA/AVHRR
; ex: AL14082799201656202815, la georeference et
; extrait une zone definie
; Fabien Lahoche, 2001
; corrections diverses par JP Berroir
; adaptee par Milton Jonathan, 2004
; *******************************************************

pro ecoair_caliwarping_doit

	envi, /restore_base_save_files

	; selects directory for processing
	current_dir = envi_pickfile(title='Select a directory with raw NOAA images for processing:', /directory, /output)
	if (current_dir eq '') then return

	; selects mask file (eg. masktaquari.inr)
	namemask = envi_pickfile(title='Select the mask file:', filter='*.inr', default='masktaquari.inr', /output)
    if (namemask eq '') then return

	; finds all NOAA raw files in directory
	rawfiles = findfile('A*.L*', count=n_raw)


	; processes each NOAA raw file
	for i_raw=0,n_raw-1 do begin

	;
	;	ouvre l'image a traiter et definit l'image de sortie
	;
		nameraw = rawfiles[i_raw]
		print, 'image selected: ', nameraw

		; obtains fid, row, line and band info
		envi_open_data_file, nameraw, r_fid=raw_fid, /avhrr
		if (raw_fid eq -1) then return
		envi_file_query, raw_fid, ns=ns, nl=nl, nb=nb



	;
	;	AVHRR CALIBRATION (bands 1 and 2)
	;
		;***MILTON***
	    out_name_calib = strjoin(['Calibrated.',nameraw])

	    ; will apply to entire image, on bands 0 and 1
		dims = [-1l, 0, ns-1, 0, nl-1]
		pos = [0,1]

		envi_check_save, /avhrr

		print, 'calibrating raw data into reflectances and radiative temperatures'

	    ; performs the operation
		envi_avhrr_calibrate_doit, fid=raw_fid, pos=pos, dims=dims, $
		out_name=out_name_calib, /correct_solarz, $
		r_fid=calib_fid

		envi_check_save, /avhrr




	;
	;	AVHRR GEOREFERENCING (warping)
	;
		;***MILTON***
	    out_name_warped = strjoin(['Warped.',nameraw])

	    ; creates a geographic projection (obs: uses WGS84)
		;***MILTON***
		;units = envi_proj_units_translate('Degrees')
		units = envi_translate_projection_units('Degrees')
		proj = envi_proj_create(/geographic, units=units)

	    ; defines pixel size for warping
		ps = [.01,.01]

		print, 'georeferencing calibrated image'

	    ; performs the operation
		envi_avhrr_warp_doit, fid=calib_fid, avhrr_fid=raw_fid, $
		dims=dims, pos=pos, out_name=out_name_warped, $
		method=6, degree=1, background=0, grid=[50,50], $
		proj=proj, pixel_size=ps, r_fid=warped_fid



	;
	;   IMAGE CROPPING
	;
		;***MILTON***
	    out_name_cropped = strjoin(['Cropped.',nameraw])

		;***MILTON***
		; obtains row, line, and band info for warped image
		envi_file_query, warped_fid, ns=ns, nl=nl, nb=nb

	    ; defines cropping window in lat/long and converts it to
	    ; image coordinates
		xmap=[-55.5,-52.5]
		ymap=[-17,-20]
		;xmap=[-59.5,-50.5]
		;ymap=[-13,-24]
		envi_convert_file_coordinates, warped_fid, xf, yf, xmap, ymap

	    ; will apply to cropped area, without scaling, on bands 0 and 1
		dims=[ -1l, max([xf[0],0]), min([xf[1],ns-1]), max([yf[0],0]), min([yf[1],nl-1]) ]
		rfact=[1.0, 1.0]
		pos=[0,1]

		envi_check_save, /utility

		print, 'cropping image'

	    ; performs the operation
		resize_doit, fid=warped_fid, pos=pos, out_name=out_name_cropped, $
		dims=dims, rfact=rfact,interp=0,r_fid=cropped_fid




	;***MILTON***
	;
	;   IMAGE PADDING
	;
	;	If part of the cropped area lies outside of the warped image, the
	;	resulting image must be completed (eg. in order to form a 301x301 image)
	;	The image is completed using 'cloudy' pixels (i.e., no info)

	    out_name_padded = strjoin(['Padded.',nameraw])

		; obtains row, line, and band info for cropped image
	   	envi_file_query, cropped_fid, ns=nsr, nl=nlr, nb=nbr

	    ; defines cloudy value
	    cloudy_val = 100

	    ; obtains desired final image size
	   	xsize = round(xf[1]-xf[0]+1)
	   	ysize = round(yf[1]-yf[0]+1)

	    ; padding info: x0 and y0 are the new pixel coordinates in the
	    ; padded image for the cropped image's (0,0) pixel
	    fid = [cropped_fid]
	    pos = [[lindgen(nbr)]]
	    dims = [[-1,0,nsr-1,0,nlr-1]]
	    x0 = [max([round(-xf[0]),0])]
	    y0 = [max([round(-yf[0]),0])]

		; retrieves the cropped image's map info and corrects it according
		; to x0 and y0
	    map_info = envi_get_map_info(fid=cropped_fid)
	    map_info.mc[0] = x0
	    map_info.mc[1] = y0

		print, 'padding image'

		; performs the operation
		envi_doit, 'mosaic_doit', fid=fid, pos=pos, dims=dims, x0=x0, y0=y0, $
		xsize=xsize, ysize=ysize, georef=1, map_info=map_info, $
		pixel_size=[1,1], out_dt=4, background=cloudy_val, $
		use_see_through=[0,0], out_name=out_name_padded, r_fid=padded_fid




	;
	;   IMAGE MASKING (eg: to Taquari area)
	;
		;***MILTON***
		; opens mask file and generates masked output file name
	   	envi_open_file, namemask, r_fid=m_fid

		out_name_masked = strjoin(['Output.',nameraw])

		; obtains row, line, and band info for padded image
		envi_file_query, padded_fid, ns=nsr, nl=nlr, nb=nbr

	    ; will apply to entire cropped image, on every band, using the mask
	    ; image's band 0 as mask
		dimsr=[-1l, 0, nsr-1, 0, nlr-1]
		posr=lindgen(nbr)
		m_pos=[0]

		print, 'masking image'

	    ; performs the operation
		envi_mask_apply_doit, fid=padded_fid, pos=posr, dims=dimsr, m_fid=m_fid, $
		m_pos=m_pos, value=0, out_name=out_name_masked, r_fid=output_fid



	;
	;   REMOVAL OF IMAGES
	;
		print, 'removing images from disk/memory'
		envi_file_mng, id=calib_fid, /remove, /delete
		envi_file_mng, id=warped_fid, /remove, /delete
		envi_file_mng, id=cropped_fid, /remove, /delete
		envi_file_mng, id=padded_fid, /remove, /delete

		envi_file_mng, id=raw_fid, /remove
		;envi_file_mng, id=output_fid, /remove

	endfor

end

