features <- read.table('profiles.ndvi.reclassif.fit8.features.allpoints.txt', header=TRUE, fill=TRUE)

# exports features
exportFeatures <- function(features, outputfile='classification.data.txt', colSelect=2:ncol(features), numeric=FALSE)
{
  if (numeric)
  {
    CLASS = features[,'CLASS']
  }
  else
  {
    classnames = c('Agriculture1','Agriculture2','Agriculture3','Forest','Pasture','Urban','Deforestation')
    CLASS = classnames[features[,'CLASS']]
  }
  values = features[,colSelect]
  write.table(cbind(values,CLASS), file=outputfile, sep='\t', quote=FALSE, row.names=FALSE, col.names=!numeric)
}

# exports sample positions and class
exportSamplePos <- function(samples, outputfile='classification.modis.train.samples.txt', colSelect=1:3, numeric=TRUE)
{
  write.table(samples[,colSelect], file=outputfile, sep='\t', quote=FALSE, row.names=FALSE, col.names=!numeric)
}


# randomly selects the nsamples from each class
getSamples <- function(features, nsamples)
{
  nclass = max(features[,'CLASS'])
  samples = NULL
  for (class in 1:nclass)
  {
    rowsclass = (features[,'CLASS'] == class)
    featclass = features[rowsclass,]
    selected = sample(1:nrow(featclass), min(nsamples,nrow(featclass)))
    samples = rbind(samples, featclass[selected,])
  }
  samples
}


# normalizes the variables so that mean(variable)=0 and var(variable)=1
normalize <- function(features, columns=c(4:ncol(features)))
{
  # separates the variable values from its classification
  CLASS = features[,'CLASS']
  values = features[,columns]
  
  # normalizes the variables
  stddev = sqrt(apply(values, 2, var, use='pairwise.complete'))
  means = apply(values, 2, mean, na.rm=TRUE)
   
  valuesNorm = t((t(values) - means)/stddev)
  valuesNorm[is.na(valuesNorm)] = rnorm(length(valuesNorm[is.na(valuesNorm)]))
  #valuesNorm[is.na(valuesNorm)] = 0
  
  cbind(CLASS, valuesNorm)
}


# computes class means for each variable (column of features)
computeClassMeans <- function(features)
{
  featuresSplit = split.data.frame(features, factor(features[,c('CLASS')]))
  
  # computes means for all columns for all classes (groups in the split)
  means = NULL
  for (class in 1:length(featuresSplit))
  {
    means = rbind(means, apply(featuresSplit[[class]], 2, mean, na.rm=TRUE))
  }
 
  means
}


# computes the R-squared statistic, which tells how much a linear model (based
# on the supposed independent columns of features) is capable of explaining a
# dependent variable (another column of features). The result is the percentage
# of variation explained by the model
computeR2 <- function(features, colDependent, colIndependent)
{
  res.lm = lm(features[,colDependent] ~ as.matrix(features)[,colIndependent])
  summary.lm(res.lm)$r.squared
}



# computes the initial F-statistic for all variables
# this statistic will be used to choose the most discriminating variable
# this statistic is the ratio between the between-class and within-class scatters,
# weighted by their respective degrees of freedom
Fglobal <- function(features, columns=c(2:ncol(features)))
{
  # computes class means and total means for each column
  classMeans = computeClassMeans(features)
  totalMeans = apply(features, 2, mean)
  nclass = nrow(classMeans)
  ntotal = nrow(features)
  
  # computes the between-class scatter
  Sb = matrix(0, 1, length(totalMeans))
  for (iclass in 1:nclass)
  {
    ni = sum(features[,'CLASS'] == iclass)
    Sb = Sb + ni*((classMeans[iclass,] - totalMeans)^2)
  }

  # computes the within-class scatter
  colClassMeans = classMeans[features[,'CLASS'],]
  Sw = apply(((features - colClassMeans)^2), 2, sum)
  
  # computes the total scatter
  #St = apply(((features - totalMeans)^2), 2, sum)
  
  # F-statistic is the ratio Sb/Sw weighted by their degrees of freedom
  # (nclass-1) and (ntotal-nclass), respectively
  F = (Sb[columns]/(nclass-1)) / (Sw[columns] / (ntotal-nclass))
  F
}


# computes the partial F-statistic for one variable (and the associated p-value)
# this statistic will be used to choose the variable that brings most information
# to the model
Fpartial <- function(features, colNew, colIncluded)
{
  nclass = max(features[,'CLASS'])
  ntotal = nrow(features)
  nincluded = length(colIncluded)
  
  # computes the classification information, given by indicative variables
  # as such, column i will be TRUE for rows of class i and FALSE otherwise
  featClassif = NULL
  for (class in 1:nclass)
  {
    featClassif = cbind(featClassif, (features[,'CLASS']==class))
  }
  ncols = ncol(features)
  colClassif = c((ncols+1):(ncols+nclass))
  
  # R-squared statistic considering classification information  
  r2.1 = computeR2(cbind(features,featClassif), colNew, c(colIncluded, colClassif))
  
  # R-squared statistic without classification information (i.e., will measure how much
  # the variables already included in the model are capable of explaining the variability
  # of the new variable)
  r2.2 = computeR2(features, colNew, colIncluded)
  
  # F-statistic is the ratio between the increase of information given by including the
  # classification info (r2.1 - r2.2), and the amount of information still missing
  # (1 - r2.1), weighted by their respective degrees of freedom
  F = ((r2.1 - r2.2)/(nclass-1)) / ((1 - r2.1)/(ntotal-nclass-nincluded))
  p = pf(F, nclass-1, ntotal-nclass-nincluded, lower.tail=FALSE)
  c(F,p)
}


# selects the new variable by picking the one that brings most new information
# to the model
# the variable selected will be the one with maximum partial F-statistic
# if the selected variable is non-significant (p-value < alpha), then NULL is
# returned
# otherwise, the return value is the column of the selected variable
selectNewVariable <- function(features, colIncluded, alpha=0.05)
{
  nclass = max(features[,'CLASS'])
  ntotal = nrow(features)
  nincluded = length(colIncluded)
  
  stats = matrix(0, 0, 3)
  colnames(stats) <- c('column','F','p')
  
  for (col in 2:ncol(features))
  {
    # checks only variables not included so far
    if (sum(colIncluded == col) == 0)
    {
      # computes partial F-statistics and p-value for the variable
      F = Fpartial(features, col, colIncluded)
      
      # adds variable stats to matrix
      stats = rbind(stats, c(col, F))
    }
  }
  
  # selects the variable with maximum F value
  variable = which.max(stats[,'F'])
  
  # checks if the variable's p-value is significant
  if (stats[variable,'p'] > alpha)
  {
    return(NULL)
  }
  
  # returns the column of the selected variable
  stats[variable,'column']
}


# checks all variables in the model to see if they are significant
# level of significance is given by the p-value of the partial F-statistic 
# computed for each variable, with all others included in the model.
# function returns the columns of the included variables, excluding all
# the ones that were considered non-significant (p-value > alpha)
cleanVariables <- function(features, colIncluded, alpha=0.05)
{
  colCleaned = NULL
  for (col in colIncluded)
  {
    colCheck = colIncluded[colIncluded != col]
    F = Fpartial(features, col, colCheck)
    if (F[2] <= alpha)
    {
      colCleaned[length(colCleaned)+1] = col
    }
  }
  colCleaned
}


# MAIN FUNCTION
# given classified data with a large number of variables, selects a subset of
# significant variables so as to provide best the discrimination between
# classes while minimizing the variables' inter-correlations
selectVariables <- function(features, alpha=0.05)
{
  # selects starting variable (the one with maximum global F-statistic)
  F = Fglobal(features)
  colIncluded = which.max(F)+1
  
  while(TRUE)
  {
    newcol = selectNewVariable(features, colIncluded, alpha)
    if (is.null(newcol))
    {
      return(colIncluded)
    }
      
    colIncluded = c(colIncluded, newcol)
    colIncluded = cleanVariables(features, colIncluded, alpha)
  }
}
