#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "file_tools.h"
#include "inr_io.h"
#include "envi2inr.h"

#define MIN(x,y) ((x<y)?x:y)

#define BUFSIZE 524288
#define FNAMESIZE 1024

static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir";

static char Udetail[] =
"\nConverts ENVI images to INRIMAGE format\n\
\nInput:\n\
\t-idir\tDirectory containing ENVI images to be converted\n\
\t-ifile\tFile name of one ENVI image to be converted\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted images\n\
\t-ofile\tOutput file name for the converted image\n\
\t\tSpecify '-ofile' only when '-ifile' is used\n";






/*
 * Reads the input Envi header file.
 */
int enviReadHeader(char *fnameFull, int *size_x, int *size_y, int *size_z,
                   InrType *type, EnviInterleave *interleave)
{
  FILE    *fpEnviHdr;
  char     fnameEnviHdr1[FNAMESIZE];
  char     fnameEnviHdr2[FNAMESIZE];
  char    *ptrEnviHdr;
  char     strEnviHdr[1024];
  size_t   nbytes;
  EnviType envitype;
  char     interleavename[10];
  int      infoPresent = 0;

  // obtains possible header file names from input envi file name
  strcpy(fnameEnviHdr1, fnameFull);
  strcat(fnameEnviHdr1, ".hdr");
  ftChangeExtension(fnameFull, fnameEnviHdr2, ".hdr");
  
  // returns in case fnameFull is an ENVI header file
  if (!strcmp(fnameFull, fnameEnviHdr2))
    return 0;

  // opens input ENVI header file (hdr)
  fpEnviHdr = fopen(fnameEnviHdr1,"rt");
  if (!fpEnviHdr)
  {
    // attempts 2nd possible envi file name
    fpEnviHdr = fopen(fnameEnviHdr2,"rt");
    if (!fpEnviHdr)
    {
      fprintf(stderr, "Error: could not locate ENVI header for file '%s'.\n", fnameFull);
      return 0;
    }
  }

  // reads input envi header file
  nbytes = fread(strEnviHdr, 1, 1023, fpEnviHdr);
  strEnviHdr[nbytes] = '\0';
  fclose(fpEnviHdr);

  // obtains info from header
  if ((ptrEnviHdr = strstr(strEnviHdr,"samples")) != NULL)
    infoPresent += sscanf(ptrEnviHdr, "samples = %d", size_x);
  if ((ptrEnviHdr = strstr(strEnviHdr,"lines")) != NULL)
    infoPresent += sscanf(ptrEnviHdr, "lines   = %d", size_y);
  if ((ptrEnviHdr = strstr(strEnviHdr,"bands")) != NULL)
    infoPresent += sscanf(ptrEnviHdr, "bands   = %d", size_z);
  if ((ptrEnviHdr = strstr(strEnviHdr,"data type")) != NULL)
    infoPresent += sscanf(ptrEnviHdr, "data type = %d", (int*)&envitype);
  if ((ptrEnviHdr = strstr(strEnviHdr,"interleave")) != NULL)
    infoPresent += sscanf(ptrEnviHdr, "interleave = %9s", interleavename);
  if (infoPresent != 5)
  {
    fprintf(stderr, "Error: could not read necessary information from ENVI header for file '%s'.\n", fnameFull);
    return 0;
  }
  
  if (!inrGetType(InrTypes[envitype], InrSizes[envitype], type))
  {
    fprintf(stderr, "Error: do not know how to convert ENVI type '%d' for file '%s'. Sorry.\n", envitype, fnameFull);
    return 0;
  }
    
  if (!strcmp(interleavename,"BSQ") || !strcmp(interleavename,"bsq"))
    *interleave = BSQ;
  else if (!strcmp(interleavename,"BIL") || !strcmp(interleavename,"bil"))
    *interleave = BIL;
  else if (!strcmp(interleavename,"BIP") || !strcmp(interleavename,"bip"))
    *interleave = BIP;
  else
  {
    fprintf(stderr, "Error: do not know how to interpret ENVI interleave '%s' for file '%s'. Sorry.\n", interleavename, fnameFull);
    return 0;
  }

  return 1;
}


int envi2inrBSQ(char *fnameFull, char *outfnameFull, int size_x, int size_y,
                int size_z, InrType type)
{
  FILE *fpEnvi;
  FILE *fpInr;
  char bandfnameFull[FNAMESIZE];
  char *outfnameFinal = outfnameFull;
  unsigned char buf[BUFSIZE];
  int z;
  int typeSize = InrTypeSizes[type]/8;

  // opens ENVI input file (img)
  fpEnvi = fopen(fnameFull,"rb");
  if (!fpEnvi)
  {
    fprintf(stderr, "Error: could not open ENVI file '%s'.\n", fnameFull);
    return 0;
  }

  // converts each band in the input ENVI data to one output INR file
  for (z=0; z<size_z; z++)
  {
    int nRemainingBytes = size_x*size_y*typeSize;
    int nBytes;
    size_t nBytesRead;

    // adjusts output file name if more than one band (channel) is present
    if (size_z > 1)
    {
      char ext[10];
      sprintf(ext, ".c%d.inr", z+1);
      ftChangeExtension(outfnameFull, bandfnameFull, ext);
      outfnameFinal = bandfnameFull;
    }
    
    // opens output file
    fpInr = fopen(outfnameFinal,"wb");
    if (!fpInr)
    {
      fprintf(stderr, "Error: could not open output file '%s' while converting ENVI file '%s'.\n", outfnameFinal, fnameFull);
      fclose(fpEnvi);
      return 0;
    }
    
    // writes INR header information
    if (!inrWriteHeaderPtr(fpInr, size_x, size_y, 1, type))
    {
      fprintf(stderr, "Error: could not write to output file '%s' while converting ENVI file '%s'.\n", outfnameFinal, fnameFull);
      fclose(fpEnvi);
      fclose(fpInr);
      return 0;
    }

    // converts data    
    while(nRemainingBytes > 0)
    {
      // checks for end of file (EOF)
      if (feof(fpEnvi))
      {
        fprintf(stderr, "Error: read error while converting ENVI file '%s'.\n", fnameFull);
        fclose(fpEnvi);
        fclose(fpInr);
        return 0;
      }

      // determines amount to be read this time
      nBytes = MIN(BUFSIZE, nRemainingBytes);

      // reads sequential data
      nBytesRead = fread(buf, 1, nBytes, fpEnvi);
      if (!nBytesRead)
      {
        fprintf(stderr, "Error: read error while converting ENVI file '%s'.\n", fnameFull);
        fclose(fpEnvi);
        fclose(fpInr);
        return 0;
      }

      // writes sequential data
      if (fwrite(buf, 1, nBytesRead, fpInr) != nBytesRead)
      {
        fprintf(stderr, "Error: write error while converting ENVI file '%s'.\n", fnameFull);
        fclose(fpEnvi);
        fclose(fpInr);
        return 0;
      }

      // computes new remaining amount to be read
      nRemainingBytes -= (int)nBytesRead;
    }

    // closes output file
    fclose(fpInr);

    // writes output envi header
    {
      InrGeorefInfo georef;
      if (inrReadGeoref(fnameFull, &georef))
        inrWriteGeoref(outfnameFinal, georef);
      else
        inrWriteEnviHdr(outfnameFinal);
    }
  }
  // closes input file and returns success
  fclose(fpEnvi);
  return 1;
}


int envi2inrBIL(char *fnameFull, char *outfnameFull, int size_x, int size_y,
                int size_z, InrType type)
{
  fprintf(stderr, "Error: could not convert file '%s' - BIL file conversion not implemented. Sorry.\n", fnameFull);
  return 0;
}


int envi2inrBIP(char *fnameFull, char *outfnameFull, int size_x, int size_y,
                int size_z, InrType type)
{
  FILE *fpEnvi;
  FILE *fpInr;
  char bandfnameFull[FNAMESIZE];
  char *outfnameFinal = outfnameFull;
  unsigned char bufIn[BUFSIZE];
  unsigned char bufOut[BUFSIZE];
  int z;
  int typeSize = InrTypeSizes[type]/8;

  // opens ENVI input file (img)
  fpEnvi = fopen(fnameFull,"rb");
  if (!fpEnvi)
  {
    fprintf(stderr, "Error: could not open ENVI file '%s'.\n", fnameFull);
    return 0;
  }

  // converts each band in the input ENVI data to one output INR file
  for (z=0; z<size_z; z++)
  {
    int nRemainingBytes = size_x*size_y*size_z*typeSize;
    int nBytes;
    size_t nBytesRead;
    size_t iByteIn, iByteOut, nBytesOut;
    
    // adjusts output file name if more than one band (channel) is present
    if (size_z > 1)
    {
      char ext[10];
      sprintf(ext, ".c%d.inr", z+1);
      ftChangeExtension(outfnameFull, bandfnameFull, ext);
      outfnameFinal = bandfnameFull;
    }
    
    // opens output file
    fpInr = fopen(outfnameFinal,"wb");
    if (!fpInr)
    {
      fprintf(stderr, "Error: could not open output file '%s' while converting ENVI file '%s'.\n", outfnameFinal, fnameFull);
      fclose(fpEnvi);
      return 0;
    }
    
    // writes INR header information
    if (!inrWriteHeaderPtr(fpInr, size_x, size_y, 1, type))
    {
      fprintf(stderr, "Error: could not write to output file '%s' while converting ENVI file '%s'.\n", outfnameFinal, fnameFull);
      fclose(fpEnvi);
      fclose(fpInr);
      return 0;
    }

    // sets Envi file pointer to the right position
    fseek(fpEnvi, 0, SEEK_SET);

    // converts data    
    while(nRemainingBytes > 0)
    {
      // checks for end of file (EOF)
      if (feof(fpEnvi))
      {
        fprintf(stderr, "Error: read error while converting ENVI file '%s'.\n", fnameFull);
        fclose(fpEnvi);
        fclose(fpInr);
        return 0;
      }

      // determines amount to be read this time
      nBytes = MIN(BUFSIZE, nRemainingBytes);
      
      // reads BIP data 
      nBytesRead = fread(bufIn, 1, nBytes, fpEnvi);
      if (!nBytesRead)
      {
        fprintf(stderr, "Error: read error while converting ENVI file '%s'.\n", fnameFull);
        fclose(fpEnvi);
        fclose(fpInr);
        return 0;
      }
      
      // copies BIP data for current band to output buffer
      for (iByteOut = 0, iByteIn = z*typeSize;
           iByteIn < nBytesRead;
           iByteOut += typeSize, iByteIn += size_z*typeSize)
      {
        memcpy(bufOut+iByteOut, bufIn+iByteIn, typeSize);
      }
      nBytesOut = iByteOut;

      // writes data in sequential form
      if (fwrite(bufOut, 1, nBytesOut, fpInr) != nBytesOut)
      {
        fprintf(stderr, "Error: write error while converting ENVI file '%s'.\n", fnameFull);
        fclose(fpEnvi);
        fclose(fpInr);
        return 0;
      }

      // computes new remaining amount to be read
      nRemainingBytes -= (int)nBytesRead;
    }

    // closes output file
    fclose(fpInr);

    // writes output envi header
    {
      InrGeorefInfo georef;
      if (inrReadGeoref(fnameFull, &georef))
      {
        //*** CAUTION: EMPIRICAL ADJUSTMENT FOR NOAA/AVHRR IMAGES GEOREFERENCED IN ENVI ***
        //georef.topleft_x += 0.01;
        inrWriteGeoref(outfnameFinal, georef);
      }
      else
        inrWriteEnviHdr(outfnameFinal);
    }
  }
  // closes input file and returns success
  fclose(fpEnvi);
  return 1;
}


int envi2inr(char *fnameFull, char *outfnameFull)
{
  int size_x, size_y, size_z;
  InrType type;
  EnviInterleave interleave;

  // reads ENVI header information
  if (!enviReadHeader(fnameFull, &size_x, &size_y, &size_z, &type, &interleave))
    return 0;

  // converts data from input envi file according to interleave
  switch (interleave)
  {
    case BSQ:
      return envi2inrBSQ(fnameFull, outfnameFull, size_x, size_y, size_z, type);
    case BIL:
      return envi2inrBIL(fnameFull, outfnameFull, size_x, size_y, size_z, type);
    case BIP:
      return envi2inrBIP(fnameFull, outfnameFull, size_x, size_y, size_z, type);
    default:
      return 0;
  }
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  char   fnameOutput[FNAMESIZE];
  int    okDirIn, okDirOut, okFileIn, okFileOut;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  , "%s", dirInput);
  okDirOut      = igetopt1("-odir"  , "%s", dirOutput);
  okFileIn      = igetopt1("-ifile" , "%s", fnameInput);
  okFileOut     = igetopt1("-ofile" , "%s", fnameOutput);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if ((okDirOut && okFileOut) || (!okDirOut && !okFileOut))
    iusage_ (Ucmd, Udetail);
  if ((okDirIn && okFileOut))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  if (okDirOut)
    ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
  if (okFileOut)
    ftDecodeFileName(fnameOutput, dirOutput, fnameOutput);

  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // converts file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be converted
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input file name (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);

    // computes the complete output file name (dirname + filename)
    if (okFileOut)
      sprintf(outfnameFull, "%s%s", dirOutput, fnameOutput);
    else
    {
      sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);
      ftChangeExtension(outfnameFull, outfnameFull, ".inr");
    }

    // converts image
    if (!envi2inr(fnameFull, outfnameFull))
      continue;
    
    // increments counter
    printf("Converted file '%s'\n", ent->d_name);
    converted++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d ENVI file(s) converted.\n", converted);
  return 1;
}
