#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "file_tools.h"
#include "str_tools.h"
#include "inr_io.h"
#include "hdfraw2inr.h"

#define MIN(x,y) ((x<y)?x:y)

#define BUFSIZE   524288
#define FNAMESIZE 1024
#define MAXBANDS  100

static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir";

static char Udetail[] =
"\nConverts HDF-RAW images to INRIMAGE format\n\
\nInput:\n\
\t-idir\tDirectory containing HDF-RAW images to be converted\n\
\t-ifile\tFile name of one HDF-RAW image to be converted (header filename)\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted images\n";



/*
 * Obtains the HDF-RAW header file name corresponding to a given HDF-RAW
 * .dat file.
 */
char *hdfrawGetHeaderName(char *fnameFull, char *fnameHdfrawHdr)
{
  char filename[FNAMESIZE];
  char dirname[FNAMESIZE];
  char *ptr;
  
  // separetes dir and file names
  if (!ftDecodeFileName(fnameFull, dirname, filename))
    return NULL;
  
  // separates base name from the rest of the file name  
  ptr = strrchr(filename, '.');
  if (!ptr)
    return 0;
    
  // appends .hdr extension
  ptr[0] = 0;
  if (!ftChangeExtension(filename, filename, ".hdr"))
    return 0;
    
  // puts back dir info for full header name
  strcat(dirname, filename);
  return strcpy(fnameHdfrawHdr, dirname);
}


/*
 * Reads the input HDF-RAW header file.
 */
int hdfrawReadHeader(char *fnameFull, int *size_x, int *size_y, int *size_z,
                     InrGeorefInfo *georef, strList *bandnames, InrType *bandtypes)
{
  FILE  *fpHeader;
  char   strHeader[16384];
  char   strnames[1024];
  char   strtypes[1024];
  char  *token;
  int    nbands;
  size_t nbytes;
    
  char *ptrHdfrawHdr;
  int infoPresent = 0;

  // opens input HDF-RAW header
  fpHeader = fopen(fnameFull,"rt");
  if (!fpHeader)
    return 0;

  // reads input HDF-RAW header file
  nbytes = fread(strHeader, 1, 16383, fpHeader);
  strHeader[nbytes] = '\0';
  fclose(fpHeader);

  // obtains info from header
  //first the projection type
  if ((ptrHdfrawHdr = strstr(strHeader,"PROJECTION_TYPE")) != NULL)
    infoPresent += sscanf(ptrHdfrawHdr, "PROJECTION_TYPE = %s",
                          georef->projection);
  
  //test to know if we work with geographic projection or Utm projection

  if(!strcmp(georef->projection,"GEOGRAPHIC") )
    {

      strcpy( georef->projection , "Geographic Lat/Lon");
      strcpy( georef->unit , "Degrees" ) ;
      georef->isUTM=0;
      
      if ((ptrHdfrawHdr = strstr(strHeader,"DATUM")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "DATUM = %s",
			      georef->datum);

      if ( !strcmp( georef->datum , "WGS84" ) )
	strcpy( georef->datum , "WGS-84" ) ;
      else if ( !strcmp( georef->datum , "WGS72" ) )
	strcpy( georef->datum , "WGS-72" ) ;
      else if ( !strcmp( georef->datum , "NAD27" ) )
	strcpy( georef->datum , "North America 1927" ) ;
      else if ( !strcmp( georef->datum , "NAD83" ) )
	strcpy( georef->datum , "North America 1983" ) ;
      else {
	fprintf(stderr,"Error! Unsupported datum %s, assuming WGS84\n",georef->datum);
	strcpy( georef->datum , "WGS-84" ) ;
      }

      if ((ptrHdfrawHdr = strstr(strHeader,"NSAMPLES")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "NSAMPLES = ( %d", size_x);
      if ((ptrHdfrawHdr = strstr(strHeader,"NLINES")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "NLINES = ( %d", size_y);
      if ((ptrHdfrawHdr = strstr(strHeader,"BANDNAMES = ")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "BANDNAMES = ( %1023[^)]", strnames);
      if ((ptrHdfrawHdr = strstr(strHeader,"DATA_TYPE")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "DATA_TYPE = ( %1023[^)]", strtypes);
      if ((ptrHdfrawHdr = strstr(strHeader,"UL_CORNER_LATLON")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "UL_CORNER_LATLON = ( %lf %lf",
			      &(georef->topleft_y), &(georef->topleft_x));
      if ((ptrHdfrawHdr = strstr(strHeader,"PIXEL_SIZE")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "PIXEL_SIZE = ( %lf",
			      &(georef->pixelsize_x));
      if (infoPresent != 9)
	return 0;
    }

  else if(!strcmp(georef->projection,"UTM"))
    {

      strcpy( georef->unit , "Meters" ) ;
      georef->isUTM=1;   
      
      if ((ptrHdfrawHdr = strstr(strHeader,"DATUM")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "DATUM = %s",
			      georef->datum);
      

      if ( !strcmp( georef->datum , "WGS84" ) )
	strcpy( georef->datum , "WGS-84" ) ;
      else if ( !strcmp( georef->datum , "WGS72" ) )
	strcpy( georef->datum , "WGS-72" ) ;
      else if ( !strcmp( georef->datum , "NAD27" ) )
	strcpy( georef->datum , "North America 1927" ) ;
      else if ( !strcmp( georef->datum , "NAD83" ) )
	strcpy( georef->datum , "North America 1983" ) ;
      else {
	fprintf(stderr,"Error! Unsupported datum %s, assuming WGS84\n",georef->datum);
	strcpy( georef->datum , "WGS-84" ) ;
      }

      if ((ptrHdfrawHdr = strstr(strHeader,"NSAMPLES")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "NSAMPLES = ( %d", size_x);
      if ((ptrHdfrawHdr = strstr(strHeader,"NLINES")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "NLINES = ( %d", size_y);
      if ((ptrHdfrawHdr = strstr(strHeader,"BANDNAMES = ")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "BANDNAMES = ( %1023[^)]", strnames);
      if ((ptrHdfrawHdr = strstr(strHeader,"DATA_TYPE")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "DATA_TYPE = ( %1023[^)]", strtypes);
      if ((ptrHdfrawHdr = strstr(strHeader,"# UL_CORNER_XY")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "# UL_CORNER_XY = ( %lf %lf",
			      &(georef->topleft_x), &(georef->topleft_y));
      if ((ptrHdfrawHdr = strstr(strHeader,"PIXEL_SIZE")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "PIXEL_SIZE = ( %lf",
			      &(georef->pixelsize_x));
      
      if ((ptrHdfrawHdr = strstr(strHeader,"UTM_ZONE")) != NULL)
	infoPresent += sscanf(ptrHdfrawHdr, "UTM_ZONE =  %d",
			      &(georef->UTMzone));
      
      if( georef->UTMzone <0 )
	{
	  strcpy(georef->UTMhemi,"South");
	  georef->UTMzone=-georef->UTMzone;
	}
      else 
	strcpy(georef->UTMhemi,"North");

      if (infoPresent != 10)
	return 0;
    } // utm
  else {
    fprintf(stderr,"Unsupported projection %s\n",georef->projection) ;
    return 0 ;
  }

  //computes pixel_x,y and pixel_mapx,mapy
  georef->pixel_x = 1.0 ;
  georef->pixel_y = 1.0 ;
  georef->pixel_mapx = georef->topleft_x ;
  georef->pixel_mapy = georef->topleft_y ;
  

 
  // reads datum and projection info, if available
  /* requires translation to INR_LIB / ENVI standards - assuming default (GEOGRAPHIC/WGS84)
  if (ptrHdfrawHdr = strstr(strHeader,"DATUM"))
    sscanf(ptrHdfrawHdr, "DATUM = %s", georef->datum);
  */
  //georef->datum[0] = 0;
  //georef->projection[0] = 0;

  // only 1 band per dat file
  *size_z = 1;

  // x and y resolutions are the same
  georef->pixelsize_y = georef->pixelsize_x;
  
  // reads band names
  token = strtok(strnames, " ");
  while (token != NULL)
  {
    strListAdd(bandnames, token);
    token = strtok(NULL, " ");
  }
  
  // reads band types
  nbands = 0;
  token = strtok(strtypes, " ");
  while (token != NULL)
  {
    if (!strcmp(token, "UINT8") || !strcmp(token,"INT8"))
      bandtypes[nbands] = InrByte;
    else if (!strcmp(token, "UINT16"))
      bandtypes[nbands] = InrUnsignedInt;
    else if (!strcmp(token, "INT16"))
      bandtypes[nbands] = InrSignedInt;
    else
    {
      char filename[FNAMESIZE];
      ftDecodeFileName(fnameFull, NULL, filename);
      printf("Error: couldn't find an INRIMAGE type for band %d of header file '%s'.\n",
              nbands+1, filename);
      printf("       assuming to be of type 'Byte'.\n");
      bandtypes[nbands] = InrByte;
    }
    nbands++;
    token = strtok(NULL, " ");
  }
  
  // checks to see if number of band names and band types coincide
  if (nbands != bandnames->count)
  {
    char filename[FNAMESIZE];
    ftDecodeFileName(fnameFull, NULL, filename);
    printf("Error: input header file '%s' has inconsistencies in number of band names\nand band types",
            filename);
    return 0;
  }

  return 1;
}


int hdfraw2inrBSQ(char *fnameFull, char *outfnameFull, int size_x, int size_y,
                  int size_z, InrType type, InrGeorefInfo georef)
{
  FILE *fpHdfraw;
  FILE *fpInr;
  char bandfnameFull[FNAMESIZE];
  char *outfnameFinal = outfnameFull;
  unsigned char buf_in[BUFSIZE];
  unsigned char buf_out[BUFSIZE];
  int z;
  int typeSize = InrTypeSizes[type]/8;

  // opens HDF-RAW input file (.dat)
  fpHdfraw = fopen(fnameFull,"rb");
  if (!fpHdfraw)
    return 0;

  // converts each band in the input HDF-RAW data to one output INR file
  for (z=0; z<size_z; z++)
  {
    int nRemainingBytes = size_x*size_y*typeSize;
    int nBytes;
    size_t iBuf;
    size_t nBytesRead;

    // adjusts output file name if more than one band (channel) is present
    if (size_z > 1)
    {
      char ext[10];
      sprintf(ext, ".c%d.inr", z+1);
      ftChangeExtension(outfnameFull, bandfnameFull, ext);
      outfnameFinal = bandfnameFull;
    }
    
    // opens output file
    fpInr = fopen(outfnameFinal,"wb");
    if (!fpInr)
    {
      fclose(fpHdfraw);
      return 0;
    }
    
    // writes INR header information
    if (!inrWriteHeaderPtr(fpInr, size_x, size_y, 1, type))
    {
      fclose(fpHdfraw);
      fclose(fpInr);
      return 0;
    }

    // converts data    
    while(nRemainingBytes > 0)
    {
      // checks for end of file (EOF)
      if (feof(fpHdfraw))
      {
        fclose(fpHdfraw);
        fclose(fpInr);
        return 0;
      }

      // determines amount to be read this time
      nBytes = MIN(BUFSIZE, nRemainingBytes);

      // reads sequential data
      nBytesRead = fread(buf_in, 1, nBytes, fpHdfraw);
      if (!nBytesRead)
      {
        fclose(fpHdfraw);
        fclose(fpInr);
        return 0;
      }
      
      // corrects byte ordering (little endian/big endian)
      for (iBuf=0; iBuf<nBytesRead; iBuf+=typeSize)
      {
        int iByte;
        for (iByte=0; iByte<typeSize; iByte++)
          buf_out[iBuf+iByte] = buf_in[iBuf+typeSize-iByte-1];
      }
      
      // writes sequential data
      if (fwrite(buf_out, 1, nBytesRead, fpInr) != nBytesRead)
      {
        fclose(fpHdfraw);
        fclose(fpInr);
        return 0;
      }

      // computes new remaining amount to be read
      nRemainingBytes -= (int)nBytesRead;
    }

    // closes output file
    fclose(fpInr);

    // writes output auxiliary ENVI header with georeference info
    inrWriteGeoref(outfnameFinal, georef);
  }
  // closes input file and returns success
  fclose(fpHdfraw);
  return 1;
}



int hdfraw2inr(char *fnameHeader, char *dirInput, char *dirOutput)
{
  int size_x, size_y, size_z;
  InrGeorefInfo georef;
  strList *bandnames;
  InrType bandtypes[MAXBANDS];
  char    fnameHeaderFull[FNAMESIZE];
  int     converted = 0;
  int     i;
  
  // creates a list of band names
  bandnames = strListCreate(MAXBANDS);
  if (!bandnames)
    return 0;

  // computes the complete header file name (dirname + filename)
  sprintf(fnameHeaderFull, "%s%s", dirInput, fnameHeader);

  // reads the header
  if (!hdfrawReadHeader(fnameHeaderFull, &size_x, &size_y, &size_z, &georef,
                        bandnames, bandtypes))
  {
    strListDestroy(bandnames);
    return 0;
  }

  for (i=0; i<bandnames->count; i++)
  {
    char  filename[FNAMESIZE];
    char  fnameInputFull[FNAMESIZE];
    char  fnameOutputFull[FNAMESIZE];
    char *bandname = strListGet(bandnames, i);
    int   channel;
    
    // computes the output file name (a .inr file representing one band)
    // this is done by replacing the extension ".hdr" for ".<bandname>.inr"
    // if band is a surface reflectance band, a conversion to "c%d" is done
    ftChangeExtension(fnameHeader, filename, ".");
    if (sscanf(bandname, "sur_refl_b0%d", &channel))
      sprintf(filename+strlen(filename), "c%d", channel);
    else
      strcat(filename, bandname);
    strcat(filename, ".inr");
    sprintf(fnameOutputFull, "%s%s", dirOutput, filename);
    
    // computes the input file name (a .dat file representing one band)
    // this is done by replacing the extension ".hdr" for ".<bandname>.dat"
    ftChangeExtension(fnameHeader, filename, ".");
    strcat(filename, bandname);
    strcat(filename, ".dat");
    sprintf(fnameInputFull, "%s%s", dirInput, filename);
    
    if (!hdfraw2inrBSQ(fnameInputFull, fnameOutputFull, size_x, size_y, size_z,
                       bandtypes[i], georef))
      continue;
                               
    // increments counter
    printf("Converted file '%s'\n", filename);
    converted++;
  }
  
  strListDestroy(bandnames);
  return converted;
}



int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  int    okDirIn, okDirOut, okFile;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  , "%s", dirInput);
  okDirOut      = igetopt1("-odir"  , "%s", dirOutput);
  okFile        = igetopt1("-ifile" , "%s", fnameInput);
  
  if ((okDirIn && okFile) || (!okDirIn && !okFile))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFile)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // crops file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char *ptr;

    // checks if file is an hdfraw header file (*.hdr)
    ptr = strrchr(ent->d_name, '.');
    if (!ptr || strcmp(ptr, ".hdr" ))
      continue;
      
    // checks if file is to be converted
    if (okFile && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // converts image
    converted += hdfraw2inr(ent->d_name, dirInput, dirOutput);
  }

  // closes directory
  closedir(dir);
    
  printf("%d HDF-RAW file(s) converted.\n", converted);
  return 1;
}
