#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "classdef.h"
#include "linearalgebra.h"
#include "inr_io.h"

#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)


/*
 * defCreate
 * 
 * Creates a new class definition from scratch
 * All features are initially considered to be continuous
 */
classdef *defCreate(int nClass, int nFeatures)
{
  int i, j;
  
  classdef *def = (classdef *)malloc(sizeof(classdef));

  def->nFeatures         = nFeatures;
  def->nClass            = nClass;
  
  def->discreteProb      = (double *) malloc (nClass*DEF_MAX_NUM_DISCRETE_VALUES * sizeof(double));
  def->discreteValues    = (double *) malloc (DEF_MAX_NUM_DISCRETE_VALUES * sizeof(double));
  def->nDiscreteValues   = 1;
  def->iDiscreteFeature  = -1;

  for (i=0; i<DEF_MAX_NUM_DISCRETE_VALUES; i++)
    def->discreteValues[i] = 0;

  for (i=0; i<nClass; i++)
  {
    def->discreteProb[i*DEF_MAX_NUM_DISCRETE_VALUES] = 1;
    for (j=1; j<DEF_MAX_NUM_DISCRETE_VALUES; j++)
      def->discreteProb[j + i*DEF_MAX_NUM_DISCRETE_VALUES] = 0;
  }


  def->nContFeatures     = nFeatures;
  def->contMeans       = (matrice **) malloc(nClass*DEF_MAX_NUM_DISCRETE_VALUES * sizeof(matrice *));
  def->contCovariances = (matrice **) malloc(nClass*DEF_MAX_NUM_DISCRETE_VALUES * sizeof(matrice *));
  def->iContFeatures   = (int *) malloc (nFeatures * sizeof(int));

  for (i=0; i<def->nContFeatures; i++)
    def->iContFeatures[i] = i;

  for (i=0; i<nClass*DEF_MAX_NUM_DISCRETE_VALUES; i++)
  {
    def->contMeans[i] = Matrice_Nulle(def->nContFeatures, 1);
    def->contCovariances[i] = Matrice_Nulle(def->nContFeatures, def->nContFeatures);
  }
  
  return def;
}



/*
 * defCreateFromFile
 * 
 * Creates a new class definition from a class definition file
 */
classdef *defCreateFromFile(char *filename)
{
  classdef *def;
  FILE *fp;
  int nClass;
  int nFeatures;
  int iClass, iDiscrete;
  int i, j;

  fp = fopen(filename, "rt");
  if (!fp)
  {
    fprintf (stderr, "Error: could not open class definition file.\n");
    return NULL;
  }

  // reads number of classes and of features
  if (fscanf(fp, "CLASSES:%d\n", &nClass) != 1)
  {
    fprintf (stderr, "Error: could not read number of classes from class definition file.\n");
    return NULL;
  }
  if (fscanf(fp, "FEATURES:%d\n", &nFeatures) != 1)
  {
    fprintf (stderr, "Error: could not read number of features from class definition file.\n");
    return NULL;
  }

  // creates a new class definition structure  
  def = defCreate(nClass, nFeatures);

  // reads discrete feature specifications
  if (fscanf(fp, "DISCRETE_FEATURE:%d\n", &def->iDiscreteFeature) != 1)
  {
    fprintf (stderr, "Error: problems reading discrete feature specifications.\n");
    return NULL;
  }
  if (fscanf(fp, "DISCRETE_VALUES:%d\n", &def->nDiscreteValues) != 1)
  {
    fprintf (stderr, "Error: problems reading discrete feature specifications.\n");
    return NULL;
  }
  for (i=0; i<def->nDiscreteValues; i++)
  {
    if (fscanf(fp, "%lf", &def->discreteValues[i]) != 1)
    {
      fprintf (stderr, "Error: problems reading discrete feature specifications.\n");
      return NULL;
    }
  }
  fscanf(fp, "\n\n");
  
  // defines discrete feature in def
  defDefineDiscreteFeature(def, def->iDiscreteFeature, def->discreteValues, def->nDiscreteValues);

  // reads the discrete probabilities
  fscanf(fp, "DISCRETE_PROBABILITES:\n");
  for (iClass=0; iClass<def->nClass; iClass++)
  {
    double *prob = defGetDiscreteProb(def, iClass);
    fscanf(fp, "class%*d\n");
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      if (fscanf(fp, "%lf", &prob[iDiscrete]) != 1)
      {
        fprintf (stderr, "Error: problems reading discrete probabilities.\n");
        return NULL;
      }
    }
    fscanf(fp, "\n\n");
  }


  // reads mean values for each class
  fscanf(fp, "\nMEANS:\n");
  for (iClass=0; iClass<def->nClass; iClass++)
  {
    fscanf(fp, "class%*d\n");
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      matrice *mean = defGetContMean(def, iClass, iDiscrete);
      for (i=0; i<mean->nbLignes; i++)
        fscanf(fp, "%lf", &mean->val[i]);
      fscanf(fp, "\n");
    }
    fscanf(fp, "\n");
  }
  fscanf(fp, "\n");


  // reads the covariance matrix for each iClass
  fscanf(fp, "\nCOVARIANCE_MATRICES:\n");
  for (iClass=0; iClass<def->nClass; iClass++)
  {
    fscanf(fp, "class%*d\n");
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      matrice *cov = defGetContCovariance(def, iClass, iDiscrete);
    
      for (i=0; i<cov->nbLignes; i++)
      {
        for (j=0; j<cov->nbColonnes; j++)
          fscanf(fp, "%lf", &cov->val[j + i*cov->nbColonnes]);
        fscanf(fp, "\n");
      }
      fscanf(fp, "\n");
    }
    fscanf(fp, "\n");
  }
  fscanf(fp, "\n");

  return def;
}



/*
 * defDestroy
 * 
 */
void defDestroy(classdef **def)
{
  int i;

  // frees memory for discrete specs
  free((*def)->discreteProb);
  free((*def)->discreteValues);

  // frees memory for continuous specs
  for (i=0; i<((*def)->nClass)*DEF_MAX_NUM_DISCRETE_VALUES; i++)
  {
    matrice *mean = (*def)->contMeans[i];
    matrice *cov  = (*def)->contCovariances[i];
    free_matrice(&mean);
    free_matrice(&cov);
  }
  free((*def)->contMeans);
  free((*def)->contCovariances);
  free((*def)->iContFeatures);
  
  free(*def);
  *def = NULL;
}



int defOutputTxt(classdef *def, FILE *fpOutput)
{
  int class;
  int val;
  int i, j;

  // outputs number of classes and of features
  fprintf(fpOutput, "CLASSES: %d\n", def->nClass);
  fprintf(fpOutput, "FEATURES: %d\n", def->nFeatures);

  fprintf(fpOutput, "DISCRETE_FEATURE: %d\n", def->iDiscreteFeature);
  fprintf(fpOutput, "DISCRETE_VALUES: %d\n", def->nDiscreteValues);
  for (val=0; val<def->nDiscreteValues; val++)
    fprintf(fpOutput, "%.10lf\t", def->discreteValues[val]);
  fprintf(fpOutput, "\n\n");

  fprintf(fpOutput, "DISCRETE_PROBABILITES:\n");
  for (class=0; class<def->nClass; class++)
  {
    double *prob = defGetDiscreteProb(def, class);
    fprintf(fpOutput, "class %d\n", class+1);
    for (val=0; val<def->nDiscreteValues; val++)
      fprintf(fpOutput, "%.10lf\t", prob[val]);
    fprintf(fpOutput, "\n\n");
  }

  // outputs mean values for each class, conditioned to the discrete variables
  fprintf(fpOutput, "\nMEANS:\n");
  for (class=0; class<def->nClass; class++)
  {
    fprintf(fpOutput, "class %d\n", class+1);
    for (val=0; val<def->nDiscreteValues; val++)
    {
      matrice *mean = defGetContMean(def, class, val);
      if (mean == NULL)
        return 0;
    
      for (i=0; i<mean->nbLignes; i++)
        fprintf(fpOutput, "%.10lf\t", mean->val[i]);
      fprintf(fpOutput, "\n");
    }
    fprintf(fpOutput, "\n");
  }
  fprintf(fpOutput, "\n");


  // outputs the covariance matrix for each class, conditioned to the discrete variables
  fprintf(fpOutput, "\nCOVARIANCE_MATRICES:\n");
  for (class=0; class<def->nClass; class++)
  {
    fprintf(fpOutput, "class %d\n", class+1);
    for (val=0; val<def->nDiscreteValues; val++)
    {
      matrice *cov = defGetContCovariance(def, class, val);
    
      for (i=0; i<cov->nbLignes; i++)
      {
        for (j=0; j<cov->nbColonnes; j++)
          fprintf(fpOutput, "%.10lf\t", cov->val[j + i*cov->nbColonnes]);
        fprintf(fpOutput, "\n");
      }
      fprintf(fpOutput, "\n");
    }
    fprintf(fpOutput, "\n");
  }
  fprintf(fpOutput, "\n");

 
  return 1;
}



double *defGetDiscreteProb(classdef *def, int iClass)
{
  int pos;
  
  if (iClass<0 || iClass>=def->nClass)
    return NULL;

  pos = iClass*DEF_MAX_NUM_DISCRETE_VALUES;
  return &(def->discreteProb[pos]);
}

matrice *defGetContMean(classdef *def, int iClass, int val)
{
  int pos;
  
  if (iClass<0 || iClass>=def->nClass || val<0 || val>=def->nDiscreteValues)
    return NULL;

  pos = val + iClass*DEF_MAX_NUM_DISCRETE_VALUES;
  return def->contMeans[pos];
}

matrice *defGetContCovariance(classdef *def, int iClass, int val)
{
  int pos;
  
  if (iClass<0 || iClass>=def->nClass || val<0 || val>=def->nDiscreteValues)
    return NULL;

  pos = val + iClass*DEF_MAX_NUM_DISCRETE_VALUES;
  return def->contCovariances[pos];
}

int defSetContCovariance(classdef *def, int iClass, int val, matrice *cov)
{
  int pos;
  
  if (iClass<0 || iClass>=def->nClass || val<0 || val>=def->nDiscreteValues)
    return 0;

  pos = val + iClass*DEF_MAX_NUM_DISCRETE_VALUES;

  free_matrice(&def->contCovariances[pos]);
  def->contCovariances[pos] = cov;
  return 1;
}

int defGetDiscreteValueIndex(classdef *def, double value)
{
  int i;
  
  for (i=0; i<def->nDiscreteValues; i++)
    if (fabs(value - def->discreteValues[i]) < 1e-10)
      return i;

  return -1;
}



/*
 * defGetPoint
 * 
 * Given input data and a class definition, returns the continuous and discrete
 * values corresponding to all features (layers) of a 2D point at position pos
 * 
 */
int defGetPoint(classdef *def, int pos, unsigned char *dataInput, int size_x, int size_y, InrType typeInput,
                double *contValues, int *iDiscrete)
{
  int i, z;
  double value;
  
  // if there is a discrete layer, checks its value (index, label) 
  *iDiscrete = 0;
  if (def->iDiscreteFeature >= 0)
  {
    z = def->iDiscreteFeature;
    value = inrGetDataValue(dataInput, typeInput, pos + z*size_x*size_y);
    *iDiscrete = defGetDiscreteValueIndex(def, value);
  }
  if (*iDiscrete == -1)
    return 0;
  
  // reads values for each layer of continous data into the values array
  for (i=0; i<def->nContFeatures; i++)
  {
    z = def->iContFeatures[i];
    contValues[i] = inrGetDataValue(dataInput, typeInput, pos + z*size_x*size_y);
  }
  return def->nContFeatures;
}


/*
 * defProcessDiscrete
 * 
 * Updates class definition related to discrete features, given input data and
 * sample classification.
 * Features are analyzed so as to identify discrete features and store the
 * number of distinct labels and their values
 */
int defProcessDiscrete(classdef *def, unsigned char *classif, unsigned char *dataInput,
                       int size_x, int size_y, int size_z,  InrType typeInput)
{
  int z;
  
  for (z=0; z<size_z; z++)
  {
    int nDiscreteValues;
    double discreteValues[DEF_MAX_NUM_DISCRETE_VALUES];

    if (!defCheckDiscreteFeature(z, classif, dataInput, size_x, size_y, size_z, typeInput, discreteValues, &nDiscreteValues))
      continue;

    // if feature is discrete, defines its specification in the classdef structure
    defDefineDiscreteFeature(def, z, discreteValues, nDiscreteValues);
  }
  
  return (def->iDiscreteFeature != -1);
}

/*
 * defComputeDiscreteProb
 * 
 * Computes probabilties of each discrete value, for each class
 */
int defComputeDiscreteProb(classdef *def, unsigned char *classif, unsigned char *dataInput,
                           int size_x, int size_y, int size_z,  InrType typeInput)
{                  
  int nValueInstances[DEF_MAX_NUM_DISCRETE_VALUES];
  int nInstances;
  int d;
  int iDiscrete;
  int iClass;
  int i, z;
  double value;
  double *prob;
  
  z = def->iDiscreteFeature;
  if (z<0 || z>=size_z)
    return 0;

  for (d=0; d<def->nDiscreteValues; d++)
  {
    for (iClass=0; iClass<def->nClass; iClass++)
    {  
      nInstances = 0;
      for (i=0; i<DEF_MAX_NUM_DISCRETE_VALUES; i++)
        nValueInstances[i] = 0;

      for (i=0; i<size_x*size_y; i++)
      {
        // checks if point is from the current class
        if (classif[i] != iClass+1)
          continue;
          
        // reads value for point i on layer z
        value = inrGetDataValue(dataInput, typeInput, i + z*size_x*size_y);;

        // obtains corresponding index (label) of the discrete value
        iDiscrete = defGetDiscreteValueIndex(def, value);
        if (iDiscrete < 0)
          continue;
          
        nValueInstances[iDiscrete]++;  
        nInstances++;
      }
      
      prob = defGetDiscreteProb(def, iClass);
      for (i=0; i<DEF_MAX_NUM_DISCRETE_VALUES; i++)
      {
        if (nValueInstances[i] > 1)
          prob[i] = (double)nValueInstances[i] / nInstances;
        else
          prob[i] = 0;
      }
    }
  }
  return 1;
}



/*
 * auxiliary methods for defProcessDiscrete
 */
int defCheckDiscreteFeature(int iFeature, unsigned char *ROI, unsigned char *dataInput, int size_x, int size_y, int size_z,
                            InrType typeInput, double *discreteValues, int *nDiscreteValues)
{
  double value;
  int i;
  int iValue;
  int z = iFeature;
  
  int nKnown = 0;

  for (i=0; i<size_x*size_y; i++)
  {
    // discards non-classified points
    if (!ROI[i])
      continue;
      
    // reads value for the layer z of the data point
    value = inrGetDataValue(dataInput, typeInput, i + z*size_x*size_y);
    
    // checks if value is known
    for (iValue=0; iValue<nKnown; iValue++)
      if (discreteValues[iValue] == value)
        break;
    
    // if known, goes on to check other values
    if (iValue < nKnown)
      continue;
     
    // if number of known values reaches limit, returns false (feature is not discrete)
    if (nKnown >= DEF_MAX_NUM_DISCRETE_VALUES)
      return 0;

    // adds new value to known values
    discreteValues[nKnown] = value;
    nKnown++;
  }
  *nDiscreteValues = nKnown;
  return 1;
}


int defDefineDiscreteFeature(classdef *def, int iFeature, double *values, int nValues)
{
  int i, iCont;

  // updates discrete variables
  for (i=0; i<nValues; i++)
    def->discreteValues[i] = values[i];

  for (i=0; i<def->nClass*DEF_MAX_NUM_DISCRETE_VALUES; i++)
    def->discreteProb[i] = 0;
  
  def->nDiscreteValues  = nValues;
  def->iDiscreteFeature = iFeature;


  // updates continuous variables
  def->nContFeatures    = def->nFeatures - 1;

  for (i=0, iCont=0; i<def->nFeatures; i++)
  {
    if (i == def->iDiscreteFeature)
      continue;
    def->iContFeatures[iCont++] = i;
  }

  for (i=0; i<def->nClass*DEF_MAX_NUM_DISCRETE_VALUES; i++)
  {
    free_matrice(&def->contMeans[i]);
    free_matrice(&def->contCovariances[i]);
    
    def->contMeans[i] = Matrice_Nulle(def->nContFeatures, 1);
    def->contCovariances[i] = Matrice_Nulle(def->nContFeatures, def->nContFeatures);
  }

  
  return 1;
}




