#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE       1024


static char Ucmd[] = "-ifile inr_input -roi roi_mask -ofile tab_output\n";

static char Udetail[] =
"Converts an INRIMAGE image into a tab-delimited text file.\n\
\nEach (x,y) point of the image within the ROI is converted to a line in the\n\
output file. Output lines will contain the x and y coordinates of the point\n\
along with the value of each entry in the z dimension.\n\
\nInput:\n\
\t-ifile\tFile name of the input INRIMAGE image file\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output tab-delimited text file\n";


/*
 * GLOBAL CONTEXT
 */
typedef struct 
{
    char fnameInput[FNAMESIZE];
    char fnameROIMask[FNAMESIZE];
    char fnameOutput[FNAMESIZE];

    InrType typeInput;
    int     typesizeInput;

    unsigned char *dataInput;
    unsigned char *ROIMask;
    FILE *fileOutput;

    int size_x, size_y, size_z;
} Inr2Tab_gc;
                                  



/*
 * outputTabLine
 *
 * Outputs a given data line to the output file in tab-delimited format
 */
int outputTabLine (Inr2Tab_gc *gc, int y)
{
    int x, z;

    // processes each point at a time
    for (x = 0; x < gc->size_x; x++)
    {
        // discards pixels outside of region of interest
        if (!gc->ROIMask[x + y*gc->size_x])
            continue;

        // outputs point (x,y) coordinates
        fprintf(gc->fileOutput, "%d\t%d", x, y);

        // outputs each z value for the point
        for (z = 0; z < gc->size_z; z++)
        {
            double val = inrGetDataValue(gc->dataInput, gc->typeInput, x+z*gc->size_x);
            if (inrTypeIsInteger(gc->typeInput))
                fprintf(gc->fileOutput, "\t%lf", val);
            else
                fprintf(gc->fileOutput, "\t%d", (int)val);
        }

        // finishes pixel line
        fprintf(gc->fileOutput, "\n");
    }

    return 1;
}


/*
 * checkLineInROI
 *
 * Checks if a given line is within the GC's region of interest.
 * Returns zero if it is not, non-zero otherwise.
 */
int checkLineInROI (Inr2Tab_gc *gc, int y)
{
    int x;
    for (x = 0; x < gc->size_x; x++)
        if (gc->ROIMask[x + y*gc->size_x])
            return 1;
    return 0;
}


/*
 * main
 *
 * Main method
 */
int main (int argc, char **argv)
{
    Inr2Tab_gc  gc_obj;           /* general context */
    Inr2Tab_gc *gc = &gc_obj;

    InrType type;

    int i, j, k;
    int y;

    // reads arguments
    inr_init (argc, argv, "1.0", Ucmd, Udetail);
    if (!igetopt1 ("-ifile", "%s", gc->fnameInput))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-roi"  , "%s", gc->fnameROIMask))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-ofile", "%s", gc->fnameOutput))
        iusage_ (Ucmd, Udetail);

    // exits program if there are unused options still present
    if (tstopts ())                
        iusage_ (Ucmd, Udetail);

    // reads header information for all input data and checks if they are appropriate
    if (!inrReadHeader(gc->fnameInput, &gc->size_x, &gc->size_y, &gc->size_z, &gc->typeInput))
    {
        fprintf(stderr, "Error: could not read header information for input file '%s'.\n", gc->fnameInput);
        return 0;
    }
    if (!inrReadHeader(gc->fnameROIMask, &i, &j, &k, &type))
    {
        fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", gc->fnameROIMask);
        return 0;
    }
    if (type != InrByte || k != 1)
    {
        fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", gc->fnameROIMask);
        return 0;
    }
    if (i != gc->size_x || j != gc->size_y)
    {
        fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
        return 0;
    }

    // reads the ROI mask image
    gc->ROIMask = inrReadByteImage (gc->fnameROIMask, &i, &j);
    if (!gc->ROIMask)
    {
        fprintf (stderr, "Error: could not read ROI mask data.\n");
        return 0;
    }

    // allocates memory for the input and output data
    gc->typesizeInput = InrTypeSizes[gc->typeInput]/8;
    gc->dataInput     = (unsigned char *) malloc (gc->size_x*gc->size_z * gc->typesizeInput * sizeof(unsigned char));


    // opens output file
    gc->fileOutput = fopen(gc->fnameOutput, "wt");
    if (!gc->fileOutput)
    {
        printf("Error: could not write to output file '%s'.\n", gc->fnameOutput);
        return 0;
    }


    // processes the data
    for (y = 0; y < gc->size_y; y++)
    {
        fprintf(stderr, "Processing... %d%%\r", 100*y/gc->size_y);

        // checks if line y is within region of interest
        if (!checkLineInROI(gc, y))
            continue;

        // reads input data line y
        if (!inrReadLine(gc->fnameInput, y, gc->dataInput))
        {
            fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, gc->fnameInput);
            return 0;
        }

        // writes data line y to tab-delimited output file
        outputTabLine(gc, y);
    }

    // closes output file
    fclose(gc->fileOutput);

    fprintf(stderr, "INRIMAGE file successfully converted to tab-delimited text format.\n");

    free(gc->dataInput);
    free(gc->ROIMask);

    return 1;
}
