#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_croplib.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE      1024
#define CUTDEG_DEFAULT 0

static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir [-mask mask_file] \
[-tl x y] [-br x y] [-cut longitude_range]";

static char Udetail[] =
"\nCrops INRIMAGE images based on a mask image or on specified corner limits\n\
\nBoth input and mask images must be in INRIMAGE format with a corresponding \
ENVI file header (.hdr) containing georeference information\n\
\nInput:\n\
\t-idir\tDirectory containing georeferenced images to be cropped\n\
\t-ifile\tFile name of one georeferenced image to be cropped\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nInput(optional):\n\
\t-mask\tFile name of a georeferenced mask image that defines the region\n\
\t\tof interest for cropping. Mask should be referenced to the same\n\
\t\tdatum and projection as the input image(s). Mask should be zero\n\
\t\tfor pixels to be masked out and non-zero otherwise\n\
\t-tl\tTop-left point of a rectangle of interest for cropping\n\
\t-br\tBottom-right point of a rectangle of interest for cropping\n\
\t-cut\tValue in longitude degrees that specifies how much of the\n\
\t\timage(s) should be always ignored left and right due to Earth\n\
\t\tcurvature and data deterioration. A cut value of 0 degrees is\n\
\t\tused by default\n\
\nOutput:\n\
\t-odir\tOutput directory for the cropped INRIMAGE image(s)\n";






/*
 *  inrCutFile
 *
 *  Crops a single file (in INR format) by cutting off a specified cut
 *  value in longitude degrees from the left and right of the image
 */
int inrCutFile(char *fnameFull, char *outfnameFull, double cutDeg)
{
  int     size_x, size_y, size_z;
  InrType type;   
  InrGeorefInfo georef;
  double  bottomright_x;
  double  bottomright_y;

  // retrieves input file dimensions
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return 0;

  // reads georef info from input file's ENVI .hdr header
  if (!inrReadGeoref(fnameFull, &georef))
  {
    printf("Error: could not read georeference information for file '%s'\n", fnameFull);
    return 0;
  }

  // computes bottom-right limits and defines crop region
  bottomright_y  = georef.topleft_y - georef.pixelsize_y*(size_y-0.5);
  bottomright_x  = georef.topleft_x + georef.pixelsize_x*(size_x-0.5) - cutDeg;
  georef.topleft_x += cutDeg;

  // crops the file
  if (!inrCropGeoref(fnameFull, outfnameFull, georef.topleft_x, georef.topleft_y,
                     bottomright_x, bottomright_y))
    return 0;

  return 1;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  char   fnameMask[FNAMESIZE];
  double topleft_x, topleft_y, bottomright_x, bottomright_y;
  double cutDeg;
  int    okDirIn, okDirOut, okFile, okMask;
  int    okTopLeft, okBottomRight, okRegion, okCut;
  DIR   *dir;
  struct dirent *ent;
  int    cropped = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  , "%s", dirInput);
  okDirOut      = igetopt1("-odir"  , "%s", dirOutput);
  okFile        = igetopt1("-ifile" , "%s", fnameInput);
  okMask        = igetopt1("-mask"  , "%s", fnameMask);
  okTopLeft     = igetopt2("-tl", "%lf", &topleft_x,  "%lf", &topleft_y);
  okBottomRight = igetopt2("-br", "%lf", &bottomright_x, "%lf", &bottomright_y);
  okCut         = igetopt1("-cut" , "%lf", &cutDeg);
  
  okRegion = (okTopLeft+okBottomRight == 4)?1:0;

  if ((okDirIn && okFile) || (!okDirIn && !okFile))
    iusage_ (Ucmd, Udetail);
  if (!okDirOut)
    iusage_ (Ucmd, Udetail);
  if (okTopLeft+okBottomRight!=0 && okTopLeft+okBottomRight!=4)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // defines default cut value
  if (!okCut)
    cutDeg = CUTDEG_DEFAULT;

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFile)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // crops file(s) in the directory
  cropped = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be cropped
    if (okFile && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input and output file names (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);
    sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);

    // checks if input is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // cuts undesired portions of image
    if (!inrCutFile(fnameFull, outfnameFull, cutDeg))
      continue;
    
    // crops to region of interest
    if (okRegion)
    {
      if (!inrCropGeoref(outfnameFull, outfnameFull, topleft_x, topleft_y,
                         bottomright_x, bottomright_y))
        continue;
    }
    
    // crops to mask file
    if (okMask)
    {
      if (!inrCropGeorefMask(outfnameFull, outfnameFull, fnameMask))
        continue;
    }

    // increments counter
    printf("Cropped file '%s'\n", ent->d_name);
    cropped++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d INRIMAGE file(s) cropped.\n", cropped);
  return 1;
}

