#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "inr_io.h"

#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)
#define ROUND(x) ((int)floor((x)+0.5))

#define BUFSIZE 1024


/*
 *  inrCropBase
 *
 *  Base method used for cropping INRIMAGE files, called by all other inr_crop
 *  methods.
 *  The mask file can remain unspecified (NULL). In case it is specified,
 *  file dimensions must match values (crop_max_x - crop_min_x) and
 *  (crop_max_y - crop_min_y).
 * 
 */
int inrCropBase(char *fnameIn, char *fnameOut, char *fnameMask,
                int crop_min_x, int crop_min_y, int crop_max_x, int crop_max_y)
{
    unsigned char *dataIn = NULL;
    unsigned char *dataMask;

    int in_size_x, in_size_y, in_size_z;
    int out_size_x, out_size_y, out_size_z;

    InrType type;
    int typeSize;
    int typeSizeMask;
    int nDataOut;
    unsigned char *dataOut;

    int temp_crop;


    // reads input data from input file 
    dataIn = inrReadDataSequence(fnameIn, &in_size_x, &in_size_y, &in_size_z, &type);
    if (!dataIn)
        return 0;
  
    // obtains data type size in bytes
    typeSize = InrTypeSizes[type]/8;

    // swaps crop_min/max values if necessary
    if (crop_min_x > crop_max_x)
    {
        temp_crop  = crop_min_x;
        crop_min_x = crop_max_x;
        crop_max_x = temp_crop;
    }
    if (crop_min_y > crop_max_y)
    {
        temp_crop  = crop_min_y;
        crop_min_y = crop_max_y;
        crop_max_y = temp_crop;
    }
  
    // defines output file dimensions
    out_size_x = crop_max_x - crop_min_x + 1;
    out_size_y = crop_max_y - crop_min_y + 1;
    out_size_z = in_size_z;

    // reads mask data, if specified
    dataMask = NULL;
    typeSizeMask = 0;
    if (fnameMask)
    {
        int mask_size_x, mask_size_y, mask_size_z;
        InrType mask_type;   
        dataMask = inrReadDataSequence(fnameMask, &mask_size_x, &mask_size_y, &mask_size_z, &mask_type);
        if (!dataMask)
        {
            free(dataIn);
            return 0;
        }
        // checks if mask data is valid
        // mask should have the same dimensions of the output data
        if (mask_size_x != out_size_x || mask_size_y != out_size_y)
        {
            free(dataIn);
            free(dataMask);
            return 0;
        }
        // obtains mask data type size in bytes
        typeSizeMask = InrTypeSizes[mask_type]/8;
    }

    // allocates memory for the output data
    nDataOut = (out_size_x)*(out_size_y)*(out_size_z);
    dataOut = (unsigned char *)malloc(nDataOut*typeSize);
    if (!dataOut)
    {
        free(dataIn);
        free(dataMask);
        return 0;
    }
        
    // checks if crop region is out-of-bounds
    // if it is, output is a blank image (all pixels are 0)
    if (crop_min_x >= in_size_x || crop_min_y >= in_size_y ||
        crop_max_x < 0 || crop_max_y < 0)
    {
        // writes output data file 
        int ret;
        memset(dataOut, 0, out_size_x*out_size_y*out_size_z);
        ret = inrWriteDataSequence(fnameOut, out_size_x, out_size_y, out_size_z, dataOut, type);

        // frees memory used and returns
        free(dataIn);
        free(dataOut);
        free(dataMask);
        return ret;
    }
  
    {
        int ret;

        // defines proper min/max values for input data
        int min_x = MIN(MAX(crop_min_x,0),in_size_x-1);
        int min_y = MIN(MAX(crop_min_y,0),in_size_y-1);
        int max_x = MIN(MAX(crop_max_x,0),in_size_x-1);
        int max_y = MIN(MAX(crop_max_y,0),in_size_y-1);
          

        // DATA CROPPING
          
        int x, y, z;
        int posIn   = 0;
        int posOut  = 0;
        int posMask = 0;
        int topPad    = (min_y - crop_min_y)*typeSize*out_size_x;
        int bottomPad = (crop_max_y - max_y)*typeSize*out_size_x;
        int leftPad   = (min_x - crop_min_x)*typeSize;
        int rightPad  = (crop_max_x - max_x)*typeSize;
          
        for (z=0; z<in_size_z; z++)
        {
            // applies top padding to crop area above input image (padding to 0)
            memset(dataOut+posOut, 0, topPad);
            posOut += topPad;

            // scans input lines
            for (y=min_y; y<=max_y; y++)
            {
                // applies left padding to crop area left of input image (padding to 0)
                memset(dataOut+posOut, 0, leftPad);
                posOut += leftPad;
                  
                // scans input samples
                for (x=min_x; x<=max_x; x++)
                {
                    int i;
                    // determines input and mask data position
                    posIn   = (x + y*in_size_x + z*in_size_x*in_size_y)*typeSize;
                    posMask = (int) (posOut*((double)typeSizeMask/typeSize));
                  
                    // if specified, applies mask to line, and then copies data to output
                    if (dataMask && dataMask[posMask]==0)
                        for (i=0; i<typeSize; i++)
                            dataOut[posOut++] = 0;
                    else
                        for (i=0; i<typeSize; i++)
                            dataOut[posOut++] = dataIn[posIn++];
                }

                // applies right padding to crop area right of input image (padding to 0)
                memset(dataOut+posOut, 0, rightPad);
                posOut += rightPad;
            }
            
            // applies bottom padding to crop area below input image (padding to 0)
            memset(dataOut+posOut, 0, bottomPad);
            posOut += bottomPad;
        }
          
        // writes output data file 
        ret = inrWriteDataSequence(fnameOut, out_size_x, out_size_y, out_size_z, dataOut, type);

        // frees memory used and returns
        free(dataIn);
        free(dataOut);
        free(dataMask);
        return ret;
    }
}



/*
 *  inrCrop
 *
 *  Crops an INRIMAGE file to a given rectangle of interest
 */
int inrCrop(char *fnameIn, char *fnameOut, int min_x, int min_y, int max_x, int max_y)
{
    return inrCropBase(fnameIn, fnameOut, NULL, min_x, min_y, max_x, max_y);
}



/*
 *  inrCropMask
 *
 *  Crops an INRIMAGE file to the rectangle of interest defined by a mask
 *  file.
 *  Mask should be zero for pixels to be masked out and non-zero otherwise
 */
int inrCropMask(char *fnameIn, char *fnameOut, char *fnameMask, int pos_x, int pos_y)
{
    // retrieves mask size
    int size_x, size_y, size_z;
    InrType type;   
    if (!inrReadHeader(fnameMask, &size_x, &size_y, &size_z, &type))
        return 0;

    // crops image
    return inrCropBase(fnameIn, fnameOut, fnameMask, pos_x, pos_y,
                       pos_x + size_x - 1, pos_y + size_y - 1);
}



/*
 *  inrCropGeoref
 *
 *  Crops an INRIMAGE file to a given rectangle of interest.
 *  Input file should be georeferenced (i.e., should have map information
 *  stored in an ENVI .hdr header file)
 */
int inrCropGeoref(char *fnameIn, char *fnameOut,
                  double crop_topleft_x, double crop_topleft_y,
                  double crop_bottomright_x, double crop_bottomright_y)
{
    InrGeorefInfo in_georef;
    InrGeorefInfo out_georef;
    int min_x, min_y, max_x, max_y;

    // reads georef info from input file's ENVI .hdr header
    if (!inrReadGeoref(fnameIn, &in_georef))
        return 0;
        
    // transforms georeferenced map coordinates into image coordinates
    // image coordinates will encompass the entire region specified
    min_x = (int)floor((crop_topleft_x - in_georef.topleft_x)/in_georef.pixelsize_x);
    min_y = (int)floor((in_georef.topleft_y - crop_topleft_y)/in_georef.pixelsize_y);
    max_x = (int)floor((crop_bottomright_x - in_georef.topleft_x)/in_georef.pixelsize_x);
    max_y = (int)floor((in_georef.topleft_y - crop_bottomright_y)/in_georef.pixelsize_y);
      
    // crops image based on image coordinates
    if (!inrCropBase(fnameIn, fnameOut, NULL, min_x, min_y, max_x, max_y))
        return 0;

    // writes effective georeference information for cropped image
    out_georef = in_georef;
    out_georef.topleft_x = in_georef.topleft_x + min_x*in_georef.pixelsize_x;
    out_georef.topleft_y = in_georef.topleft_y - min_y*in_georef.pixelsize_y;
    inrWriteGeoref(fnameOut, out_georef);
      
    return 1;
}


/*
 *  inrCropGeorefMask
 *
 *  Crops an INRIMAGE file to the rectangle of interest defined by a mask
 *  file.
 *  Mask should be zero for pixels to be masked out and non-zero otherwise.
 *  Both the input and mask files should be georeferenced (i.e., should
 *  have map information stored in an ENVI .hdr header file), use the same
 *  reference system (i.e., same datum and projection) and have the same
 *  spatial resolution (i.e., same pixel size in the x and y dimensions)
 */
int inrCropGeorefMask(char *fnameIn, char *fnameOut, char *fnameMask)
{
    InrGeorefInfo in_georef;
    InrGeorefInfo mask_georef;
    int pos_x, pos_y;

    // reads georef info from mask file's ENVI .hdr header
    if (!inrReadGeoref(fnameMask, &mask_georef))
    {
        printf("Error: could not read georeference information for mask '%s'\n", fnameMask);
        return 0;
    }
        
    // reads georef info from input file's ENVI .hdr header
    if (!inrReadGeoref(fnameIn, &in_georef))
    {
        printf("Error: could not read georeference information for file '%s'\n", fnameIn);
        return 0;
    }
        
    // asserts that pixelsize info are the same for input and mask images
    if (in_georef.pixelsize_x != mask_georef.pixelsize_x ||
        in_georef.pixelsize_y != mask_georef.pixelsize_y)
    {
        printf("Error: pixel sizes for file '%s' and mask '%s' do not match.\n",
                fnameIn, fnameMask);
        return 0;
    }
        
    // asserts that datum/proj info are the same for input and mask images
    if (strcmp(in_georef.datum, mask_georef.datum) || 
        strcmp(in_georef.projection,  mask_georef.projection))
    {
        printf("Error: datum/projection for file '%s' and mask '%s' do not match.\n",
            fnameIn, fnameMask);
        return 0;
    }

    // finds out mask position in input image coordinates (i.e., transforms
    // mask's georeferenced top-left coordinates into input image coordinates)
    // - coordinates are rounded to nearest integer value:
    //   leads to error of up to 1/2 an input image pixel in final cropped image
    //   georeference information
    pos_x = ROUND((mask_georef.topleft_x - in_georef.topleft_x)/in_georef.pixelsize_x);
    pos_y = ROUND((in_georef.topleft_y - mask_georef.topleft_y)/in_georef.pixelsize_y);
      
    // crops image based on mask's position in input image coordinates
    if (!inrCropMask(fnameIn, fnameOut, fnameMask, pos_x, pos_y))
        return 0;

    // writes effective georeference information for cropped image
    inrWriteGeoref(fnameOut, mask_georef);
    return 1;
}

