#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "inr_io.h"
#include "file_tools.h"

#define HEADERSIZE 256
#define FNAMESIZE  1024
#define BUFSIZE    16384


// INR data type names sizes (corresponding to InrType enum)
char *InrTypeNames[] = {
  "unsigned fixed", // byte
  "unsigned fixed", // unsigned int
  "unsigned fixed", // unsigned long int
  "signed fixed",   // signed int
  "signed fixed",   // signed long int
  "float",          // float
  "float"           // double
};

int InrTypeSizes[] = {
  8,  // byte
  16, // unsigned int
  32, // unsigned long int
  16, // signed int
  32, // signed long int
  32, // float
  64  // double
};

int Inr2EnviTypes[] = {
  1,  // byte
  12, // unsigned short
  13, // unsigned long
  2,  // signed int
  3,  // signed long int
  4,  // float
  5   // double
};



/*
 *  inrTypeIsInteger
 *
 *  Indicates whether the given type is an integer type (byte, unsigned short,
 *  signed int, etc.) or a floating point type (float, double).
 *  Returns zero if not integer, non-zero otherwise
 */
int inrTypeIsInteger (InrType type)
{
    if (type == InrFloat || type == InrDouble)
        return 1;
    else
        return 0;
}


/*
 *  inrGetType
 *
 *  Determines the INR type corresponding to the name and size provided
 */
int inrGetType(char *typeName, int typeSize, InrType *type)
{
  // determines total number of types available
  int nTypes = sizeof(InrTypeSizes)/sizeof(int);

  int i;
  for (i=0; i<nTypes; i++)
  {
    // checks if type i is the one
    if ( InrTypeSizes[i]==typeSize && !strncmp(InrTypeNames[i],typeName,strlen(InrTypeNames[i])) )
    {
      *type = (InrType)i;
      return 1;
    }
  }
  // returns 0 in case of failure
  return 0;
}



/*
 *  inrGetDataValue
 *
 *  Returns the value of the data of any INR type at a given position
 */
double inrGetDataValue(unsigned char *data, InrType type, int pos)
{
  switch (type)
  {
    case InrByte:
      return ((double) (data[pos]));
      break;
      
    case InrUnsignedInt:
      return ((double) ( ((unsigned short *)data)[pos] ) );
      break;
       
    case InrUnsignedLongInt:
      return ((double) ( ((unsigned long *)data)[pos] ) );
      break;
       
    case InrSignedInt:
      return ((double) ( ((signed short *)data)[pos] ) );
      break;
       
    case InrSignedLongInt:
      return ((double) ( ((signed long *)data)[pos] ) );
      break;
       
    case InrFloat:
      return ((double) ( ((float *)data)[pos] ) );
      break;
      
    case InrDouble:
      return ((double) ( ((double *)data)[pos] ) );
      break;
      
    default:
      return -1;
  }
}



/*
 *  inrPutDataValue
 *
 *  Inserts a value in data of any INR type at a given position
 */
int inrPutDataValue(unsigned char *data, InrType type, int pos, double value)
{
  switch (type)
  {
    case InrByte:
      data[pos] = (unsigned char)value;
      return 1;
      
    case InrUnsignedInt:
      ((unsigned short *)data)[pos] = (unsigned short)value;
      return 1;
       
    case InrUnsignedLongInt:
      ((unsigned long *)data)[pos] = (unsigned long)value;
      return 1;
       
    case InrSignedInt:
      ((signed short *)data)[pos] = (signed short)value;
      return 1;
       
    case InrSignedLongInt:
      ((signed long *)data)[pos] = (signed long)value;
      return 1;
       
    case InrFloat:
      ((float *)data)[pos] = (float)value;
      return 1;
      
    case InrDouble:
      ((double *)data)[pos] = value;
      return 1;
      
    default:
      return 0;
  }
}



/*
 *  inrReadHeader, inrReadHeaderPtr
 *
 *  Reads header info of a INR file.
 */
int inrReadHeader(char *filename, int *size_x, int *size_y, int *size_z, InrType *type)
{
  int ret;
  FILE *fpInr = fopen(filename, "rb");
  if (!fpInr)
    return 0;

  ret = inrReadHeaderPtr(fpInr, size_x, size_y, size_z, type);
  fclose(fpInr);
  return ret;
}

int inrReadHeaderPtr(FILE *fpInr, int *size_x, int *size_y, int *size_z, InrType *type)
{
  int    infoPresent;
  int    typeSize;
  char   typeName[HEADERSIZE];
  char  *ptrHeader;
  long   offset;
  size_t nbytes;

  // reads header data into a buffer
  char strHeader[INRHDRBUFSIZE];
  nbytes = fread(strHeader, 1, INRHDRBUFSIZE-1, fpInr);
  strHeader[nbytes] = '\0';

  // checks file format (should be INRIMAGE v4)
  if (strncmp(strHeader, "#INRIMAGE-4#", 12))
    return 0;

  // reads header info
  infoPresent = 0;
  if ((ptrHeader = strstr(strHeader,"XDIM"))!=NULL)
    infoPresent += sscanf(ptrHeader, "XDIM=%d", size_x);
  if ((ptrHeader = strstr(strHeader,"YDIM"))!=NULL)
    infoPresent += sscanf(ptrHeader, "YDIM=%d", size_y);
  if ((ptrHeader = strstr(strHeader,"ZDIM"))!=NULL)
    infoPresent += sscanf(ptrHeader, "ZDIM=%d", size_z);
  if ((ptrHeader = strstr(strHeader,"TYPE"))!=NULL)
    infoPresent += sscanf(ptrHeader, "TYPE=%255[^\n]s", typeName);
  if ((ptrHeader = strstr(strHeader,"PIXSIZE"))!=NULL)
    infoPresent += sscanf(ptrHeader, "PIXSIZE=%d bits", &typeSize);

  // asserts that all required info is present and valid
  if (infoPresent != 5)
    return 0;

  // finds the type of the file data (eg. InrByte, InrFloat)
  if (!inrGetType(typeName, typeSize, type))
    return 0;

  // skips file pointer to the first byte after the end of the header
  // obs: INR header end is indicated by a '##}' string
  if ((ptrHeader = strstr(strHeader,"##}"))==NULL)
    return 0;
  offset = (long)(ptrHeader - strHeader) + 4;
  fseek(fpInr, offset, SEEK_SET);

  return 1;
}



/*
 *  inrWriteHeader, inrWriteHeaderPtr
 *
 *  Writes the header of a INR file. This header has to be written before the
 *  rest of the data.
 */
int inrWriteHeader(char *filename, int size_x, int size_y, int size_z, InrType type)
{
  int ret;
  FILE *fpInr = fopen(filename, "wb");
  if (!fpInr)
    return 0;

  ret = inrWriteHeaderPtr(fpInr, size_x, size_y, size_z, type);
  fclose(fpInr);
  return ret;
}


int inrWriteHeaderPtr(FILE *fpInr, int size_x, int size_y, int size_z, InrType type)
{
  // writes core header information to output inr file
  int bytesWritten = 0;
  bytesWritten += fprintf(fpInr, "#INRIMAGE-4#{\n");
  bytesWritten += fprintf(fpInr, "XDIM=%d \n", size_x);
  bytesWritten += fprintf(fpInr, "YDIM=%d \n", size_y);
  bytesWritten += fprintf(fpInr, "ZDIM=%d \n", size_z);
  bytesWritten += fprintf(fpInr, "VDIM=1 \n");
  bytesWritten += fprintf(fpInr, "TYPE=%s\n", InrTypeNames[type]);
  bytesWritten += fprintf(fpInr, "PIXSIZE=%d bits\n", InrTypeSizes[type]);
  if (type != InrFloat)
    bytesWritten += fprintf(fpInr, "SCALE=2**0\n");
  bytesWritten += fprintf(fpInr, "CPU=decm");

  // completes the header
  for (; bytesWritten < HEADERSIZE-4; bytesWritten++)
    fputc('\n', fpInr);
  fprintf(fpInr, "##}\n");

  return 1;
}

/*
 *  inrCheckFile, inrCheckFilePtr
 * 
 *  Checks if a given file is an INRIMAGE image
 */
int inrCheckFile(char *filename)
{
  int ret;
  FILE *fp = fopen(filename, "rb");
  if (!fp)
    return 0;

  ret = inrCheckFilePtr(fp);
  fclose(fp);
  return ret;
}

int inrCheckFilePtr(FILE *fp)
{
  int size_x, size_y, size_z;
  InrType type;   

  // checks file pointer and stores current position
  fpos_t pos;
  if (!fp || fgetpos(fp, &pos)!=0)
    return 0;

  // checks if it is possible to read header info
  if (!inrReadHeaderPtr(fp, &size_x, &size_y, &size_z, &type))
  {
    fsetpos(fp, &pos);
    return 0;
  }

  // could check if size of the file is enough given the header info..

  // restores position
  fsetpos(fp, &pos);
  return 1;
}



/*
 *  inrReadDataSequence
 *
 *  Reads file containing an INR image sequence of any data type.
 */
unsigned char *inrReadDataSequence(char *filename, int *size_x, int *size_y, int *size_z, InrType *type)
{
  int typeSize;
  int nData;
  unsigned char *data;

  // opens inr input file
  FILE *fpInr = fopen(filename,"rb");
  if (!fpInr)
    return NULL;

  // reads header info
  if (!inrReadHeaderPtr(fpInr, size_x, size_y, size_z, type))
  {
    fclose(fpInr);
    return NULL;
  }

  // obtains data type size in bytes
  typeSize = InrTypeSizes[*type]/8;

  // allocates memory for the data
  nData = (*size_x)*(*size_y)*(*size_z);
  data = (unsigned char *)malloc(nData*typeSize);
  if (!data)
  {
    // error: could not allocate memory
    fclose(fpInr);
    return 0;
  }

  // reads the data
  if ((int)fread(data, typeSize, nData, fpInr) != nData)
  {
    // read error: frees memory
    free(data);
    fclose(fpInr);
    return 0;
  }

  // returns the data read    
  fclose(fpInr);
  return data;
}



/*
 *  inrReadByteSequence
 *
 *  Reads file containing an INR image sequence.
 *  Each image in the sequence should be composed of 8bit pixels (bytes)
 */
unsigned char *inrReadByteSequence(char *filename, int *size_x, int *size_y, int *size_z)
{
  InrType type;   
  int nData;
  unsigned char *data;

  // opens inr input file
  FILE *fpInr = fopen(filename,"rb");
  if (!fpInr)
  {
    fclose(fpInr);
    return NULL;
  }

  // reads header info
  if (!inrReadHeaderPtr(fpInr, size_x, size_y, size_z, &type))
  {
    fclose(fpInr);
    return NULL;
  }

  // checks if file type is correct (i.e., InrByte)
  if (type != InrByte)
  {
    fclose(fpInr);
    return NULL;
  }

  // allocates memory for the data
  nData = (*size_x)*(*size_y)*(*size_z);
  data = (unsigned char *)malloc(nData*sizeof(unsigned char));
  if (!data)
  {
    // error: could not allocate memory
    fclose(fpInr);
    return 0;
  }

  // reads the data
  if ((int)fread(data, sizeof(unsigned char), nData, fpInr) != nData)
  {
    // read error: frees memory
    free(data);
    fclose(fpInr);
    return 0;
  }

  // returns the data read    
  fclose(fpInr);
  return data;
}



/*
 *  inrReadFloatSequence
 *
 *  Reads file containing an INR image sequence.
 *  Each image in the sequence should be composed of 32bit pixels (floats)
 */
float *inrReadFloatSequence(char *filename, int *size_x, int *size_y, int *size_z)
{
  InrType type;   
  int nData;
  float *data;

  // opens inr input file
  FILE *fpInr = fopen(filename,"rb");
  if (!fpInr)
    return NULL;

  // reads header info
  if (!inrReadHeaderPtr(fpInr, size_x, size_y, size_z, &type))
  {
    fclose(fpInr);
    return NULL;
  }
    
  // checks if file type is correct (i.e., InrFloat)
  if (type != InrFloat)
  {
    fclose(fpInr);
    return NULL;
  }

  // allocates memory for data
  nData = (*size_x)*(*size_y)*(*size_z);
  data = (float *)malloc(nData*sizeof(float));
  if (!data)
  {
    // error: could not allocate memory
    fclose(fpInr);
    return 0;
  }
  
  // reads the data
  if ((int)fread(data, sizeof(float), nData, fpInr) != nData)
  {
    // read error: frees memory
    free(data);
    fclose(fpInr);
    return 0;
  }

  // returns the data read    
  fclose(fpInr);
  return data;
}



/*
 *  inrWriteDataSequence
 * 
 *  Writes file containing an INR image sequence of any data type based on
 *  supplied data.
 */
int inrWriteDataSequence(char *filename, int size_x, int size_y, int size_z, unsigned char *data, InrType type)
{
  int typeSize;
  int nData;

  // opens inr output file
  FILE *fpInr = fopen(filename,"wb");
  if (!fpInr)
    return 0;

  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;

  // writes header info
  if (!inrWriteHeaderPtr(fpInr, size_x, size_y, size_z, type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // writes data to file
  nData = size_x*size_y*size_z;
  if ((int)fwrite(data, typeSize, nData, fpInr) != nData)
  {
    fclose(fpInr);
    return 0;
  }

  fclose(fpInr);
  inrWriteEnviHdr(filename);
  return 1;
}



/*
 *  inrWriteByteSequence
 * 
 *  Writes file containing an INR image sequence based on supplied data.
 *  Each image in the sequence will be composed of 8bit pixels (bytes)
 */
int inrWriteByteSequence(char *filename, int size_x, int size_y, int size_z, unsigned char *data)
{
  int nData;

  // opens inr output file
  FILE *fpInr = fopen(filename,"wb");
  if (!fpInr)
    return 0;
 
  // writes header info
  if (!inrWriteHeaderPtr(fpInr, size_x, size_y, size_z, InrByte))
  {
    fclose(fpInr);
    return 0;
  }
    
  // writes data to file
  nData = size_x*size_y*size_z;
  if ((int)fwrite(data, sizeof(unsigned char), nData, fpInr) != nData)
  {
    fclose(fpInr);
    return 0;
  }

  fclose(fpInr);
  inrWriteEnviHdr(filename);
  return 1;
}



/*
 *  inrWriteFloatSequence
 * 
 *  Writes file containing an INR image sequence based on supplied data.
 *  Each image in the sequence will be composed of 32bit pixels (floats)
 */
int inrWriteFloatSequence(char *filename, int size_x, int size_y, int size_z, float *data)
{
  int nData;

  // opens inr output file
  FILE *fpInr = fopen(filename,"wb");
  if (!fpInr)
    return 0;
 
  // writes header info
  if (!inrWriteHeaderPtr(fpInr, size_x, size_y, size_z, InrFloat))
  {
    fclose(fpInr);
    return 0;
  }
    
  // writes data to file
  nData = size_x*size_y*size_z;
  if ((int)fwrite(data, sizeof(float), nData, fpInr) != nData)
  {
    fclose(fpInr);
    return 0;
  }

  // writes ENVI .hdr file
  fclose(fpInr);
  inrWriteEnviHdr(filename);
  return 1;
}



/*
 *  inrBuildDataSequence
 * 
 *  Writes a sequence based on a list of INR image files of any data type.
 *  Files should be of the same data type and have the same x and y dimensions.
 */
int inrBuildDataSequence(char *filelist, char *outname)
{
  int     out_size_x, out_size_y, out_size_z;
  InrType out_type;
  int     okGeoref;
  int     typeSize;
  FILE   *fpOut;
  FILE   *fpData;
  unsigned char *data;
  InrGeorefInfo  georef;

  // returns failure if no files specified
  char *filename = strtok(filelist, "\n");
  if (!filename)
    return 0;
    
  // defines output sequence dimensions and data type
  if (!inrReadHeader(filename, &out_size_x, &out_size_y, &out_size_z, &out_type))
    return 0;
    
  // attempts to read georef info from input file's ENVI .hdr header
  okGeoref = inrReadGeoref(filename, &georef);

  // obtains data type size in bytes
  typeSize = InrTypeSizes[out_type]/8;
  
  // opens temporary file for writing sequence data
  fpData = tmpfile();

  // builds data sequence
  out_size_z = 0;
  for (; filename!=NULL; filename=strtok(NULL,"\n"))
  {
    int size_x, size_y, size_z;
    InrType type;
    int nData;

    // reads data of next file in list
    data = inrReadDataSequence(filename, &size_x, &size_y, &size_z, &type);
    if (!data)
      return 0;
      
    // checks if data is compatible with previously read files
    if (size_x!=out_size_x || size_y!=out_size_y || type!=out_type)
    {
      free(data);
      return 0;
    }

    // writes data to file
    nData = size_x*size_y*size_z;
    if ((int)fwrite(data, typeSize, nData, fpData) != nData)
    {
      free(data);
      fclose(fpData);
      return 0;
    }
    
    // frees data buffer
    free(data);

    // increments z dimension counter
    out_size_z += size_z;
  }
  
  // opens output sequence file
  fpOut = fopen(outname,"wb");
  if (!fpOut)
    return 0;
 
  // writes output sequence header info
  if (!inrWriteHeaderPtr(fpOut, out_size_x, out_size_y, out_size_z, out_type))
  {
    fclose(fpOut);
    return 0;
  }
  
  // copies data from temporary file to output sequence file
  fseek(fpData, 0, SEEK_SET);
  while(!feof(fpData))
  {
    char buf[BUFSIZE];
    size_t numbytes = fread(buf, 1, BUFSIZE, fpData);
    fwrite(buf, 1, numbytes, fpOut);
  }
  
  // writes ENVI .hdr header for sequence file
  if (okGeoref)
    inrWriteGeoref(outname, georef);
  else
    inrWriteEnviHdr(outname);

  // closes files and returns success
  fclose(fpOut);
  fclose(fpData);
  return 1;
}


/*
 *  inrResetData
 * 
 *  Resets all data values of a given INR image.
 *  All points are set to a specified value
 */
int inrResetData(char *filename, double value)
{
  unsigned char *frame;
  int     size_x, size_y, size_z;
  int     i, z;
  InrType type;
  int     typeSize;
  
  // reads parameters from input file  
  if (!inrReadHeader(filename, &size_x, &size_y, &size_z, &type))
    return 0;

  // allocates memory for frame data buffer 
  typeSize = InrTypeSizes[type]/8;
  frame = (unsigned char *)malloc(size_x*size_y * typeSize * sizeof(unsigned char));

  // sets data frame to specified value  
  for (i=0; i<size_x*size_y; i++)
    inrPutDataValue(frame, type, i, value);

  // writes data frames to file
  for (z=0; z<size_z; z++)
  {
    if (!inrWriteFrame(filename, z, frame))
    {
      free(frame);
      return 0;
    }
  }
  free(frame);
  return 1;
}



/*
 *  inrReadByteImage
 * 
 *  Reads file containing an INR image.
 *  The image should be composed of 8bit pixels (bytes)
 */
unsigned char *inrReadByteImage(char *filename, int *size_x, int *size_y)
{
  int size_z;
  return inrReadByteSequence(filename, size_x, size_y, &size_z);
}



/*
 *  inrReadFloatImage
 * 
 *  Reads file containing an INR image.
 *  The image should be composed of 32bit pixels (floats)
 */
float *inrReadFloatImage(char *filename, int *size_x, int *size_y)
{
  int size_z;
  return inrReadFloatSequence(filename, size_x, size_y, &size_z);
}



/*
 *  inrWriteByteImage
 * 
 *  Writes file containing an INR image.
 *  The image will be composed of 8bit pixels (bytes)
 */
int inrWriteByteImage (char *filename, int size_x, int size_y, unsigned char *data)
{
  return inrWriteByteSequence(filename, size_x, size_y, 1, data);
}



/*
 *  inrWriteFloatImage
 * 
 *  Writes file containing an INR image.
 *  The image will be composed of 32bit pixels (floats)
 */
int inrWriteFloatImage (char *filename, int size_x, int size_y, float *data)
{
  return inrWriteFloatSequence(filename, size_x, size_y, 1, data);
}



/*
 *  inrReadPoint
 *
 * 
 */
int inrReadPoint(char *filename, int x, int y, unsigned char *data)
{
  int size_x, size_y, size_z;
  int z;
  InrType        type;   
  int            typeSize;
  unsigned char *ptrData;
  long           posBegin;
  long           posPoint;

  // opens file
  FILE *fpInr = fopen(filename, "rb");
  if (!fpInr)
    return 0;

  // reads file header
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;
  
  // stores initial data position (after the header)
  posBegin = ftell(fpInr);
  if (posBegin == -1)
  {
    fclose(fpInr);
    return 0;
  }
 
  // reads point x,y for each layer z
  for (z=0; z<size_z; z++)
  {
    // positions the file pointer at the desired point x,y, at layer z
    posPoint = posBegin + typeSize*size_x*size_y * z + typeSize*size_x * y + typeSize * x;
    fseek(fpInr, posPoint, SEEK_SET);

    // positions the buffer pointer at the desired layer z
    ptrData = data + typeSize * z;
    
    // reads the data for the desired Point
    if ((int)fread(ptrData, typeSize, 1, fpInr) != 1)
    {
      fclose(fpInr);
      return 0;
    }
  }
  
  // closes files and returns success
  fclose(fpInr);
  return 1;
}



/*
 *  inrWritePoint
 *
 *  
 */
int inrWritePoint(char *filename, int x, int y, unsigned char *data)
{
  int size_x, size_y, size_z;
  int z;
  InrType        type;   
  int            typeSize;
  unsigned char *ptrData;
  long           posBegin;
  long           posPoint;

  // opens file
  FILE *fpInr = fopen(filename, "rb+");
  if (!fpInr)
    return 0;

  // reads file header
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;
  
  // stores initial data position (after the header)
  posBegin = ftell(fpInr);
  if (posBegin == -1)
  {
    fclose(fpInr);
    return 0;
  }
 
  // writes Point y for each layer z
  for (z=0; z<size_z; z++)
  {
    // positions the file pointer at the desired Point y, at layer z
    posPoint = posBegin + typeSize*size_x*size_y * z + typeSize*size_x * y + typeSize * x;
    fseek(fpInr, posPoint, SEEK_SET);

    // positions the buffer pointer at the desired layer z
    ptrData = data + typeSize * z;
    
    // writes data for the desired Point
    if ((int)fwrite(ptrData, typeSize, 1, fpInr) != 1)
    {
      fclose(fpInr);
      return 0;
    }
  }

  // closes file and returns success
  fclose(fpInr);
  return 1;
}



/*
 *  inrReadLine
 *
 * 
 */
int inrReadLine(char *filename, int y, unsigned char *data)
{
  int size_x, size_y, size_z;
  int z;
  InrType        type;   
  int            typeSize;
  unsigned char *ptrData;
  long           posBegin;
  long           posLine;

  // opens file
  FILE *fpInr = fopen(filename, "rb");
  if (!fpInr)
    return 0;

  // reads file header
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;
  
  // stores initial data position (after the header)
  posBegin = ftell(fpInr);
  if (posBegin == -1)
  {
    fclose(fpInr);
    return 0;
  }
 
  // reads line y for each layer z
  for (z=0; z<size_z; z++)
  {
    // positions the file pointer at the desired line y, at layer z
    posLine = posBegin + typeSize*size_x*size_y * z + typeSize*size_x * y;
    fseek(fpInr, posLine, SEEK_SET);

    // positions the buffer pointer at the desired layer z
    ptrData = data + typeSize*size_x * z;
    
    // reads the data for the desired line
    if ((int)fread(ptrData, typeSize, size_x, fpInr) != size_x)
    {
      fclose(fpInr);
      return 0;
    }
  }
  
  // closes files and returns success
  fclose(fpInr);
  return 1;
}



/*
 *  inrWriteLine
 *
 *  
 */
int inrWriteLine(char *filename, int y, unsigned char *data)
{
  int size_x, size_y, size_z;
  int z;
  InrType        type;   
  int            typeSize;
  unsigned char *ptrData;
  long           posBegin;
  long           posLine;

  // opens file
  FILE *fpInr = fopen(filename, "rb+");
  if (!fpInr)
    return 0;

  // reads file header
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;
  
  // stores initial data position (after the header)
  posBegin = ftell(fpInr);
  if (posBegin == -1)
  {
    fclose(fpInr);
    return 0;
  }
 
  // writes line y for each layer z
  for (z=0; z<size_z; z++)
  {
    // positions the file pointer at the desired line y, at layer z
    posLine = posBegin + typeSize*size_x*size_y * z + typeSize*size_x * y;
    fseek(fpInr, posLine, SEEK_SET);

    // positions the buffer pointer at the desired layer z
    ptrData = data + typeSize*size_x * z;
    
    // writes data for the desired line
    if ((int)fwrite(ptrData, typeSize, size_x, fpInr) != size_x)
    {
      fclose(fpInr);
      return 0;
    }
  }

  // closes file and returns success
  fclose(fpInr);
  return 1;
}



/*
 *  inrReadFrame
 *
 * 
 */
int inrReadFrame (char *filename, int z, unsigned char *data)
{
  int size_x, size_y, size_z;
  InrType type;   
  int     typeSize;
  int     nData;
  long    offsetFrame;

  // opens file
  FILE *fpInr = fopen(filename, "rb");
  if (!fpInr)
    return 0;

  // reads file header
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;
  
  // defines total number of data items in one frame
  nData = size_x*size_y;
  
  // positions the file pointer at the desired layer z
  offsetFrame = typeSize*size_x*size_y * z;
  if (offsetFrame > 0)
    fseek(fpInr, offsetFrame, SEEK_CUR);
  
  // reads the data
  if ((int)fread(data, typeSize, nData, fpInr) != nData)
  {
    fclose(fpInr);
    return 0;
  }

  // closes file and returns success
  fclose(fpInr);
  return 1;
}



/*
 *  inrWriteFrame
 *
 *  
 */
int inrWriteFrame(char *filename, int z, unsigned char *data)
{
  int size_x, size_y, size_z;
  InrType type;   
  int     typeSize;
  int     nData;
  long    offsetFrame;

  // opens file
  FILE *fpInr = fopen(filename, "rb+");
  if (!fpInr)
    return 0;

  // reads file header
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr);
    return 0;
  }
    
  // obtains data type size in bytes
  typeSize = InrTypeSizes[type]/8;
  
  // defines total number of data items in one frame
  nData = size_x*size_y;
  
  // positions the file pointer at the desired layer z
  offsetFrame = typeSize*size_x*size_y * z;
  if (offsetFrame > 0)
    fseek(fpInr, offsetFrame, SEEK_CUR);
  
  // reads the data
  if ((int)fwrite(data, typeSize, nData, fpInr) != nData)
  {
    fclose(fpInr);
    return 0;
  }

  // closes file and returns success
  fclose(fpInr);
  return 1;
}



/*
 *  inrBuildEnviHdr
 * 
 *  Given an INR image, reads its header info and creates a corresponding string
 *  that describes the .hdr ENVI header file
 */
char *inrBuildEnviHdr(char *filename, char *strHdr)
{
  int size_x, size_y, size_z;
  InrType type;   
  char *pos;

  // opens INR file
  FILE *fpInr = fopen(filename, "rb");
  if (!fpInr)
    return NULL;

  // reads header info from INR file 
  if (!inrReadHeaderPtr(fpInr, &size_x, &size_y, &size_z, &type))
  {
    fclose(fpInr); 
    return NULL;
  }

  // builds .hdr data    
  pos  = strHdr;
  pos += sprintf(pos, "ENVI\n");  
  pos += sprintf(pos, "description = {\n");  
  pos += sprintf(pos, "  INRIMAGE file header generated by INRLIB}\n");  
  pos += sprintf(pos, "samples = %d\n", size_x);  
  pos += sprintf(pos, "lines   = %d\n", size_y);  
  pos += sprintf(pos, "bands   = %d\n", size_z);  
  pos += sprintf(pos, "header offset = %d\n", HEADERSIZE);  
  pos += sprintf(pos, "file type = ENVI Standard\n");  
  pos += sprintf(pos, "data type = %d\n", Inr2EnviTypes[type]);  
  pos += sprintf(pos, "interleave = bsq\n");  
  pos += sprintf(pos, "sensor type = unknown\n");  
  pos += sprintf(pos, "byte order = 0\n");  
  
  fclose(fpInr);
  return strHdr;
}


/*
 *  inrWriteEnviHdr
 * 
 *  Writes an ENVI header file (.hdr) for an INR image.
 *  'filename' should refer to the INR file, not the .hdr file to be written
 */
int inrWriteEnviHdr(char *filename)
{
  FILE *fpHdr;
  char strHdr[INRHDRBUFSIZE];

  // defines .hdr file name
  char hdrname[FNAMESIZE];
  ftChangeExtension(filename, hdrname, ".hdr");

  // opens .hdr file for writing  
  fpHdr = fopen(hdrname, "wt");
  if (!fpHdr)
    return 0;

  // builds .hdr data
  if (!inrBuildEnviHdr(filename, strHdr))
  {
    fclose(fpHdr);
    return 0;
  }

  // writes .hdr data to file
  if (fputs(strHdr, fpHdr) == EOF)
  {
    fclose(fpHdr);
    return 0;
  }

  fclose(fpHdr);
  return 1;
}



/*
 *  inrReadGeoref
 * 
 *  Reads an ENVI header file (.hdr) containing map info for an INR image
 *  'filename' should refer to the INR file, not the .hdr file to be read
 *
 *  Georeference info is given by:
 *  . coordinates for the top-left corner of the image
 *  . pixel size in the x and y directions.
 *  . datum and projection
 *
 *  Notes:
 *  - The corner coordinates given refer to the top-left corner of the top-left
 *    pixel, not the center of that pixel.
 */
int inrReadGeoref (char *filename, InrGeorefInfo *georef)
{
  FILE  *fpHdr;
  char   strHdr[INRHDRBUFSIZE];
  int    infoPresent;
  char  *ptrHdr;
  size_t nbytes;
  int nfield=0;
  
  // defines .hdr file name
  char hdrname[FNAMESIZE];
  ftChangeExtension(filename, hdrname, ".hdr");
  
  // opens .hdr file for reading
  fpHdr = fopen(hdrname, "rt");
  if (!fpHdr)
    {
      // attempts alternate header name "*.inr.hdr"
      strcpy(hdrname, filename);
      strcat(hdrname, ".hdr");
      fpHdr = fopen(hdrname, "rt");
      if (!fpHdr)
	return 0;
    }
  
  // reads file data into a buffer
  nbytes = fread(strHdr, 1, INRHDRBUFSIZE-1, fpHdr);
  strHdr[nbytes] = '\0';
  fclose(fpHdr);
  
  // reads info
  infoPresent = 0;
  if ((ptrHdr = strstr(strHdr,"map info"))!=0)
    {
      
      // first check projection
      infoPresent += sscanf(ptrHdr, "map info = {%[^,]",georef->projection);
      if (!infoPresent) return 0;
      infoPresent=0;
      
      
      // if utm, must also read zone and hemisphere
      if (!strcmp(georef->projection,"UTM")) {
	

	georef->isUTM=1;
	nfield=11;
      
	infoPresent += sscanf(ptrHdr, "map info = {%[^,], %lf, %lf, %lf, %lf, %lf, %lf, %d, %[^,], %[^,], units=%[^}]",
			      georef->projection, 
			      &(georef->pixel_x),&(georef->pixel_y),
			      &(georef->pixel_mapx), &(georef->pixel_mapy),
			      &(georef->pixelsize_x), &(georef->pixelsize_y), 
			      &(georef->UTMzone),georef->UTMhemi,
			      georef->datum,georef->unit);
      } // utm
      else { // other projections
	
	georef->isUTM=0;
	nfield=9;
	infoPresent += sscanf(ptrHdr, "map info = {%[^,], %lf, %lf, %lf, %lf, %lf, %lf, %[^,], units=%[^}]",
			      georef->projection, 
			      &(georef->pixel_x),&(georef->pixel_y),
			      &(georef->pixel_mapx), &(georef->pixel_mapy),
			      &(georef->pixelsize_x), &(georef->pixelsize_y), 
			      georef->datum,georef->unit);
	
      }
      // computes map coordinates of 1,1 pixel
      georef->topleft_x = georef->pixel_mapx - (georef->pixel_x - 1.0)*georef->pixelsize_x ;
      georef->topleft_y = georef->pixel_mapy - (georef->pixel_y - 1.0)*georef->pixelsize_y ;
    }
  // asserts that all required info is present and valid
  if (infoPresent != nfield || georef->pixelsize_x <= 0 || georef->pixelsize_y <= 0)
    return 0;
  
  return 1;
}



/*
 *  inrWriteGeoref
 * 
 *  Writes an ENVI header file (.hdr) containing map info for an INR image
 *  'filename' should refer to the INR file, not the .hdr file to be written
 *
 *  Georeference info is given by:
 *  . coordinates for the top-left corner of the image
 *  . pixel size in the x and y directions.
 *  . datum and projection
 *
 *  Notes:
 *  - The corner coordinates given should refer to the top-left corner
 *    of the top-left pixel, not the center of that pixel.
 *  - Datum and projection can remain unspecified (empty string)
 *    In this case, "WGS-84" and "Geographic Lat/Lon" are respectively assumed
 */
int inrWriteGeoref(char *filename, InrGeorefInfo georef)
{
  FILE *fpHdr;
  char  strHdr[INRHDRBUFSIZE];
  //char *datum_final;
  //char *proj_final;

  // defines .hdr file name
  char hdrname[FNAMESIZE];
  ftChangeExtension(filename, hdrname, ".hdr");

  // opens .hdr file for writing  
  fpHdr = fopen(hdrname, "wt");
  if (!fpHdr)
    return 0;

  // builds basic .hdr data
  if (!inrBuildEnviHdr(filename, strHdr))
  {
    fclose(fpHdr);
    return 0;
  }
  
  // builds datum and projection info
  //datum_final = georef.datum;
  //proj_final  = georef.projection;
  //if (!georef.datum[0])
    // datum_final = "WGS-84";
  //if (!georef.projection[0])
    //   proj_final = "Geographic Lat/Lon";

  // builds map info data and appends it to the basic info
  {
    char mapinfo[INRHDRBUFSIZE];
    if ( georef.isUTM )
      sprintf(mapinfo, "map info = {%s, %lf, %lf, %lf, %lf, %lf, %lf, %d, %s, %s, units=%s}\n",
	      georef.projection,
	      georef.pixel_x,georef.pixel_y,
	      georef.pixel_mapx,georef.pixel_mapy,
	      georef.pixelsize_x,georef.pixelsize_y,
	      georef.UTMzone,georef.UTMhemi,
	      georef.datum,georef.unit);
    else
      sprintf(mapinfo, "map info = {%s, %lf, %lf, %lf, %lf, %lf, %lf, %s, units=%s}\n",
	      georef.projection, 
	      georef.pixel_x,georef.pixel_y,
	      georef.pixel_mapx, georef.pixel_mapy, 
	      georef.pixelsize_x, georef.pixelsize_y, 
	      georef.datum,georef.unit);
    strcat(strHdr, mapinfo);
  }
  
  // writes .hdr data to file
  if (fputs(strHdr, fpHdr) == EOF)
  {
    fclose(fpHdr);
    return 0;
  }

  fclose(fpHdr);
  return 1;
}

/*
 *
 *inrVerifGeoref
 *Chek if 2 georef are compatible(i.e same projection (if UTM same zone and hemi), same Datum and same Units)
*/
    
int inrVerifGeorefInfo(InrGeorefInfo georef1,  InrGeorefInfo georef2)
{
  int verif=0;
 
  if(!strcmp(georef1.projection, georef2.projection))
    {
    if(georef1.isUTM)
      {
      if(georef1.UTMzone== georef2.UTMzone && !strcmp(georef1.UTMhemi, georef2.UTMhemi))
	verif++;}
      
    else
        verif++;
    }

  if(!strcmp(georef1.datum, georef2.datum) && !strcmp(georef1.unit, georef2.unit))
    verif++;

  if(verif==2)
    return 1;
  else 
     return 0;
}
/*

 * inrReadBandName
 *
 * read band name information from an ENVI header
 * 'filename' should refer to the INR file, not the .hdr file to be written
 *
 */
InrBandNameInfo * inrReadBandName(char *filename  )
{  
  FILE  *fpHdr;
  char *ptrHdr ;
  size_t nbytes ;
  char strHdr[INRHDRBUFSIZE] ;
  int i,n,nbands ;
  int nread ;
  InrBandNameInfo *bn ;

  // defines .hdr file name
  char hdrname[FNAMESIZE];
  ftChangeExtension(filename, hdrname, ".hdr");
  // opens .hdr file for reading
  fpHdr = fopen(hdrname, "rt");
  if (!fpHdr)
  {
    // attempts alternate header name "*.inr.hdr"
    strcpy(hdrname, filename);
    strcat(hdrname, ".hdr");
    fpHdr = fopen(hdrname, "rt");
    if (!fpHdr)
      return NULL;
  }

  // reads file data into a buffer
  nbytes = fread(strHdr, 1, INRHDRBUFSIZE-1, fpHdr);
  strHdr[nbytes] = '\0';
  fclose(fpHdr);

  // reads band name info
  // 1st find band name
  if ( (ptrHdr = strstr(strHdr,"bands   =")) != 0 ){

    nread=sscanf(ptrHdr,"bands   = %d",&nbands);
    //assert info present and valid
    if ( nread != 1 || nbands<=0 ) return NULL ;
  } 
  else return NULL ; // nbands not present: should not happen



  // alloc bn struct
  bn = (InrBandNameInfo*)malloc(sizeof(InrBandNameInfo));
  bn->nBands = nbands;
  bn->bandName=(char**)malloc(nbands*sizeof(char*));
  for (n=0;n<nbands;n++)
    bn->bandName[n]=(char*)malloc(50);

  // find band names
  if ( (ptrHdr=strstr(strHdr,"band names")) != 0) {
    // go to first {
    for (;*ptrHdr!='{';ptrHdr++) ;
    ptrHdr++;
    // if next char is \n skip it
    if ( *ptrHdr=='\n') ptrHdr++ ;
    n=0;
    do {
      // scan until , or }
      for (i=0 ; *ptrHdr!=','&&*ptrHdr!='}';i++,ptrHdr++) {
	bn->bandName[n][i]= *ptrHdr ;
      }

      // add final 0
      bn->bandName[n][i]=0;
      // if next char is \n skip it
      ptrHdr++;
      if ( *ptrHdr=='\n') ptrHdr++ ;
      n++;
    } while(n<nbands) ;

    

    return bn;
  }
  else {
    for (n=0;n<nbands;n++) free( bn->bandName[n]);
    free(bn->bandName);
    free(bn);    return NULL ;
  }
   
}

/*
 * inrAppendBandName
 *
 * append band name to an ENVI header, assumed existing
 * 'filename' should refer to the INR file, not the .hdr file to be written
 * brute force: it is the responsibility of user to test that the band number
 * coherent with information present in header
 *
 * Addition by JPaul, april 2006
 */
int inrAppendBandName( char *filename, InrBandNameInfo *bn )
{
  int  i;
  FILE *fpHdr;
  char hdrname[FNAMESIZE];

  // first check that the envi header exists
  // dont even check if it is non empty!
  // just open for reading and close
  // if ok reopen in append mode
  ftChangeExtension(filename, hdrname, ".hdr");
  fpHdr = fopen( hdrname, "r") ;
  if ( !fpHdr ) {
    // attempts alternate header name "*.inr.hdr"
    strcpy(hdrname, filename);
    strcat(hdrname, ".hdr");
    fpHdr = fopen(hdrname, "rt");
    if (!fpHdr){
      fprintf(stderr,"error in inrAppendBandName: cannot find envi header %s\n",hdrname);
      return 0 ;
    }
  }
  fclose(fpHdr);
  fpHdr = fopen( hdrname, "a") ;
  if ( !fpHdr ) {
    fprintf(stderr,"error in inrAppendBandName: no write permission on envi header %s\n",hdrname);
    return 0 ;
  }

  // just write band names
  fprintf(fpHdr,"band names = {\n");
  for (i=0 ; i< bn->nBands-1 ; i++ ) 
    fprintf( fpHdr , "%s,\n",bn->bandName[i]);
  fprintf(fpHdr,"%s}\n",bn->bandName[i]);

  //that's all
  fclose(fpHdr);
  return 1;

}
 
void  printBandName( InrBandNameInfo * bn ) 
{
  int  n ;


  printf("Nb Bands = %d\n" , bn->nBands ) ;
  for (n=0;n<bn->nBands;n++)
   printf("Band %d: %s\n",n,bn->bandName[n]);
}
