#include <stdio.h>
#include <stdlib.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024
#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)



static char Ucmd[] = "-ifile input_mask -ofile inverted_mask [-roi roi_mask]\n";

static char Udetail[] =
"Given an input mask image, outputs a new mask image inverting which pixels\n\
are masked and which ones are not.\n\
All files are interpreted as having value zero for masked pixels and non-zero\n\
totherwise.\n\
\nInput:\n\
\t-ifile\tFile name of the input mask image\n\
\nOutput:\n\
\t-ofile\tFile name of the output inverted mask image\n\
\nOptional:\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tPixels outside of this region will be masked out in the output\n\
\t\timage\n";





int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  int  okROI;
  InrGeorefInfo georef;
  InrType typeInput;
  InrType type;
  int     typesizeInput;

  unsigned char *dataInput;
  unsigned char *dataOutput;
  unsigned char *ROIMask = NULL;

  int size_x, size_y, size_z;
  int i, j, k;
  
  double value;
  

  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okROI = igetopt1 ("-roi", "%s", fnameROIMask);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (size_z != 1)
  {
    fprintf(stderr, "Error: input mask image '%s' should have only one layer.\n", fnameInput);
    return 0;
  }
  if (okROI)
  {
    if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
    {
      fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
      return 0;
    }
    if (type != InrByte || k != 1)
    {
      fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
      return 0;
    }
    if (i != size_x || j != size_y)
    {
      fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
      return 0;
    }
  }
 
  // reads the input image
  dataInput = inrReadDataSequence(fnameInput, &size_x, &size_y, &size_z, &typeInput);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not read input file '%s'.\n", fnameInput);
    return 0;
  }
  typesizeInput = InrTypeSizes[typeInput]/8;
  
  // reads the ROI mask image
  if (okROI)
  {
    ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
    if (!ROIMask)
    {
      fprintf (stderr, "Error: could not read ROI mask data.\n");
      return 0;
    }
  }

  // allocates memory for output data
  dataOutput = (unsigned char *)  malloc (size_x*size_y * sizeof(unsigned char));


  // processes the data
  for (i=0; i<size_x*size_y; i++)
  {
    // checks if point is within region of interest
    if (okROI && !ROIMask[i])
    {
      dataOutput[i] = 0;
      continue;
    }

    // gets data value    
    value = inrGetDataValue(dataInput, typeInput, i);

    // inverts value
    dataOutput[i] = (value==0);
  }

  
  // writes output translated classification
  if (!inrWriteByteImage(fnameOutput, size_x, size_y, dataOutput))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  
  // copies georreference info if present
  if (inrReadGeoref(fnameInput, &georef))
    inrWriteGeoref(fnameOutput, georef);
  
  fprintf(stderr, "Mask image inverted.\n");

  free(dataInput);
  free(dataOutput);
  free(ROIMask);
  return 1;
}
