/***************************************************************************
                          inr_memberships.c  -  description
                             -------------------
    Created              : September, 2007
    By                   : (C) 2007 by Carlos de la Torre
    Email                : Carlos.de_la_Torre@inria.fr
 ***************************************************************************/
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "inr_memberships.h"
#include "classdef.h"
#include "inr_io.h"
#include "file_tools.h"
#include "linearalgebra.h"
#include "dcdflib/src/cdflib.h"
#include "modis_classes_names.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

/*
 *
 * Computes memberships based on a Mahalanobis distance.
 * It uses a Chi-square cumulative distribution function.
 * 
 * An INRIMAGE is created with the memberships to every class, for every pixel.
 */
int mahalachi(char* fnameInput, unsigned char *ROIMask, classdef* def, char* fnameOutputMemb, InrType typeInput, int size_x, int size_y, int size_z)
{
  unsigned char *dataInput;

  int cnFeatures;
  int iClass, iFeat;
  
  int     typesizeInput;
  InrGeorefInfo georef;

  int i;
  int x, y, z;
  
  matrice *mean;
  matrice *cov;
  
  matrice *point;
  matrice *point_transp;
  float  *values=NULL;
  float  *dataMemb=NULL;
  
  // Variables for the chi-square computation
  double mahalanobis=0.0;
  int w=1;
  double memb=0.0;
  int status=0;
  double p=0.0;
  double dof=0.0;
  double bound=0.0;
  float val;
  
  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  values           = (float *) malloc (size_z * sizeof(float));
  dataMemb         = (float *) malloc (size_x * def->nClasses * sizeof(float));

  // initializes data line
  for (i=0; i<size_x*def->nClasses; i++)
    dataMemb[i] = 0.0;
  
  // inverts the covariance matrices, for the Mahalanobis distance
  if (!defInvertCov(def))
    return 0;

  // PROCESSES THE DATA: line by line
  for (y=0; y<size_y; y++)
  {
    if(y%10==0) fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);

    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    // writes empty output membership data line
    if (x == size_x)
    {
      if (!inrWriteLine(fnameOutputMemb, y, (unsigned char *)dataMemb))
      {
        fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputMemb);
        return 0;
      }

      continue;
    }

    // reads input data line y
    if (!inrReadLine(fnameInput, y, dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }
    
    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      matrice *mult1, *mult2;
      // pixels outside of region of interest are rated unclassified
      if (!ROIMask[x + y*size_x]) continue;
        
      // gets point values for the position x of the input line
      // Get the values of all features...
      for(iFeat=0; iFeat<def->nFeatures; iFeat++)
      {
        values[iFeat] = inrGetDataValue(dataInput, typeInput, x + size_x*iFeat);
      }

      // evaluates membership for all classes
      for (iClass=1; iClass<=def->nClasses; iClass++)
      {
        cnFeatures = defGetClassNfeatures(def,iClass);
        mean = def->mean[iClass-1];       // 1 x cnFeatures
        cov  = defGetCovariance(def, iClass); // cnFeatures x cnFeatures
 
        // From the values of the features in that point,
        // keep only the features interesting for the class
        point = Matrice_Nulle(cnFeatures, 1);
        defGetClassPoint(def, iClass, values, point); // cnFeatures x 1
        point_transp = Matrice_Nulle(1, cnFeatures);
        // defines vectors: (values-means), t(values-means)
        for (z=0; z<cnFeatures; z++)
        {
          point->val[z] -= mean->val[z]; // Manual substraction...
          point_transp->val[z] = point->val[z];
        }

        // computes mahanalobis distance:
        //   mahal = t(x-m) * COV^-1 * (x-m)
        // adds point per-class memberships
        mult1 = Multiplier_Matrice(point_transp, cov); // 1 x ClassNfeatures
        mult2 = Multiplier_Matrice(mult1, point);      // 1 x 1

        // The membership of pixel x to class c, m(c,x), is defined as:
        // m(c,x) = 1 - Fn(Z(x))
        // Where:
        // Z(c,x)  is the squared Mahalanobis distance of pixel x to class c
        // Fn    is the cumulative chi-square distribution function 
        //       with n-degrees of freedom.
        // n     is the number of features used for class c
        mahalanobis = mult2->val[0];
        w=1;
        memb=0.0;
        status=0;
        p=0.0;
        dof = (double) cnFeatures;
//printf("C%i: mahalanobis=%lf ",iClass,mahalanobis);
        // void cdfchi(int *which, double *p, double *q, double *x, double *df, int *status, double *bound)
        cdfchi(&w, &p, &memb, &mahalanobis, &dof, &status, &bound);
        
        // Ugly way to fit a double into a float...
        if(memb < 0.0000001)
        {
          val = 0.0;
        }else
        {
          val = (float) memb;
        }
        dataMemb[x+(iClass-1)*size_x] = val;
// if(status!=0){
// printf("bound=%.20lf in pos %i,%i\n",bound,x,y);
// bound = 0.0;
// }
        free_matrice(&mult1);
        free_matrice(&mult2);
        free_matrice(&point);
        free_matrice(&point_transp);
      }
    }
    // writes the line of class memberships
    if (!inrWriteLine(fnameOutputMemb, y, (unsigned char *)dataMemb))
    {
      fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputMemb);
      return 0;
    }
    // re-initializes output data line
    for (i=0; i<size_x*def->nClasses; i++)
      dataMemb[i] = 0.0;
  }

  // transfers georeference info if available
  if (inrReadGeoref(fnameInput, &georef))
  {
    inrWriteGeoref(fnameOutputMemb, georef);
  }

  if (dataInput) free(dataInput);
  if (values)    free(values);
  if (dataMemb)  free(dataMemb);

  return 1;
}



/*
 * Computes memberships based on a weighted average approach.
 * For every class/feature pair (i,j), we have a 'weight' that measures how much that feature
 * discriminates that class ('weights'): w(i,j)
 * 
 * For every class, we have a membership value for every feature ('fcMemberships'): m(i,j)
 *
 * The membership to a class 'i' is computed as : 1/N * SUM( w(i,j)*m(i,j) )
 * where j=1..N
 * 
 * An INRIMAGE is created with the memberships to every class, for every pixel.
 *
 */
int weighted_averaging(char* fnameInput, unsigned char *ROIMask, classdef* def, char* fnameOutputMemb, InrType typeInput, int size_x, int size_y, int size_z)
{
  unsigned char *dataInput;

  int iClass, iFeat, nFeats_c;
  
  int     typesizeInput;
  InrGeorefInfo georef;

  int i;
  int x, y;
  /* Here we need a classdef variable that will read the class/feature mean and sd from a file*/
  float  *values=NULL;
  float  *memberships=NULL;
  float   memb, mass_value, weight_cf;


  // writes output file header (output type is always float)
  if (!inrWriteHeader(fnameOutputMemb, size_x, size_y, def->nClasses, InrFloat))
  {
    fprintf(stderr, "Error: could not write header information for memberships file '%s'.\n", fnameOutputMemb);
    return 0;
  }
  
  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  values           = (float *) malloc (size_z * sizeof(float));
  memberships      = (float *) malloc (size_x * def->nClasses * sizeof(float));

  // PROCESSES THE DATA: line by line
  for (y=0; y<size_y; y++)
  {
    if(y%10==0) fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);
    // initializes data line
    for (i=0; i<size_x*def->nClasses; i++)
      memberships[i] = 0.0;
    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    // writes empty output membership data line
    if (x == size_x)
    {
      if (!inrWriteLine(fnameOutputMemb, y, (unsigned char *)memberships))
      {
        fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputMemb);
        return 0;
      }
      continue;
    }

    // reads input data line y
    if (!inrReadLine(fnameInput, y, dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }
    
    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      // pixels outside of region of interest are rated unclassified
      if (!ROIMask[x + y*size_x]) continue;

      // gets point values for the position x of the input line
      // Get the values of all features...
      for(iFeat=0; iFeat<def->nFeatures; iFeat++)
      {
        values[iFeat] = inrGetDataValue(dataInput, typeInput, x + size_x*iFeat);
      }

      // evaluates membership for all classes
      for (iClass=0; iClass<def->nClasses; iClass++)
      {
        memb     = 0.0;
        nFeats_c = 0;

        for(iFeat=0;iFeat<def->nFeatures;iFeat++)
        {
          weight_cf = def->weight[iClass][iFeat];
          if(weight_cf > 0)
          {
            nFeats_c++;
            mass_value = mass(values, def, iClass+1, iFeat+1);
            memb += mass_value * weight_cf;
          }
        }
        memberships[x+(iClass)*size_x] = (nFeats_c>0 ? memb / (float)nFeats_c : 0.0);
      }
    }
    // writes the line of class memberships
    if (!inrWriteLine(fnameOutputMemb, y, (unsigned char *)memberships))
    {
      fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputMemb);
      return 0;
    }
  }

  // transfers georeference info if available
  if (inrReadGeoref(fnameInput, &georef))
  {
    inrWriteGeoref(fnameOutputMemb, georef);
  }

  if (dataInput) free(dataInput);
  if (values)    free(values);
  if (memberships)  free(memberships);

  return 1;
}


/*
 * Computes memberships based on Dempster rule of combination.
 */
int dempster_rule(char* fnameInput, unsigned char *ROIMask, classdef* def, char* fnameOutputMemb, InrType typeInput, int size_x, int size_y, int size_z)
{
  unsigned char *dataInput;

  int     typesizeInput;
  InrGeorefInfo georef;

  int iFeat, iClass;
  int x, y;
  /* Here we need a classdef variable that will read the class/feature mean and sd from a file*/
  float  *values=NULL;
  float  *point=NULL;
  float  *memberships=NULL;


  // writes output file header (output type is always float)
  if (!inrWriteHeader(fnameOutputMemb, size_x, size_y, def->nClasses, InrFloat))
  {
    fprintf(stderr, "Error: could not write header information for memberships file '%s'.\n", fnameOutputMemb);
    return 0;
  }
  
  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  values           = (float *) malloc (size_z * sizeof(float));
  point            = (float *) malloc (def->nClasses * sizeof(float));
  memberships      = (float *) malloc (size_x * def->nClasses * sizeof(float));

  // PROCESSES THE DATA: line by line
  for (y=0; y<size_y; y++)
  {
    if(y%10==0) fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);

    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    // writes empty output membership data line
    if (x == size_x)
    {
      if (!inrWriteLine(fnameOutputMemb, y, (unsigned char *)memberships))
      {
        fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputMemb);
        return 0;
      }
      continue;
    }

    // reads input data line y
    if (!inrReadLine(fnameInput, y, dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }
    
    //clean the memberships line
    for (iClass=0; iClass<size_x*def->nClasses; iClass++)
      memberships[iClass] = 0.0;

    // processes each point of the line at a time
    for (x=0; x<size_x; x++)
    {
      // pixels outside of region of interest are rated unclassified
      if (!ROIMask[x + y*size_x]) continue;

      // gets point values for the position x of the input line
      // Get the values of all features...
      for(iFeat=0; iFeat<def->nFeatures; iFeat++)
      {
        values[iFeat] = inrGetDataValue(dataInput, typeInput, x + size_x*iFeat);
      }

      // Compute membership to ALL classes for point in x using
      // Dempster rule of combination, where the sources of information
      // are the features
      if(!DR(def, values, point))
      {
        fprintf(stderr, "Error: could not compute Dempster rule for pixel (%d,%d).\n", x,y);
        continue;
      }

// // // //DEBUG
// // // fprintf(stderr, " point");      
// // // for (iClass=0; iClass<def->nClasses; iClass++){
// // //   fprintf(stderr, ", %f",point[iClass]);
// // // }
// // // fprintf(stderr, "\n");
// // // //DEBUG

      // Put the point in the memberships line
      for (iClass=0; iClass<def->nClasses; iClass++)
      {
        memberships[x+(iClass)*size_x] = point[iClass];
      }
    }
    // writes the memberships line when it's complete
//fprintf(stderr, "\t\tWriting line (%d).\n",y);
    if (!inrWriteLine(fnameOutputMemb, y, (unsigned char *)memberships))
    {
      fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputMemb);
      return 0;
    }
  }

  // transfers georeference info if available
  if (inrReadGeoref(fnameInput, &georef))
  {
    inrWriteGeoref(fnameOutputMemb, georef);
  }

  if (dataInput)   free(dataInput);
  if (values)      free(values);
  if (point)       free(point);
  if (memberships) free(memberships);

  return 1;
}


/*
 * Computes the mass of the feature 'iFeat' and class 'iClass'.
 * Asumption: Features have a Gaussian distribution in every class.
 * returned value = N(mu,sigma)(x) / N(mu,sigma)(mu)
 *  Where
 *    'N(mu,sigma)(x)' is the Normal probability distribution function 
 *    with mean 'mu' and standard deviation 'sigma' in point 'x'.
 *
 *    'mu' is the mean of feature 'iFeat' in class 'iClass'.
 *
 *    'sigma' is the standard deviation of feature 'iFeat' in class 'iClass'.
 */
float mass(float *values, classdef* def, int iClass, int iFeat)
{
  float mu, sigma, x, N_x, N_mu;
  float ret=0.0, modes;
  int is_wood          = (iClass == WOOD1          || iClass == WOOD2 || iClass == CILIAR);
  int is_deforestation = (iClass == DEFORESTATION1 || iClass == DEFORESTATION2);
//  if(is_deforestation) fprintf(stderr,"class %i feat %i\n",iClass, iFeat);
  if(iFeat==5) // Number of modes
  {
    modes = values[iFeat-1];
    switch(iClass){
      case(AGRICULTURE1): //A1
        if(modes==1) 
          ret = 1;
        else 
          ret = 0;
        break; 
      case(AGRICULTURE2): //A2
        if(modes==1) 
          ret = 1;
        else 
          ret = 0;
        break;
      case(AGRICULTURE3): //A3
        if(modes==2) 
          ret = 1;
        else 
          ret = 0;
        break;
      case(WOOD1): // Wood1
        if(modes==0) 
          ret = 1;
        else 
          ret = 0;
        break;
      case(WOOD2): // Wood2
        if(modes==0) 
          ret = 1;
        else 
          ret = 0;
        break;
      case(DEFORESTATION1): // DEforestation 1
        if(modes==0) 
          ret = 1;
        else 
          ret = 0;
        break;
      case(DEFORESTATION2): // DEforestation 2
        if(modes==0) 
          ret = 1.0;
        else 
          ret = 0.0;
        break;
      case(CILIAR): // Ciliar Vegetation
        if(modes==0) 
          ret = 1.0;
        else 
          ret = 0;
        break;
      case(PASTURE1): // Pasture1
        if(modes==0) 
          ret = 1.0;
        else 
          ret = 0;
        break;
      case(PASTURE2): // Pasture
        if(modes==0) 
          ret = 1.0;
        else 
          ret = 0;
        break;
      case(URBAN): // Urban
        if(modes==0) 
          ret = 1.0;
        else 
          ret = 0;
        break;
      default:
        break;
    }
  }
  else if(iClass==URBAN) // Urban
  {
    if(iFeat==7) // Amplitude
    {
      ret = MIN(sigmoid(-50.0, 0.4, values[iFeat-1]),sigmoid(50.0, 0.18, values[iFeat-1]));
    }
    if(iFeat==9 || iFeat==10 || iFeat==11 || iFeat==12) // Mode Features: Value 0 or nothing...
    {
      ret = sigmoid(-50.0, 0.1, values[iFeat-1]);
    }
    else
    {
      mu    = defGetClassFeatMean(def, iClass, iFeat);
      sigma = defGetClassFeatStdDev(def, iClass, iFeat);
      if(sigma==0.0)
        ret = 0.0;
      else 
      {
        x     = values[iFeat-1];
        N_x   = N(mu, sigma, x);
        N_mu  = N(mu, sigma, mu);
        ret   = (N_x / N_mu);
      }
    }
  }
  else if(is_wood && iFeat==1) // Wood / mean
  {
    ret = sigmoid(20.0, 0.6, values[iFeat-1]);
  }
  else if(is_wood && iFeat==6) // Wood / Min
  {
    ret = sigmoid(20.0, 0.45, values[iFeat-1]);
  }
  else if(is_deforestation && iFeat==3) // Deforestation / delta45days
  {
    ret = sigmoid(-20.0, -0.07, values[iFeat-1]);
  }
  else if(is_deforestation && iFeat==7) // Deforestation / deltaMaxMin
  {
    ret = sigmoid(20.0, 0.33, values[iFeat-1]);
  }
// //   else if(iClass==7 && iFeat==1) // Deforestation / mean
// //   {
// //     ret = sigmoid(30.0, 0.5, values[iFeat-1]);
// //   }
  else // All the other classes / features
  {
    mu    = defGetClassFeatMean(def, iClass, iFeat);
    sigma = defGetClassFeatStdDev(def, iClass, iFeat);
    if(sigma==0.0)
      ret = 0.0;
    else 
    {
      x     = values[iFeat-1];
      N_x   = N(mu, sigma, x);
      N_mu  = N(mu, sigma, mu);
      ret = (N_x / N_mu);
    }
  }
  
  return ret;
}


/*
 * Sigmoid function
 * Refer to http://www.computing.dcu.ie/~humphrys/Notes/Neural/sigmoid.html
 * for it's properties.
 */
float sigmoid(float factor, float despl, float x)
{
  return 1.0 / ( 1.0 + expf( -factor * ( x - despl ) ) );
}


/*
 *
 * Normal probability distribution function with mean 'mu' and
 * standard deviation 'sigma' at point 'x'.
 * Used to compute the masses...
 */
float N(float mu, float sigma, float x)
{
  return 1.0/(sqrtf(2.0*M_PI) * sigma) * expf(-1.0 * powf(x-mu,2.0) / (2.0*sigma*sigma) );
}


/*
 *
 * Dempster Rule of combination: Merges the information that every feature
 * provides for every class.
 */
int DR(classdef* def, float* values, float* memb)
{
  float D=0.0;
  float** m;
  int c, f;

  // m is used to compute every claff/feature mass, only once
  m = (float**) malloc(def->nClasses*sizeof(float*));
  for(c=0;c<def->nClasses;c++)
  {
    m[c] = (float*) malloc(def->nFeatures*sizeof(float));
  }

  // Compute all the masses for the first time
  masses(def, values, m);
  for(c=0;c<def->nClasses;c++)
  {
    memb[c] = m[c][0]; //mass(values, def, c+1, 1);
  }

  //'memb' keeps the previous computation, at the begining of the loop.  
  for(f=2;f<=def->nFeatures;f++)
  {
    D = conflict(def, f, values, memb, m);
    for(c=0;c<def->nClasses;c++)
    {
      memb[c] = memb[c] * m[c][f-1] / fabsf(1.0 - D); // m[][]= mass(values, def, c+1, f)
// if(c==6) fprintf(stderr, "memb %f\n",memb[c]);
    }
    masses(def, values, m);
  }

  if (m)
  {
    for(c=0;c<def->nClasses;c++) if(m[c]) free(m[c]);
    free(m);
  }
  return 1;
}

void masses(classdef* def, float* values, float** m)
{
  int c,f;

  for(c=1;c<=def->nClasses;c++)
  {
    for(f=1;f<=def->nFeatures;f++)
    {
      m[c-1][f-1] = mass(values, def, c, f);
// fprintf(stderr,"c=%i , f=%i, mass = %.10f\n",c,f,m[c-1][f-1]);
    }
  }
}

/*
 *
 * Conflict between two sources (features)
 * For Demster Rule
 */
float conflict(classdef* def, int f, float* values, float* prev, float** ms)
{
  float confl = 0.0;
  int c1, c2;

  for(c1=0;c1<def->nClasses;c1++)
  {
    for(c2=0;c2<def->nClasses;c2++)
    {
//fprintf(stderr, "mass %f\t",mass(values, def, c2+1, f));
      if(c1 != c2) confl += prev[c1] * ms[c2][f-1];  //m[][] == mass(values, def, c2+1, f);
    }    
  }

  return confl;
}


