/***************************************************************************
                          inr_memberships.h  -  description
                             -------------------
    Created              : August, 2007
    By                   : (C) 2007 by Carlos de la Torre
    Email                : Carlos.de_la_Torre@inria.fr
 ***************************************************************************/

#ifndef __INR_MEMBERSHIPS_H__
#define __INR_MEMBERSHIPS_H__

#include "inr_io.h"
#include "classdef.h"

/*
 * Computes memberships based on a Mahalanobis distance.
 * It uses a Chi-square cumulative distribution function.
 * 
 * An INRIMAGE is created with the memberships to every class, for every pixel.
 *
 */
int mahalachi(  char* fnameInput,
                unsigned char *ROIMask, 
                classdef* def, 
                char* fnameOutputMemb, 
                InrType typeInput, 
                int size_x, 
                int size_y, 
                int size_z);

/*
 * Computes memberships based on a weighted average approach.
 * For every class/feature pair (i,j), we have a 'weight' that measures how much that feature
 * discriminates that class ('weights'): w(i,j)
 * 
 * For every class, we have a membership value for every feature ('fcMemberships'): m(i,j)
 *
 * The membership to a class 'i' is computed as : 1/N * SUM( w(i,j)*m(i,j) )
 * where j=1..N
 * 
 * An INRIMAGE is created with the memberships to every class, for every pixel.
 *
 */
int weighted_averaging( char* fnameInput,
                        unsigned char *ROIMask, 
                        classdef* def, 
                        char* fnameOutputMemb, 
                        InrType typeInput, 
                        int size_x, 
                        int size_y, 
                        int size_z);

/*
 * Computes memberships based on a weighted average approach.
 * For every class/feature pair (i,j), we have a 'weight' that measures how much that feature
 * discriminates that class ('weights'): w(i,j)
 * 
 * For every class, we have a membership value for every feature ('fcMemberships'): m(i,j)
 *
 * The membership to a class 'i' is computed with Dempster Rule of combination.
 * (http://en.wikipedia.org/wiki/Dempster-Shaffer_theory)
 * 
 * An INRIMAGE is created with the memberships to every class, for every pixel.
 *
 * TODO!!!
 */
int dempster_rule(char* fnameInput,
                        unsigned char *ROIMask, 
                        classdef* def, 
                        char* fnameOutputMemb, 
                        InrType typeInput, 
                        int size_x, 
                        int size_y, 
                        int size_z);


/*
 * Computes the mass of the feature 'iFeat' and class 'iClass'
 */
float mass(float *values, classdef* def, int iClass, int iFeat);


/*
 * Normal probability distribution function with mean 'mu' and
 * standard deviation 'sigma' at point 'x'.
 * Used to compute the masses...
 */
float N(float mu, float sigma, float x);


/*
 *
 * Dempster Rule of combination
 */
int DR(classdef* def, float* values, float* memb);


/*
 *
 * Computes all the massses for pixel in values...
 */
void masses(classdef* def, float* values, float** m);


/*
 *
 * Conflict between two sources (features)
 * For Demster Rule
 */
float conflict(classdef* def, int f, float* values, float* prev, float** ms);


/*
 * Sigmoid function
 * Refer to http://www.computing.dcu.ie/~humphrys/Notes/Neural/sigmoid.html
 * for it's properties.
 */
float sigmoid(float factor, float despl, float x);

#endif //__INR_MEMBERSHIPS_H__

