#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE       1024


static char Ucmd[] = "-ifile1 inr_file1 -ifile2 inr_file2 -ofile inr_merged \
[-nodata value]\n";

static char Udetail[] =
"Given two input images, outputs a new image merging its values.\n\
\nMerging is based on File1. If a given point has no valid data (i.e., value\n\
for File1 equals 'nodata'), then data from File2 is used.\n\
\nInput:\n\
\t-ifile1\tFile name of input INRIMAGE image file 1\n\
\t-ifile2\tFile name of input INRIMAGE image file 2\n\
\nOutput:\n\
\t-ofile\tFile name of the output merged INRIMAGE file\n\
\nOptional:\n\
\t-nodata\tValue to be considered as missing data\n\
\t\tDefault value is 0\n";



int main (int argc, char **argv)
{
  char fnameInput1[FNAMESIZE];
  char fnameInput2[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  InrType typeInput1;
  InrType typeInput2;
  InrType typeOutput;
  int     typesizeInput1;
  int     typesizeInput2;
  int     typesizeOutput;
  InrGeorefInfo georef;
  
  unsigned char *dataInput1;
  unsigned char *dataInput2;
  unsigned char *dataOutput;
  
  int    okNoData;
  double nodata;
  
  int size_x, size_y, size_z;
  int i, j, k;
  int z;
  
  double value1;
  double value2;

  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile1", "%s", fnameInput1))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ifile2", "%s", fnameInput2))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okNoData = igetopt1 ("-nodata", "%lf", &nodata);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okNoData)
    nodata = 0;
    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput1, &size_x, &size_y, &size_z, &typeInput1))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput1);
    return 0;
  }
  if (!inrReadHeader(fnameInput2, &i, &j, &k, &typeInput2))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput2);
    return 0;
  }
  if (i != size_x || j != size_y || k != size_z)
  {
    fprintf (stderr, "Error: dimensions of input files do not match.\n");
    return 0;
  }
  
  
  // output file type follows input file 1's type
  typeOutput = typeInput1;

  // allocates memory for the input and output data
  typesizeInput1   = InrTypeSizes[typeInput1]/8;
  typesizeInput2   = InrTypeSizes[typeInput2]/8;
  typesizeOutput   = InrTypeSizes[typeOutput]/8;
  dataInput1       = (unsigned char *) malloc (size_x*size_y * typesizeInput1 * sizeof(unsigned char));
  dataInput2       = (unsigned char *) malloc (size_x*size_y * typesizeInput2 * sizeof(unsigned char));
  dataOutput       = (unsigned char *) malloc (size_x*size_y * typesizeOutput * sizeof(unsigned char));


  // writes output file header
  if (!inrWriteHeader(fnameOutput, size_x, size_y, size_z, typeOutput))
  {
    fprintf(stderr, "Error: could not write header information for output file '%s'.\n", fnameOutput);
    return 0;
  }
  

  // processes the data
  for (z=0; z<size_z; z++)
  {
    fprintf(stderr, "Processing layer %d of %d...\r", z+1, size_z);
    if (!inrReadFrame(fnameInput1, z, dataInput1))
    {
      fprintf(stderr, "Error: could not read frame %d for input image '%s'.\n", z, fnameInput1);
      return 0;
    }
    if (!inrReadFrame(fnameInput2, z, dataInput2))
    {
      fprintf(stderr, "Error: could not read frame %d for input image '%s'.\n", z, fnameInput2);
      return 0;
    }

    for (i=0; i<size_x*size_y; i++)
    {
      // gets data values for input files
      value1 = inrGetDataValue(dataInput1, typeInput1, i);
      value2 = inrGetDataValue(dataInput2, typeInput2, i);

      // defines output merged data value
      if (value1 != nodata)
        inrPutDataValue(dataOutput, typeOutput, i, value1);
      else if (value2 != nodata)
        inrPutDataValue(dataOutput, typeOutput, i, value2);
      else
        inrPutDataValue(dataOutput, typeOutput, i, nodata);
    }
    
    // writes output data frame
    if (!inrWriteFrame(fnameOutput, z, (unsigned char*)dataOutput))
    {
      fprintf(stderr, "Error: could not write frame %d for output image '%s'.\n", z, fnameOutput);
      return 0;
    }
  }
  
  // transfers georeference info from file1, if available
  if (inrReadGeoref(fnameInput1, &georef))
    inrWriteGeoref(fnameOutput, georef);
  else
    inrWriteEnviHdr(fnameOutput);
  
    
  fprintf(stderr, "INRIMAGE files successfully merged.\n");

  free(dataInput1);
  free(dataInput2);
  free(dataOutput);
  
  return 0;
}
