#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "profile_features.h"

#define FNAMESIZE       1024


static char Ucmd[] = "-ifile inr_input -roi roi_mask -ofile inr_normalized \
[-nodata ignore_value]\n";

static char Udetail[] =
"Given an input image, normalizes its values so that they have mean 0 and\n\
variance 1.\n\
Data is normalized independently frame by frame (i.e., each image in the Z-\n\
dimension is processed separately)\n\
\nInput:\n\
\t-ifile\tFile name of the input INRIMAGE image file\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output normalized INRIMAGE file\n\
\nOptional:\n\
\t-inorm\tFile name containing normalization information (mean and\n\
\t\tstddev) for each frame or layer of the input image\n\
\t\tFile must be in INRIMAGE format\n\
\t-onorm\tFile name for outputting normalization information (mean and\n\
\t\tstddev) for each frame or layer of the input image\n\
\t-nodata\tValue to be considered as missing data\n\
\t\tMissing data will be given random values in the output file\n\
\t\tDefault is to consider all values within ROI as valid\n";


/*
 * GLOBAL CONTEXT
 */
typedef struct 
{
    char fnameInput[FNAMESIZE];
    char fnameROIMask[FNAMESIZE];
    char fnameOutput[FNAMESIZE];
    char fnameInputNormParams[FNAMESIZE];
    char fnameOutputNormParams[FNAMESIZE];

    InrType typeInput;
    int     typesizeInput;

    unsigned char *dataInput;
    unsigned char *ROIMask;
    float         *dataOutput;

    int    okInputNorm;
    int    okOutputNorm;
    int    okNoData;
    double nodata;

    int size_x, size_y, size_z;
} InrNorm_gc;
                                  



/*
 * outputNormParamsInit
 *
 * Initializes the output file for storing normalization parameters
 */
int outputNormParamsInit (InrNorm_gc *gc)
{
	// writes output file header (output type is always double)
    if (!inrWriteHeader(gc->fnameOutputNormParams, 2, gc->size_z, 1, InrDouble))
	{
		printf("Error: could not write header information for output file '%s'.\n", gc->fnameOutputNormParams);
		return 0;
	}

	// writes output ENVI file header (output type is always double)
    if (!inrWriteEnviHdr(gc->fnameOutputNormParams))
	{
		printf("Error: could not write ENVI header information for output file '%s'.\n", gc->fnameOutputNormParams);
		return 0;
	}

    return 1;
}


/*
 * outputNormParams
 *
 * Outputs normalization parameter information of a given frame to the appropriate file
 */
int outputNormParams (InrNorm_gc *gc, int iFrame, double mean, double stddev)
{
    // outputs normalization mean information to the output file
    if (!inrWritePoint(gc->fnameOutputNormParams, 0, iFrame, (unsigned char *) &mean))
	{
		printf("Error: could not write normalization information for frame %d on output file '%s'.\n", iFrame, gc->fnameOutputNormParams);
		return 0;
	}
    // outputs normalization stddev information to the output file
	if (!inrWritePoint(gc->fnameOutputNormParams, 1, iFrame, (unsigned char *) &stddev))
	{
		printf("Error: could not write normalization information for frame %d on output file '%s'.\n", iFrame, gc->fnameOutputNormParams);
		return 0;
	}
	return 1;
}


/*
 * readNormParams
 *
 * Reads input normalization parameter information for a given frame from the appropriate file
 */
int readNormParams (InrNorm_gc *gc, int iFrame, double *mean, double *stddev)
{
    // reads normalization mean information for the specified frame
    if (!inrReadPoint(gc->fnameInputNormParams, 0, iFrame, (unsigned char *) mean))
	{
        printf("Error: could not read normalization information for frame %d from input file '%s'.\n", iFrame, gc->fnameInputNormParams);
		return 0;
	}
    // reads normalization stddev information for the specified frame
    if (!inrReadPoint(gc->fnameInputNormParams, 1, iFrame, (unsigned char *) stddev))
	{
        printf("Error: could not read normalization information for frame %d from input file '%s'.\n", iFrame, gc->fnameInputNormParams);
		return 0;
	}
    return 1;
}


/*
 * computeNormParams
 *
 * Computes mean and standard deviation values for the current data frame of a GC
 */
int computeNormParams (InrNorm_gc *gc, double *mean, double *stddev)
{
    double sum = 0;
    double sum2 = 0;
    int n = 0;
    int i;
    double value;

    // computes mean and standard deviation for the data frame 
    for (i = 0; i < gc->size_x * gc->size_y; i++)
    {
        // checks if point within region of interest
        if (!gc->ROIMask[i])
            continue;
        
        // gets data value    
        value = inrGetDataValue(gc->dataInput, gc->typeInput, i);

        // checks if data value is valid
        if (gc->okNoData && value==gc->nodata)
            continue;
        
        // updates values
        n++;
        sum  += value;
        sum2 += value*value;
    }
    *mean   = sum/n;
    *stddev = sqrt((sum2 - sum*sum/n)/(n-1));

    return 1;
}


/*
 * normalize
 *
 * Normalizes the current data frame of a GC, given the mean and standard deviation parameters
 */
int normalize (InrNorm_gc *gc, double mean, double stddev)
{
    int i;
    double value;

    for (i = 0; i < gc->size_x * gc->size_y; i++)
    {
        // checks if point within region of interest
        if (!gc->ROIMask[i])
        {
            gc->dataOutput[i] = 0;
            continue;
        }

        // gets data value    
        value = inrGetDataValue(gc->dataInput, gc->typeInput, i);

        // checks if data value is valid
        // missing data are given random values in the output file
        if (gc->okNoData && value==gc->nodata)
        {
            gc->dataOutput[i] = (((float)rand())/RAND_MAX)*2 - 1;
            continue;
        }
        
        // normalizes data value
        if (stddev != 0)
            gc->dataOutput[i] = (float) ((value-mean)/stddev);
        else
            gc->dataOutput[i] = (float) (value-mean);
    }

    return 1;
}


/*
 * main
 *
 * Main method
 */
int main (int argc, char **argv)
{
    InrNorm_gc  gc_obj;           /* general context */
    InrNorm_gc *gc = &gc_obj;

    InrType type;
    InrGeorefInfo georef;

    time_t t;

    int i, j, k;
    int z;

    double mean, stddev;
    InrBandNameInfo * bn;
    /* lecture options */
    inr_init (argc, argv, "1.0", Ucmd, Udetail);
    if (!igetopt1 ("-ifile", "%s", gc->fnameInput))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-roi"  , "%s", gc->fnameROIMask))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-ofile", "%s", gc->fnameOutput))
        iusage_ (Ucmd, Udetail);
    gc->okInputNorm = igetopt1 ("-inorm", "%s", gc->fnameInputNormParams);
    gc->okOutputNorm = igetopt1 ("-onorm", "%s", gc->fnameOutputNormParams);
    gc->okNoData = igetopt1 ("-nodata", "%lf", &gc->nodata);

    if (tstopts ())                /* s'il reste des options : exit avec message */
        iusage_ (Ucmd, Udetail);

    if (!gc->okNoData)
        gc->nodata = 0;

    // reads header information for all input data and checks if they are appropriate
    if (!inrReadHeader(gc->fnameInput, &gc->size_x, &gc->size_y, &gc->size_z, &gc->typeInput))
    {
        fprintf(stderr, "Error: could not read header information for input file '%s'.\n", gc->fnameInput);
        return 0;
    }
    if (!inrReadHeader(gc->fnameROIMask, &i, &j, &k, &type))
    {
        fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", gc->fnameROIMask);
        return 0;
    }
    if (type != InrByte || k != 1)
    {
        fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", gc->fnameROIMask);
        return 0;
    }
    if (i != gc->size_x || j != gc->size_y)
    {
        fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
        return 0;
    }
    if (gc->okInputNorm)
    {
        if (!inrReadHeader(gc->fnameInputNormParams, &i, &j, &k, &type))
        {
            fprintf(stderr, "Error: could not read header information for input normalization parameter file '%s'.\n", gc->fnameInputNormParams);
            return 0;
        }
        if (type != InrDouble || i != 2 || k != 1)
        {
            fprintf(stderr, "Error: input normalization parameter file '%s' has to be a single image with two columns of type 'Double'.\n", gc->fnameInputNormParams);
            return 0;
        }
        if (j != gc->size_z)
        {
            fprintf (stderr, "Error: dimensions of input data and input normalization parameter file do not match.\n");
            return 0;
        }
    }

    // reads the ROI mask image
    gc->ROIMask = inrReadByteImage (gc->fnameROIMask, &i, &j);
    if (!gc->ROIMask)
    {
        fprintf (stderr, "Error: could not read ROI mask data.\n");
        return 0;
    }

    // allocates memory for the input and output data
    gc->typesizeInput       = InrTypeSizes[gc->typeInput]/8;
    gc->dataInput           = (unsigned char *) malloc (gc->size_x*gc->size_y * gc->typesizeInput * sizeof(unsigned char));
    gc->dataOutput          = (float *)  malloc (gc->size_x*gc->size_y * sizeof(float));


    // writes output file header (output type is always float)
    if (!inrWriteHeader(gc->fnameOutput, gc->size_x, gc->size_y, gc->size_z, InrFloat))
    {
        printf("Error: could not write header information for output file '%s'.\n", gc->fnameOutput);
        return 0;
    }

    // initializes optional I/O files as appropriate
    if (gc->okOutputNorm && !outputNormParamsInit(gc))
        return 0;


    // initializes random number generator
    srand((unsigned) time(&t));


    // processes the data
    for (z = 0; z < gc->size_z; z++)
    {
        fprintf(stderr, "Processing layer %d of %d...\r", z+1, gc->size_z);

        // reads input data frame
        if (!inrReadFrame(gc->fnameInput, z, gc->dataInput))
        {
            fprintf(stderr, "Error: could not read frame %d for input image '%s'.\n", z, gc->fnameInput);
            return 0;
        }

        // reads or computes mean and standard deviation for the data frame 
        if (gc->okInputNorm)
            readNormParams(gc, z, &mean, &stddev);
        else
            computeNormParams(gc, &mean, &stddev);

        // outputs normalization info to normalization file if appropriate
        if (gc->okOutputNorm && !outputNormParams(gc, z, mean, stddev))
            return 0;

        // normalizes values
        normalize(gc, mean, stddev);

        // writes output data frame
        if (!inrWriteFrame(gc->fnameOutput, z, (unsigned char*)gc->dataOutput))
        {
            fprintf(stderr, "Error: could not write frame %d for output image '%s'.\n", z, gc->fnameOutput);
            return 0;
        }
    }

    // transfers georeference info if available
    if (inrReadGeoref(gc->fnameInput, &georef))
        inrWriteGeoref(gc->fnameOutput, georef);
    else
        inrWriteEnviHdr(gc->fnameOutput);
    
    
    bn= getBandNameCFS(z);
    
    inrAppendBandName(gc->fnameOutput,bn );

    fprintf(stderr, "INRIMAGE file successfully normalized.\n");

    free(gc->dataInput);
    free(gc->dataOutput);
    free(gc->ROIMask);

    return 0;
}
