#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "file_tools.h"
#include "str_tools.h"
#include<string.h>

#define FNAMESIZE  1024
#define FLISTSIZE  10000
#define BUFSIZE    16384


static char Ucmd[] =
"-idir input_dir -ofile output_file";

static char Udetail[] =
"\nBuilds an INRIMAGE sequence out of the INRIMAGE images contained in a given\n\
directory.\n\
\nImages should be of the same data type and have the same x and y dimensions.\n\
\nInput:\n\
\t-idir\tInput directory containing INRIMAGE images\n\
\nOutput:\n\
\t-ofile\tOutput file name for the INRIMAGE image sequence\n";




/*
 *  inrStrListBuildDataSequence
 * 
 *  Writes a sequence based on a list of INR image files of any data type.
 *  Files should be of the same data type and have the same x and y dimensions.
 */
int inrStrListBuildDataSequence(strList *filelist, char *outname)
{
  int     out_size_x, out_size_y, out_size_z;
  InrType out_type;
  int     okGeoref;
  int     typeSize;
  FILE   *fpOut;
  FILE   *fpData;
  char   *filename;
  int     i;
  unsigned char *data;
  InrGeorefInfo  georef;

  // returns failure if no files specified
  if (!filelist || filelist->count < 1)
    return 0;
    
  // gets name of the first file in the list
  filename = strListGet(filelist, 0);
    
  // defines output sequence dimensions and data type
  if (!inrReadHeader(filename, &out_size_x, &out_size_y, &out_size_z, &out_type))
    return 0;
    
  // attempts to read georef info from input file's ENVI .hdr header
  okGeoref = inrReadGeoref(filename, &georef);

  // obtains data type size in bytes
  typeSize = InrTypeSizes[out_type]/8;
  
  // opens temporary file for writing sequence data
  fpData = tmpfile();

  // builds data sequence
  out_size_z = 0;
  for (i=0; i<filelist->count; i++)
  {
    int size_x, size_y, size_z;
    InrType type;
    int nData;
    
    // gets name of the ith file in the list
    filename = strListGet(filelist, i);

    // reads data for the file
    data = inrReadDataSequence(filename, &size_x, &size_y, &size_z, &type);
    if (!data)
      return 0;
      
    // checks if data is compatible with previously read files
    if (size_x!=out_size_x || size_y!=out_size_y || type!=out_type)
    {
      free(data);
      return 0;
    }

    // writes data to temporary file
    nData = size_x*size_y*size_z;
    if ((int)fwrite(data, typeSize, nData, fpData) != nData)
    {
      free(data);
      fclose(fpData);
      return 0;
    }
    
    // frees data buffer
    free(data);

    // increments z dimension counter
    out_size_z += size_z;
  }
  
  // opens output sequence file
  fpOut = fopen(outname,"wb");
  if (!fpOut)
    return 0;
 
  // writes output sequence header info
  if (!inrWriteHeaderPtr(fpOut, out_size_x, out_size_y, out_size_z, out_type))
  {
    fclose(fpOut);
    return 0;
  }
  
  // copies data from temporary file to output sequence file
  fseek(fpData, 0, SEEK_SET);
  while(!feof(fpData))
  {
    char buf[BUFSIZE];
    size_t numbytes = fread(buf, 1, BUFSIZE, fpData);
    fwrite(buf, 1, numbytes, fpOut);
  }
  
  // writes ENVI .hdr header for sequence file
  if (okGeoref)
    inrWriteGeoref(outname, georef);
  else
    inrWriteEnviHdr(outname);

  // closes files and returns success
  fclose(fpOut);
  fclose(fpData);
  return 1;
}



int main(int argc, char *argv[])
{
  char     dirInput[FNAMESIZE];
  char     fnameOutput[FNAMESIZE];
  char     strFilter[FNAMESIZE];
  DIR     *dir;
  struct   dirent *ent;
  strList *filelist;
  int      okDir, okFile, okFilter;
  

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDir    = igetopt1 ("-idir"   , "%s", dirInput);
  okFile   = igetopt1 ("-ofile"  , "%s", fnameOutput);
  okFilter = igetopt1 ("-ifilter", "%s", strFilter);
  
  if (!okDir || !okFile)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // appends final slash if necessary
  ftFixDirName(dirInput);
    
  // opens directory
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // parses file(s) in the directory
  filelist = strListCreate(FLISTSIZE);
  while ((ent = readdir(dir)) != NULL)
  {
    // computes the complete input and output file names (dirname + filename)
    char fnameInput[FNAMESIZE];
    sprintf(fnameInput, "%s%s", dirInput, ent->d_name);
    
    // checks filter to see if file is selected to be processed
    if (okFilter && !strstr(fnameInput, strFilter))
      continue;

    // checks if file is of type INR
    if (!inrCheckFile(fnameInput))
      continue;

    // checks if list size is still within limits
    if (filelist->count >= FLISTSIZE)
    {
      printf("Error: file list too big for designated buffer\n");
      strListDestroy(filelist);
      return 0;
    }

    // adds file to list
    strListAdd(filelist, fnameInput);
  }
  // closes directory
  closedir(dir);
  
  // puts list in alphabetical order
  strListSort(filelist);
    
  // builds the sequence
  if (!inrStrListBuildDataSequence(filelist, fnameOutput))
  {
    printf("Error building INRIMAGE sequence.\n");
    return 0;
  }

  printf("INRIMAGE sequence successfully built with %d images.\n", filelist->count);
  strListDestroy(filelist);
  return 1;
}

