#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE       1024


static char Ucmd[] = "-ifile1 inr_file1 -ifile2 inr_file2 -ofile inr_merged \
[-nodata value]\n";

static char Udetail[] =
"Given two input images, outputs a new image merging its values.\n\
\nMerging is based on File1. If a given point has no valid data (i.e., value\n\
for File1 equals 'nodata'), then data from File2 is used.\n\
\nInput:\n\
\t-ifile\tFile name of input INRIMAGE image file\n\
\t-roi\tFile name of input ROI mask\n\
\nOutput:\n\
\t-ofile\tFile name of the output merged INRIMAGE file\n\
\nOptional:\n\
\t-nodata\tValue to be considered as missing data\n\
\t\tDefault value is 0\n";




int main (int argc, char **argv)
{
  char fnameInput1[FNAMESIZE];
  char fnameInput2[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  InrType type;
  InrType typeInput1;
  InrType typeInput2;
  InrType typeOutput;
  int     typesizeOutput;
  InrGeorefInfo georef;
  
  unsigned char *dataInput1;
  unsigned char *dataInput2;
  unsigned char *dataROIMask;
  unsigned char *dataOutput;
  
  int    okNoData;
  double nodata;
  
  int size_x, size_y, size_z;
  int i, j, k;
  int z;
  
  double value1;
  double value2;
  int perc;

  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile1", "%s", fnameInput1))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ifile2", "%s", fnameInput2))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi", "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okNoData = igetopt1 ("-nodata", "%lf", &nodata);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okNoData)
    nodata = 0;
    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput1, &size_x, &size_y, &size_z, &typeInput1))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput1);
    return 0;
  }
  if (!inrReadHeader(fnameInput1, &i, &j, &k, &typeInput2))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput2);
    return 0;
  }
  if (i != size_x || j != size_y || k != size_z)
  {
    fprintf (stderr, "Error: dimensions of input data files do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
  
  // reads input images
  dataInput1 = inrReadDataSequence(fnameInput1, &i, &j, &k, &type);
  if (!dataInput1)
    return 0;
  dataInput2 = inrReadDataSequence(fnameInput2, &i, &j, &k, &type);
  if (!dataInput2)
    return 0;

  // reads the ROI mask image
  dataROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!dataROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }
  
  // output file type follows input file's type
  typeOutput = typeInput1;

  // allocates memory for the output data
  typesizeOutput  = InrTypeSizes[typeOutput]/8;
  dataOutput = (unsigned char *) malloc (size_x*size_y*size_z * typesizeOutput * sizeof(unsigned char));


  // processes the data
  perc = -1;
  for (i=0; i<size_x*size_y; i++)
  {
    int new_perc = (int)(100 * ((double)i/(size_x*size_y)));
    if (new_perc > perc)
    {
      perc = new_perc;
      fprintf (stderr, "Processing... %d%%\r", perc);
    }

    if (!dataROIMask[i])
    {
      for (z=0; z<size_z; z++)
      {
        int pos = i + z * size_x*size_y;
        inrPutDataValue(dataOutput, typeOutput, pos, 0);
      }
    }
    else
    {
      for (z=0; z<size_z; z++)
      {
        int pos = i + z * size_x*size_y;

        // processes data value from input files to output file
        value1 = inrGetDataValue(dataInput1, typeInput1, pos);
        value2 = inrGetDataValue(dataInput2, typeInput2, pos);

        if (value2 != 0)
            inrPutDataValue(dataOutput, typeOutput, pos, nodata);
        else
            inrPutDataValue(dataOutput, typeOutput, pos, value1);
      }
    }
  }
  
  // writes output data file 
  inrWriteDataSequence(fnameOutput, size_x, size_y, size_z, dataOutput, typeOutput);

  // transfers georeference info from input, if available
  if (inrReadGeoref(fnameInput1, &georef))
    inrWriteGeoref(fnameOutput, georef);
  else
    inrWriteEnviHdr(fnameOutput);
  
    
  fprintf(stderr, "INRIMAGE file successfully masked.\n");

  free(dataInput1);
  free(dataInput2);
  free(dataROIMask);
  free(dataOutput);
  
  return 0;
}
