#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024
#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)



static char Ucmd[] = "-ifile image -trans translation_file -ofile new_image\n";

static char Udetail[] =
"Translates a given input INRIMAGE image file, given a text translation file.\n\
\nThe translation file should contain the following information at each line:\n\
\t'input_value  output_value'\n\
indicating that all pixels of value input_value should have value output_value\n\
\nInput:\n\
\t-ifile\tFile name of the input image\n\
\t-trans\tFile name of the text file defining the new values\n\
\nOutput:\n\
\t-ofile\tFile name of the output new image\n";





int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameTrans[FNAMESIZE];
  InrGeorefInfo georef;
  InrType       type;
  int           typesize;
  int           has_georef;

  FILE *fpTrans;

  unsigned char *dataInput;
  unsigned char *dataOutput;

  int size_x, size_y, size_z;
  int i;
  double value;
  double input_value;
  double output_value;
  

  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-trans"  , "%s", fnameTrans))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &type))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  fpTrans = fopen(fnameTrans, "rt");
  if (!fpTrans)
  {
    fprintf(stderr, "Error: could not read translation file '%s'.\n", fnameTrans);
    return 0;
  }
  has_georef = inrReadGeoref(fnameInput, &georef);
 
  // reads the input classification image
  dataInput = inrReadDataSequence(fnameInput, &size_x, &size_y, &size_z, &type);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not read input file '%s'.\n", fnameInput);
    fclose(fpTrans);
    return 0;
  }
  typesize = InrTypeSizes[type]/8;
  
  // allocates memory for output data
  dataOutput = (unsigned char *)malloc(size_x*size_y*size_z * typesize * sizeof(unsigned char));
  memcpy(dataOutput, dataInput, size_x*size_y*size_z * typesize);
    
  

  // processes the data
  while (!feof(fpTrans))
  {
    int nitems = fscanf(fpTrans, "%lf %lf", &input_value, &output_value);
    if (nitems != 2)
      continue;

    for (i=0; i<size_x*size_y*size_z; i++)
    {
      value = inrGetDataValue(dataInput, type, i);
      if (value == input_value)
        inrPutDataValue(dataOutput, type, i, output_value);
    }
  }
  fclose(fpTrans);

  
  // writes output translated classification
  if (!inrWriteDataSequence(fnameOutput, size_x, size_y, size_z, dataOutput, type))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    free(dataInput);
    free(dataOutput);
    return 0;
  }
  free(dataInput);
  free(dataOutput);
  
  // copies georreference info if present
  if (has_georef)
    inrWriteGeoref(fnameOutput, georef);
  
  //inrWriteColor (fnameOutput, 7);
  
  fprintf(stderr, "Image successfully translated.\n");
  return 1;
}
