#include <stdio.h>
#include <stdlib.h>
#include<string.h>
#include <math.h>
#include "linearalgebra.h"
#include "profile_features_original.h"

#define MIN(x,y)    ((x<y)?x:y)
#define MAX(x,y)    ((x>y)?x:y)


// profile feature names
char *profFeaturesNames[] = {
  "mean",
  "min",
  "max",
  "maxDate",
  "begin",
  "deltaMaxMin",
  "deltaBeginEnd",
  "delta150days",
  "slopeMin",
  "slopeAbsMean",
  "valueStddev",
  "valueSlopeMin",
  "valueSlopeAbsMean",
  "valueInterannualLoss",
  "premaxMean",
  "premaxMin",
  "premaxDeltaMaxMin",
  "premaxSlopeAbsMean"
};




/*
 *  profInitStruct
 * 
 *  Initializes a profFeatures stucture
 */
int profInitStruct(profFeatures *features)
{
  int i;
  
  features->nModes = 0;
  features->nModesStats = 0;

  for (i=0; i<PROF_NUM_FEATURES; i++)
    features->val[i] = 0;
  
  for (i=0; i<PROF_MAX_MODES; i++)
  {
    features->modes[i].datePeak        = 0;
    features->modes[i].dateWidthPeak   = 0;
    features->modes[i].dateWidthLeft   = 0;
    features->modes[i].dateWidthRight  = 0;
    features->modes[i].valuePeak       = 0;
    features->modes[i].valueDeltaLeft  = 0;
    features->modes[i].valueDeltaRight = 0;
    features->modes[i].integral        = 0;
    features->modes[i].slopeLeft       = 0;
    features->modes[i].slopeRight      = 0;
    features->modes[i].nSamples        = 0;
  }
  
  return 1;
}



/*
 *  profPolyValue
 * 
 *  Evaluates a polynome at a specified point
 */
double profPolyValue(double *coef, int degree, double point)
{
  int i;
  double value;
  
  value = 0;
  for (i=0; i<degree+1; i++)
    value += coef[i]*pow(point,i);

  return value;
}


/*
 *  profPolyDerivate
 * 
 *  Derivates a polynome specified by its coefficients
 */
double *profPolyDerivate(double *coef, int degree, double *output)
{
  int i;
  
  if (!coef || !output)
    return NULL;
    
  for (i=0; i<degree; i++)
    output[i] = coef[i+1]*(i+1);

  return output;
}



/*
 *  profFeaturesSum
 * 
 *  Adds a profFeature structure to a sum of features. Optionally, if 'sum2' is
 *  not NULL, it will also add the squared values of a structure to a squared
 *  sum of features
 */
int profFeaturesSum(profFeatures *sum, profFeatures *sum2, profFeatures term)
{
  profFeatures sum_temp;
  int i;
  int iSum;
  int nSums;
  int match[PROF_MAX_MODES];
  
  // prepares sum_temp structure for mode-matching purposes
  sum_temp = *sum;
  for (i=0; i<sum_temp.nModes; i++)
    sum_temp.modes[i].datePeak /= sum_temp.modes[i].nSamples;
    
  // matches modes
  profModesMatch(sum_temp, term, match);
  

  // defines number of sums to be done ('sum' and 'sum2' or just 'sum')
  nSums = (sum2 == NULL)?1:2;


  // sums values
  for (iSum=0; iSum<nSums; iSum++)
  {
    // operates on regular features      
    sum->nModesStats += term.nModesStats;
  
    for (i=0; i<PROF_NUM_FEATURES; i++)
      sum->val[i] += term.val[i];

#ifndef PROF_MATCH_MODES

    /*** NO MODE MATCHING! ***/
    for (i=0; i<PROF_MAX_MODES; i++)
      profModesAdd(sum->modes[i], term.modes[i], &sum->modes[i]);
    sum->nModes = MAX(sum->nModes, term.nModes);
    
#else
    /*** MATCHES MODES! ***/
    // sums up matching modes
    for (i=0; i<sum_temp.nModes; i++)
      if (match[i] >= 0)
        profModesAdd(sum->modes[i], term.modes[match[i]], &sum->modes[i]);
    
    // checks for unmatched modes from term
    for (j=0; j<term.nModes; j++)
    {
      // searches if mode j has a match
      for (i=0; i<sum_temp.nModes; i++)
        if (match[i] == j)
          break;
          
      // if there is no match, mode j is transfered "as is" to sum
      // (provided it doesn't exceed maximum number of modes)
      if (i == sum_temp.nModes && sum->nModes < PROF_MAX_MODES)
      {
        sum->modes[sum->nModes] = term.modes[j];
        sum->nModes++;
      }
    }
#endif
    
    // prepares for summing squared values
    sum = sum2;
    profFeaturesSquare(term, &term);
  }
  return 1;
}



/*
 *  profFeaturesSquare
 * 
 *  Performs profFeature^2
 */
int profFeaturesSquare(profFeatures term1, profFeatures *result)
{
  int i;
  
  result->nModes      = term1.nModes;
  result->nModesStats = term1.nModesStats * term1.nModesStats;

  for (i=0; i<PROF_NUM_FEATURES; i++)
    result->val[i] = term1.val[i] * term1.val[i];

  for (i=0; i<term1.nModes; i++)
    profModesSquare(term1.modes[i], &result->modes[i]);
  
  return 1;
}



/*
 *  profFeaturesMean
 * 
 *  Computes the mean of a group of profFeature structures based on their
 *  sum and on the number of elements in the group
 */
int profFeaturesMean(profFeatures sum, int n, profFeatures *result)
{
  int i;
  
  if (n < 1)
    return 0;

  result->nModes        = sum.nModes;
  result->nModesStats   = sum.nModesStats / n;

  for (i=0; i<PROF_NUM_FEATURES; i++)
    result->val[i] = sum.val[i] / n;

  for (i=0; i<sum.nModes; i++)
    profModesMean(sum.modes[i], &result->modes[i]);
  
  return 1;
}



/*
 *  profFeaturesStddev
 * 
 *  Computes the stddev of a group of profFeature structures based on their
 *  sum, squared sum, and on the number of elements in the group
 */
int profFeaturesStddev(profFeatures sum, profFeatures sum2, int n, profFeatures *result)
{
  int i;

  if (n < 2)
    return 0;

  result->nModes      = sum.nModes;
  result->nModesStats = sqrt((sum2.nModesStats - sum.nModesStats*sum.nModesStats/n)/(n-1));
  
  for (i=0; i<PROF_NUM_FEATURES; i++)
    result->val[i] = sqrt((sum2.val[i] - sum.val[i]*sum.val[i]/n)/(n-1));

  for (i=0; i<sum.nModes; i++)
    profModesStddev(sum.modes[i], sum2.modes[i], &result->modes[i]);
  
  return 1;
}



/*
 *  profModesAdd
 * 
 *  Adds two profMode structures
 */
int profModesAdd (profMode mode1, profMode mode2, profMode *result)
{
  result->datePeak        = mode1.datePeak        + mode2.datePeak;
  result->dateWidthPeak   = mode1.dateWidthPeak   + mode2.dateWidthPeak;
  result->dateWidthLeft   = mode1.dateWidthLeft   + mode2.dateWidthLeft;
  result->dateWidthRight  = mode1.dateWidthRight  + mode2.dateWidthRight;
  result->valuePeak       = mode1.valuePeak       + mode2.valuePeak;
  result->valueDeltaLeft  = mode1.valueDeltaLeft  + mode2.valueDeltaLeft;
  result->valueDeltaRight = mode1.valueDeltaRight + mode2.valueDeltaRight;
  result->integral        = mode1.integral        + mode2.integral;
  result->slopeLeft       = mode1.slopeLeft       + mode2.slopeLeft;
  result->slopeRight      = mode1.slopeRight      + mode2.slopeRight;
  result->nSamples        = mode1.nSamples        + mode2.nSamples;
  
  return 1;
}



/*
 *  profModesSquare
 * 
 *  Performs profMode^2
 */
int profModesSquare (profMode mode1, profMode *result)
{
  result->datePeak        = mode1.datePeak        * mode1.datePeak;
  result->dateWidthPeak   = mode1.dateWidthPeak   * mode1.dateWidthPeak;
  result->dateWidthLeft   = mode1.dateWidthLeft   * mode1.dateWidthLeft;
  result->dateWidthRight  = mode1.dateWidthRight  * mode1.dateWidthRight;
  result->valuePeak       = mode1.valuePeak       * mode1.valuePeak;
  result->valueDeltaLeft  = mode1.valueDeltaLeft  * mode1.valueDeltaLeft;
  result->valueDeltaRight = mode1.valueDeltaRight * mode1.valueDeltaRight;
  result->integral        = mode1.integral        * mode1.integral;
  result->slopeLeft       = mode1.slopeLeft       * mode1.slopeLeft;
  result->slopeRight      = mode1.slopeRight      * mode1.slopeRight;
  result->nSamples        = mode1.nSamples;
  
  return 1;
}



/*
 *  profModesMean
 * 
 *  Computes the mean of a mode for a group of profFeature structures based on
 *  their sum and on the number of mode samples present in the group
 */
int profModesMean (profMode sum, profMode *result)
{
  if (sum.nSamples < 1)
    return 0;
    
  result->datePeak        = sum.datePeak        / sum.nSamples;
  result->dateWidthPeak   = sum.dateWidthPeak   / sum.nSamples;
  result->dateWidthLeft   = sum.dateWidthLeft   / sum.nSamples;
  result->dateWidthRight  = sum.dateWidthRight  / sum.nSamples;
  result->valuePeak       = sum.valuePeak       / sum.nSamples;
  result->valueDeltaLeft  = sum.valueDeltaLeft  / sum.nSamples;
  result->valueDeltaRight = sum.valueDeltaRight / sum.nSamples;
  result->integral        = sum.integral        / sum.nSamples;
  result->slopeLeft       = sum.slopeLeft       / sum.nSamples;
  result->slopeRight      = sum.slopeRight      / sum.nSamples;
  result->nSamples        = sum.nSamples;
  
  return 1;
}



/*
 *  profModesStddev
 * 
 *  Computes the stddev of a mode for a group of profFeature structures based on
 *  their sum, squared sum, and on the number of mode samples present in the group
 */
int profModesStddev (profMode sum, profMode sum2, profMode *result)
{
  int n = sum.nSamples;
  
  if (n < 2)
    return 0;
    
  result->datePeak        = sqrt((sum2.datePeak        - sum.datePeak*sum.datePeak/n)/(n-1));
  result->dateWidthPeak   = sqrt((sum2.dateWidthPeak   - sum.dateWidthPeak*sum.dateWidthPeak/n)/(n-1));
  result->dateWidthLeft   = sqrt((sum2.dateWidthLeft   - sum.dateWidthLeft*sum.dateWidthLeft/n)/(n-1));
  result->dateWidthRight  = sqrt((sum2.dateWidthRight  - sum.dateWidthRight*sum.dateWidthRight/n)/(n-1));
  result->valuePeak       = sqrt((sum2.valuePeak       - sum.valuePeak*sum.valuePeak/n)/(n-1));
  result->valueDeltaLeft  = sqrt((sum2.valueDeltaLeft  - sum.valueDeltaLeft*sum.valueDeltaLeft/n)/(n-1));
  result->valueDeltaRight = sqrt((sum2.valueDeltaRight - sum.valueDeltaRight*sum.valueDeltaRight/n)/(n-1));
  result->integral        = sqrt((sum2.integral        - sum.integral*sum.integral/n)/(n-1));
  result->slopeLeft       = sqrt((sum2.slopeLeft       - sum.slopeLeft*sum.slopeLeft/n)/(n-1));
  result->slopeRight      = sqrt((sum2.slopeRight      - sum.slopeRight*sum.slopeRight/n)/(n-1));
  result->nSamples        = n;
  
  return 1;
}



/*
 *  profModesMatch
 * 
 *  Finds out which modes in features2 correspond to the modes present in
 *  features1. Result is given in an array of integers, so that 'j=result[i]'
 *  specifies that the mode features1.modes[i] corresponds to the mode
 *  features2.modes[j]
 */
int profModesMatch (profFeatures features1, profFeatures features2, int *result)
{
  int i;
  int match_in_1;
  int match_in_2;
  
  if (!result)
    return -1;

  for (i=0; i<features1.nModes; i++)
  {
    // finds matching mode for the ith mode in features1
    match_in_2 = profModeFind(features2, features1.modes[i]);
    if (match_in_2 < 0)
    {
      // if nothing found, no match for mode i
      result[i] = -1;
      continue;
    }
    
    // does the inverse operation for the selected mode from features2
    match_in_1 = profModeFind(features1, features2.modes[match_in_2]);
    
    // match is only valid if it is stable
    // (i.e. the match of the match should be the original mode)
    if (match_in_1 == i)
      result[i] = match_in_2;
    else
      result[i] = -1;
  }
  return 1;  
}


/*
 *  profModeFind
 * 
 *  Searches a profFeatures structure for the mode closest to the one specified
 *  Search criterion is minimin peak date distance
 */
int profModeFind(profFeatures features, profMode mode)
{
  int i;
  int iMode;
  double dist;
  double distMin;
  
  
  if (features.nModes < 1)
    return -1;
    
  if (mode.datePeak == 0)
    return -1;
  
  iMode = 0;
  distMin = fabs(mode.datePeak - features.modes[0].datePeak);
  
  for (i=1; i<features.nModes; i++)
  {
    // ignores invalid modes
    if (features.modes[i].datePeak == 0)
      continue;
      
    dist = fabs(mode.datePeak - features.modes[i].datePeak);
    if (dist < distMin)
    {
      distMin = dist;
      iMode = i;
    }
  }
  return iMode;
}



/*
 *  profDefineMode
 * 
 *  Computes modes based on a polynomial representation of a temporal profile
 */
int profDefineMode (double *coef, int degree, double *profile, double *profile_deriv,
                    double  dateBegin, double dateEnd, double dateInc,
                    int iBegin, int iPeak, int iEnd, profFeatures *features)
{
  int       iDate;
  double    date;
  double    deriv2Prev;
  double    deriv2Curr;
  double   *coef_deriv2;
  double    slopeTotal;
  int       slopeCounter;
  int       peakWidthCounter;
  profMode *mode = &(features->modes[features->nModes]);
  
  
  // computes the 2nd derivative of the polynome
  coef_deriv2 = (double *)malloc((degree)*sizeof(double));
  profPolyDerivate(coef, degree, coef_deriv2);
  profPolyDerivate(coef_deriv2, degree-1, coef_deriv2);

  // defines general mode values
  mode->datePeak        = dateBegin + dateInc*iPeak;
  mode->dateWidthLeft   = dateInc*(iPeak - iBegin);
  mode->dateWidthRight  = dateInc*fabs(iEnd  - iPeak);
  mode->valuePeak       = profile[iPeak];
  mode->valueDeltaLeft  = profile[iPeak] - profile[iBegin];
  mode->valueDeltaRight = profile[iPeak] - profile[iEnd];
  mode->nSamples        = 1;
  
  // defines datePeakWidth
  peakWidthCounter = 0;
  for (iDate=iBegin; iDate<iPeak; iDate++)
  {
    // increments peak width counter if value is within peak tolerance
    if ((profile[iPeak]-profile[iDate]) < PROF_PEAK_TOLERANCE) 
      peakWidthCounter++;
  }
  if (iEnd == iBegin)
    peakWidthCounter *= 2;
  else
  {
    for (iDate=iPeak+1; iDate<iEnd; iDate++)
    {
      // increments peak width counter if value is within peak tolerance
      if ((profile[iPeak]-profile[iDate]) < PROF_PEAK_TOLERANCE) 
        peakWidthCounter++;
    }
  }
  peakWidthCounter++;
  mode->dateWidthPeak = peakWidthCounter * dateInc;
  
  // defines integral
  mode->integral = 0;
  for (iDate=iBegin; iDate<iPeak; iDate++)
    mode->integral += profile[iDate];
  if (iEnd == iBegin)
    mode->integral *= 2;
  else    
    for (iDate=iPeak+1; iDate<iEnd; iDate++)
      mode->integral += profile[iDate];
  mode->integral += profile[iPeak];

  // defines slopeLeft
  deriv2Prev = profPolyValue(coef_deriv2, degree-2, iBegin);
  slopeTotal   = 0;
  slopeCounter = 0;
  for (iDate=iBegin+1; iDate<iPeak; iDate++)
  {
    date = dateBegin + dateInc*iDate;

    // computes 2nd derivative at this date
    deriv2Curr = profPolyValue(coef_deriv2, degree-2, date);
   
    // identifies a point where the 2nd derivative equals zero (change of signal)
    if (deriv2Curr * deriv2Prev < 0)
    {
      // increments slope values
      slopeTotal += profile_deriv[iDate];
      slopeCounter++;
    }
    deriv2Prev = deriv2Curr;
  }
  if (slopeCounter)
    mode->slopeLeft = slopeTotal / slopeCounter;
  else
    mode->slopeLeft = 0;

  // defines slopeRight
  if (iEnd == iBegin)
    mode->slopeRight = -mode->slopeLeft;
  else
  {
    deriv2Prev = profPolyValue(coef_deriv2, degree-2, iPeak+1);
    slopeTotal   = 0;
    slopeCounter = 0;
    for (iDate=iPeak+2; iDate<iEnd; iDate++)
    {
      date = dateBegin + dateInc*iDate;
  
      // computes 2nd derivative at this date
      deriv2Curr = profPolyValue(coef_deriv2, degree-2, date);
     
      // identifies a point where the 2nd derivative equals zero (change of signal)
      if (deriv2Curr * deriv2Prev < 0)
      {
        // increments slope values
        slopeTotal += profile_deriv[iDate];
        slopeCounter++;
      }
      deriv2Prev = deriv2Curr;
    }
    if (slopeCounter)
      mode->slopeRight = slopeTotal / slopeCounter;
    else
      mode->slopeRight = 0;
  }
      
  // updates feature mode counters
  features->nModes++;
  features->nModesStats = features->nModes;
  
  free(coef_deriv2);
  return 1;
}


/*
 *  profComputeModes
 * 
 *  Computes modes based on a polynomial representation of a temporal profile
 */
int profComputeModes(double *coef, int degree, double *profile, double *profile_deriv, 
                     double dateBegin, double dateEnd, double dateInc,
                     profFeatures *features)
{
  int    state;
  int    iDate;
  int    nDates;
  int    iBegin;
  int    iPeak;
  
  double tolerance = PROF_DEFAULT_MODE_TOLERANCE;
  

  // computes number of dates (size of 'profile')
  nDates = (int)((dateEnd - dateBegin)/dateInc + 1);
  

  // initializes values
  features->nModes = 0;
  iBegin = 0;
  iPeak  = -1;
  state  = 1;

  // searches profile for modes
  for (iDate=1; iDate<nDates; iDate++)
  {
    // identifies a point where the derivative equals zero (change of signal)
    if ((profile_deriv[iDate] * profile_deriv[iDate-1]) < 0)
    {
      switch (state)
      {
        case 0:
          // begins a search for a mode
          iBegin = iDate;
          state  = 1;
          break;
          
        case 1:
          // searches for a peak value for the mode
          if ((profile[iDate] - profile[iBegin]) > tolerance)
          {
            iPeak = iDate;
            state = 2;
          }
          else
          {
            // values too close: re-starts search from here
            iBegin = iDate;
          }
          break;
        
        case 2:
          // searches for an ending value for the mode
          if (profile[iDate] > profile[iPeak])
          {
            // current value is higher than current peak: defines new peak
            iPeak = iDate;
            state = 2;
          }
          else if ((profile[iPeak] - profile[iDate]) < tolerance)
          {
            // values too close: keeps current peak and goes on searching
            state = 2;
          }
          else
          {
            // mode is complete
            profDefineMode(coef, degree, profile, profile_deriv, dateBegin, dateEnd, dateInc,
                           iBegin, iPeak, iDate, features);

            // re-starts search            
            iBegin = iDate;
            state  = 1;
          }
          break;
          
        default:
          state = 0;
          break;
      }
    }  
  }
  
  if (state == 2)
  {
    // if a mode is halfway done, finishes it up by estimating that the right
    // half is equal to the left half (i.e., mirrors the left half)
    profDefineMode(coef, degree, profile, profile_deriv, dateBegin, dateEnd, dateInc,
                   iBegin, iPeak, iBegin, features);
  }
  
  return 1;
}


/*
 *  profComputePolyFeatures_PreMax
 * 
 *  Auxiliary method for computing premax features: mean, deltaMaxMin, etc
 *  
 */
void profComputePolyFeatures_PreMax(double *profile, double *profile_deriv,
                                    double dateBegin, double dateEnd, double dateInc,
                                    profFeatures *features)
{
  double sum, min, max, maxDate;
  double slopeAbsSum;
  double date;
  int    iDate, nDates;
  
  // computes number of dates (size of 'profile')
  nDates = (int)((dateEnd - dateBegin)/dateInc + 1);

  min  = profile[0];
  max  = profile[0];
  sum  = 0;
  maxDate = features->val[prof_maxDate];
  slopeAbsSum = 0;
  
  // does nothing if date of maximum is less or equal than beginning date
  // (there is nothing to compute!)
  if (maxDate <= dateBegin)
    return;

  for (iDate=0; iDate<nDates; iDate++)
  {
    date = dateBegin + dateInc*iDate;
    
    // goes on only until maxDate
    if (date > maxDate)
      break;
      
    sum += profile[iDate];
    min  = MIN(min, profile[iDate]);
    max  = MAX(max, profile[iDate]);
    slopeAbsSum += fabs(profile_deriv[iDate]);
  }

  features->val[prof_premaxMean]         = sum/iDate;
  features->val[prof_premaxMin]          = min;
  features->val[prof_premaxDeltaMaxMin]  = max - min;
  features->val[prof_premaxSlopeAbsMean] = slopeAbsSum/iDate;
}


/*
 *  profGetMean
 * 
 *  Auxiliary method for computing the mean profile value given a date center and
 *  width in days. The mean value will thus refer to all data contained between
 *  dates (center - width/2) and (center + width/2)
 */
double profGetMean(int dateCenter, int dateWidth,
                   double *profile, int *dates, int nDates,
                   double *profilePrev, int *datesPrev, int nDatesPrev)
{
    double sum = 0;
    int nSum = 0;

    int iDate;
    int iDatePrev;

    // computes begin and end dates for the computation
    int dateBegin = dateCenter - (int) ceil((double)dateWidth/2);
    int dateEnd   = dateCenter + (int) ceil((double)dateWidth/2);

    // checks if profile values from the previous year must be included
    if (nDatesPrev > 0 && dateBegin < dates[0])
    {
        // computes begin date from which previous profile values must be included 
        int dateBeginPrev = datesPrev[nDatesPrev-1] - (dates[0] - dateBegin);

        // accumulates previous profile values backwards until the begin point
        for (iDatePrev = nDatesPrev-1; iDatePrev >= 0; iDatePrev--)
        {
            if (datesPrev[iDatePrev] < dateBeginPrev)
                break;

            // accumulates value
            sum += profilePrev[iDatePrev];
            nSum++;
        }
    }

    // accumulates profile values between the begin and end dates
    for (iDate = 0; iDate < nDates; iDate++)
    {
        // skips date if it is prior to the begin date
        if (dates[iDate] < dateBegin)
            continue;

        // breaks iteration if date is beyond the end date
        if (dates[iDate] > dateEnd)
            break;

        // accumulates value
        sum += profile[iDate];
        nSum++;
    }

    // returns mean value computed
    return (nSum > 0 ? sum/nSum : 0);
}
                         

/*
 *  profGetPrevMin
 * 
 *  Auxiliary method for computing the mean previous profile value for a given date,
 *  considering a width tolerance in days. The mean value will thus refer to all
 *  data from the previous profile contained between dates (center - width/2) and
 *  (center + width/2)
 */
double profGetPrevMean(int dateCenter, int dateWidth,
                       double *profile, int *dates, int nDates,
                       double *profilePrev, int *datesPrev, int nDatesPrev)
{
    double sum = 0;
    int nSum = 0;

    int iDate;
    int iDatePrev;

    // computes begin and end dates for the computation
    int dateBegin = dateCenter - (int) ceil((double)dateWidth/2);
    int dateEnd   = dateCenter + (int) ceil((double)dateWidth/2);

    // checks if previous date information is available
    if (nDatesPrev <= 0)
        return 0;

    // evaluates previous profile values between the begin and end dates
    for (iDatePrev = 0; iDatePrev < nDatesPrev; iDatePrev++)
    {
        // skips date if it is prior to the begin date
        if (datesPrev[iDatePrev] < dateBegin)
            continue;

        // breaks iteration if date is beyond the end date
        if (datesPrev[iDatePrev] > dateEnd)
            break;

        //*** DEBUG *** incomplete profiles have zero values
        // ignores null values
        if (profilePrev[iDatePrev] == 0)
            continue;

        // accumulates value
        sum += profilePrev[iDatePrev];
        nSum++;
    }

    // checks if profile values from the current year must also be evaluated
    if (dateEnd > datesPrev[nDatesPrev-1])
    {
        // computes end date up to which profile values must be considered
        int dateEndCurr = dates[0] + (dateEnd - datesPrev[nDatesPrev-1]);

        // evaluates profile values until the end point
        for (iDate = 0; iDate < nDates; iDate++)
        {
            if (dates[iDate] > dateEndCurr)
                break;

            // accumulates value
            sum += profile[iDate];
            nSum++;
        }
    }

    // returns mean value computed
    return (nSum > 0 ? sum/nSum : 0);
}



/*
 *  profGetPrevMin
 * 
 *  Auxiliary method for computing the minimum previous value for a given date,
 *  considering a width tolerance in days. The minimum value will thus refer to all
 *  data from the previous profile contained between dates (center - width/2) and
 *  (center + width/2)
 */
double profGetPrevMin(int dateCenter, int dateWidth,
                      double *profile, int *dates, int nDates,
                      double *profilePrev, int *datesPrev, int nDatesPrev)
{
    double min = 0;
    int isMinValid = 0;

    int iDate;
    int iDatePrev;

    // computes begin and end dates for the computation
    int dateBegin = dateCenter - (int) ceil((double)dateWidth/2);
    int dateEnd   = dateCenter + (int) ceil((double)dateWidth/2);

    // checks if previous date information is available
    if (nDatesPrev <= 0)
        return 0;

    // evaluates previous profile values between the begin and end dates
    for (iDatePrev = 0; iDatePrev < nDatesPrev; iDatePrev++)
    {
        // skips date if it is prior to the begin date
        if (datesPrev[iDatePrev] < dateBegin)
            continue;

        // breaks iteration if date is beyond the end date
        if (datesPrev[iDatePrev] > dateEnd)
            break;

        //*** DEBUG *** incomplete profiles have zero values
        // ignores null values
        if (profilePrev[iDatePrev] == 0)
            continue;

        // checks value and resets minimum if appropriate
        if (!isMinValid || profilePrev[iDatePrev] < min)
        {
            min = profilePrev[iDatePrev];
            isMinValid = 1;
        }
    }

    // checks if profile values from the current year must also be evaluated
    if (dateEnd > datesPrev[nDatesPrev-1])
    {
        // computes end date up to which profile values must be considered
        int dateEndCurr = dates[0] + (dateEnd - datesPrev[nDatesPrev-1]);

        // evaluates profile values until the end point
        for (iDate = 0; iDate < nDates; iDate++)
        {
            if (dates[iDate] > dateEndCurr)
                break;

            // checks value and resets minimum if appropriate
            if (!isMinValid || profile[iDate] < min)
            {
                min = profile[iDate];
                isMinValid = 1;
            }
        }
    }

    // returns minimum value computed
    return min;
}



/**** MAJOR METHODS ****/

/*
 *  profComputeFeatures
 * 
 *  Computes the features of a temporal profile by means of fitting and
 *  analyzing a polynomial curve of a specified degree
 */
int profComputeFeatures(double *values, int *dates, int nDates,
                        double *valuesPrev, int *datesPrev, int nDatesPrev,
                        int degree, profFeatures *features)
{
  double *coef = (double *) malloc ((degree+1) * sizeof(double));
  
  // computes the polynomial coefficients that best fit the profile values
  if (!profFitPoly(values, dates, nDates, coef, degree))
    return 0;

  // initializes feature structure
  profInitStruct(features);
  // computes profile features based on the polynomial coefficients
  profComputePolyFeatures(coef, degree, dates[0], dates[nDates-1], 1, features);
  // computes profile features based on the raw values
  profComputeValueFeatures(values, dates, nDates, valuesPrev, datesPrev, nDatesPrev, features);
  
  free(coef);
  return 1;  
}


/*
 *  profComputePolyFeatures
 * 
 *  Computes features based on a polynomial representation of a temporal profile
 */
int profComputePolyFeatures(double *coef, int degree, double dateBegin, double dateEnd, 
                            double dateInc, profFeatures *features)
{
  double *profile;
  double *profile_deriv;
  double *coef_deriv;
  int     iDate;
  int     nDates;
  double  date;

  int     iDate150;
  double  mean150;
  double  mean150rest;
  


  // initializes feature structure
  //profInitStruct(features);
  
  // allocates memory for the profile
  nDates        = (int)((dateEnd - dateBegin)/dateInc + 1);
  profile       = (double *)malloc(nDates * sizeof(double));
  profile_deriv = (double *)malloc(nDates * sizeof(double));

  // derivates the polynome
  coef_deriv = (double *)malloc(degree*sizeof(double));
  profPolyDerivate(coef, degree, coef_deriv);


  // computes profile and profile_deriv
  for (iDate=0; iDate<nDates; iDate++)
  {
    date = dateBegin + dateInc*iDate;
    profile[iDate]       = profPolyValue(coef, degree, date);
    profile_deriv[iDate] = profPolyValue(coef_deriv, degree-1, date);
  }

  
  // computes mean, min, max, maxDate, begin, deltaMaxMin, deltaBeginEnd, delta150days 
  features->val[prof_min] = features->val[prof_max] = profile[0];
  features->val[prof_maxDate] = dateBegin;
  features->val[prof_mean] = 0;
  for (iDate=0; iDate<nDates; iDate++)
  {
    features->val[prof_mean] += profile[iDate];
    features->val[prof_min]   = MIN(features->val[prof_min], profile[iDate]);
    if (features->val[prof_max] < profile[iDate])
    {
      features->val[prof_max]     = profile[iDate];
      features->val[prof_maxDate] = dateBegin + dateInc*iDate;
    }
  }
  features->val[prof_mean]         /= nDates;
  features->val[prof_begin]         = profile[0];
  features->val[prof_deltaMaxMin]   = features->val[prof_max] - features->val[prof_min];
  features->val[prof_deltaBeginEnd] = profile[nDates-1] - profile[0];


  // computes delta150days 
  iDate150 = (int)(150/dateInc);
  mean150 = 0;
  mean150rest = 0;
  for (iDate=0; iDate<iDate150; iDate++)
    mean150 += profile[iDate];
  mean150 /= iDate;
  for (; iDate<nDates; iDate++)
    mean150rest += profile[iDate];
  mean150rest /= nDates - iDate150;
  features->val[prof_delta150days] = mean150rest - mean150;
  

  // computes slope features
  features->val[prof_slopeMin]     = profile_deriv[0];
  features->val[prof_slopeAbsMean] = fabs(profile_deriv[0]);
  for (iDate=1; iDate<nDates; iDate++)
  {
    features->val[prof_slopeMin]      = MIN(features->val[prof_slopeMin], profile_deriv[iDate]);
    features->val[prof_slopeAbsMean] += fabs(profile_deriv[iDate]);
  }
  features->val[prof_slopeAbsMean] /= nDates;
  
  
  // computes premax features
  profComputePolyFeatures_PreMax(profile, profile_deriv, dateBegin, dateEnd, dateInc, features);

  // computes features for the modes present in the temporal profile
  profComputeModes(coef, degree, profile, profile_deriv, dateBegin, dateEnd, dateInc, features);


  free(profile); 
  free(profile_deriv); 
  free(coef_deriv);
  return 1;
}



/*
 *  profComputeValueFeatures
 * 
 *  Computes features based on the raw values of a temporal profile
 */
int profComputeValueFeatures(double *profile, int *dates, int nDates,
                             double *profilePrev, int *datesPrev, int nDatesPrev,
                             profFeatures *features)
{
  int iDate;
  double slope;
  
  // computes valueStddev
  features->val[prof_valueStddev] = 0;
  for (iDate=1; iDate<nDates; iDate++)
    features->val[prof_valueStddev] += (profile[iDate] - profile[iDate-1])*(profile[iDate] - profile[iDate-1]);
  features->val[prof_valueStddev] = sqrt(features->val[prof_valueStddev]/(nDates-1));
  

  // computes valueSlope features
  slope = (profile[1] - profile[0])/(dates[1] - dates[0]);
  features->val[prof_valueSlopeMin]     = slope;
  features->val[prof_valueSlopeAbsMean] = fabs(slope);
  for (iDate=1; iDate<nDates; iDate++)
  {
    slope = (profile[iDate] - profile[iDate-1])/(dates[iDate] - dates[iDate-1]);
    features->val[prof_valueSlopeMin]      = MIN(features->val[prof_valueSlopeMin], slope);
    features->val[prof_valueSlopeAbsMean] += fabs(slope);
  }
  features->val[prof_valueSlopeAbsMean] /= nDates;


  // computes valueInterannualLoss (as long as there is previous date information)
  features->val[prof_valueInterannualLoss] = 0;
  if (nDatesPrev > 0)
  {
      int middleDate = dates[0] + (dates[nDates-1] - dates[0])/2;

      // computes mean value for current profile, and mean value for previous profile
      double mean     = profGetMean(middleDate, 180, profile, dates, nDates, profilePrev, datesPrev, nDatesPrev);
      double meanPrev = profGetPrevMean(middleDate, 180, profile, dates, nDates, profilePrev, datesPrev, nDatesPrev);

      // final interannual feature is equivalent to the loss observed
      features->val[prof_valueInterannualLoss] = meanPrev - mean;
  }
 

  return 1;
}



/*
 *  profFitPoly
 * 
 *  Fits a polynomial curve of a specified degree to a sequence of values.
 * 
 *  Fitting is done considering the dates (i.e. the x coordinate) of each value.
 *  'Coef' corresponds to an array previously allocated that will contain
 *  the coefficients of the fitted polynomial.
 *  For an arbitrary date, the corredponding value for the fitted curve can 
 *  then be computed as follows:
 *     value = sum( coef[i]*(date^i) )   ,  where  0 < i <= degree
 * 
 */
int profFitPoly(double *values, int *dates, int nDates, double *coef, int degree)
{
  int size_matrix;
  int i, j;
  int iDate;
  matrice *A;
  matrice *b;
  matrice *A_inv;
  matrice *result;
  
  size_matrix = degree + 1;


  A = Matrice_Nulle(size_matrix, size_matrix);
  b = Matrice_Nulle(size_matrix, 1);
  
  // fills up matrix A
  for (i=0; i<size_matrix; i++)
    for (j=0; j<size_matrix; j++)
      for (iDate=0; iDate<nDates; iDate++)
        A->val[j + i*size_matrix] += pow(dates[iDate],i+j);

  // fills up vector b
  for (i=0; i<size_matrix; i++)
    for (iDate=0; iDate<nDates; iDate++)
      b->val[i] += pow(dates[iDate],i) * values[iDate];
  
  
  // solves system and obtains function coefficients
  A_inv  = Inverser_Matrice(A);
  result = Multiplier_Matrice(A_inv, b);


  // puts result into output variable
  for (i=0; i<size_matrix; i++)
    coef[i] = result->val[i];
  
  // frees memory and returns success
  free_matrice(&A);
  free_matrice(&b);
  free_matrice(&A_inv);
  free_matrice(&result);

  return 1;
}


/*
 *  profGetPolyValue
 * 
 *  Given a polynomial profile (defined by its degree and coefficients),
 *  computes its value at an arbitrary date
 *     value = sum( coef[i]*(date^i) )   ,  where  0 < i <= degree
 */
double profGetPolyValue(double *coef, int degree, int date)
{
  double value = 0;
  int i;
  
  for (i=0; i<(degree+1); i++)
    value += coef[i]*pow(date,i);
    
  return value;
}

/*
 * getBandNameCFS
 *
 * provide band name of CFS features 
 * for use in ENVI headers
 */
InrBandNameInfo * getBandNameCFS()
{
  InrBandNameInfo* bn ;
  int i;
  int nFeatures=11 ;

  bn=(InrBandNameInfo*)malloc(sizeof(InrBandNameInfo));
  bn->nBands = nFeatures ;


  bn->bandName=(char**)malloc(nFeatures*sizeof(char*));
  for (i=0;i<nFeatures;i++)
    bn->bandName[i] = (char*)malloc(30);
    
  
  strcpy(bn->bandName[0],"mean");
  strcpy(bn->bandName[1],"maxDate");
  strcpy(bn->bandName[2],"delta150days");
  strcpy(bn->bandName[3],"slopeAbsMean");
  strcpy(bn->bandName[4],"nModes");
  strcpy(bn->bandName[5],"min");
  strcpy(bn->bandName[6],"deltaMaxMin");
  strcpy(bn->bandName[7],"valuesstddev");
  strcpy(bn->bandName[8],"datePeak");
  strcpy(bn->bandName[9],"dateWidthPeak");
  strcpy(bn->bandName[10],"Noise");

  return bn ;
} 


 
