#include <stdlib.h>
#include "std.h"



/*
 * stdCreate
 * 
 * Creates an empty STD given a number of states and a number of class profiles
 */
stdDiagram *stdCreate(int nStates, int nProfiles)
{
    int iState;

    // creates basic STD and sets number of states and profile classes
    stdDiagram *std = (stdDiagram *) malloc (sizeof(stdDiagram));
    std->nStates = nStates;
    std->nProfiles = nProfiles;

    // allocates memory for STD's states
    std->states = (stdState *) malloc (nStates * sizeof(stdState));

    // intitializes per-state structures
    for (iState = 0; iState < nStates; iState++)
    {
        // allocates memory for all transitions: for each state, there will
        // potentially be one transition for each profile class
        std->states[iState].transitions = (stdTransition *) malloc (nProfiles * sizeof(stdTransition));

        // at creation, no transitions are defined
        // (they are added later using 'stdAddTransition()' method)
        std->states[iState].nTransitions = 0;
    }

    return std;
}


/*
 * stdDestroy
 * 
 * Destroys a given STD
 */
void stdDestroy(stdDiagram **std)
{
    int iState;

    // frees memory for all transitions
    for (iState = 0; iState < (*std)->nStates; iState++)
        free((*std)->states[iState].transitions);

    // frees memory for all state structures
    free((*std)->states);

    // frees memory for STD diagram itself and points it to NULL
    free(*std);
    *std = NULL;
}



/*
 * stdAddTransition
 * 
 * Adds a transition to an existing STD
 * Transition is defined for a given state so that, given a profile class, a
 * consequent state is specified.
 * This method returns false (zero) if the specified states or profile classes
 * are not contemplated within the STD (i.e., if indices are out of bounds)
 */
int stdAddTransition(stdDiagram *std, int iState, int iProfile, int iNextState)
{
    stdState *state;
    int iTrans;

    // checks if iState, iProfile and iNextState are not out of bounds
    if (iState < 0 || iState >= std->nStates)
        return 0;
    if (iProfile < 0 || iProfile >= std->nProfiles)
        return 0;
    if (iNextState < 0 || iNextState >= std->nStates)
        return 0;

    // retrieves corresponding state object
    state = &std->states[iState];

    // defines transition index: checks if there is already a transition defined
    // for the given profile class
    // - if there is, it is to be substituted
    // - if there isn't, then the transition is appended (index == nTransitions)
    for (iTrans = 0; iTrans < state->nTransitions; iTrans++)
        if (state->transitions[iTrans].iProfile == iProfile)
            break;

    // defines new transition
    state->transitions[iTrans].iProfile = iProfile;
    state->transitions[iTrans].iNextState = iNextState;

    // increments number of transitions if appropriate
    if (iTrans == state->nTransitions)
        state->nTransitions++;

    return 1;
}



/*
 * stdGetStateEvaluation
 * 
 * Given a transition evaluation for a state, returns the corresponding evaluation for
 * the following state, as specified by the STD.
 * 'transitionEval' should contain the evaluations of the transitions defined for the
 * given state; the evaluations are to be returned in the 'stateEval' array, which is
 * to be as large as there are states within the STD.
 * Obs: if for the given state there are two distinct transitions towards the same next
 *      state, those evaluations are merged (summed up)
 */
int stdGetStateEvaluation(stdDiagram *std, int iState, double *transitionEval, double *stateEval)
{
    stdState *state;
    int iTrans;
    int i;
    
    // checks if iState is not out of bounds
    if (iState < 0 || iState >= std->nStates)
        return 0;

    // retrieves corresponding state object
    state = &std->states[iState];

    // initializes all state evaluations to zero
    for (i = 0; i < std->nStates; i++)
        stateEval[i] = 0;

    // defines next state evaluations
    for (iTrans = 0; iTrans < state->nTransitions; iTrans++)
    {
        int iNextState = state->transitions[iTrans].iNextState;
        stateEval[iNextState] += transitionEval[iTrans];
    }
    return 1;
}
