#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"

#define round(x) (int)(floor((x)+0.5))

#define FNAMESIZE    1024


static char Ucmd[] =
"-ifile1 image1 -ifile2 image2";

static char Udetail[] =
"\nOutputs the mean reflectance value for the corresponding non-masked pixels\n\
of each georeferenced INRIMAGE image\n";



int inrPixelCompare(char *fnameImg1, char *fnameImg2)
{
  unsigned char *data1;
  unsigned char *data2;
  InrType type1;
  InrType type2;
  int size1_x, size1_y, size1_z;
  int size2_x, size2_y, size2_z;
  InrGeorefInfo georef1;
  InrGeorefInfo georef2;
  int z;
  int x1, y1;
  int x2, y2;
  
  double long_begin;
  double long_end;
  double lat_begin;
  double lat_end;
  
  int x2_begin;
  int x2_end;
  int y2_begin;
  int y2_end;
  
  int pos1;
  int pos2;
  double val1;
  double val2;
  
  double val2mean;
  int    val2nPixels;


  if (!inrReadGeoref(fnameImg1, &georef1) || !inrReadGeoref(fnameImg2, &georef2))
    return 0;
 
  data1 = inrReadDataSequence(fnameImg1, &size1_x, &size1_y, &size1_z, &type1);
  if (!data1)
  {
    fprintf(stderr, "Error: could not read data for image '%s'.\n", fnameImg1);
    return 0;
  }

  data2 = inrReadDataSequence(fnameImg2, &size2_x, &size2_y, &size2_z, &type2);
  if (!data2)
  {
    fprintf(stderr, "Error: could not read data for image '%s'.\n", fnameImg2);
    free(data1);
    return 0;
  }
  
  if (size1_z!=size2_z)
  {
    fprintf(stderr, "Error: image sizes in direction Z do not match!\n");
    return 0;
  }
  
  for (z=0; z<size1_z; z++)
  {
    for (y1=0; y1<size1_y; y1++)
    {
      for (x1=0; x1<size1_x; x1++)
      {
        long_begin = georef1.topleft_x + x1*georef1.pixelsize_x;
        long_end   = long_begin + georef1.pixelsize_x;
        lat_begin  = georef1.topleft_y - y1*georef1.pixelsize_y;
        lat_end    = lat_begin  - georef1.pixelsize_y;
        
        x2_begin = round((long_begin - georef2.topleft_x)/georef2.pixelsize_x);
        x2_end   = round((long_end   - georef2.topleft_x)/georef2.pixelsize_x);
        y2_begin = round((georef2.topleft_y -  lat_begin)/georef2.pixelsize_y);
        y2_end   = round((georef2.topleft_y -  lat_end  )/georef2.pixelsize_y);
        
        pos1 = x1 + y1*size1_x + z*size1_x*size1_y;
        val1 = inrGetDataValue(data1, type1, pos1);
        
        val2mean = 0;
        val2nPixels = 0;

        if (val1 <= 0)
          continue;
        
        if (x2_begin<0 || x2_end>size2_x || y2_begin<0 || y2_end>size2_y)
          continue;

        for (y2=y2_begin; y2<y2_end; y2++)
        {
          for (x2=x2_begin; x2<x2_end; x2++)
          {
            pos2 = x2 + y2*size2_x + z*size2_x*size2_y;
            val2 = inrGetDataValue(data2, type2, pos2);
            
            if (val2 <= 0)
              continue;
              
            val2mean += val2;
            val2nPixels++;
          }
        }
        if (val2nPixels < (y2_end-y2_begin)*(x2_end-x2_begin))
          continue;
          
        val2mean /= val2nPixels;
        printf("%10.5lf\t%10.5lf\n", val1, val2mean);
      }
    }
  }
  free(data1);
  free(data2);
  
  return 1;
}



int main(int argc, char *argv[])
{
  int      okFile1, okFile2;
  char     fnameImg1[FNAMESIZE];
  char     fnameImg2[FNAMESIZE];
  InrGeorefInfo georef1;
  InrGeorefInfo georef2;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okFile1 = igetopt1 ("-ifile1", "%s", fnameImg1);
  okFile2 = igetopt1 ("-ifile2", "%s", fnameImg2);
  
  if (!(okFile1 && okFile2))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  if (!inrReadGeoref(fnameImg1, &georef1) || !inrReadGeoref(fnameImg2, &georef2))
  {
    fprintf(stderr, "Error: could not read georeference info for the images.\n");
    return 1;
  }
    
  if (georef1.pixelsize_x >= georef2.pixelsize_x)
    return inrPixelCompare(fnameImg1, fnameImg2);
  else
    return inrPixelCompare(fnameImg2, fnameImg1);
}

