#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"

#define round(x) (int)(floor((x)+0.5))
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE    1024



static char Ucmd[]="-ifile classification_image -mask mask_file -ofile proportion_image -nclass num_classes\n";

static char Udetail[]= 
"Given an input classification image and a mask image, computes the\n\
proportions of each class within each pixel of the mask image. The output\n\
corresponds to a proportion image with nclass bands, where each band contains\n\
the proportions computed for each class.\n\
\nBoth input and mask images must be in INRIMAGE format with a corresponding\n\
ENVI file header (.hdr) containing georeference information\n\
\nInput:\n\
\t-ifile\tFile name of a georeferenced input classification image\n\
\t-mask\tFile name of a georeferenced mask image that defines the region\n\
\t\tof interest and the spatial resolution for proportion\n\
\t\tcomputation. Mask should be zero for pixels to be masked out\n\
\t\tand non-zero otherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output proportion image\n";




int main(int argc, char *argv[])
{
  int           okFileIn, okMask, okFileOut;
  char          fnameClassif[FNAMESIZE];
  char          fnameMask[FNAMESIZE];
  char          fnameOutput[FNAMESIZE];

  unsigned char *dataMask;
  unsigned char *dataClassif;
  float         *dataOut;
  InrType        typeMask;
  InrType        typeClassif;
  int            sizeMask_x, sizeMask_y, sizeMask_z;
  int            sizeClassif_x, sizeClassif_y, sizeClassif_z;
  InrGeorefInfo  georefMask;
  InrGeorefInfo  georefClassif;

  int x_mask, y_mask;
  int x_classif, y_classif;
  int nClass;
  int class;
  
  double long_begin;
  double long_end;
  double lat_begin;
  double lat_end;
  
  int x_classif_begin;
  int x_classif_end;
  int y_classif_begin;
  int y_classif_end;
  
  int posMask;
  int posClassif;
  int nPixels;


  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okFileIn  = igetopt1 ("-ifile" , "%s", fnameClassif);
  okMask    = igetopt1 ("-mask"  , "%s", fnameMask);
  okFileOut = igetopt1 ("-ofile" , "%s", fnameOutput);
  
  if (!(okFileIn && okMask && okFileOut))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);


  // reads georeference info from files
  if (!inrReadGeoref(fnameClassif, &georefClassif))
  {
    printf("Could not read georeference info for input file '%s'\n", fnameClassif);
    return 0;
  }
  if (!inrReadGeoref(fnameMask, &georefMask))
  {
    printf("Could not read georeference info for mask file '%s'\n", fnameMask);
    return 0;
  }

  // checks if datum/unit/proj(in case of UTM projection checks also UTM zone and hemisphere) info are the same for input and mask images
 
  if( !inrVerifGeorefInfo(georefMask,georefClassif))
  {
    fprintf(stderr,"\nFatal Error: datum/projection/unit information does not match!\n");
    printf("  Input file: '%s'\n", fnameClassif);
    printf("  Mask file : '%s'\n", fnameMask);
    return 0;
  }


  // reads data for input and mask files 
  dataClassif = inrReadDataSequence(fnameClassif, &sizeClassif_x, &sizeClassif_y, &sizeClassif_z, &typeClassif);
  if (!dataClassif)
  {
    fprintf(stderr, "Error: could not read data for input image '%s'.\n", fnameClassif);
    return 0;
  }
  dataMask = inrReadDataSequence(fnameMask, &sizeMask_x, &sizeMask_y, &sizeMask_z, &typeMask);
  if (!dataMask)
  {
    fprintf(stderr, "Error: could not read data for mask image '%s'.\n", fnameMask);
    free(dataClassif);
    return 0;
  }
  
  // checks if data is ok
  if (sizeMask_z!=1 || sizeClassif_z!=1)
  {
    fprintf(stderr, "Error: input and mask images should have size Z == 0!\n");
    return 0;
  }
  if (typeMask != InrByte)
  {
    fprintf(stderr, "Error: mask image should be of type 'Byte'.\n");
    return 0;
  }
  if (typeClassif!=InrByte && typeClassif!=InrUnsignedInt && typeClassif!=InrUnsignedLongInt)
  {
    fprintf(stderr, "Error: input image should be of type 'Byte', 'UnsignedInt' or 'UnsignedLongInt'!\n");
    return 0;
  }

  // verifies total number of classes in the classification image
  nClass = 0;
  for (posClassif=0; posClassif<sizeClassif_x*sizeClassif_y; posClassif++)
    nClass = MAX(nClass, (int)inrGetDataValue(dataClassif, typeClassif, posClassif));
  
  
  // allocates memory for output frame
  dataOut = (float *)malloc(sizeMask_x*nClass*sizeof(float));
  if (!dataOut)
  {
    fprintf(stderr, "Error: could not allocate memory for output image!\n");
    return 0;
  }
  
  // writes output file header
  if (!inrWriteHeader(fnameOutput, sizeMask_x, sizeMask_y, nClass, InrFloat))
  {
    fprintf(stderr, "Error: could not write data to output file '%s'.\n", fnameOutput);
    return 0;
  }
  
 
  for (y_mask=0; y_mask<sizeMask_y; y_mask++)
  {
    for (x_mask=0; x_mask<sizeMask_x; x_mask++)
    {
      posMask = x_mask + y_mask*sizeMask_x;

      // intializes output data to zero (0%)
      for (class=0; class<nClass; class++)
        dataOut[x_mask + class*sizeMask_x] = 0;
      
      // checks if pixel is not masked
      if (dataMask[posMask] == 0)
        continue;
      
      // if it is not masked, computes the geographical range of the pixel
      long_begin = georefMask.topleft_x + x_mask*georefMask.pixelsize_x;
      long_end   = long_begin + georefMask.pixelsize_x;
      lat_begin  = georefMask.topleft_y - y_mask*georefMask.pixelsize_y;
      lat_end    = lat_begin  - georefMask.pixelsize_y;
      
      // computes the corresponding initial and final indices in the
      // classification image
      x_classif_begin = round((long_begin - georefClassif.topleft_x)/georefClassif.pixelsize_x);
      x_classif_end   = round((long_end   - georefClassif.topleft_x)/georefClassif.pixelsize_x);
      y_classif_begin = round((georefClassif.topleft_y -  lat_begin)/georefClassif.pixelsize_y);
      y_classif_end   = round((georefClassif.topleft_y -  lat_end  )/georefClassif.pixelsize_y);
      

      // checks if indices are valid for the classification image
      // if indices out of bounds, output is zero (0%)
      if (x_classif_begin<0 || x_classif_end>sizeClassif_x || y_classif_begin<0 || y_classif_end>sizeClassif_y)
        continue;

      // computes total number of pixels in the given area
      nPixels = ((y_classif_end-y_classif_begin)*(x_classif_end-x_classif_begin));
      if (nPixels <= 0)
        continue;

      // computes the proportion of pixels in the classification image which
      // are labeled with each class
      for (y_classif=y_classif_begin; y_classif<y_classif_end; y_classif++)
      {
        for (x_classif=x_classif_begin; x_classif<x_classif_end; x_classif++)
        {
          posClassif = x_classif + y_classif*sizeClassif_x;
          class = (int)inrGetDataValue(dataClassif, typeClassif, posClassif);
          
          // if not unclassified (zero), increments counter at the right position
          // classification must also be within valid range (1..nClass)
          if (class>0 && class<=nClass)
            dataOut[x_mask + (class-1)*sizeMask_x]++;
        }
      }
      // normalizes proportions in output data buffer
      for (class=0; class<nClass; class++)
        dataOut[x_mask + class*sizeMask_x] /= nPixels;
    }

    // writes data to output file
    if (!inrWriteLine(fnameOutput, y_mask, (unsigned char *)dataOut))
    {
      fprintf(stderr, "Error: could not write data to output file '%s'.\n", fnameOutput);
      return 0;
    }
  }
  free(dataMask);
  free(dataClassif);
  free(dataOut);

  // writes output georeference info  
  inrWriteGeoref(fnameOutput, georefMask);

  fprintf(stderr, "Proportion image '%s' created successfully.\n", fnameOutput);
  
  return 1;
}

