#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024



static char Ucmd[] = "-ifile proportions_file -ofile purepixels_file [-min value]\n";

static char Udetail[] =
"Converts a proportions image to a classification image containing only pure\n\
pixels.\n\
The proportion image should contain one layer for each class, with each layer\n\
defining the percentage of that class within each pixel. The final output\n\
image will consist of a single layer with the label (index) of each pixel's\n\
class. All mixed pixels will be labeled as 'unclassified' (index 0) in the\n\
output image.\n\
\nInput:\n\
\t-ifile\tFile name of the input proportion image\n\
\nOutput:\n\
\t-ofile\tFile name of the output pure-pixels classification image\n\
\nOptional:\n\
\t-min\tMinimum proportion of the major class within a pixel in order\n\
\t\tto consider the pixel as 'pure'\n\
\t\tDefault value is 1\n";





int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  InrType       type;
  InrGeorefInfo georef;

  int    okPureMinProp;
  double pureMinProp;
  
  float         *proportions;
  unsigned char *classif;

  int size_x, size_y;
  int i;
  int class;
  int nClass;
  int major_class;
  
  double prop;
  double sum_prop;
  double major_prop;
  int    nPure, nTotal;
  


  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okPureMinProp = igetopt1 ("-min", "%lf", &pureMinProp);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okPureMinProp)
    pureMinProp = 1;

  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &nClass, &type))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (type != InrFloat)
  {
    fprintf(stderr, "Error: proportion image '%s' has to be a image of type 'Float'.\n", fnameInput);
    return 0;
  }
  if (nClass >= 255)
  {
    fprintf(stderr, "Error: proportion image '%s' has 255 classes or more - cannot convert to a classification image of type 'Byte'.\n", fnameInput);
    return 0;
  }
  

 
  // reads the proportions image
  proportions = inrReadFloatSequence (fnameInput, &size_x, &size_y, &nClass);
  if (!proportions)
  {
    fprintf (stderr, "Error: could not read proportion image data.\n");
    return 0;
  }

  // allocates memory for the output data
  classif = (unsigned char *) malloc (size_x*size_y * sizeof(unsigned char));
  


  // processes the data
  nPure = 0;
  nTotal = 0;
  for (i=0; i<size_x*size_y; i++)
  {
    // initializes pixel's classification to zero (unclassified)
    classif[i] = 0;

    // defines a classification if this is a pure-pixel
    sum_prop = 0;
    major_prop = 0;
    major_class = 0;
    for (class=0; class<nClass; class++)
    {
      prop = proportions[i + class*size_x*size_y];
      sum_prop += prop;

      if (prop > major_prop)
      {
        major_class = class+1;
        major_prop = prop;
      }
    }
    // disregards pixel if there are no proportions defined for it
    if (sum_prop == 0)
      continue;

    // counts pixel as pure if the proportion of the majority class reaches
    // the minimum
    if (major_prop >= pureMinProp) 
    {
      classif[i] = major_class;
      nPure++;
    }
    nTotal++;
  }
  

  // writes output classification data
  inrWriteByteImage (fnameOutput, size_x, size_y, classif);
  
  // copies georreference info if present
  if (inrReadGeoref(fnameInput, &georef))
    inrWriteGeoref(fnameOutput, georef);
  
  
  fprintf(stderr, "Proportions image successfully converted: %5.2lf%% of pure pixels.\n", ((double)nPure*100)/nTotal);

  free(proportions);
  free(classif);
  
  return 0;
}
