#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024
#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)



static char Ucmd[] = "-ifile proportions -set new_class_file -ofile new_proportions\n";

static char Udetail[] =
"Sets pixels in an INRIMAGE proportions image based on a text file.\n\
\nThe text file should contain the following information at each line:\n\
\t'x  y  new_class'\n\
  indicating that the pixel at position (x,y) should belong to new_class\n\
Pixel will have proportions = 1 for the new class, and 0 for all other\n\
classes.\n\
\nInput:\n\
\t-ifile\tFile name of the input classification proportions image\n\
\t\tImage should contain one layer for each class\n\
\t-set\tFile name of the text file defining the new classes\n\
\nOutput:\n\
\t-ofile\tFile name of the output proportions image\n";





int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameSet[FNAMESIZE];
  InrGeorefInfo georef;
  InrType       type;
  int           typesize;

  FILE *fpSet;

  unsigned char *data;

  int size_x, size_y, size_z;
  int x, y;
  int pos;
  int nClass;
  int class;
  int new_class;
  

  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-set"  , "%s", fnameSet))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &type))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  fpSet = fopen(fnameSet, "rt");
  if (!fpSet)
  {
    fprintf(stderr, "Error: could nor read text file '%s'.\n", fnameSet);
    return 0;
  }
 
  // reads the input classification image
  data = inrReadDataSequence(fnameInput, &size_x, &size_y, &size_z, &type);
  if (!data)
  {
    fprintf (stderr, "Error: could not read input file '%s'.\n", fnameInput);
    fclose(fpSet);
    return 0;
  }
  typesize = InrTypeSizes[type]/8;
  nClass = size_z;
  

  // processes the data
  while (!feof(fpSet))
  {
    int nitems = fscanf(fpSet, "%d %d %d", &x, &y, &new_class);
    if (nitems != 3)
      continue;
    if (x<0 || x>=size_x || y<0 || y>size_y)
      continue;

    if (new_class > nClass)
    {
      // reallocates memory for 'data' if new_class is beyond nClass
      unsigned char *old_data = data;
      data = realloc(old_data, size_x*size_y*new_class * typesize * sizeof(unsigned char));
      if (!data)
      {
        fprintf (stderr, "Error: could not reallocate memory to incorporate class %d.\n", new_class);
        free(old_data);
        return 0;
      }
      // initializes all new values to zero
      for (pos = nClass*size_x*size_y; pos < new_class*size_x*size_y; pos++)
        inrPutDataValue(data, type, pos, 0);
      
      nClass = new_class;
    }
      

    pos = x + y*size_x;
    
    for (class=0; class<nClass; class++)
       inrPutDataValue(data, type, pos + class*size_x*size_y, 0);

    inrPutDataValue(data, type, pos + (new_class-1)*size_x*size_y, 1);
  }
  fclose(fpSet);

  
  // writes output proportions image
  if (!inrWriteDataSequence(fnameOutput, size_x, size_y, nClass, data, type))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    free(data);
    return 0;
  }
  free(data);
  
  // copies georreference info if present
  if (inrReadGeoref(fnameInput, &georef))
    inrWriteGeoref(fnameOutput, georef);
  
  fprintf(stderr, "Proportions image pixels successfuly set.\n");
  return 1;
}
