#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024
#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)



static char Ucmd[] = "-ifile proportions -trans translation_file -ofile new_proportions\n";

static char Udetail[] =
"Translates a given input INRIMAGE proportions image, given a text translation\n\
file.\n\
\nThe translation file should contain the following information at each line:\n\
\t'input_class  output_class'\n\
indicating that all proportions of input_class should now belong to output_class\n\
\nInput:\n\
\t-ifile\tFile name of the input proportions image\n\
\t-trans\tFile name of the text file defining the class translations\n\
\nOutput:\n\
\t-ofile\tFile name of the output new proportions image\n";


// initializes list of class translations
int *initTransClassList(int nClass)
{
  int *transClassList;
  int i;
  
  transClassList = (int *) malloc (nClass * sizeof(int));
  for (i=0; i<nClass; i++)
    transClassList[i] = i+1;
    
  return transClassList;
}

// updates list of classes present in a list of class translations
void updateTransClassList(int *transClassList, int input_class, int output_class)
{
  transClassList[input_class-1] = output_class;
}

// returns maximum class index present in a list of class translations
int getMaxTransClassList(int *transClassList, int nClass)
{
  int i;
  int maxclass = transClassList[0];
  
  for (i=1; i<nClass; i++)
    maxclass = MAX(maxclass, transClassList[i]);
    
  return maxclass;
}



int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameTrans[FNAMESIZE];
  InrGeorefInfo georef;
  InrType       type;
  int           typesize;

  FILE *fpTrans;

  unsigned char *dataInput;
  unsigned char *dataOutput;
  
  int *transClassList;

  int size_x, size_y, size_z;
  int input_class, output_class;
  int input_class_pos, output_class_pos;
  int nClassInput;
  int nClassOutput;
  int i;
  
  double propCurr, propTrans;
  

  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-trans", "%s", fnameTrans))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &type))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  fpTrans = fopen(fnameTrans, "rt");
  if (!fpTrans)
  {
    fprintf(stderr, "Error: could nor read translation file '%s'.\n", fnameTrans);
    return 0;
  }
 
  // reads the input classification image
  dataInput = inrReadDataSequence(fnameInput, &size_x, &size_y, &size_z, &type);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not read input file '%s'.\n", fnameInput);
    fclose(fpTrans);
    return 0;
  }
  // output data starts as a copy of input data
  dataOutput = inrReadDataSequence(fnameInput, &size_x, &size_y, &size_z, &type);
  if (!dataOutput)
  {
    fprintf (stderr, "Error: could not read input file '%s'.\n", fnameInput);
    fclose(fpTrans);
    return 0;
  }
  typesize = InrTypeSizes[type]/8;
  nClassInput = nClassOutput = size_z;

  // initializes list of class translations
  transClassList = initTransClassList(nClassInput);
  

  // processes the data
  while (!feof(fpTrans))
  {
    int nitems = fscanf(fpTrans, "%d %d", &input_class, &output_class);
    if (nitems != 2)
      continue;
      
    // ignores translation if input_class is not valid
    if (input_class < 1 || input_class > nClassInput)
      continue;

    // ignores translation if output_class is not valid
    if (output_class < 1)
      continue;
      
    // extends output proportions image if output_class is beyond nClass
    if (output_class > nClassOutput)
    {
      // reallocates memory for 'dataOutput'
      dataOutput = realloc(dataOutput, size_x*size_y*output_class * typesize * sizeof(unsigned char));
      if (!dataOutput)
      {
        fprintf (stderr, "Error: could not reallocate memory to incorporate class %d.\n", output_class);
        return 0;
      }
      // initializes all new values to zero
      for (i = nClassOutput*size_x*size_y; i < output_class*size_x*size_y; i++)
        inrPutDataValue(dataOutput, type, i, 0);
      
      nClassOutput = output_class;
    }
      
    // updates list of classes present in output proportions  
    updateTransClassList(transClassList, input_class, output_class);

    // transfers proportions from input_class to output_class
    for (i=0; i<size_x*size_y; i++)
    {
      input_class_pos  = i + (input_class-1)*size_x*size_y;
      output_class_pos = i + (output_class-1)*size_x*size_y;
   
      // obtains value to be transferred from the input data   
      propTrans = inrGetDataValue(dataInput, type, input_class_pos);
   
      // updates proportion value for the input class in the output data
      // updated value is (current_value - transfer_value)
      propCurr = inrGetDataValue(dataOutput, type, input_class_pos);
      inrPutDataValue(dataOutput, type, input_class_pos, propCurr - propTrans);

      // updates proportions value for the output class in the output data
      // updated value is (current_value + transfer_value)
      propCurr = inrGetDataValue(dataOutput, type, output_class_pos);
      inrPutDataValue(dataOutput, type, output_class_pos, propCurr + propTrans);
    }
  }
  fclose(fpTrans);

  // verifies actual final number of classes in output file
  // number will be equivalent to the maximum class index present in the output
  // proportions image, which is given by the maximum number in the full
  // class translation list
  nClassOutput = getMaxTransClassList(transClassList, nClassInput);
  
  
  // writes output translated classification
  if (!inrWriteDataSequence(fnameOutput, size_x, size_y, nClassOutput, dataOutput, type))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  
  // copies georreference info if present
  if (inrReadGeoref(fnameInput, &georef))
    inrWriteGeoref(fnameOutput, georef);
  
  fprintf(stderr, "Proportions image successfully translated\n");

  free(dataInput);
  free(dataOutput);
  free(transClassList);
  return 1;
}
