/*
  visualisation of NDVI profiles
  from DLR sequences
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include "dirent-universal.h"
#include <inrimage/image.h>
#include "inr_io.h"
#include "file_tools.h"
#include "dlr.h"
#include "dlr_dates.h"

static char Ucmd[] = "\n"
"-ifile DLR_NOAANDVI sequence\n"
"-p proportion_image | -pix x y\n"
"-c class_num\n"
"[-m mask]\n"
"[-dates dateliste_file]\n"
"[-dmin date_min -dmax date_max]\n"
"[-pure seuil_purete]\n"
"[-stat a|m|s]\n"
"[-nodata value]";

static char Udetail[] =
"Prints on stdout date/NDVI couples from a DLR-NOAA NDVI sequence\n"
"Inputs:\n"
"\t-ifile: NDVI sequence, output of dlr_buildsequence (without -ndvi flag)\n"
"\t2 modes: pixel profiles, or class profiles.\n"
"Pixel profile mode:\n"
"\t-pix: activates pixel profile mode for pixel coordinates x and y\n"
"Class profile mode:\n"
"\t-p: activates class profile mode. Specifies the name of a proportion image\n"
"\tthe proportion image is obtained from a HR classification (see xxx)\n"
"\t-c: specifies the class number\n"
"Inputs (optionnal):\n"
"\t-dates (all modes): specifies date file. If absent date=number of frame\n"
"\t-stat (class mode): a for all pixels of class, m for mean, s for mean +/- std dev\n"
"\t-pure (class mode): proportion threshold for pure pixels (default 0.9)\n"
"\t-m (class mode): specifies a mask of the region of interest (0 outside ROI)\n"
"\t-nodata (all modes): nodata code (default -9999)\n";



int main (int argc, char **argv)
{
  char nom_sequence[200],nom_prop[200],datefile[200],nom_mask[200] ; 
  int classnum , nclass ;
  unsigned char *buf_mask=NULL ;
  float * bufndvi=NULL , *bufprop=NULL ;
  int i,t,sx,sy,sxx,syy,sz;
  int *dates , *outdates, datemode;
  int *bufpur = NULL ;
  int dmin,dmax,idmin,idmax;
  float pure ;
  int npur,nn;
  char affichage ;
  float moy,var;
  int ismaskimage;
  float nodata ;
  int modeprop,modepix ;
  int pixi,pixj;


  inr_init( argc, argv , "1.0", Ucmd, Udetail ) ;

  // options
  if ( !igetopt1("-ifile","%s",nom_sequence) ) 
    iusage_(Ucmd,Udetail) ;
  modeprop = igetopt1("-p","%s",nom_prop) ;
  modepix = igetopt2("-pix","%d",&pixi,"%d",&pixj) ;
  if (!modeprop && !modepix ) {
    fprintf(stderr,"at least one of -p or -pix must be present\n");
    iusage_(Ucmd,Udetail);
  }
  if (modepix&&modeprop) {
    fprintf(stderr,"-p and -pix are incompatible\n");
    iusage_(Ucmd,Udetail);
  }

  if ( modeprop && !igetopt1("-c","%d",&classnum) ) 
    iusage_(Ucmd,Udetail) ;
  
  if ( !igetopt1("-nodata","%f",&nodata) ) nodata=-9999 ;

   bufndvi = inrReadFloatSequence( nom_sequence, &sx, &sy, &sz ) ;
   if (!bufndvi) { 
    fprintf(stderr,"Error: cannot read image '%s'.\n",nom_sequence);
      return 1;
   }
   
  // lecture dates
  dates=(int*)malloc(sz*sizeof(int));
  outdates=(int*)malloc(sz*sizeof(int));
  if ( !dates || !outdates ) {
    fprintf(stderr,"Error: memory allocation failed\n");
    return 1;
  }

  if ( igetopt1("-dates","%s",datefile) ) {
    datemode=1;
    FILE *fd ;
    fd = fopen( datefile , "r" ) ;
    if ( !fd ) {
      fprintf(stderr,"Error: cannot open date file %s\n",datefile) ;
      return 1;
    }
    for (i=0;i<sz;i++) {
      if ( fscanf(fd,"%d",&dates[i]) != 1 ) {
	fprintf(stderr,"Error reading in %s\n",datefile) ;
	return 1;
      }
    }
  }
  else {
    datemode=0;
    for (i=0;i<sz;i++) dates[i]=i;
  }
  if (!igetopt1("-dmin","%d",&dmin)) dmin=dates[0] ;
  if (!igetopt1("-dmax","%d",&dmax)) dmax=dates[sz-1] ;
  if ( dmin < dates[0] ) dmin = dates[0] ;
  if ( dmax > dates[sz-1] ) dmax = dates[sz-1] ;
  if ( dmax<=dmin) {
    fprintf(stderr,"warning, all dates considered\n");
    dmin=dates[0];
    dmax=dates[sz-1];
  }
  for (idmin=0; dmin>dates[idmin];idmin++) ;
  for (idmax=sz-1; dmax<dates[idmax];idmax--) ;
  // compute outdates in days
   if ( !datemode ) {
    for (t=idmin;t<=idmax;t++) outdates[t]=dates[t];
     }
   else
    setoutdates(dates,outdates,idmin,idmax);
  

  /* CLASS MODE */
  if (modeprop) {
    // read proportion image

    bufprop = inrReadFloatSequence( nom_prop, &sxx, &syy, &nclass );
    if (!bufprop) {
      fprintf(stderr,"Error: cannot read proportion image '%s'.\n",nom_prop);
      return 1;
    }
    //check dimensions
    if (sxx!=sx||syy!=sy) {
      fprintf(stderr,"Error: proportion image should be %d x %d pixels\n",sx,sy);
      return 1;
    }
    if (classnum>nclass||classnum<1) {
      fprintf(stderr,"Error: classnum (-c) should lie between 1 and %d\n",nclass);
      return 1;
    }

    // mask
    if ( igetopt1("-m","%s",nom_mask) ) {
      buf_mask = inrReadByteImage(nom_mask,&sxx,&syy);
      if ( sxx != sx || syy != sy ) {
	fprintf(stderr,"Error: mask image should be %d x %d pixels\n",sx,sy);
	return 1;
      } 
      ismaskimage=1;
    }
    else {
      buf_mask = (unsigned char*)malloc(sx*sy) ;
      if (!buf_mask) {
	fprintf(stderr,"Error: memory allocation failed\n");
	return 1;
      }
      for (i=0;i<sx*sy;i++) buf_mask[i]=255 ;
      ismaskimage=0;
    }
    

    // seuil purete
    if (!igetopt1("-pure","%f",&pure) ) pure=0.9 ;
    if (pure <0 || pure>1 ) pure=0.9 ;
    
    
    // display
    if ( !igetopt1("-stat","%c",&affichage) ) affichage=MOYENNE ;
    if ( affichage!=ALLPIX && affichage != MOYENNE && affichage != MOYSTD ) 
      affichage = MOYENNE ;
    
    
    // parcours de l'image de proportions pour selection des pixels purs
    classnum--;
    npur=0;
    bufpur=(int*)malloc(sx*sy*sizeof(int));
    if (!bufpur){
      fprintf(stderr,"Error: memory allocation failed\n");
      return 1;
    }
    for (i=0;i<sx*sy;i++) {
      if (bufprop[i+sx*sy*classnum]>=pure && buf_mask[i] ) bufpur[npur++]=i;
    }
    fprintf(stderr,"Class %d: %d pure pixels (proportion above %f)\n",classnum+1,npur,pure);
    if (!npur) return 0;
    
    // affichage
    switch (affichage) {
    case ALLPIX: // on garde les nodata pour nbre constant de colonnes
      for (t=idmin;t<=idmax;t++) {
	printf("%d",outdates[t]);
	for (i=0;i<npur;i++) 
	  printf(" %f",bufndvi[bufpur[i]+sx*sy*t]);
	printf("\n");
	
      }
      break;
    case MOYENNE: // il faut supprimer les nodata

      for (t=idmin;t<=idmax;t++) {
	moy=0.;nn=0 ;
	for (i=0;i<npur;i++) if (bufndvi[bufpur[i]+sx*sy*t]!=nodata){
	  moy += bufndvi[bufpur[i]+sx*sy*t];
	  nn++;
	}
	if (nn) {
	  moy/=(float)nn;
	  printf("%d %f\n",outdates[t],moy);
	}
      }
      break;
    case MOYSTD: 
      for (t=idmin;t<=idmax;t++) {
	moy=0.;
	nn=0;
	for (i=0;i<npur;i++) if (bufndvi[bufpur[i]+sx*sy*t]!=nodata){
	  moy += bufndvi[bufpur[i]+sx*sy*t];
	  nn++;
	}
	if (nn) 
	  moy/=(float)nn;
	
	var=0.;
	for (i=0;i<npur;i++) if (bufndvi[bufpur[i]+sx*sy*t]!=nodata)
	  var += SQR( bufndvi[bufpur[i]+sx*sy*t] - moy ) ;
	var /= (float)(nn);
	var = sqrt(var);
	if (nn)
	  printf("%d %f %f %f\n",outdates[t],moy-var,moy,moy+var);
	
      }
    }
  } // fin mode proportion
  
  else { /* PIXEL MODE */
    // modepixel
    int coord ;
    // check that pixi and pixj are within limits
    if ( pixi <1 || pixi>sx ) {
      fprintf(stderr,"error: x coordinate should be between %d and %d inclusive\n",1,sx);
      exit(1);
    }    
    if ( pixj <1 || pixj>sy ) {
      fprintf(stderr,"error: y coordinate should be between %d and %d inclusive\n",1,sy);
      exit(1);
    }
    pixi --;
    pixj -- ;


    for (t=idmin;t<=idmax;t++) {
      coord = pixi+sx*pixj +sx*sy*t ;
      if ( bufndvi[coord] != nodata )
	printf("%d %f\n",outdates[t],bufndvi[coord]) ;
    }
    
  }
  free( dates ) ;
  free( outdates ) ;
  free( bufndvi ) ;
  
  if (modeprop) {
    free( bufpur ) ;
    free( bufprop ) ;
    free( buf_mask ) ; 
  }
  
  return 0 ;
}
