#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE             1024
#define MINCOVERAGE_DEFAULT   20


static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir [-coverage min_percentage] [-ifilter string]";

static char Udetail[] = 
"\nConverts MOD09GQK 250m orbit/coverage images to masks that indicate which\n\
data can be considered useful.\n\
\nThe output mask will have value 1 for pixels that present observation coverage\n\
equal or greater than a given minimum percentage, and 0 otherwise. Observation\n\
coverage is defined as the area of intersection of observation footprint and\n\
cell divided by the area of the observation\n\
Input images must be in INRIMAGE format.\n\
\nInput:\n\
\t-idir\tDirectory containing MOD09GQK quality images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD09GQK quality image in INRIMAGE format\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted masks\n\
\nOptional:\n\
\t-coverage\tMinimum percentage criterion for obervation coverage\n\
\t\t\tDefault value is 20%\n\
\t-ifilter\tFilter string for identifying MOD09GQK quality images\n\
\t\t\tDefault identifier is 'orb_cov'\n";


/*
    MOD09GQK orb_cov_1 specification
    NOTE: bits are numbered in the opposite way by this official specification!!

    Bit     Description
    
    0-3     Orbit number: 4 bit unsigned integer; range 0 to 13
            key: 0000 ( 0) to 1011 (13)
    4       Scan half flag; which half of the scan the observation
             came from:
            key: 0 - top half
                 1 - bottom half
    5-7     Observation coverage: 3 bit unsigned integer;
            range 0 to 7
            key: 000 (0) -  0.0 to  12.5%
                 001 (1) - 12.5 to  25.0%
                 010 (2) - 25.0 to  37.5%
                 011 (3) - 37.5 to  50.0%
                 100 (4) - 50.0 to  62.5%
                 101 (5) - 62.5 to  75.0%
                 110 (6) - 75.0 to  87.5%
                 111 (7) - 87.5 to 100.0%
*/


/*
 *  modisCovertGQKFile
 *
 *  
 *  
 */
int modisCovertGQKFile(char *fnameFull, char *outfnameFull, double minCoverage)
{
  int             size_x, size_y, size_z;
  int             i;
  int             nData;
  InrType         type;   
  InrGeorefInfo   georef;   
  int             hasGeoref;
  unsigned char  *data_in;
  unsigned char  *data_out;
  

  // retrieves input file dimensions and georeference information
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return 0;
  hasGeoref = inrReadGeoref(fnameFull, &georef);
    
  // checks if data_in type is correct for a MOD09GQK orbit/coverage file
  if (type != InrByte)
  {
    fprintf(stderr, "Error: data type is not 'Byte' for file '%s'.\n", fnameFull);
    return 0;
  }
  
  // retrieves input file data
  data_in = inrReadByteSequence(fnameFull, &size_x, &size_y, &size_z);
  if (!data_in)
  {
    fprintf(stderr, "Error: could not read data for file '%s'.\n", fnameFull);
    return 0;
  }
    
  // allocates memory for output file data
  nData = size_x*size_y*size_z;
  data_out = (unsigned char *)malloc(nData * sizeof(unsigned char));
  if (!data_out)
  {
    fprintf(stderr, "Error: could not allocate memory.\n");
    free(data_in);
    return 0;
  }
  

  for (i=0; i<nData; i++)
  {
    int mask_val = 1;
     
    // masked out portions of the image have value zero
    if (data_in[i] == 0)
      mask_val = 0;
     
    // areas with value 15 are no-data areas: coverage is zero!
    else if (data_in[i] == 15)
      mask_val = 0;
       
    else
    {
      // checks the observation coverage percentage for the pixel (bits 5-7)
      // coverage values correspond to indices from 0 to 7, with 0 meaning
      // 0%-12% coverage and 7 meaning 87.5%-100% of coverage for the pixel
      int    index = data_in[i]>>5;
      double coverage = 0.125*index + 0.0625;
       
      // if coverage is less than minimum, pixel is masked out
      if (coverage < minCoverage)
        mask_val = 0;
      else
        mask_val = 1;
    }
       
    data_out[i] = mask_val;
  }
  free(data_in);
  
  // writes output file
  if (!inrWriteByteSequence(outfnameFull, size_x, size_y, size_z, data_out))
  {
    fprintf(stderr, "Error: could not write to output file '%s'.\n", outfnameFull);
    free(data_out);
    return 0;
  }
  free(data_out);

  // writes georeference information, if applicable
  if (hasGeoref)
    inrWriteGeoref(outfnameFull, georef);
  
  return 1;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  char   strFilter[FNAMESIZE];
  double minCoverage;
  int    okDirIn, okDirOut, okFileIn, okCoverage, okFilter;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"    , "%s" , dirInput);
  okDirOut      = igetopt1("-odir"    , "%s" , dirOutput);
  okFileIn      = igetopt1("-ifile"   , "%s" , fnameInput);
  okFilter      = igetopt1("-ifilter" , "%s" , strFilter);
  okCoverage    = igetopt1("-coverage", "%lf", &minCoverage);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (!okDirOut)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // specifies default file name filter if necessary
  if (!okFilter)
    strcpy(strFilter, "orb_cov");

  // specifies default minimum observation coverage percentage if necessary 
  if (!okCoverage)
    minCoverage = MINCOVERAGE_DEFAULT;
  minCoverage /= 100;

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // processes file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be cropped
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input and output file names (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);
    sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);
    ftChangeExtension(outfnameFull, outfnameFull, "");
    ftChangeExtension(outfnameFull, outfnameFull, ".coveragemask.inr");

    // checks if input is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // checks if input is a MOD09GQK orbit/coverage file
    if (!strstr(ent->d_name, strFilter))
      continue;

    // converts file
    if (!modisCovertGQKFile(fnameFull, outfnameFull, minCoverage))
      continue;
    
    // increments counter
    printf("Converted file '%s'\n", ent->d_name);
    converted++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d MOD09GQK coverage file(s) converted.\n", converted);
  return 1;
}

