#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE             1024
#define MINQUALITY_DEFAULT    40
#define MINCOVERAGE_DEFAULT   20

static char Ucmd[] =
"-idir input_dir [-ifile input_file] -ofile badlist_file [-quality min_percentage] \
[-coverage min_percentage]";

static char Udetail[] = 
"\nProcesses MODIS MOD09GQK images in order to evaluate which data is not\n\
appropriate for further analysis\n\
\nImages must be in INRIMAGE format, one file per band, and should be named\n\
according to the Modis Reprojection Tool nomenclature for HDF-RAW files (.hdr)\n\
Criteria used for evaluation is:\n\
1. A minimum percentage of good quality of pixels, as specified by the quality\n\
   control bit flag band (eg. QC_250m)\n\
2. A minimum percentage for mean observation coverage, as specified by the\n\
   orbit/coverage band (orb_cov). Observation coverage is defined as the area\n\
   of intersection of observation footprint and cell divided by the area of\n\
   the observation\n\
\nInput:\n\
\t-idir\tDirectory containing MOD09GQK images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD09GQK image (note that this will correspond to\n\
\t      \tseveral files, one for each band)\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nInput(optional):\n\
\t-quality\tMinimum percentage criterion for good quality pixels.\n\
\t        \tDefault value is 40%\n\
\t-coverage\tMinimum percentage criterion for mean obervation coverage\n\
\t        \tDefault value is 20%\n\
\nOutput:\n\
\t-ofile\tOutput text file containing all images rejected by the evaluation\n";




/*
 *  modisCheckQuality
 *
 *  Checks the quality bit flag to evaluate if the image is acceptable
 *  Quality bit flag should be 4096 for clear pixels with ideal atmospheric
 *  correction
 */
double modisCheckQuality(char *fnameFull)
{
  int     size_x, size_y, size_z;
  int     x, y;
  int     nData;
  int     nGood;
  InrType type;
  unsigned short *data;

  // reads input file file header
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return -1;

  // checks if data type is correct for a quality bit flag
  if (type != InrUnsignedInt || size_z != 1)
    return -1;
  
  // retrieves input file data
  data = (unsigned short *)inrReadDataSequence(fnameFull, &size_x, &size_y, &size_z, &type);
  if (!data)
    return -1;

  nData = nGood = 0;
  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
       int pos = x + y*size_x;
       
       // masked out portions of the image have value zero
       if (data[pos] == 0)
         continue;
       
       // checks if pixel is bad or not  
       if (data[pos] == 4096)
         nGood++;
         
       nData++;
    }
  }
  free(data);
 
  // returns percentage of good pixels in image
  return (((double)nGood)/nData);
}


/*
 *  modisCheckCoverage
 *
 *  Computes the mean observation coverage to evaluate if an image is acceptable
 *  Coverage corresponds to the area of intersection of observation footprint
 *  and cell divided by the area of the observation
 *  Coverage values correspond to indices from 0 to 7, with 0 meaning 0%-12%
 *  coverage and 7 meaning 87.5%-100% of coverage for the pixel
 */
double modisCheckCoverage(char *fnameFull)
{
  int     size_x, size_y, size_z;
  int     x, y;
  int     nData;
  InrType type;
  double  meanCoverage;
  unsigned char *data;

  // reads input file file header
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return -1;

  // checks if data type is correct for a observation/coverage file
  if (type != InrByte || size_z != 1)
    return -1;
  
  // retrieves input file data
  data = inrReadByteImage(fnameFull, &size_x, &size_y);
  if (!data)
    return -1;

  nData = 0;
  meanCoverage = 0;
  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
       int index;
       int pos = x + y*size_x;
       
       // masked out portions of the image have value zero
       if (data[pos] == 0)
         continue;
       
       // areas with value 15 are no-data areas: coverage is zero!
       if (data[pos] == 15)
       {
         nData++;
         continue;
       }
       
       // adds up the estimated mean coverage for the pixel (bits 5-7)
       index = data[pos]>>5;
       meanCoverage += 0.125*index + 0.0625;
       nData++;
    }
  }
  free(data);
 
  // divides by total number of pixels to compute mean and returns
  meanCoverage /= nData;
  return meanCoverage;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   fnameOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  double minQuality;
  double minCoverage;
  int    okDirIn, okFileIn, okFileOut, okQuality, okCoverage;
  FILE  *fpOut;
  DIR   *dir;
  struct dirent *ent;
  int    nBadQuality = 0;
  int    nBadCoverage = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  ,   "%s",  dirInput);
  okFileIn      = igetopt1("-ifile" ,   "%s",  fnameInput);
  okFileOut     = igetopt1("-ofile" ,   "%s",  fnameOutput);
  okQuality     = igetopt1("-quality",  "%lf", &minQuality);
  okCoverage    = igetopt1("-coverage", "%lf", &minCoverage);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (!okFileOut)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // defines threshold values
  if (!okQuality)
    minQuality = MINQUALITY_DEFAULT;
  if (!okCoverage)
    minCoverage = MINCOVERAGE_DEFAULT;
  minQuality /= 100;
  minCoverage /= 100;

  // opens output file
  fpOut = fopen(fnameOutput, "wt");
  if (!fpOut)
    printf("Error: unable to write to output file '%s'\n", fnameOutput);

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens output directory
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    fclose(fpOut);
    return 0;
  }

  printf("\nProcessing files with the following options:\n");
  printf("- Minimum quality criterion:  %5.2lf%% of good quality pixels\n", minQuality*100);
  printf("- Minimum coverage criterion: %5.2lf%% of observation coverage\n\n", minCoverage*100);

  // processes file(s) in the directory
  nBadQuality = nBadCoverage = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char basename[FNAMESIZE];
    double pGoodQuality = 0;
    double meanCoverage = 0;

    // checks if file is the one selected to be processed
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input file name (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);

    // checks if it is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // obtains basename by taking out the last two extensions (eg. '.QC_250m.inr')
    ftChangeExtension(ftChangeExtension(ent->d_name, basename, ""), basename, "");

    // checks if it is a quality bit flag file or an orbit/coverage file
    if (strstr(ent->d_name, "QC") || strstr(ent->d_name, "qc"))
    {
      // quality file
      // checks if image quality is acceptable (mostly for clouds)
      pGoodQuality = modisCheckQuality(fnameFull);
      if (pGoodQuality < minQuality)
      {
        // adds file to bad list
        fprintf(fpOut, "%s\n", basename);
        printf("REJECTED file '%s': %6.2lf%% of good quality pixels.\n", basename, pGoodQuality*100);
        nBadQuality++;
      }
      else
        printf("Accepted file '%s': %6.2lf%% of good quality pixels.\n", basename, pGoodQuality*100);
    }
    else if (strstr(ent->d_name, "ORB") || strstr(ent->d_name, "orb"))
    {
      // orbit/coverage file
      // checks if image coverage is acceptable (depends on the view angle)
      meanCoverage = modisCheckCoverage(fnameFull);
      if (meanCoverage < minCoverage)
      {
        // adds file to bad list
        fprintf(fpOut, "%s\n", basename);
        printf("REJECTED file '%s': %6.2lf%% of observation coverage.\n", basename, meanCoverage*100);
        nBadCoverage++;
      }
      else
        printf("Accepted file '%s': %6.2lf%% of observation coverage.\n", basename, meanCoverage*100);
    }
  }

  // closes directory and output file
  closedir(dir);
  fclose(fpOut);
    
  printf("%d MODIS images with bad quality in bad image list.\n", nBadQuality);
  printf("%d MODIS images with bad coverage in bad image list.\n", nBadCoverage);
  return 1;
}

