#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE             1024


static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir [-ifilter string]";

static char Udetail[] = 
"\nConverts MOD09GQK 250m quality images to masks that indicate which data can\n\
be considered useful.\n\
\nThe output mask will have value 1 for good quality, clear-sky pixels, and 0\n\
otherwise, as indicated by the quality flags.\n\
Namely, the criteria used for evaluating a good pixel is the following:\n\
1. Bits 0-1  must be '00'    (ideal quality all bands)\n\
2. Bits 2-3  must be '00'    (cloud state: clear)\n\
3. Bits 4-7  must be '0000'  (band 1 highest data quality)\n\
4. Bits 8-11 must be '0000'  (band 2 highest data quality)\n\
5. Bit  12   must be '1'     (atmospheric correction performed)\n\
(being 0 the least significant bit):\n\
Input images must be in INRIMAGE format.\n\
\nInput:\n\
\t-idir\tDirectory containing MOD09GQK quality images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD09GQK quality image in INRIMAGE format\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted masks\n\
\nOptional:\n\
\t-ifilter\tFilter string for identifying MOD09GQK quality images\n\
\t\t\tDefault identifier is 'QC_250'\n";


/*
    MOD09GQK QC_250m_1 specification
    NOTE: bits are numbered in the opposite way by this official specification!!
    
    Bits are numbered from left (MSB) to right (LSB):
    Bit    Description
    0-1    spare (unused);
    2      adjacency correction performed;
           1 -- yes
           0 -- no
    3      atmospheric correction performed;
           1 -- yes
           0 -- no
    4-7    band 2 data quality four bit range;
           0000 -- highest quality
           1000 -- dead detector; data interpolated in L1B
           1001 -- solar zenith >= 86 degrees
           1010 -- solar zenith >= 85 and < 86 degrees
           1011 -- missing input
           1100 -- internal constant used in place of climatological data
                   for at least one atmospheric constant
           1101 -- correction out of bounds pixel constrained to extreme
                   allowable value
           1110 -- L1B data faulty
           1111 -- not processed due to deep ocean or clouds
    8-11   band 1 data quality four bit range;
           SAME AS ABOVE
    12-13  cloud state;
           00 -- clear
           01 -- cloudy
           10 -- mixed
           11 -- not set, assumed clear
    14-15  MODLAND QA bits;
           corrected product produced at
           00 -- ideal quality all bands
           01 -- less than ideal quality some or all bands
           corrected product not produced due to
           10 -- cloud effects all bands
           11 -- other reasons some or all bands may be fill value
           [Note that a value of (11) overrides a value of (01)].

*/


/*
 *  modisCovertGQKFile
 *
 *  
 *  
 */
int modisCovertGQKFile(char *fnameFull, char *outfnameFull)
{
  int             size_x, size_y, size_z;
  int             i;
  int             nData;
  InrType         type;   
  InrGeorefInfo   georef;   
  int             hasGeoref;
  unsigned short *data_in;
  unsigned char  *data_out;
  

  // retrieves input file dimensions and georeference information
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return 0;
  hasGeoref = inrReadGeoref(fnameFull, &georef);
    
  // checks if data_in type is correct for a MOD09GQK quality file
  if (type != InrUnsignedInt)
  {
    fprintf(stderr, "Error: data type is not 'UnsignedInt' for file '%s'.\n", fnameFull);
    return 0;
  }
  
  // retrieves input file data
  data_in = (unsigned short *)inrReadDataSequence(fnameFull, &size_x, &size_y, &size_z, &type);
  if (!data_in)
  {
    fprintf(stderr, "Error: could not read data for file '%s'.\n", fnameFull);
    return 0;
  }
    
  // allocates memory for output file data
  nData = size_x*size_y*size_z;
  data_out = (unsigned char *)malloc(nData * sizeof(unsigned char));
  if (!data_out)
  {
    fprintf(stderr, "Error: could not allocate memory.\n");
    free(data_in);
    return 0;
  }
  

  for (i=0; i<nData; i++)
  {
     int mask_val = 1;
     
     // masked out portions of the image have value zero
     if (data_in[i] == 0)
       mask_val = 0;
     
     // checks criterion 1: bits 0-1  must be '00'    (ideal quality all bands)
     else if ((data_in[i] & 3) != 0)
       mask_val = 0;
       
     // checks criterion 2: bits 2-3  must be '00'    (cloud state: clear)
     else if ((data_in[i] & 3<<2) != 0)
       mask_val = 0;

     // checks criterion 3: bits 4-7  must be '0000'  (band 1 highest data quality )
     else if ((data_in[i] & 15<<4) != 0)
       mask_val = 0;

     // checks criterion 4: bits 8-11 must be '0000'  (band 2 highest data quality )
     else if ((data_in[i] & 15<<8) != 0)
       mask_val = 0;

     // checks criterion 5: bit  12   must be '1'     (atmospheric correction performed)
     else if ((data_in[i] & 1<<12) != 1<<12)
       mask_val = 0;

     // pixel passed all tests: considered to be a good quality pixel
     else
       mask_val = 1;
       
     data_out[i] = mask_val;
  }
  free(data_in);
  
  // writes output file
  if (!inrWriteByteSequence(outfnameFull, size_x, size_y, size_z, data_out))
  {
    printf("Error: could not write to output file '%s'.\n", outfnameFull);
    free(data_out);
    return 0;
  }
  free(data_out);

  // writes georeference information, if applicable
  if (hasGeoref)
    inrWriteGeoref(outfnameFull, georef);
  
  return 1;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  char   strFilter[FNAMESIZE];
  int    okDirIn, okDirOut, okFileIn, okFilter;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"   , "%s", dirInput);
  okDirOut      = igetopt1("-odir"   , "%s", dirOutput);
  okFileIn      = igetopt1("-ifile"  , "%s", fnameInput);
  okFilter      = igetopt1("-ifilter", "%s", strFilter);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (!okDirOut)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // specifies default file name filter if necessary
  if (!okFilter)
    strcpy(strFilter, "QC_250");

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // processes file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be cropped
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input and output file names (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);
    sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);
    ftChangeExtension(outfnameFull, outfnameFull, "");
    ftChangeExtension(outfnameFull, outfnameFull, ".qualitymask.inr");

    // checks if input is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // checks if input is a MOD09GQK quality file
    if (!strstr(ent->d_name, strFilter))
      continue;

    // converts file
    if (!modisCovertGQKFile(fnameFull, outfnameFull))
      continue;
    
    // increments counter
    printf("Converted file '%s'\n", ent->d_name);
    converted++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d MOD09GQK quality file(s) converted.\n", converted);
  return 1;
}

