#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "modis.h"
#include "file_tools.h"
#include "str_tools.h"

#define FLISTSIZE             65536
#define FNAMESIZE             1024
#define MINCLEARSKY_DEFAULT   40
#define MINCOVERAGE_DEFAULT   20

static char Ucmd[] =
"-idir input_dir [-ifile input_file] -ofile badlist_file [-clearsky min_percentage] \
[-coverage min_percentage]";

static char Udetail[] = 
"\nProcesses MODIS MOD09GST images in order to evaluate which data is not\n\
appropriate for further analysis\n\
\nImages must be in INRIMAGE format, one file per band, and should be named\n\
according to the Modis Reprojection Tool nomenclature for HDF-RAW files (.hdr)\n\
Criteria used for evaluation is:\n\
1. A minimum percentage of clear sky pixels, as specified by the state band.\n\
   Clear sky pixels will also consider areas of cloud shadow as well as areas\n\
   of high aerosol and cirrus.\n\
   Namely, the criteria used for evaluating a clear pixel is the following:\n\
     1. Bits 0-1 must be '00'     (clear cloud condition)\n\
     2. Bit  2   must be '0'      (no cloud shadow detected)\n\
     3. Bits 6-7 must not be '11' (high aerosol quantity)\n\
     4. Bits 8-9 must not be '11' (high cirrus detected)\n\
     5. Bit  10  must be '0'      (no clouds detected - internal flag)\n\
     (being 0 the least significant bit):\n\
2. A minimum percentage for mean observation coverage, as specified by the\n\
   orbit/coverage band (orb_cov). Observation coverage is defined as the area\n\
   of intersection of observation footprint and cell divided by the area of\n\
   the observation\n\
\nInput:\n\
\t-idir\tDirectory containing MOD09GST images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD09GST band image\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nInput(optional):\n\
\t-clearsky\tMinimum percentage criterion for clear sky pixels.\n\
\t        \tDefault value is 40%\n\
\t-coverage\tMinimum percentage criterion for mean obervation coverage\n\
\t        \tDefault value is 20%\n\
\nOutput(optional):\n\
\t-ofile\tOutput text file containing all images accepted by the evaluation\n\
\t-ofilebad\tOutput text file containing all images rejected by the evaluation\n";




/*
 *  modisCheckClearSky
 *
 *  Checks the state band to evaluate the percentage of clear sky pixels
 *  State bit flag should be 4096 for clear pixels with ideal atmospheric
 *  correction
 */
double modisCheckClearSky(char *fnameFull)
{
  int     size_x, size_y, size_z;
  int     x, y;
  int     nData;
  int     nClearSky;
  InrType type;
  unsigned short *data;

  // reads input file file header
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return -1;

  // checks if data type is correct for a state band
  if (type != InrUnsignedInt || size_z != 1)
    return -1;
  
  // retrieves input file data
  data = (unsigned short *)inrReadDataSequence(fnameFull, &size_x, &size_y, &size_z, &type);
  if (!data)
    return -1;

  nData = nClearSky = 0;
  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
       int pos = x + y*size_x;
       int clearsky;
       
       // masked out portions of the image have value zero
       if (data[pos] == 0)
         continue;

       // checks criterion 1: bits 0-1 must be '00'     (clear cloud condition)
       else if ((data[pos] & 3) != 0)
         clearsky = 0;
         
       // checks criterion 2: bit  2   must be '0'      (no cloud shadow detected)
       else if ((data[pos] & 1<<2) != 0)
         clearsky = 0;

       // checks criterion 3: bits 6-7 must not be '11' (high aerosol quantity)
       else if ((data[pos] & 3<<6) == 3<<6)
         clearsky = 0;

       // checks criterion 4: bits 8-9 must not be '11' (high cirrus detected)
       else if ((data[pos] & 3<<8) == 3<<8)
         clearsky = 0;

       // checks criterion 5: bit  10  must be '0'      (no clouds detected - internal flag)
       else if ((data[pos] & 1<<10) != 0)
         clearsky = 0;
         
       // pixel passed all tests: considered to be a clear sky pixel
       else
         clearsky = 1;

       // checks if pixel is bad or not  
       if (clearsky)
         nClearSky++;
         
       nData++;
    }
  }
  free(data);
 
  // returns percentage of clear sky pixels in image
  return (((double)nClearSky)/nData);
}


/*
 *  modisCheckCoverage
 *
 *  Computes the mean observation coverage to evaluate if an image is acceptable
 *  Coverage corresponds to the area of intersection of observation footprint
 *  and cell divided by the area of the observation
 *  Coverage values correspond to indices from 0 to 7, with 0 meaning 0%-12%
 *  coverage and 7 meaning 87.5%-100% of coverage for the pixel
 */
double modisCheckCoverage(char *fnameFull)
{
  int     size_x, size_y, size_z;
  int     x, y;
  int     nData;
  InrType type;
  double  meanCoverage;
  unsigned char *data;

  // reads input file file header
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return -1;

  // checks if data type is correct for a observation/coverage file
  if (type != InrByte || size_z != 1)
    return -1;
  
  // retrieves input file data
  data = inrReadByteImage(fnameFull, &size_x, &size_y);
  if (!data)
    return -1;

  nData = 0;
  meanCoverage = 0;
  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
       int index;
       int pos = x + y*size_x;
       
       // masked out portions of the image have value zero
       if (data[pos] == 0)
         continue;
       
       // areas with value 15 are no-data areas: coverage is zero!
       if (data[pos] == 15)
       {
         nData++;
         continue;
       }
       
       // adds up the estimated mean coverage for the pixel (bits 5-7)
       index = data[pos]>>5;
       meanCoverage += 0.125*index + 0.0625;
       nData++;
    }
  }
  free(data);
 
  // divides by total number of pixels to compute mean and returns
  meanCoverage /= nData;
  return meanCoverage;
}


int main(int argc, char *argv[])
{
  char     dirInput[FNAMESIZE];
  char     fnameGoodOutput[FNAMESIZE];
  char     fnameBadOutput[FNAMESIZE];
  char     fnameInput[FNAMESIZE];
  double   minClearSky;
  double   minCoverage;
  int      okDirIn, okFileIn, okGoodOut, okBadOut, okClearSky, okCoverage;
  DIR     *dir;
  struct   dirent *ent;
  strList *fileList;
  strList *badList;
  int      iList;
  int      nCloudy = 0;
  int      nBadCoverage = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  ,   "%s",  dirInput);
  okFileIn      = igetopt1("-ifile" ,   "%s",  fnameInput);
  okGoodOut     = igetopt1("-ofile" ,   "%s",  fnameGoodOutput);
  okBadOut      = igetopt1("-ofilebad", "%s",  fnameBadOutput);
  okClearSky    = igetopt1("-clearsky", "%lf", &minClearSky);
  okCoverage    = igetopt1("-coverage", "%lf", &minCoverage);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // defines threshold values
  if (!okClearSky)
    minClearSky = MINCLEARSKY_DEFAULT;
  if (!okCoverage)
    minCoverage = MINCOVERAGE_DEFAULT;
  minClearSky /= 100;
  minCoverage /= 100;

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens output directory
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  printf("\nProcessing files with the following options:\n");
  printf("- Minimum clear sky criterion: %5.2lf%% of clear sky pixels\n", minClearSky*100);
  printf("- Minimum coverage criterion:  %5.2lf%% of observation coverage\n\n", minCoverage*100);

  // processes file(s) in the directory
  nCloudy = nBadCoverage = 0;
  fileList = strListCreate(FLISTSIZE);
  badList  = strListCreate(FLISTSIZE);
  while ((ent = readdir(dir)) != NULL)
  {
    char      fnameFull[FNAMESIZE];
    char      basename[FNAMESIZE];
    char      strDate[10];
    ModisInfo info;
    double    pClearSky = 0;
    double    meanCoverage = 0;

    // checks if file is the one selected to be processed
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input file name (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);

    // checks if it is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // parses modis filename
    if (!modisParseName(ent->d_name, &info))
      continue;
    // obtains basename by taking out the last two extensions (eg. '.orb_cov.inr')
    if (!modisGetBaseName(ent->d_name, basename))
      continue;
    
    // writes date info to a string
    sprintf(strDate, "%04d%03d", info.year, info.day);

    // checks if it is a state file or an orbit/coverage file
    if (strstr(ent->d_name, "STATE") || strstr(ent->d_name, "state"))
    {
      // adds file's date to file list if it is not already there
      if (strListFind(fileList, strDate) == -1)
        strListAdd(fileList, strDate);

      // state file
      // checks if there is an acceptable percentage of clear sky pixels
      pClearSky = modisCheckClearSky(fnameFull);
      if (pClearSky < minClearSky)
      {
        // adds file's date to bad list if it is not already there
        if (strListFind(badList, strDate) == -1)
          strListAdd(badList, strDate);
        printf("Image '%s': %6.2lf%% of clear sky pixels.     [FAILED]\n", basename, pClearSky*100);
        nCloudy++;
      }
      else
        printf("Image '%s': %6.2lf%% of clear sky pixels.\n", basename, pClearSky*100);
    }

    else if (strstr(ent->d_name, "ORB") || strstr(ent->d_name, "orb"))
    {
      // adds file's date to file list if it is not already there
      if (strListFind(fileList, strDate) == -1)
        strListAdd(fileList, strDate);

      // orbit/coverage file
      // checks if image coverage is acceptable (depends on the view angle)
      meanCoverage = modisCheckCoverage(fnameFull);
      if (meanCoverage < minCoverage)
      {
        // adds file's date to bad list if it is not already there
        if (strListFind(badList, strDate) == -1)
          strListAdd(badList, strDate);
        printf("Image '%s': %6.2lf%% of observation coverage. [FAILED]\n", basename, meanCoverage*100);
        nBadCoverage++;
      }
      else
        printf("Image '%s': %6.2lf%% of observation coverage.\n", basename, meanCoverage*100);
    }
  }

  // puts lists in alphabetical order
  strListSort(fileList);
  strListSort(badList);

  // writes dates of rejected images to output file, if desired
  if (okBadOut)
  {
    // opens output file
    FILE *fpOut = fopen(fnameBadOutput, "wt");
    if (!fpOut)
      printf("Error: unable to write to output file '%s'\n", fnameBadOutput);

    for (iList=0; iList<badList->count; iList++)
      fprintf(fpOut, "%s\n", strListGet(badList,iList));
    
    // closes output file
    fclose(fpOut);
  }

  // writes dates of accepted images to output file, if desired
  if (okGoodOut)
  {
    // opens output file
    FILE *fpOut = fopen(fnameGoodOutput, "wt");
    if (!fpOut)
      printf("Error: unable to write to output file '%s'\n", fnameGoodOutput);

    // writes dates of all images in the file list that are not in the bad list
    for (iList=0; iList<fileList->count; iList++)
    {
      if (strListFind(badList, strListGet(fileList,iList)) == -1)
        fprintf(fpOut, "%s\n", strListGet(fileList,iList));
    }

    // closes output file
    fclose(fpOut);
  }

  // prints final results
  printf("%d MODIS images have less than %5.2lf%% of clear sky pixels.\n", nCloudy, minClearSky*100);
  printf("%d MODIS images have less than %5.2lf%% of observation coverage.\n", nBadCoverage, minCoverage*100);
  printf("A total of %d out of %d MODIS images met the specified criteria.\n",
         fileList->count - badList->count, fileList->count);
  
  // closes directory and destroys lists
  closedir(dir);
  strListDestroy(fileList);
  strListDestroy(badList);
  return 1;
}

