#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE             1024


static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir";

static char Udetail[] = 
"\nConverts MOD09GST 1km state images to a proper cloud mask at 250m\n\
\nThe output masks will be valid for MOD09GQK products, and will have value 1\n\
for clear pixels and 0 for cloudy/contaminated pixels. Cloudy/contaminated\n\
pixels will include areas of cloud shadow as well as areas of high aerosol and\n\
cirrus.\n\
Namely, the criteria used for evaluating a clear pixel is the following:\n\
1. Bits 0-1 must be '00'     (clear cloud condition)\n\
2. Bit  2   must be '0'      (no cloud shadow detected)\n\
3. Bits 6-7 must not be '11' (high aerosol quantity)\n\
4. Bits 8-9 must be '00'     (no cirrus detected)\n\
5. Bit  10  must be '0'      (no clouds detected - internal flag)\n\
(being 0 the least significant bit):\n\
Input images must be in INRIMAGE format.\n\
\nInput:\n\
\t-idir\tDirectory containing MOD09GST state images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD09GST state image in INRIMAGE format\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted cloud masks\n";


/*
    MOD09GST state_1km_1 specification example
    NOTE: bits are numbered in the opposite way by this official specification!!
    
    Bits are numbered from left (MSB) to right (LSB):
    Bit    Description
    0      internal snow algorithm flag; 
           1 -- yes, 0.19%
           0 -- no, 99.81%
    1-2    BRDF correction performed;
           00 -- none, 100.00%
           01 -- Montana methodology, 0.00%
           10 -- Boston methodology, 0.00%
    3      MOD35 snow/ice flag;
           1 -- yes, 0.00%
           0 -- no, 100.00%
    4      internal fire algorithm flag;
           1 -- fire, 0.00%
           0 -- no fire, 100.00%
    5      internal cloud algorithm flag;
           1 -- cloud, 60.16%
           0 -- no cloud, 39.84%
    6-7    cirrus detected;
           00 -- none, 91.56%
           01 -- small, 0.77%
           10 -- average, 0.76%
           11 -- high, 6.91%
    8-9    aerosol quantity;
           00 -- climatology, 50.62%
           01 -- low, 13.10%
           10 -- average, 20.46%
           11 -- high, 15.82%
    10-12  land/water flag;
           000 -- shallow ocean, 1.70%
           001 -- land, 68.19%
           010 -- ocean coastlines and lake shorelines, 1.66%
           011 -- shallow inland water, 0.84%
           100 -- ephemeral water, 0.00%
           101 -- deep inland water, 0.00%
           110 -- continental/moderate ocean, 1.63%
           111 -- deep ocean, 25.97%
    13     cloud shadow;
           1 -- yes, 0.67%
           0 -- no, 99.33%
    14-15  cloud state;
           00 -- clear, 39.13%
           01 -- cloudy, 53.24%
           10 -- mixed, 7.62%
           11 -- not set, assumed clear, 0.00%
*/


/*
 *  modisCovertGSTFile
 *
 *  
 *  
 */
int modisCovertGSTFile(char *fnameFull, char *outfnameFull)
{
  int             size_x, size_y, size_z;
  int             x, y;
  int             i, j;
  InrType         type;   
  InrGeorefInfo   georef;   
  int             hasGeoref;
  unsigned short *data_in;
  unsigned char  *data_out;
  

  // retrieves input file dimensions and georeference information
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return 0;
  hasGeoref = inrReadGeoref(fnameFull, &georef);
    
  // checks if data_in type is correct for a MOD09GST state file
  if (type != InrUnsignedInt || size_z != 1)
    return 0;
  
  // retrieves input file data
  data_in = (unsigned short *)inrReadDataSequence(fnameFull, &size_x, &size_y, &size_z, &type);
  if (!data_in)
    return 0;
    
  // allocates memory for output file data
  // output is 16 times bigger due to 1km to 250m conversion
  data_out = (unsigned char *)malloc(size_x*4 * size_y*4 * sizeof(unsigned char));
  if (!data_out)
  {
    printf("Error: could not allocate memory.\n");
    free(data_in);
    return 0;
  }
  

  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
       int pos = x + y*size_x;
       int mask_val = 1;
       
       // masked out portions of the image have value zero
       if (data_in[pos] == 0)
         mask_val = 0;
       
       // checks criterion 1: bits 0-1 must be '00'     (clear cloud condition)
       else if ((data_in[pos] & 3) != 0)
         mask_val = 0;
         
       // checks criterion 2: bit  2   must be '0'      (no cloud shadow detected)
       else if ((data_in[pos] & 1<<2) != 0)
         mask_val = 0;

       // checks criterion 3: bits 6-7 must not be '11' (high aerosol quantity)
       else if ((data_in[pos] & 3<<6) == 3<<6)
         mask_val = 0;

       // checks criterion 4: bits 8-9 must be '00'     (no cirrus detected)
       else if ((data_in[pos] & 3<<8) != 0)
         mask_val = 0;

       // checks criterion 5: bit  10  must be '0'      (no clouds detected - internal flag)
       else if ((data_in[pos] & 1<<10) != 0)
         mask_val = 0;
         
       // pixel passed all tests: considered to be a clear pixel
       else
         mask_val = 1;
         
       for (j=y*4; j<(y+1)*4; j++)
         for (i=x*4; i<(x+1)*4; i++)
           data_out[i + j*size_x*4] = mask_val;
    }
  }
  free(data_in);
  
  // writes output file
  if (!inrWriteByteImage(outfnameFull, size_x*4, size_y*4, data_out))
  {
    printf("Error: could not write to output file '%s'.\n", outfnameFull);
    free(data_out);
    return 0;
  }
  free(data_out);

  // writes georeference information, if applicable
  if (hasGeoref)
  {
    georef.pixelsize_x /= 4;
    georef.pixelsize_y /= 4;
    inrWriteGeoref(outfnameFull, georef);
  }
  
  return 1;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  int    okDirIn, okDirOut, okFileIn;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  , "%s", dirInput);
  okDirOut      = igetopt1("-odir"  , "%s", dirOutput);
  okFileIn      = igetopt1("-ifile" , "%s", fnameInput);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (!okDirOut)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // processes file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be cropped
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input and output file names (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);
    sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);
    ftChangeExtension(outfnameFull, outfnameFull, "");
    ftChangeExtension(outfnameFull, outfnameFull, ".cloudmask_250m.inr");

    // checks if input is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // checks if input is a MOD09GST state file
    if (!strstr(ent->d_name, "MOD09GST") || !strstr(ent->d_name, "state"))
      continue;

    // converts file
    if (!modisCovertGSTFile(fnameFull, outfnameFull))
      continue;
    
    // increments counter
    printf("Converted file '%s'\n", ent->d_name);
    converted++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d MOD09GST state file(s) converted.\n", converted);
  return 1;
}

