#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE             1024


static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir";

static char Udetail[] = 
"\nConverts MOD35 1km band 3 images to a proper cloud mask at 250m\n\
\nThe output masks will be valid for MOD09GQK products, and will have value 1\n\
for clear pixels and 0 for cloudy/contaminated pixels. Cloudy/contaminated\n\
pixels will include areas of suspended dust as well as areas that did not\n\
pass the final confidence confirmation test.\n\
Namely, the criteria used for evaluating a clear pixel is the following:\n\
1. Bit 2 must be '0' ('yes' - final confidence confirmation test)\n\
2. Bit 4 must be '1' ('no'  - suspended dust flag)\n\
(being 0 the least significant bit):\n\
Input images must be in INRIMAGE format.\n\
\nInput:\n\
\t-idir\tDirectory containing MOD35 band 3 images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD35 band 3 image in INRIMAGE format\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted cloud masks\n";




/*
 *  modisCovertMOD35File
 *
 *  
 *  
 */
int modisCovertMOD35File(char *fnameFull, char *outfnameFull)
{
  int             size_x, size_y, size_z;
  int             x, y;
  int             i, j;
  InrType         type;   
  InrGeorefInfo   georef;   
  int             hasGeoref;
  unsigned char  *data_in;
  unsigned char  *data_out;
  

  // retrieves input file dimensions and georeference information
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return 0;
  hasGeoref = inrReadGeoref(fnameFull, &georef);
    
  // checks if data_in type is correct for a MOD35 band 3 file
  if (type != InrByte || size_z != 1)
    return 0;
  
  // retrieves input file data
  data_in = inrReadByteImage(fnameFull, &size_x, &size_y);
  if (!data_in)
    return 0;
    
  // allocates memory for output file data
  // output is 16 times bigger due to 1km to 250m conversion
  data_out = (unsigned char *)malloc(size_x*4 * size_y*4 * sizeof(unsigned char));
  if (!data_out)
  {
    printf("Error: could not allocate memory.\n");
    free(data_in);
    return 0;
  }
  

  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
       int pos = x + y*size_x;
       int mask_val = 1;
       
       // masked out portions of the image have value zero
       if (data_in[pos] == 0)
         mask_val = 0;
       
       // checks criterion 1: bit  2   must be '0' ('yes' - final confidence confirmation test)
       else if ((data_in[pos] & 1<<2) != 0)
         mask_val = 0;
         
       // checks criterion 2: bit  4   must be '1' ('no'  - suspended dust flag)
       else if ((data_in[pos] & 1<<4) != 1<<4)
         mask_val = 0;

       // pixel passed all tests: considered to be a clear pixel
       else
         mask_val = 1;
         
       for (j=y*4; j<(y+1)*4; j++)
         for (i=x*4; i<(x+1)*4; i++)
           data_out[i + j*size_x*4] = mask_val;
    }
  }
  free(data_in);
  
  // writes output file
  if (!inrWriteByteImage(outfnameFull, size_x*4, size_y*4, data_out))
  {
    printf("Error: could not write to output file '%s'.\n", outfnameFull);
    free(data_out);
    return 0;
  }
  free(data_out);

  // writes georeference information, if applicable
  if (hasGeoref)
  {
    georef.pixelsize_x /= 4;
    georef.pixelsize_y /= 4;
    inrWriteGeoref(outfnameFull, georef);
  }
  
  return 1;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  int    okDirIn, okDirOut, okFileIn;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"  , "%s", dirInput);
  okDirOut      = igetopt1("-odir"  , "%s", dirOutput);
  okFileIn      = igetopt1("-ifile" , "%s", fnameInput);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (!okDirOut)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // processes file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be cropped
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input and output file names (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);
    sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);
    ftChangeExtension(outfnameFull, outfnameFull, "");
    ftChangeExtension(outfnameFull, outfnameFull, ".mod35test_250m.inr");

    // checks if input is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // checks if input is a MOD35 band 3 file
    if (!strstr(ent->d_name, "MOD35") || !strstr(ent->d_name, "b3"))
      continue;

    // converts file
    if (!modisCovertMOD35File(fnameFull, outfnameFull))
      continue;
    
    // increments counter
    printf("Converted file '%s'\n", ent->d_name);
    converted++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d MOD35 band 3 file(s) converted.\n", converted);
  return 1;
}

