#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "inr_io.h"
#include "file_tools.h"

#define FNAMESIZE             1024


static char Ucmd[] =
"-idir input_dir [-ifile input_file] -odir output_dir [-ifilter string]";

static char Udetail[] = 
"\nConverts MOD13Q1 250m VI quality images to masks that indicate which data can\n\
be considered useful.\n\
\nThe output mask will have value 1 for good quality, clear-sky pixels, and 0\n\
otherwise, as indicated by the quality flags.\n\
Namely, the criteria used for evaluating a good pixel is the following:\n\
1. Bit  1    must be '0'      (VI produced and probably non-cloudy)\n\
2. Bit  5    must be '0'      (VI usefulness average or better)\n\
3. Bits 6-7  must not be '11' (high aerosol quantity)\n\
4. Bit  10   must be '0'      (no mixed clouds)\n\
5. Bit  14   must be '0'      (no possible shadow)\n\
(being 0 the least significant bit):\n\
Input images must be in INRIMAGE format.\n\
\nInput:\n\
\t-idir\tDirectory containing MOD13Q1 VI quality images in INRIMAGE format\n\
\t-ifile\tFile name of one MOD13Q1 VI quality image in INRIMAGE format\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-odir\tOutput directory for the converted masks\n\
\nOptional:\n\
\t-ifilter\tFilter string for identifying MOD13Q1 VI quality images\n\
\t\t\tDefault identifier is 'NDVI_Quality'\n";


/*
    MOD13Q1 VI Quality specification
    
    Bit     Description
    0-1     VI quality
            00 VI produced, good quality;
            01 VI produced, but check QA
            10 produced but most likely cloudy pixel
            11 pixel not produced due to other reasons than clouds
    2-5     VI usefulness four bit range 0= highest quality
            13 no atmospheric correction performed
            14 quality too low to be useful
            15 not useful for any other reason
    6-7     aerosol quantity:
            00 climatology
            01 low
            10 average
            11 high
    8       1 yes adjacency correction performed
            0 no adjacency correction performed
    9       1 yes atmosphere BRDF correction performed
            0 no atmosphere BRDF correction performed
    10      1 yes mixed clouds
            0 no mixed clouds
    11-12   land/water flag
            00 ocean
            01 coast
            10 wetland
            11 land
    13      1 yes possible snow/ice
            0 no possible snow/ice
    14      1 yes possible shadow
            0 no possible shadow
    15      composite method for VI
            0 BRDF model based nadir equivalent VI
            1 CVMVC (constraint view angle maximum value VI)

*/


/*
 *  modisConvertVIQualityFile
 *
 *  
 *  
 */
int modisConvertVIQualityFile(char *fnameFull, char *outfnameFull)
{
  int             size_x, size_y, size_z;
  int             i;
  int             nData;
  InrType         type;   
  InrGeorefInfo   georef;   
  int             hasGeoref;
  unsigned short *data_in;
  unsigned char  *data_out;
  

  // retrieves input file dimensions and georeference information
  if (!inrReadHeader(fnameFull, &size_x, &size_y, &size_z, &type))
    return 0;
  hasGeoref = inrReadGeoref(fnameFull, &georef);
    
  // checks if data_in type is correct for a MOD13Q1 VI quality file
  if (type != InrUnsignedInt)
  {
    fprintf(stderr, "Error: data type is not 'UnsignedInt' for file '%s'.\n", fnameFull);
    return 0;
  }
  
  // retrieves input file data
  data_in = (unsigned short *)inrReadDataSequence(fnameFull, &size_x, &size_y, &size_z, &type);
  if (!data_in)
  {
    fprintf(stderr, "Error: could not read data for file '%s'.\n", fnameFull);
    return 0;
  }
    
  // allocates memory for output file data
  nData = size_x*size_y*size_z;
  data_out = (unsigned char *)malloc(nData * sizeof(unsigned char));
  if (!data_out)
  {
    fprintf(stderr, "Error: could not allocate memory.\n");
    free(data_in);
    return 0;
  }
  

  for (i=0; i<nData; i++)
  {
     int mask_val = 1;
     
     // masked out portions of the image have value zero
     if (data_in[i] == 0)
       mask_val = 0;
     
     // checks criterion 1: bit  1    must be '0'  (VI produced and probably non-cloudy)
     else if ((data_in[i]>>1 & 1) != 0)
       mask_val = 0;
       
     // checks criterion 2: bit  5    must be '0'  (VI usefulness average or better)
     else if ((data_in[i]>>5 & 1) != 0)
       mask_val = 0;

     // checks criterion 3: bits 6-7  must not be '11' (high aerosol quantity)
     else if ((data_in[i]>>6 & 3) == 3)
       mask_val = 0;

     // checks criterion 4: bit  10   must be '0'  (no mixed clouds)
     else if ((data_in[i]>>10 & 1) != 0)
       mask_val = 0;

     // checks criterion 5: bit  14   must be '0'  (no shadow)
     else if ((data_in[i]>>14 & 1) != 0)
       mask_val = 0;

     // pixel passed all tests: considered to be a good quality pixel
     else
       mask_val = 1;
       
     data_out[i] = mask_val;
  }
  free(data_in);
  
  // writes output file
  if (!inrWriteByteSequence(outfnameFull, size_x, size_y, size_z, data_out))
  {
    printf("Error: could not write to output file '%s'.\n", outfnameFull);
    free(data_out);
    return 0;
  }
  free(data_out);

  // writes georeference information, if applicable
  if (hasGeoref)
    inrWriteGeoref(outfnameFull, georef);
  
  return 1;
}


int main(int argc, char *argv[])
{
  char   dirInput[FNAMESIZE];
  char   dirOutput[FNAMESIZE];
  char   fnameInput[FNAMESIZE];
  char   strFilter[FNAMESIZE];
  int    okDirIn, okDirOut, okFileIn, okFilter;
  DIR   *dir;
  struct dirent *ent;
  int    converted = 0;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDirIn       = igetopt1("-idir"   , "%s", dirInput);
  okDirOut      = igetopt1("-odir"   , "%s", dirOutput);
  okFileIn      = igetopt1("-ifile"  , "%s", fnameInput);
  okFilter      = igetopt1("-ifilter", "%s", strFilter);
  
  if ((okDirIn && okFileIn) || (!okDirIn && !okFileIn))
    iusage_ (Ucmd, Udetail);
  if (!okDirOut)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // specifies default file name filter if necessary
  if (!okFilter)
    strcpy(strFilter, "NDVI_Quality");

  // appends final slash to directories, if necessary
  if (okDirIn)
    ftFixDirName(dirInput);
  ftFixDirName(dirOutput);

  // if working with a single file, extracts file and directory info
  if (okFileIn)
    ftDecodeFileName(fnameInput, dirInput, fnameInput);
    
  // opens directories
  if ((dir = opendir(dirOutput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirOutput);
    return 0;
  }
  closedir(dir);
  if ((dir = opendir(dirInput)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirInput);
    return 0;
  }

  // processes file(s) in the directory
  converted = 0;
  while ((ent = readdir(dir)) != NULL)
  {
    char fnameFull[FNAMESIZE];
    char outfnameFull[FNAMESIZE];

    // checks if file is to be cropped
    if (okFileIn && strcmp(ent->d_name, fnameInput)!=0)
      continue;
      
    // computes the complete input and output file names (dirname + filename)
    sprintf(fnameFull, "%s%s", dirInput, ent->d_name);
    sprintf(outfnameFull, "%s%s", dirOutput, ent->d_name);
    ftChangeExtension(outfnameFull, outfnameFull, "");
    ftChangeExtension(outfnameFull, outfnameFull, ".qualitymask.inr");

    // checks if input is an INRIMAGE image
    if (!inrCheckFile(fnameFull))
      continue;

    // checks if input is a MOD13Q1 VI quality file
    if (okDirIn && !strstr(ent->d_name, strFilter))
      continue;

    // converts file
    if (!modisConvertVIQualityFile(fnameFull, outfnameFull))
      continue;
    
    // increments counter
    printf("Converted file '%s'\n", ent->d_name);
    converted++;
  }

  // closes directory
  closedir(dir);
    
  printf("%d MOD13Q1 VI quality file(s) converted.\n", converted);
  return 1;
}

