#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inrimage/image.h>
#include "dirent-universal.h"
#include "file_tools.h"
#include "str_tools.h"
#include "modis.h"
#include "inr_io.h"

#define FLISTSIZE    65536
#define FNAMESIZE    1024


static char Ucmd[] =
"-idir dirname [-ifile filename] -ofile date_file";

static char Udetail[] =
"\nOutputs a date list for MODIS images based on its file name.\n\
\nFile names should follow the standard MODIS naming convention for Level 2\n\
data, namely:\
  MODxxxxx.AYYYYDDD.*\n\
where:\n\
  MODxxxxx      represents the MODIS data product name\n\
  A             represents the platform code (e.g. A=Terra)\n\
  YYYY          represents the year of acquisition\n\
  DDD           represents the julian day of acquisition\n\
for instance:\n\
  MOD09GQK.A2001211.h12v10.004.2003131015044\n\
\nInput:\n\
\t-idir\tDirectory containing MODIS images\n\
\t-ifile\tFile name of one MODIS image\n\
\t\tSpecify '-idir' for batch processing\n\
\t\tSpecify '-ifile' for single file processing\n\
\nOutput:\n\
\t-ofile\tFile name for a date list of the MODIS images\n\
\nOptional:\n\
\t-ddd\tOutputs list in 'DDD' format instead of 'YYYYDDD' format\n\
\t\t(dates in subsequent years are considered to be nYears*366\n\
\t\tahead)\n";




int main(int argc, char *argv[])
{
  int      okDir, okFile, okDate, okDDD;
  char     dirModis[FNAMESIZE];
  char     fnameModis[FNAMESIZE];
  char     fnameDate[FNAMESIZE];
  FILE    *fpDate;
  DIR     *dir;
  struct   dirent *ent;
  strList *fileList;
  int      iList;

  // initializes inrimage
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  // checks arguments
  okDir  = igetopt1 ("-idir" , "%s", dirModis);
  okFile = igetopt1 ("-ifile", "%s", fnameModis);
  okDate = igetopt1 ("-ofile", "%s", fnameDate);
  okDDD  = igetopt0 ("-ddd");
  
  if ((okDir && okFile) || (!okDir && !okFile))
    iusage_ (Ucmd, Udetail);
  if (!okDate)
    iusage_ (Ucmd, Udetail);
  if (tstopts ())
    iusage_ (Ucmd, Udetail);

  // opens date file
  fpDate = fopen(fnameDate, "wt");
  if (!fpDate)
    printf("Error: unable to write to date file '%s'\n", argv[2]);

  // if working with a directory, appends final slash if necessary
  if (okDir)
    ftFixDirName(dirModis);
  // if working with a single file, extracts file and directory info
  if (okFile)
    ftDecodeFileName(fnameModis, dirModis, fnameModis);
    
  // opens directory
  if ((dir = opendir(dirModis)) == NULL)
  {
    printf("Error: could not open directory '%s'\n", dirModis);
    return 0;
  }

  // processes file(s) in the directory
  fileList = strListCreate(FLISTSIZE);
  while ((ent = readdir(dir)) != NULL)
  {
    ModisInfo info;
    char strDate[10];
    
    // checks if file is to be included
    if (okFile && strcmp(ent->d_name, fnameModis)!=0)
      continue;
      
    // parses modis filename
    if (!modisParseName(ent->d_name, &info))
      continue;
    
    // writes date info to a string
    sprintf(strDate, "%04d%03d", info.year, info.day);

    // adds date info to list if it is not already there
    if (strListFind(fileList, strDate) == -1)
      strListAdd(fileList, strDate);
  }

  // puts list in alphabetical order
  strListSort(fileList);

  // checks to see if dates are to be written in YYYYDDD format or DDD format
  if (!okDDD)
  {
      // writes dates as they are (YYYYDDD format)
      for (iList=0; iList<fileList->count; iList++)
          fprintf(fpDate, "%s\n", strListGet(fileList,iList));
  }
  else
  {
      // converts dates to DDD format and writes them
      int firstDate = atoi(strListGet(fileList,0));
      int firstYear = (firstDate - firstDate%1000)/1000;
      for (iList=0; iList<fileList->count; iList++)
      {
          int date = atoi(strListGet(fileList,iList));
          int year = (date - date%1000)/1000;
          int yeardiff = year - firstYear;
          int day = (date - year*1000) + yeardiff*366;
          fprintf(fpDate, "%d\n", day);
      }
  }

  // closes directory and date file
  closedir(dir);
  fclose(fpDate);
    
  printf("%d MODIS file(s) in date list.\n", fileList->count);
  return 1;
}

