#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "classdef.h"
#include "inr_io.h"
#include "file_tools.h"
#include "linearalgebra.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define PI  3.14159265358979323846

#define FNAMESIZE         1024
#define TOLERANCE_DEFAULT 0



static char Ucmd[] = "-ifile inr_input -roi roi_mask -def class_definitions \
-ofile inr_classification [-oprob] [-tol min_probability]\n";

static char Udetail[] =
"Given input data and a class definition file, applies a maximum likelihood\n\
classification to the data\n\
\nClasses are assumed to have a gaussian probability distribution.\n\
Input class definitions file should contain the mean values and covariance\n\
matrices for all classes.\n\
Input and output files are in INRIMAGE format.\n\
\nInput:\n\
\t-ifile\tFile name of the input data\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\t-def\tFile name of a text file containing the class definitions\n\
\nOutput:\n\
\t-ofile\tFile name of the output classificantion image\n\
\nOptional:\n\
\t-oprob\tOutput an image with nClass layers, where layer i corresponds\n\
\t\tto the probability that each point belongs to class i+1\n\
\t-tol\tMinimum probability tolerance value for assigning a class to a\n\
\t\tpoint.\n\
\t\tIf, for one point, the class with maximum probability does\n\
\t\tnot exceed this value, the point is left unclassified (class 0)\n\
\t\tDefault value is 0 (all points with ROI are classified)\n";




int computeCovDeterminants(classdef *def, double *cov_determinants)
{
  matrice *cov;
  double  *prob;
  double   det;
  int iClass; 
  int iDiscrete;

  for (iClass=0; iClass<def->nClass; iClass++)
  {
    prob = defGetDiscreteProb(def, iClass);
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      // if class probability given the discrete variable is zero, then there
      // the covariance matrix is singular and its determinant is zero
      if (prob[iDiscrete] == 0)
        det = 0;
      else
      {
        cov = defGetContCovariance(def, iClass, iDiscrete);
        det = Determinant(cov);
        // if the determinant computed is too low, the matrix will be
        // near-singular, making matrix inversion problematic
        if (det < 1e-30)
        {
          //fprintf (stderr, "Warning: found near-singular covariance matrix for class %d , discrete index %d\n", iClass+1, iDiscrete);
          //fprintf (stderr, "         discrete probability set to 0 (det = %g)\n", det);
          prob[iDiscrete] = 0;
          det = 0;
        }
      }
      cov_determinants[iDiscrete + iClass*def->nDiscreteValues] = det;
    }
  }
  return 1; 
}


int invertCov(classdef *def)
{
  matrice *cov;
  matrice *inv;
  double  *prob;
  int iClass; 
  int iDiscrete;

  // inverts covariance matrices
  for (iClass=0; iClass<def->nClass; iClass++)
  {
    prob = defGetDiscreteProb(def, iClass);
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      cov = defGetContCovariance(def, iClass, iDiscrete);

      if (prob[iDiscrete] == 0)
        inv = Matrice_Nulle(cov->nbLignes, cov->nbColonnes);
      else
        inv = Inverser_Matrice(cov);

      defSetContCovariance(def, iClass, iDiscrete, inv);
    }
  }
  return 1;
}


int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameClassDef[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameOutputProb[FNAMESIZE];
  InrType typeInput;
  InrType type;
  int     typesizeInput;
  InrGeorefInfo georef;
  
  int    okOutputProb;
  int    okTolerance;
  double tolerance;
  
  unsigned char *dataInput;
  unsigned char *ROIMask;
  unsigned char *dataClassif;
  float         *dataProb;
  
  matrice  *point;
  matrice  *point_transp;
  double   *cov_determinants;
  double   *values;
  double   *evaluation;
  double    evaluation_max;
  double    evaluation_sum;
  
  matrice *mean;
  matrice *cov;
  double  *prob;
  double   det;

  classdef *def;
  int iDiscrete;
  
  int nClass;
  int nFeatures;
  int iClass;
  
  int size_x, size_y, size_z;
  int i, j, k;
  int x, y, z;
  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"  , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-def"  , "%s", fnameClassDef))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okOutputProb = igetopt0 ("-oprob");
  okTolerance  = igetopt1 ("-tol", "%lf", &tolerance);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okTolerance)
    tolerance = TOLERANCE_DEFAULT;

  dataProb = NULL;
  if (okOutputProb)
    ftChangeExtension(fnameOutput, fnameOutputProb, ".prob.inr");
    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
 

  // reads the class definitions
  def = defCreateFromFile(fnameClassDef);
  if (!def)
    return 0;
  nClass    = def->nClass;
  nFeatures = def->nFeatures;

  if (nFeatures != size_z)
  {
    fprintf (stderr, "Error: number of features in class definition file (%d) does not match number of layers in the input data (%d).\n",
             nFeatures, size_z);
    return 0;
  }
    
  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }
  

  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  dataClassif      = (unsigned char *) malloc (size_x*size_y * sizeof(unsigned char));
  cov_determinants = (double *) malloc (nClass*def->nDiscreteValues * sizeof(double));
  evaluation       = (double *) malloc (nClass * sizeof(double));
  values           = (double *) malloc (size_z * sizeof(double));
  point            = Matrice_Nulle(def->nContFeatures, 1);
  point_transp     = Matrice_Nulle(1, def->nContFeatures);

  // inititalizes variables related to outputting the posterior probabilities (if specified)  
  if (okOutputProb)
  {
    dataProb = (float *) malloc (size_x*nClass * sizeof(float));

    // writes output file header (output type is always float)
    if (!inrWriteHeader(fnameOutputProb, size_x, size_y, nClass, InrFloat))
    {
      fprintf(stderr, "Error: could not write header information for output file '%s'.\n", fnameOutputProb);
      return 0;
    }
    // initializes data line
    for (i=0; i<size_x*nClass; i++)
      dataProb[i] = 0;
  }
  
  // computes the determinants of the covariance matrices
  if (!computeCovDeterminants(def, cov_determinants))
    return 0;

  // inverts the covariance matrices
  if (!invertCov(def))
    return 0;
 


  // PROCESSES THE DATA
  
  for (y=0; y<size_y; y++)
  {
    fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);

    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    if (x == size_x)
    {
      // all points in the line are rated unclassified
      for (x=0; x<size_x; x++)
        dataClassif[x + y*size_x] = 0;
        
      if (okOutputProb)
      {
        // writes empty output probability data line
        if (!inrWriteLine(fnameOutputProb, y, (unsigned char *)dataProb))
        {
          fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputProb);
          return 0;
        }
      }
      continue;
    }

    // reads input data line y
    if (!inrReadLine(fnameInput, y, dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }
    
    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      // pixels outside of region of interest are rated unclassified
      if (!ROIMask[x + y*size_x])
      {
        dataClassif[x + y*size_x] = 0;
        continue;
      }
        
      // gets point values for the position x of the input line
      if (!defGetPoint(def, x, dataInput, size_x, 1, typeInput, values, &iDiscrete))
        continue;
      
      // evaluates likelihood for all classes
      for (iClass=0; iClass<nClass; iClass++)
      {
        matrice *mult1, *mult2;
        
        prob = defGetDiscreteProb(def, iClass);
        mean = defGetContMean(def, iClass, iDiscrete);
        cov  = defGetContCovariance(def, iClass, iDiscrete);
        det  = cov_determinants[iDiscrete + iClass*def->nDiscreteValues];
 
        // likelihood is zero if class has no probability of having the point's
        // discrete value (will have a singular covariance matrix)
        if (prob[iDiscrete] == 0)
        {
          evaluation[iClass] = 0;
          continue;
        }
        
        // defines vectors: (values-means), t(values-means)
        for (z=0; z<def->nContFeatures; z++)
        {
          point->val[z] = values[z] - mean->val[z];
          point_transp->val[z] = point->val[z];
        }

        // computes mahanalobis distance:
        //   mahal = t(x-m) * COV^-1 * (x-m)
        mult1 = Multiplier_Matrice(point_transp, cov);
        mult2 = Multiplier_Matrice(mult1, point);

        // computes likelihood for the class, given by:
        //   likelihood = p(x|wj,disc)*P(wj|disc)*P(wj)
        // where
        //   P(wj)        is the prior probability for class wj
        //   P(wj|disc)   is the conditional probability related to the discrete value
        //   p(x|wj,disc) is the pdf of the class related to the discrete value
        // so:
        //   likelihood = 1/( ((2*PI)^DIM/2)(det(COV)^1/2) ) * exp(-0.5*mahal) * prob(disc) * prior
        //              = exp (- 0.5*DIM*log(2*PI) - 0.5*log(det(COV)) - 0.5*mahal + log(prob(disc)) + log(prior) )
        evaluation[iClass] = - 0.5*nFeatures*log(2*PI) - 0.5*log(det) - 0.5*mult2->val[0] + log(prob[iDiscrete]) + log(1./nClass);
        evaluation[iClass] = exp(evaluation[iClass]);
        
        free_matrice(&mult1);
        free_matrice(&mult2);
      }
      
      // converts likelihood into posterior probabilities for each class
      // posterior probabilities are given by:
      //   P(wj|x,disc) = likelihood / sum(likelihood)
      evaluation_sum = 0;
      for (iClass=0; iClass<nClass; iClass++)
        evaluation_sum += evaluation[iClass];
      for (iClass=0; iClass<nClass; iClass++)
        evaluation[iClass] /= evaluation_sum;
      
      // assigns the class with highest likelihood to the point
      // (as long as it exceeds the minimum probability tolerance value)
      dataClassif[x + y*size_x] = 0;
      evaluation_max = evaluation[0];
      for (iClass=0; iClass<nClass; iClass++)
      {
        if (evaluation[iClass]>=tolerance && evaluation[iClass]>=evaluation_max)
        {
          dataClassif[x + y*size_x] = iClass+1;
          evaluation_max = evaluation[iClass];
        }
      }
      
      // adds point class evaluations (posterior probabilities) if specified
      if (okOutputProb)
        for (iClass=0; iClass<nClass; iClass++)
          dataProb[x+iClass*size_x] = (float) evaluation[iClass];
    }
        

    // writes line class evaluations (posterior probabilities) if specified
    if (okOutputProb)
    {
      if (!inrWriteLine(fnameOutputProb, y, (unsigned char *)dataProb))
      {
        fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, fnameOutputProb);
        return 0;
      }
      // re-initializes output data line
      for (i=0; i<size_x*nClass; i++)
        dataProb[i] = 0;
    }
  }


  // outputs results
  inrWriteByteImage(fnameOutput, size_x, size_y, dataClassif);

   

  // transfers georeference info if available
   if (inrReadGeoref(fnameInput, &georef))
    {
     inrWriteGeoref(fnameOutput, georef);
     if (okOutputProb)
       inrWriteGeoref(fnameOutputProb, georef);
    }
  else if (okOutputProb)

    inrWriteEnviHdr(fnameOutputProb);

  //write color info
   //   inrWriteColor(fnameOutput,nClass);

  fprintf(stderr, "Image classification successfully completed.\n");

  free(dataInput);
  free(dataClassif);
  free(values);
  free(evaluation);
  free_matrice(&point);
  free_matrice(&point_transp);
  free(cov_determinants);
  defDestroy(&def);
  
  return 0;
}
