#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "classdef.h"
#include "inr_io.h"
#include "file_tools.h"
#include "linearalgebra.h"
#include "std.h"
#include "modis_std.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define PI  3.14159265358979323846

#define FNAMESIZE         1024
#define TOLERANCE_DEFAULT 0



static char Ucmd[] = "-ifile inr_input -class inr_classification -roi roi_mask \
-def class_definitions -ofile inr_classification [-oprob] [-tol min_probability]\
[-keep]\n";

static char Udetail[] =
"Given input data, a previous state classification and a class definition\n\
file, applies a maximum likelihood classification to the data based on the\n\
transitions specified by a state-transition diagram (STD).\n\
\nClasses are assumed to have a gaussian probability distribution.\n\
Input class definitions file should contain the mean values and covariance\n\
matrices for all temporal profile classes.\n\
OBS: STD is hard-coded internally.\n\
Input and output files are in INRIMAGE format.\n\
\nInput:\n\
\t-ifile\tFile name of the input data\n\
\t-class\tFile name of a classification image representing the previous\n\
\t\tstate of the area\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\t-def\tFile name of a text file containing the class definitions\n\
\nOutput:\n\
\t-ofile\tFile name of the output classificantion image\n\
\nOptional:\n\
\t-oprob\tOutput an image with nStates layers, where layer i corresponds\n\
\t\tto the probability that each point belongs to state i+1\n\
\t-tol\tMinimum probability tolerance value for assigning a valid state\n\
\t\tto a point.\n\
\t\tDefault value is 0 (all points with ROI are classified)\n\
\t-keep\tSpecifies that the previous state should be kept whenever the\n\
\t\tminimum tolerance probability is not reached.\n\
\t\tIf 'keep' is not specified and, for a given point, the state\n\
\t\twith maximum probability does not exceed this tolerance value,\n\
\t\tthen that point is left unclassified (state 0)\n";


/*
 * GLOBAL CONTEXT
 */
typedef struct 
{
    char fnameInput[FNAMESIZE];
    char fnameClassif[FNAMESIZE];
    char fnameClassDef[FNAMESIZE];
    char fnameROIMask[FNAMESIZE];
    char fnameOutput[FNAMESIZE];
    char fnameOutputProb[FNAMESIZE];

    InrType typeInput;
    int     typesizeInput;

    InrGeorefInfo georef;

    int    okOutputProb;
    int    okKeep;
    int    okTolerance;
    double tolerance;

    unsigned char *dataInput;
    unsigned char *dataClassif;
    unsigned char *ROIMask;
    unsigned char *dataOutput;
    float         *dataOutputProb;

    matrice  *point;
    matrice  *point_transp;
    double   *cov_determinants;
    double   *values;
    double   *stateEval;
    double   *transitionEval;

    classdef *def;

    stdDiagram *std;

    int size_x, size_y, size_z;
}
ModisClassifApplyStd_gc;
                                  

/*
 * EVALUATE_POINT
 *
 * Given an (x,y) point of the input image and the previous state information, performs
 * a maximum-likelihood evaluation in order to retrieve the posterior probabilities for
 * each possible subsequent state (returned in gc->stateEval)
 */
int evaluatePoint(ModisClassifApplyStd_gc *gc, int x, int y)
{
    double evaluation_sum;
    int iDiscreteVal;
    int iState;
    int iTrans;
    int nTrans;
    stdState *state;

    // gets point values for the position x of the input line
    if (!defGetPoint(gc->def, x, gc->dataInput, gc->size_x, 1, gc->typeInput, gc->values, &iDiscreteVal))
        return 0;

    // retrieves the pixel's previous state from the classification data
    iState = gc->dataClassif[x + y*gc->size_x];
    state = &gc->std->states[iState];
    nTrans = state->nTransitions;

    // evaluates likelihood for all transitions specified for the given state
    // obs: each transition specifies a profile class that leads to a subsequent state
    for (iTrans = 0; iTrans < nTrans; iTrans++)
    {
        int iProfile = state->transitions[iTrans].iProfile;
        matrice *mult1, *mult2;
        int z;

        // retrieves profile class index for the transitions

        // retrives statistical information for the profile class:
        // - probabilities for discrete variable, mean values for continuous variables,
        //   inverted covariance matrices for continuous variables, and their determinantes
        //   obs: covariance matrices should have been previously inverted using invertCov()
        double  *prob = defGetDiscreteProb(gc->def, iProfile);
        matrice *mean = defGetContMean(gc->def, iProfile, iDiscreteVal);
        matrice *cov  = defGetContCovariance(gc->def, iProfile, iDiscreteVal);
        double   det  = gc->cov_determinants[iDiscreteVal + iProfile*gc->def->nDiscreteValues];
        int inputDim  = gc->def->nFeatures;

        // likelihood is zero if profile class has no probability of having the point's
        // discrete value (will have a singular covariance matrix)
        if (prob[iDiscreteVal] == 0)
        {
            gc->transitionEval[iTrans] = 0;
            continue;
        }

        // defines vectors: (values-means), t(values-means)
        for (z = 0; z < gc->def->nContFeatures; z++)
        {
            gc->point->val[z] = gc->values[z] - mean->val[z];
            gc->point_transp->val[z] = gc->point->val[z];
        }

        // computes mahanalobis distance:
        //   mahal = t(x-m) * COV^-1 * (x-m)
        mult1 = Multiplier_Matrice(gc->point_transp, cov);
        mult2 = Multiplier_Matrice(mult1, gc->point);

        // computes likelihood for the profile class, given by:
        //   likelihood = p(x|wj,disc)*P(wj|disc)*P(wj)
        // where
        //   P(wj)        is the prior probability for class wj
        //   P(wj|disc)   is the conditional probability related to the discrete value
        //   p(x|wj,disc) is the pdf of the class related to the discrete value
        // so:
        //   likelihood = 1/( ((2*PI)^DIM/2)(det(COV)^1/2) ) * exp(-0.5*mahal) * prob(disc) * prior
        //              = exp (- 0.5*DIM*log(2*PI) - 0.5*log(det(COV)) - 0.5*mahal + log(prob(disc)) + log(prior) )
        gc->transitionEval[iTrans] = - 0.5*inputDim*log(2*PI) - 0.5*log(det) - 0.5*mult2->val[0] + log(prob[iDiscreteVal]) + log(1./nTrans);
        gc->transitionEval[iTrans] = exp(gc->transitionEval[iTrans]);

        free_matrice(&mult1);
        free_matrice(&mult2);
    }

    // converts likelihood into posterior probabilities for each transition
    // posterior probabilities are given by:
    //   P(wj|x,disc) = likelihood / sum(likelihood)
    evaluation_sum = 0;
    for (iTrans = 0; iTrans < nTrans; iTrans++)
        evaluation_sum += gc->transitionEval[iTrans];
    for (iTrans = 0; iTrans < nTrans; iTrans++)
        gc->transitionEval[iTrans] /= evaluation_sum;

    // based on the transition evaluations, retrieves the final state evaluation for the
    // point
    return stdGetStateEvaluation(gc->std, iState, gc->transitionEval, gc->stateEval);
}



/*
 * COMPUTE_COV_DETERMINANTS
 *
 * Computes the determinants of the covariant matrices specified by a class definition
 */
int computeCovDeterminants(classdef *def, double *cov_determinants)
{
  matrice *cov;
  double  *prob;
  double   det;
  int iClass; 
  int iDiscrete;

  for (iClass = 0; iClass < def->nClass; iClass++)
  {
    prob = defGetDiscreteProb(def, iClass);
    for (iDiscrete = 0; iDiscrete < def->nDiscreteValues; iDiscrete++)
    {
      // if class probability given the discrete variable is zero, then there
      // the covariance matrix is singular and its determinant is zero
      if (prob[iDiscrete] == 0)
        det = 0;
      else
      {
        cov = defGetContCovariance(def, iClass, iDiscrete);
        det = Determinant(cov);
        // if the determinant computed is too low, the matrix will be
        // near-singular, making matrix inversion problematic
        if (det < 1e-30)
        {
          //fprintf (stderr, "Warning: found near-singular covariance matrix for class %d , discrete index %d\n", iClass+1, iDiscrete);
          //fprintf (stderr, "         discrete probability set to 0 (det = %g)\n", det);
          prob[iDiscrete] = 0;
          det = 0;
        }
      }
      cov_determinants[iDiscrete + iClass*def->nDiscreteValues] = det;
    }
  }
  return 1; 
}



/*
 * INVERT_COV
 *
 * Inverts the covariance matrices of a class definition.
 * Obs: the original matrices are substituted by the inverted ones 
 */
int invertCov(classdef *def)
{
  matrice *cov;
  matrice *inv;
  double  *prob;
  int iClass; 
  int iDiscrete;

  // inverts covariance matrices
  for (iClass=0; iClass<def->nClass; iClass++)
  {
    prob = defGetDiscreteProb(def, iClass);
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      cov = defGetContCovariance(def, iClass, iDiscrete);

      if (prob[iDiscrete] == 0)
        inv = Matrice_Nulle(cov->nbLignes, cov->nbColonnes);
      else
        inv = Inverser_Matrice(cov);

      defSetContCovariance(def, iClass, iDiscrete, inv);
    }
  }
  return 1;
}



/*
 * MAIN METHOD
 *
 */
int main (int argc, char **argv)
{
    ModisClassifApplyStd_gc gc_obj;           /* general context */
    ModisClassifApplyStd_gc *gc = &gc_obj;

    InrType type;
  
    int iState;
    int nStates;
    int nProfiles;

    int i, j, k;
    int x, y;
  
    // reads arguments
    inr_init (argc, argv, "1.0", Ucmd, Udetail);
    if (!igetopt1 ("-ifile", "%s", gc->fnameInput))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-class", "%s", gc->fnameClassif))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-roi"  , "%s", gc->fnameROIMask))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-def"  , "%s", gc->fnameClassDef))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-ofile", "%s", gc->fnameOutput))
        iusage_ (Ucmd, Udetail);
    gc->okOutputProb = igetopt0 ("-oprob");
    gc->okTolerance  = igetopt1 ("-tol", "%lf", &gc->tolerance);
    gc->okKeep       = igetopt0 ("-keep");

    // exits program if there are unused options still present
    if (tstopts ())                
        iusage_ (Ucmd, Udetail);

    if (!gc->okTolerance)
        gc->tolerance = TOLERANCE_DEFAULT;

    gc->dataOutputProb = NULL;
    if (gc->okOutputProb)
        ftChangeExtension(gc->fnameOutput, gc->fnameOutputProb, ".prob.inr");
    
    // reads header information for all input data and checks if they are appropriate
    if (!inrReadHeader(gc->fnameInput, &gc->size_x, &gc->size_y, &gc->size_z, &gc->typeInput))
    {
        fprintf(stderr, "Error: could not read header information for input file '%s'.\n", gc->fnameInput);
        return 0;
    }
    if (!inrReadHeader(gc->fnameClassif, &i, &j, &k, &type))
    {
        fprintf(stderr, "Error: could not read header information for the input classification image '%s'.\n", gc->fnameClassif);
        return 0;
    }
    if (type != InrByte || k != 1)
    {
        fprintf(stderr, "Error: input classification image '%s' has to be a single image of type 'Byte'.\n", gc->fnameClassif);
        return 0;
    }
    if (i != gc->size_x || j != gc->size_y)
    {
        fprintf (stderr, "Error: dimensions of input data and classification image do not match.\n");
        return 0;
    }
    if (!inrReadHeader(gc->fnameROIMask, &i, &j, &k, &type))
    {
        fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", gc->fnameROIMask);
        return 0;
    }
    if (type != InrByte || k != 1)
    {
        fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", gc->fnameROIMask);
        return 0;
    }
    if (i != gc->size_x || j != gc->size_y)
    {
        fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
        return 0;
    }


    // reads the ROI mask image
    gc->ROIMask = inrReadByteImage(gc->fnameROIMask, &i, &j);
    if (!gc->ROIMask)
    {
        fprintf (stderr, "Error: could not read ROI mask data.\n");
        return 0;
    }
  
    // reads the classification data
    gc->dataClassif = inrReadByteImage (gc->fnameClassif, &i, &j);
    if (!gc->dataClassif)
    {
        fprintf (stderr, "Error: could not read classification data.\n");
        return 0;
    }

    // reads the class definitions
    gc->def = defCreateFromFile(gc->fnameClassDef);
    if (!gc->def)
    {
        return 0;
    }
    if (gc->def->nFeatures != gc->size_z)
    {
        fprintf (stderr, "Error: number of features in class definition file (%d) does not match number of layers in the input data (%d).\n",
                 gc->def->nFeatures, gc->size_z);
        return 0;
    }

    // creates MODIS STD
    gc->std = modisStdCreate();
    nStates = gc->std->nStates;
    nProfiles = gc->std->nProfiles;

    // checks if MODIS STD matches class definition file
    if (nProfiles != gc->def->nClass)
    {
        fprintf (stderr, "Error: number of class in class definition file (%d) does not match number of profile classes in the STD (%d).\n",
                 gc->def->nClass, nProfiles);
        return 0;
    }
    
    // allocates memory for the input and output data
    gc->typesizeInput    = InrTypeSizes[gc->typeInput]/8;
    gc->dataInput        = (unsigned char *) malloc (gc->size_x*gc->size_z * gc->typesizeInput * sizeof(unsigned char));
    gc->dataOutput       = (unsigned char *) malloc (gc->size_x*gc->size_y * sizeof(unsigned char));
    gc->cov_determinants = (double *) malloc (nProfiles*gc->def->nDiscreteValues * sizeof(double));
    gc->stateEval        = (double *) malloc (nStates * sizeof(double));
    gc->transitionEval   = (double *) malloc (nProfiles * sizeof(double));
    gc->values           = (double *) malloc (gc->size_z * sizeof(double));
    gc->point            = Matrice_Nulle(gc->def->nContFeatures, 1);
    gc->point_transp     = Matrice_Nulle(1, gc->def->nContFeatures);

    // inititalizes variables related to outputting the posterior probabilities (if specified)  
    if (gc->okOutputProb)
    {
        gc->dataOutputProb = (float *) malloc (gc->size_x*nStates * sizeof(float));
    
        // writes output file header (output type is always float)
        if (!inrWriteHeader(gc->fnameOutputProb, gc->size_x, gc->size_y, nStates, InrFloat))
        {
            fprintf(stderr, "Error: could not write header information for output file '%s'.\n", gc->fnameOutputProb);
            return 0;
        }
        // initializes data line
        for (i = 0; i < gc->size_x*nStates; i++)
            gc->dataOutputProb[i] = 0;
    }
  
    // computes the determinants of the covariance matrices
    if (!computeCovDeterminants(gc->def, gc->cov_determinants))
        return 0;

    // inverts the covariance matrices
    if (!invertCov(gc->def))
        return 0;
 


    // PROCESSES THE DATA
  
    for (y = 0; y < gc->size_y; y++)
    {
        fprintf (stderr, "Processing... %d%%\r", 100*y/gc->size_y);
    
        // checks if line y is within region of interest
        for (x = 0; x < gc->size_x; x++)
            if (gc->ROIMask[x + y*gc->size_x])
                break;

        // if no unmasked point was detected, line is not within ROI and is skipped 
        if (x == gc->size_x)
        {
            // all points in the line are rated unclassified
            for (x = 0; x < gc->size_x; x++)
                gc->dataOutput[x + y*gc->size_x] = 0;
            
            if (gc->okOutputProb)
            {
                // writes empty output probability data line
                if (!inrWriteLine(gc->fnameOutputProb, y, (unsigned char *)gc->dataOutputProb))
                {
                    fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, gc->fnameOutputProb);
                    return 0;
                }
            }
            continue;
        }

        // reads input data line y
        if (!inrReadLine(gc->fnameInput, y, gc->dataInput))
        {
            fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, gc->fnameInput);
            return 0;
        }
    
        // processes each point at a time
        for (x = 0; x < gc->size_x; x++)
        {
            double evaluation_max;

            // pixels outside of region of interest are rated unclassified
            if (!gc->ROIMask[x + y*gc->size_x])
            {
                gc->dataOutput[x + y*gc->size_x] = 0;
                continue;
            }
        

            // evaluates likelihood for the point
            evaluatePoint(gc, x, y);
      

            // assigns the class with highest likelihood to the point
            // (as long as it exceeds the minimum probability tolerance value)
            gc->dataOutput[x + y*gc->size_x] = 0;
            evaluation_max = gc->stateEval[0];
            for (iState = 1; iState < nStates; iState++)
            {
                if (gc->stateEval[iState] >= gc->tolerance && gc->stateEval[iState] > evaluation_max)
                {
                    gc->dataOutput[x + y*gc->size_x] = iState;
                    evaluation_max = gc->stateEval[iState];
                }
            }
            // if no evaluation exceeded the tolerance value and 'keep' option was
            // specified, then previous classification's state is kept as the current
            // state
            if (gc->okKeep && gc->dataOutput[x + y*gc->size_x] == 0)
                gc->dataOutput[x + y*gc->size_x] = gc->dataClassif[x + y*gc->size_x];;

      
            // adds point class evaluations (posterior probabilities) if specified
            if (gc->okOutputProb)
                for (iState = 1; iState < nStates; iState++)
                    gc->dataOutputProb[x + (iState-1)*gc->size_x] = (float) gc->stateEval[iState];
        }
        

        // writes line class evaluations (posterior probabilities) if specified
        if (gc->okOutputProb)
        {
            if (!inrWriteLine(gc->fnameOutputProb, y, (unsigned char *)gc->dataOutputProb))
            {
                fprintf(stderr, "Error: could not write line %d to output file '%s'.\n", y, gc->fnameOutputProb);
                return 0;
            }
            // re-initializes output data line
            for (i = 0; i < gc->size_x*nStates; i++)
                gc->dataOutputProb[i] = 0;
        }
    }


    // outputs results
    inrWriteByteImage(gc->fnameOutput, gc->size_x, gc->size_y, gc->dataOutput);

    // transfers georeference info if available
    if (inrReadGeoref(gc->fnameInput, &gc->georef))
    {
        inrWriteGeoref(gc->fnameOutput, gc->georef);
        if (gc->okOutputProb)
            inrWriteGeoref(gc->fnameOutputProb, gc->georef);
    }
    else if (gc->okOutputProb)
        inrWriteEnviHdr(gc->fnameOutputProb);
    
    fprintf(stderr, "Image classification successfully completed.\n");

    free(gc->dataInput);
    free(gc->dataClassif);
    free(gc->dataOutput);
    free(gc->values);
    free(gc->transitionEval);
    free(gc->stateEval);
    free_matrice(&gc->point);
    free_matrice(&gc->point_transp);
    free(gc->cov_determinants);
    defDestroy(&gc->def);
  
    return 0;
}
