#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE       1024



static char Ucmd[] = "-ifile inr_classification -truth inr_ground_truth -roi roi_mask\n";

static char Udetail[] =
"Given a classification image, compares it to a specified ground truth image\n\
so as to evaluate its accuracy.\n\
\nStatistics computed for comparison are:\n\
  . Confusion matrix\n\
  . Producer's, user's and overall accuracies\n\
  . Kappa coefficient\n\
These statistics are computed twice:\n\
  1. For evaluating misclassification (i.e., if class indices agree). Pixels\n\
     left unclassified are ignored in these statistics\n\
  2. For evaluating the relation between classified/unclassified pixels and\n\
     pixels with/without ground truth (i.e., if pure pixels and mixed pixels\n\
     were correctly identified)\n\
All input files must be in INRIMAGE format.\n\
Output is given in stdout.\n\
\nInput:\n\
\t-ifile\tFile name of the input classification image\n\
\t-truth\tFile name of a classification image to be used as ground truth\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n";



void computeProdAcc(int *conf_matrix, double *prod_acc, int nClass)
{
  int i, j;
  int sum;
  
  for (j=0; j<nClass; j++)
  {
    sum = 0;
    for (i=0; i<nClass; i++)
      sum += conf_matrix[j + i*nClass];
      
    if (sum == 0)
      prod_acc[j] = 0;
    else
      prod_acc[j] = (double)conf_matrix[j + j*nClass]/sum;
  }
}

void computeUserAcc(int *conf_matrix, double *user_acc, int nClass)
{
  int i, j;
  int sum;
  
  for (i=0; i<nClass; i++)
  {
    sum = 0;
    for (j=0; j<nClass; j++)
      sum += conf_matrix[j + i*nClass];
      
    if (sum == 0)
      user_acc[i] = 0;
    else
      user_acc[i] = (double)conf_matrix[i + i*nClass]/sum;
  }
}

double computeOverallAcc(int *conf_matrix, int nClass)
{
  int i;
  int sum_diag  = 0;
  int sum_total = 0;
  
  // total number of correctly classified pixels
  for (i=0; i<nClass; i++)
    sum_diag += conf_matrix[i + i*nClass];
  
  // total number of pixels
  for (i=0; i<nClass*nClass; i++)
    sum_total += conf_matrix[i];

  return (double)sum_diag/sum_total;
}

double computeKappa(int *conf_matrix, int nClass)
{
  int class, i,j;
  double sum_diag  = 0;
  double sum_total = 0;
  double sum_mult = 0;
  double sum_line = 0;
  double sum_col  = 0;
  
  // total number of correctly classified pixels
  for (i=0; i<nClass; i++)
    sum_diag += conf_matrix[i + i*nClass];
  
  // total number of pixels
  for (i=0; i<nClass*nClass; i++)
    sum_total += conf_matrix[i];
    
  // sum of sum_line*sum_col
  for (class=0; class<nClass; class++)
  {
    sum_line = 0;
    for (j=0; j<nClass; j++)
      sum_line += conf_matrix[j + class*nClass];

    sum_col = 0;
    for (i=0; i<nClass; i++)
      sum_col += conf_matrix[class + i*nClass];
      
    sum_mult += sum_line*sum_col;
  }
  return (sum_total*sum_diag - sum_mult)/(sum_total*sum_total - sum_mult);
}

double computePercClassified(int *nPurePixels, int *nClassified, double *perc_classified, int nClass)
{
   int i;
   double sum_purepixels = 0;
   double sum_classified = 0;
   
   for (i=0; i<nClass; i++)
   {
     perc_classified[i] = (double)nClassified[i]/nPurePixels[i];
     sum_classified += nClassified[i];
     sum_purepixels += nPurePixels[i];
   }
   
   return sum_classified/sum_purepixels;
}



int outputConfMatrix(int *conf_matrix, int nClass)
{
  int i, j;
  int sum;
  int totalsum;

  printf("\n");
  printf("Confusion Matrix\n");
  printf("----------------\n");

  // column numbers (classes)
  printf("\tGround Truth\n");
  printf("\t");
  for (j=0; j<nClass; j++)
    printf("%d\t", j+1);
  printf("\tTotal");
  printf("\n");
  
  // separator
  for (j=0; j<nClass+3; j++)
    printf("--------");
  printf("\n");
  
  // matrix itself
  for (i=0; i<nClass; i++)
  {
    // line numbers (classes)
    printf("%d\t", i+1);

    // line values
    for (j=0; j<nClass; j++)
      printf("%d\t", conf_matrix[j + i*nClass]);

    // line sums
    sum = 0;
    for (j=0; j<nClass; j++)
      sum += conf_matrix[j + i*nClass];
    printf("|\t%d\n", sum);
  }
  
  // separator
  for (j=0; j<nClass+3; j++)
    printf("--------");
  printf("\n");
  
  // column sums and total sum
  printf("Total\t");
  totalsum = 0;
  for (j=0; j<nClass; j++)
  {
    sum = 0;
    for (i=0; i<nClass; i++)
      sum += conf_matrix[j + i*nClass];
    printf("%d\t", sum);
    totalsum += sum;
  }
  printf("|\t%d\n\n", totalsum);

  return 1;
}


void outputAccuracies(double *prod_acc, double *user_acc, double overall_acc, double kappa, int nClass)
{
  int i;
  
  printf("Accuracies\n");
  printf("----------\n");
  
  printf("\tProducer's Accuracies\n");
  for (i=0; i<nClass; i++)
    printf("\t%d", i+1);
  printf("\n");
  for (i=0; i<nClass; i++)
    printf("\t%.2lf%%", prod_acc[i]*100);
  printf("\n\n");
  
  printf("\tUser's Accuracies\n");
  for (i=0; i<nClass; i++)
    printf("\t%d", i+1);
  printf("\n");
  for (i=0; i<nClass; i++)
    printf("\t%.2lf%%", user_acc[i]*100);
  printf("\n\n");
  
  printf("\tOVERALL ACCURACY:  %.2lf%%\n\n", overall_acc*100);
  printf("\tKAPPA COEFFICIENT: %.4lf\n\n", kappa);
}

void outputPercClassified(double *perc_classified, double overall_perc_classified, int nClass)
{
  int i;
  
  printf("Percentage of pixels with truth that were classified (per class)\n");
  printf("----------------------------------------------------------------\n");
  for (i=0; i<nClass; i++)
    printf("\t%d", i+1);
  printf("\n");
  for (i=0; i<nClass; i++)
    printf("\t%.2lf%%", perc_classified[i]*100);
  printf("\n\n");
  printf("\tOVERALL: %.2lf%%\n\n", overall_perc_classified*100);
}



int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameTruth[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  InrType type;
  
  unsigned char *dataInput;
  unsigned char *dataTruth;
  unsigned char *ROIMask;
  
  int nClass;
  int classInput;
  int classTruth;
  
  int size_x, size_y, size_z;
  int i, j, k;
  
  // statistics
  int    *classif_conf_matrix;
  int    *puremixed_conf_matrix;
  double *prod_acc;
  double *user_acc;
  double  overall_acc = 0;
  double  kappa;
  double *perc_classified;
  double  overall_perc_classified = 0;
  
  // auxiliary variables for computing statistics
  int *nPurePixels;
  int *nClassified;
  int ispurepixel;
  int classified;
  int pos;
  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-truth", "%s", fnameTruth))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"  , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &type))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (type != InrByte || size_z != 1)
  {
    fprintf(stderr, "Error: Input classification image '%s' has to be a single image of type 'Byte'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameTruth, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ground truth image '%s'.\n", fnameTruth);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: Ground truth classification image '%s' has to be a single image of type 'Byte'.\n", fnameTruth);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input and ground truth images do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
 
  // reads the input classification data
  dataInput = inrReadByteImage (fnameInput, &i, &j);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not read input classification data.\n");
    return 0;
  }
  // reads the ground truth classification data
  dataTruth = inrReadByteImage (fnameTruth, &i, &j);
  if (!dataTruth)
  {
    fprintf (stderr, "Error: could not read ground truth classification data.\n");
    return 0;
  }
  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }


  // verifies total number of classes in the input classification image
  nClass = 0;
  for (i=0; i<size_x*size_y; i++)
    nClass = MAX(nClass, dataInput[i]);
  for (i=0; i<size_x*size_y; i++)
    nClass = MAX(nClass, dataTruth[i]);


  // allocates memory 
  classif_conf_matrix   = (int *) malloc (nClass*nClass * sizeof(int));
  puremixed_conf_matrix = (int *) malloc (4 * sizeof(int));
  prod_acc              = (double *) malloc (nClass * sizeof(double));
  user_acc              = (double *) malloc (nClass * sizeof(double));
  perc_classified       = (double *) malloc (nClass * sizeof(double));
  nPurePixels           = (int *) malloc (nClass * sizeof(int));
  nClassified           = (int *) malloc (nClass * sizeof(int));
  
  // initializes variables
  for (i=0; i<nClass*nClass; i++)
    classif_conf_matrix[i] = 0;
  for (i=0; i<4; i++)
    puremixed_conf_matrix[i] = 0;
  
  for (i=0; i<nClass; i++)
  {
    nPurePixels[i] = 0;
    nClassified[i] = 0;
  }


  // PROCESSES THE DATA

  for (i=0; i<size_x*size_y; i++)
  {
    // skips pixels outside of the region of interest
    if (!ROIMask[i])
      continue;
      
    // pixels within ROI with truth are pure pixels, being mixed otherwise
    ispurepixel = (dataTruth[i] != 0);
    classified  = (dataInput[i] != 0);
    classTruth = dataTruth[i]-1;
    classInput = dataInput[i]-1;
      
    // increments pure/mixed confusion matrix at the appropriate position
    // (line 'classified/unclassified', column 'purepixel/mixedpixel')
    pos = (ispurepixel) + (classified)*2;
    puremixed_conf_matrix[pos]++;
    
    // increments counters for per-class identification of pure pixels
    if (ispurepixel)
    {
      nPurePixels[classTruth]++;
      if (dataInput[i])
        nClassified[classTruth]++;
    }

    // remaining stats apply only if pixel has truth and was classified
    if (!dataTruth[i] || !dataInput[i])
      continue;

    // increments classification/misclassification confusion matrix at the
    // appropriate position (line 'input', column 'truth')
    pos = (classTruth) + (classInput)*nClass;
    classif_conf_matrix[pos]++;
    
    //*** TEMP - OUTPUT MISCLASSIFIED DEFOREST POINTS ***/
    /*
    if (classTruth != classInput)
    {
      int x = i%size_x;
      int y = i/size_x;
      char strfile[100];
      
      sprintf(strfile, "deforest.error.class%02d.txt", classTruth+1);

      FILE *fp = fopen(strfile, "at+");
      if (fp != NULL)
      {
        fprintf(fp, "%d\t%d\t1\n", x, y);
        fclose(fp);
      }
    }
    */
  }


  // COMPUTES STATISTICS

  // pure/mixed pixel identification
  printf("\n");
  printf("PIXELS WITH/WITHOUT TRUTH VS. CLASSIFIED/NOT-CLASSIFIED\n");
  printf("-------------------------------------------------------\n");
  computeProdAcc(puremixed_conf_matrix, prod_acc, 2);
  computeUserAcc(puremixed_conf_matrix, user_acc, 2);
  overall_acc = computeOverallAcc(puremixed_conf_matrix, 2);
  kappa = computeKappa(puremixed_conf_matrix, 2);
  overall_perc_classified = computePercClassified(nPurePixels, nClassified, perc_classified, nClass);

  outputConfMatrix(puremixed_conf_matrix, 2);
  outputAccuracies(prod_acc, user_acc, overall_acc, kappa, 2);
  outputPercClassified(perc_classified, overall_perc_classified, nClass);


  // classification/misclassification
  printf("\n\n");
  printf("CLASSIFICATION AGREEMENT\n");
  printf("------------------------\n");
  computeProdAcc(classif_conf_matrix, prod_acc, nClass);
  computeUserAcc(classif_conf_matrix, user_acc, nClass);
  overall_acc = computeOverallAcc(classif_conf_matrix, nClass);
  kappa = computeKappa(classif_conf_matrix, nClass);

  outputConfMatrix(classif_conf_matrix, nClass);
  outputAccuracies(prod_acc, user_acc, overall_acc, kappa, nClass);
    

/* TEMP    
  free(dataInput);
  free(dataTruth);
  free(classif_conf_matrix);
  free(puremixed_conf_matrix);
  free(prod_acc);
  free(user_acc);
  free(perc_classified);
  free(nPurePixels);
  free(nClassified);
  */
  return 0;
}
