#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "linearalgebra.h"
#include "classdef.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE               1024


static char Ucmd[] = "-ifile inr_input -class inr_classification -ofile class_definitions\n";

static char Udetail[] =
"Given input data and a training classification file, outputs a file that\n\
describes the classes' probability distributions.\n\
\nClasses are assumed to have a gaussian probability distribution.\n\
Output file will contain the mean value for each layer in the input file for\n\
each class, along with the covariance matrices for each class.\n\
All input files must be in INRIMAGE format.\n\
\nInput:\n\
\t-ifile\tFile name of the input data\n\
\t-class\tFile name of a classification image\n\
\nOutput:\n\
\t-ofile\tFile name of the output text file containing the class\n\
\t\tdefinitions\n";







int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameClassif[FNAMESIZE];
  InrType typeInput;
  InrType type;
  FILE   *fpOutput;
  
  unsigned char *dataInput;
  unsigned char *classif;
  
  int nClass;
  int iClass;
  int class;
  
  matrice  *mean;
  matrice  *cov;
  int      *nInstances;
  
  double *contValues;
  int     iDiscrete;
  
  int size_x, size_y, size_z;
  int i, j, k;
  int pos;
  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-class", "%s", fnameClassif))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameClassif, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for classification image '%s'.\n", fnameClassif);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: Classification image '%s' has to be a single image of type 'Byte'.\n", fnameClassif);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and classification image do not match.\n");
    return 0;
  }
  fpOutput = fopen(fnameOutput, "wt");
  if (fpOutput == NULL)
  {
    fprintf (stderr, "Error: could not open output file '%s' for writing.\n", fnameOutput);
    return 0;
  }
 
  // reads the input data
  dataInput = inrReadDataSequence(fnameInput, &i, &j, &k, &type);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not input data.\n");
    return 0;
  }
  // reads the classification training data
  classif = inrReadByteImage (fnameClassif, &i, &j);
  if (!classif)
  {
    fprintf (stderr, "Error: could not read classification data.\n");
    return 0;
  }
  // verifies total number of classes in the classification image
  nClass = 0;
  for (i=0; i<size_x*size_y; i++)
    nClass = MAX(nClass, classif[i]);

 
  // PROCESSES THE DATA
  
  // 1. finds discrete features and computes their statistics for each class
  classdef *def = defCreate(nClass, size_z);
  defProcessDiscrete(def, classif, dataInput, size_x, size_y, size_z, typeInput);
  defComputeDiscreteProb(def, classif, dataInput, size_x, size_y, size_z, typeInput);

  // allocates memory for step 2 and 3
  nInstances      = (int *) malloc (nClass*def->nDiscreteValues * sizeof(int));
  contValues      = (double *) malloc (size_z * sizeof(double));
  for (i=0; i<nClass*def->nDiscreteValues; i++)
    nInstances[i] = 0;

  // 2. computes means for continuous layers for all classes and all discrete values
  for (pos=0; pos<size_x*size_y; pos++)
  {
    if (pos%size_x == 0)
      fprintf (stderr, "Computing mean values... %d%%\r", 100*pos/(size_x*size_y));

    // reads classification of the point
    class = classif[pos];

    // discards non-classified points
    if (!class)
      continue;
    iClass = class -1;

    // gets point values for the position pos
    if (!defGetPoint(def, pos, dataInput, size_x, size_y, typeInput, contValues, &iDiscrete))
      continue;

    // reads values for each layer of continous data into the appropriate mean
    mean = defGetContMean(def, iClass, iDiscrete);
    for (j=0; j<def->nContFeatures; j++)
      mean->val[j] += contValues[j];

    // increments instance counter for point's class and discrete value
    nInstances[iDiscrete + iClass*def->nDiscreteValues]++;
  }
  // finalizes mean computation by dividing total sum by nInstances
  for (iClass=0; iClass<nClass; iClass++)
  {
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      int n = nInstances[iDiscrete + iClass*def->nDiscreteValues];
      mean = defGetContMean(def, iClass, iDiscrete);
      if (n > 0)
        Multiplication_Scalaire_Matrice(1./n, mean);
    }
  }

  // 3. computes covariance matrices for all classes
  for (pos=0; pos<size_x*size_y; pos++)
  {
    if (pos%size_x == 0)
      fprintf (stderr, "Computing covariance values... %d%%\r", 100*pos/(size_x*size_y));

    // reads classification of the point
    class = classif[pos];

    // discards non-classified points
    if (!class)
      continue;
    iClass = class -1;

    // gets point values for the position pos
    if (!defGetPoint(def, pos, dataInput, size_x, size_y, typeInput, contValues, &iDiscrete))
      continue;

    // subtracts the mean for the continuous values: (x-m)
    mean = defGetContMean(def, iClass, iDiscrete);
    for (j=0; j<def->nContFeatures; j++)
      contValues[j] -= mean->val[j];
        
    // updates covariance matrix: sigma += (x-m) * t(x-m)
    cov = defGetContCovariance(def, iClass, iDiscrete);
    for (i=0; i<def->nContFeatures; i++)
      for (j=0; j<def->nContFeatures; j++)
        cov->val[j + i*def->nContFeatures] += contValues[i]*contValues[j];
  }
  // finalizes covariances computation by dividing total sum by nInstances
  for (iClass=0; iClass<nClass; iClass++)
  {
    for (iDiscrete=0; iDiscrete<def->nDiscreteValues; iDiscrete++)
    {
      int n = nInstances[iDiscrete + iClass*def->nDiscreteValues];
      cov = defGetContCovariance(def, iClass, iDiscrete);
      if (n > 0)
        Multiplication_Scalaire_Matrice(1./n, cov);
    }
  }


  // outputs results
  defOutputTxt(def, fpOutput);
    
  fprintf(stderr, "Classification training successfully completed.\n");

  free(dataInput);
  free(classif);
  free(contValues);
  free(nInstances);
  defDestroy(&def);
  
  return 0;
}
