/***************************************************************************
                          modis_classification_train_ds.c  -  description
                             -------------------
    Created              : 7 August, 2007
    By                   : (C) 2007 by Carlos de la Torre
    Email                : Carlos.de_la_Torre@inria.fr
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <inrimage/image.h>
#include "infra/inr_io.h"
#include "infra/linearalgebra.h"
#include "infra/fuzzyclassdef.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE         1024
#define STRATSIZE         32

// If this constants change, be sure to update the Udetail
#define FUSION_DR  "fusionDR"

static char Ucmd[] = "-ifile inr_input -class inr_classification -strategy strat -config config_file -ofile class_definitions\n";

static char Udetail[] =
"Outputs a file that describes the classes present in the input classification.\n\
In the description of every class, a specific subset of features are used\n\
(Those that better represent the class)\n\
The features to be used in the representation of every class are specified in the config_file\n\
The description of a class comprises the features means and covariance matrix.\n\
Classes are assumed to have a gaussian probability distribution.\n\
All input files must be in INRIMAGE format.\n\
\nInput:\n\
\t-ifile\tFile name of the input data\n\
\t-class\tFile name of a classification image\n\
\t-config\tName of the text file with the specification of the features to be used to represent every class.\n\
\nOutput:\n\
\t-ofile\tFile name of the output text file containing the classes definitions\n";






int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameClassif[FNAMESIZE];
  char fnameConfig[FNAMESIZE];
  InrType typeInput;
  InrType type;
  FILE   *fpOutput;
  
  unsigned char *dataInput;
  unsigned char *classif;
  
  int    nClass;
  
  int size_x, size_y, size_z;
  int i, j, k;

  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-class", "%s", fnameClassif))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-config", "%s", fnameConfig))
    iusage_ (Ucmd, Udetail);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameClassif, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for classification image '%s'.\n", fnameClassif);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: Classification image '%s' has to be a single image of type 'Byte'.\n", fnameClassif);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and classification image do not match.\n");
    return 0;
  }
  fpOutput = fopen(fnameOutput, "wt");
  if (fpOutput == NULL)
  {
    fprintf (stderr, "Error: could not open output file '%s' for writing.\n", fnameOutput);
    return 0;
  }
 
  // reads the input data
  dataInput = inrReadDataSequence(fnameInput, &i, &j, &k, &type);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not input data.\n");
    return 0;
  }
  // reads the classification training data
  classif = inrReadByteImage (fnameClassif, &i, &j);
  if (!classif)
  {
    fprintf (stderr, "Error: could not read classification data.\n");
    return 0;
  }
  // verifies total number of classes in the classification image
  nClass = 0;
  for (i=0; i<size_x*size_y; i++)
    nClass = MAX(nClass, classif[i]);

 
  // PROCESSES THE DATA

  //Retrieve information from the configuration file
  classdef *def = defCreate(nClass, size_z, FusionDR);
  if(!defConfigFromFile(def, fnameConfig))
  {
    fprintf (stderr, "Error: Problems with the configuration file %s.\n", fnameConfig);
    return 0;
  }
  //Compute the definitions, based on the strategy
  defComputeDefinitions(def, dataInput, type, classif, size_x, size_y, size_z);

  // outputs definitions into a file
  if(!defToTextFile(def, fpOutput))
  {
    fprintf(stderr, "Error: Couldn't write output file %s.\n",fnameOutput);
    return 0;
  }

  fprintf(stderr, "Classification training successfully completed.\n");

  fclose(fpOutput);
  free(dataInput);
  free(classif);
  defDestroy(def);
  
  return 0;
}



// // /*
// // // In the fusion strategy, every class is represented by two arrays;
// // //  - The mean of every feature in that class,
// // //  - The standard deviation of every feature in that class.
// // //  
// // // This function computes those arrays and outputs them to a file.
// // void fusionTrain(data, classif, x, y, nFeatures, nClasses, FILE* fpOutput)
// // {
// //   float mean[nClasses][nFeatures];
// //   float S[nClasses][nFeatures];
// //   int   size[nClasses];
// // 
// //   int pos=0, iClass=0, iFeat=0;
// //   float fValue=0.0;
// // 
// //   // PROCESSES THE DATA
// //   //
// //   for (pos=0; pos < x*y; pos++)
// //   {
// //     if (pos%x == 0)
// //       fprintf (stderr, "Computing mean values... %d%%\r", 100*pos/(size_x*size_y));
// // 
// //     // reads classification of the point
// //     iClass = classif[pos];
// // 
// //     // discards non-classified points
// //     if (!iClass)
// //       continue;
// // 
// //     // Initialize arrays
// //     for (iClass=0; iClass<nClasses; iClass++)
// //     {
// //       size[iClass]=0;
// //       for (iFeat=0; iFeat<nFeatures; iFeat++)
// //       {
// //         mean[iClass][iFeat] = 0.0;
// //         S[iClass][iFeat] = 0.0;
// //       }
// //     }
// //     
// //     // Recursive calculus of mean and stddev.
// //     // Algorithm III in http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance.
// //     //  foreach x in data:
// //     //    n = n + 1
// //     //    delta = x - mean
// //     //    mean = mean + delta/n
// //     //    S = S + delta*(x - mean)      // This expression uses the new value of mean
// //     //  end for
// //     //  variance = S/(n - 1)
// //     size[iClass]++;
// //     for (iFeat=0; iFeat<nFeatures; iFeat++)
// //     {
// //       fValue = data[pos + x*y*iFeat]
// //       delta = fValue - mean[iClass][iFeat];
// //       mean[iClass][iFeat] += delta/(float)size[iClass];
// //       S[iClass][iFeat] += delta*(fValue-mean[iClass][iFeat]);
// //     }
// //   }
// // 
// //   for (iClass=0; iClass<nClasses; iClass++)
// //   {
// //     for (iFeat=0; iFeat<nFeatures; iFeat++)
// //     {
// //       S[iClass][iFeat] = sqrt(S[iClass][iFeat]/(size[iClass] - 1.0));
// //     }
// //   }
// //   // outputs results into a file
// //   if(!fusionDefToTextFile(mean, S, fpOutput))
// //   {
// //     return 0;
// //   }
// //   
// //   return 1;
// // }
// // 
// // int fusionDefToTextFile(float** mean, float** S, FILE *fpOutput)
// // {
// //   
// // }*/

