#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "noaa_dates.h"

#define HDRBUFSIZE 4096
#define FNAMESIZE  1024

static char Ucmd[] = "-ifile inr_sequence -mask cloud_mask -roi roi_mask \
-ofile corrected_inr_sequence [filter options]";

static char Udetail[] = 
"\nRemoves cloudy pixels through temporal interpolation and optionnaly also\n\
filters the temporal signal.\n\
\nInput:\n\
\t-ifile\tFile name of the image sequence to be corrected\n\
\t-mask\tFile name of the corresponding cloud mask sequence\n\
\t\tMask should be zero for cloudy pixels and non-zero otherwise\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name for the output sequence containing corrected data\n\
\nOptional:\n\
\t-clean\t\tSuppress values outside mean +- 2stdev.\n\
\t-median win\tPerforms a median filtering using a moving window\n\
\t\t\tTotal window size will be (2*win + 1)\n\
\t-localvar win n\tPerforms local variance filtering using a moving window\n\
\t\t\tTotal window size will be (2*win + 1)\n\
\t\t\tValues outside [mean +/- n*stddev] will be filtered out\n\
\t-atmo [u|l]\tAtmosphere filtering. Does a morphological filtering of\n\
\t\t\tatmosphere effects.\n\
\t\t\t-atmo u\tPerforms a closing of the signal (if atmosphere\n\
\t\t\t\t tends to raise signal (visible))\n\
\t\t\t-atmo l\tPerforms an opening of the signal (if atmosph.\n\
\t\t\t\t tends to lower signal (other channels)).\n\
\t\t\tDefault=no atmosphere filtering,\n\
\t-win nb\t\tIf morphological filtering of atmosphere is chosen,\n\
\t\t\ttunes size of structuring element (see documentation\n\
\t\t\tfor more details)\n\
\t\t\tDefault=2.\n\
\t-daily datefile\tInterpolation of the signal on a daily basis instead of\n\
\t\t\ton data basis.\n\
\t\t\t'datefile' specifies a list of dates for each image in\n\
\t\t\tthe sequence\n\
\t\t\tDefault=data basis\n";

float *work;

#define MAX( X, Y ) ((X>Y)?X:Y)
#define MIN( X, Y ) ((X<Y)?X:Y)



/* dilatation d'un signal, 1/2 taille=win */
void dilate( float  * ori , float  * dil , int win , int sz) ;

/* erosion d'un signal de n float, 1/2 taille=win */
void erode( float  * ori , float  * ero , int win , int sz);


/* ouverture ... ori et ouv peuvent etre les meme */
void ouverture( float  *ori , float  *ouv , int win , int sz ) ;


/* fermeture ... ori et fer peuvent etre les meme */
void fermeture( float  *ori , float  *fer , int win , int sz ) ;

/* compute mean and stdev of a signal */
void signal_stats( float * sig , unsigned char * signalCloudMask , float * m , float * s , int sz ) ;

/* suppress values of a signal outside mean +- 2*stdev */
void signal_clean_extrema( float * sig , unsigned char * signalCloudMask , int sz ) ;


/* copy one temporal signal onto another */
void copy_signal( float * in , float * out , int sz ) ;

/* set the temporal signal for one pixel of a sequence */
void insert_signal( float * sig , float  * image_line , int x , int sx, int sz );

/* same as above, with signal=0 */
void insert_null_signal(float  * image_line , int x , int sx, int sz);

/* get the temporal signal for one pixel of a sequence */
int extract_signal( float * sig , unsigned char  * image_line , int x , int sx, int sz , InrType type );

/* same as above for byte images, adapted to masks */
int extract_signal_byte (unsigned char *sig, unsigned char *sequence_line, int x, int sx, int sz);


/* create a sequence of indices of clear sky value, return number of clear sky values */
int mark_clouds( unsigned char * signalCloudMask , int * indicesClearSky , int sz );


/* met a 0 les dates nuageuses 
   RESULTAT DANS LE MEME SIGNAL */
void suppress_cloudy_dates(float  *sig, unsigned char  *signalCloudMask, int sz ) ;

/* pack a signal retaining only non masked values 
 to be used before morphological filters are applied */
void pack_signal( float * sig , int  * indicesClearSky , int sz , int nClearSky ) ;

/* opposite of above */
void unpack_signal( float * sig , int  * indicesClearSky , int sz , int nClearSky );

/* expands a signal from data to day basis */
void expand_signal_per_day( float * signal , int * indicesClearSky , float * signalOutput, int nClearSky , int szdaily );

/* interpolation of missing data */
void interpolate_cloudy_dates( float * data  , int * indicesClearSky , int nClearSky , int sz ) ;



/* median filter */
void signal_filter_median (float *sig, float *sig_filtered, int win, int sz)
{
  int iSig, iWin, iWork, iWork2;
  int wmin, wmax, wsize;
  float swap_temp;
  
  for (iSig=0; iSig<sz; iSig++)
  {
    wmin  = MAX (0, iSig - win);
    wmax  = MIN (sz - 1, iSig + win);
    wsize = wmax - wmin + 1;

    // copies values within window to a working area
    iWork = 0;    
    for (iWin=wmin; iWin<=wmax; iWin++)
      work[iWork++] = sig[iWin];

    // bubblesorts the values within the working window
    for (iWork=0; iWork<wsize; iWork++)
    {
      for (iWork2=iWork+1; iWork2<wsize; iWork2++)
      {
        if (work[iWork] > work[iWork2])
        {
          swap_temp    = work[iWork];
          work[iWork]  = work[iWork2];
          work[iWork2] = swap_temp;
        }
      }
    }
    
    // filtered value is the median of the sorted values    
    sig_filtered[iSig] = work[wsize/2];
  }
}


/* local variance filter */
void signal_filter_localvar (float *sig, float *sig_filtered, int win, double n_stddev, int sz)
{
  int    iSig, iWin;
  int    wmin, wmax, wsize;
  double localmean, localvar, localstd;
  
  for (iSig=0; iSig<sz; iSig++)
  {
    wmin  = MAX (0, iSig - win);
    wmax  = MIN (sz - 1, iSig + win);
    wsize = wmax - wmin + 1;
    
    if (wsize < 3)
    {
      sig_filtered[iSig] = sig[iSig];
      continue;
    }

    // computes local mean (without current value)
    localmean = 0;
    for (iWin=wmin; iWin<iSig; iWin++)
      localmean += sig[iWin];
    for (iWin=iSig+1; iWin<=wmax; iWin++)
      localmean += sig[iWin];
    localmean /= (wsize-1);
  
    // computes local standard deviation (without current value)
    localvar  = 0;
    for (iWin=wmin; iWin<iSig; iWin++)
      localvar += (sig[iWin] - localmean) * (sig[iWin] - localmean);
    for (iWin=iSig+1; iWin<=wmax; iWin++)
      localvar += (sig[iWin] - localmean) * (sig[iWin] - localmean);
    localvar /= (wsize-2);
    localstd = sqrt(localvar);

    if ((sig[iSig] > (localmean + n_stddev*localstd)) ||
        (sig[iSig] < (localmean - n_stddev*localstd)))
      sig_filtered[iSig] = localmean;
    else
      sig_filtered[iSig] = sig[iSig];
  }
}


/* dilatation d'un signal, 1/2 taille=win */
void dilate (float *ori, float *dil, int win, int sz)
{
  int i, j;
  int wmin, wmax;

  for (i = 0; i < sz; i++)
  {
    wmin = MAX (0, i - win);
    wmax = MIN (sz - 1, i + win);
    dil[i] = ori[i];
    for (j = wmin; j <= wmax; j++)
      dil[i] = MAX (dil[i], ori[j]);
  }
}

/* erosion d'un signal de n float, 1/2 taille=win */
void erode (float *ori, float *ero, int win, int sz)
{
  int i, j;
  int wmin, wmax;

  for (i = 0; i < sz; i++)
  {
    wmin = MAX (0, i - win);
    wmax = MIN (sz - 1, i + win);
    ero[i] = ori[i];
    for (j = wmin; j <= wmax; j++)
      ero[i] = MIN (ero[i], ori[j]);
  }
}

/* ouverture ... ori et ouv peuvent etre les meme */
void ouverture (float *ori, float *ouv, int win, int sz)
{
  dilate (ori, work, win, sz);
  erode (work, ouv, win, sz);
}

/* fermeture ... ori et fer peuvent etre les meme */
void fermeture (float *ori, float *fer, int win, int sz)
{
  erode (ori, work, win, sz);
  dilate (work, fer, win, sz);
}

/* compute mean and stdev of a signal */
void signal_stats (float *sig, unsigned char *signalCloudMask, float *m, float *s, int sz)
{
  int z, n = 0;
  float mean = 0.0, var = 0.0;

  for (z = 0; z < sz; z++)
  {
    if (signalCloudMask[z])
    {
      mean += sig[z];
      n++;
    }
  }
  mean /= (float) n;

  for (z = 0; z < sz; z++)
    if (signalCloudMask[z])
      var += (sig[z] - mean) * (sig[z] - mean);
  var /= (float) (n - 1);

  *m = mean;
  *s = (float) sqrt (var);
}

/* pack a signal retaining only non masked values 
 to be used before morphological filters are applied */
void pack_signal (float *sig, int *indicesClearSky, int sz, int nClearSky)
{
  int i;
  for (i = 0; i < nClearSky; i++)
    work[i] = sig[indicesClearSky[i]];
  copy_signal (work, sig, sz);
}

/* opposite of above */
void unpack_signal (float *sig, int *indicesClearSky, int sz, int nClearSky)
{
  int i;
  for (i = 0; i < sz; i++)
    work[i] = 0;
  for (i = 0; i < nClearSky; i++)
    work[indicesClearSky[i]] = sig[i];
  copy_signal (work, sig, sz);

}

/* expands a signal from data to day basis */
void expand_signal_per_day (float *signal, int *indicesClearSky, float *signalOutput,
                       int nClearSky, int szdaily)
{
  int i, day;
  /* set out to 0 */
  for (i = 0; i < szdaily; i++)
    signalOutput[i] = 0;
  /* scan non masked values */
  for (i = 0; i < nClearSky; i++)
  {
    day = day_since_beginning (indicesClearSky[i]);
    signalOutput[day] = signal[indicesClearSky[i]];
    indicesClearSky[i] = day;
  }
}

/* interpolation of missing data */
void interpolate_cloudy_dates (float *data, int *indicesClearSky, int nClearSky, int sz)
{
  int i, j;
  float pente;

  /* dates until 1st clear sky day: constant */
  for (i = 0; i < indicesClearSky[0]; i++)
    data[i] = data[indicesClearSky[0]];
  /* idem en fin de signal */
  for (i = indicesClearSky[nClearSky - 1] + 1; i < sz; i++)
    data[i] = data[indicesClearSky[nClearSky - 1]];

  /* interpolate in between */
  for (i = 0; i < nClearSky - 1; i++)
  {
    /* if no missing data, do nothing */
    if (indicesClearSky[i + 1] == indicesClearSky[i] + 1)
      continue;
    pente = (data[indicesClearSky[i + 1]] - data[indicesClearSky[i]]) /
      (float) (indicesClearSky[i + 1] - indicesClearSky[i]);
    for (j = indicesClearSky[i] + 1; j < indicesClearSky[i + 1]; j++)
      data[j] = data[indicesClearSky[i]] + pente * (float) (j - indicesClearSky[i]);
  }
}



/* suppress values of a signal outside mean +- 2*stdev */
void signal_clean_extrema (float *sig, unsigned char *signalCloudMask, int sz)
{
  float mean, stdev;
  int z;

  signal_stats (sig, signalCloudMask, &mean, &stdev, sz);
  for (z = 0; z < sz; z++)
  {
    if (signalCloudMask[z])
    {
      if (sig[z] < mean - 2.0 * stdev || sig[z] > mean + 2.0 * stdev)
      {
        sig[z] = 0.0;
        signalCloudMask[z] = 0;
      }
    }
  }
}

/* copy one temporal signal onto another */
void copy_signal (float *in, float *out, int sz)
{
  int i;
  for (i = 0; i < sz; i++)
    out[i] = in[i];
}

/* set the temporal signal for one pixel of a sequence */
void insert_signal (float *sig, float *sequence_line, int x, int sx, int sz)
{
  int z;

  for (z = 0; z < sz; z++)
    sequence_line[x + z * sx] = sig[z];

}

/* same as above, with signal=0 */
void insert_null_signal (float *sequence_line, int x, int sx, int sz)
{
  int z;

  for (z = 0; z < sz; z++)
    sequence_line[x + z * sx] = 0.0;

}

/* get the temporal signal for one pixel of a sequence */
int extract_signal (float *sig, unsigned char *sequence_line, int x, int sx, int sz, InrType type)
{
  int pos, z;

  for (z = 0; z < sz; z++)
  {
    pos = x + z * sx;
    sig[z] = (float)inrGetDataValue(sequence_line, type, pos);
  }
  return 1;
}

/* get the temporal signal for one pixel of a sequence (unsigned char signal) */
int extract_signal_byte (unsigned char *sig, unsigned char *sequence_line, int x, int sx, int sz)
{
  int pos, z;

  for (z = 0; z < sz; z++)
  {
    pos = x + z * sx;
    sig[z] = sequence_line[pos];
  }
  return 1;
}



/* create a sequence of indices of clear sky value, return number of clear sky values */
int mark_clouds (unsigned char *signalCloudMask, int *indicesClearSky, int sz)
{
  int z, i;
  for (z = 0, i = 0; z < sz; z++)
    if (signalCloudMask[z] != 0)
      indicesClearSky[i++] = z;

  return i;
}

/* met a 0 les dates nuageuses 
   RESULTAT DANS LE MEME SIGNAL */
void suppress_cloudy_dates (float *sig, unsigned char *signalCloudMask, int sz)
{
  int i;

  for (i = 0; i < sz; i++)
    if (!signalCloudMask[i])
      sig[i] = 0.0;

}


int main (int argc, char **argv)
{
  unsigned char *lineInput; 
  InrType        typeInput;
  int            typesizeInput;
  
  float         *lineOutput;
  float         *signalOutput;
  float         *signal;
  
  unsigned char *ROIMask;
  unsigned char *lineCloudMask;
  InrType        type;

  unsigned char *signalCloudMask;
  int           *indicesClearSky;
  int            nClearSky;

  char fnameInput[FNAMESIZE];
  char fnameCloudMask[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameDates[FNAMESIZE];

  int sz_out;
  int sx, sy, sz, sxsy, szdaily = 0;
  int i, j, k, px, py;

  /* filter flags */
  int  do_clean;
  int  do_atmo;
  int  do_median;
  int  do_localvar;
  char atmo_effect;
  int  win_atmo;
  int  win_median;
  int  win_localvar;
  int  do_daily;

  double  n_stddev_localvar;
  
  /*
     LECTURE OPTIONS
   */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput) ||
      !igetopt1 ("-roi"  , "%s", fnameROIMask) ||
      !igetopt1 ("-ofile", "%s", fnameOutput) ||
      !igetopt1 ("-mask", "%s", fnameCloudMask))
    iusage_ (Ucmd, Udetail);

  do_clean = igetopt0 ("-clean");

  do_atmo = igetopt1 ("-atmo", "%c", &atmo_effect);
  if (do_atmo)
  {
    if (atmo_effect != 'u' && atmo_effect != 'l')
    {
      fprintf (stderr, "atmosphere effect must be u or l\n");
      exit (1);
    }
    if (!igetopt1 ("-win", "%d", &win_atmo))
      win_atmo = 2;
  }

  do_median   = igetopt1 ("-median"  , "%d", &win_median);
  do_localvar = igetopt2 ("-localvar", "%d", &win_localvar, "%lf", &n_stddev_localvar);

  do_daily = igetopt1 ("-daily", "%s", fnameDates);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);
  kdopts_ ();

  /*
     LECTURE IMAGES
   */
  // reads header information for all input data and checks if they are
  // appropriate
  if (!inrReadHeader(fnameInput, &sx, &sy, &sz, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    exit(0);
  }
  if (!inrReadHeader(fnameCloudMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for cloud mask '%s'.\n", fnameCloudMask);
    exit(0);
  }
  if (type != InrByte)
  {
    fprintf(stderr, "Error: cloud mask '%s' has to be of type 'Byte'.\n", fnameCloudMask);
    exit(0);
  }
  if (i != sx || j != sy || k != sz)
  {
    fprintf (stderr, "Error: dimensions of input data and cloud mask do not match.\n");
    exit (0);
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    exit(0);
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    exit(0);
  }
  if (i != sx || j != sy)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    exit (0);
  }

  // allocates memory for the input and cloud mask data
  typesizeInput = InrTypeSizes[typeInput]/8;
  lineInput     = (unsigned char *) malloc (sx*sz * typesizeInput * sizeof(unsigned char));
  lineCloudMask = (unsigned char *) malloc (sx*sz * sizeof(unsigned char));
  sxsy = sx * sy;

  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    i_Free ((void *) &lineInput);
    i_Free ((void *) &lineCloudMask);
    exit (0);
  }

  /*
     DEBUT DES HOSTILITES
   */

  /* images de resultat */
  if (do_daily)
  {
    if (!init_date (fnameDates))
      exit(-1);
    szdaily = day_since_beginning (sz - 1) + 1;
    sz_out = szdaily;
  }
  else
    sz_out = sz;

  // writes output file header (output type is always float)
  if (!inrWriteHeader(fnameOutput, sx, sy, sz_out, InrFloat))
  {
    printf("Error: could not write header information for output file '%s'.\n", fnameOutput);
    exit(0);
  }
  lineOutput = (float *) i_malloc (sx * sz_out * sizeof (float));

  /* sequences signal, signalCloudMask and signalOutput for one pixel */
  signal          = (float *)         i_malloc (sz * sizeof (float));
  signalCloudMask = (unsigned char *) i_malloc (sz * sizeof (unsigned char));
  signalOutput    = (float *)         i_malloc (sz_out * sizeof (float));

  /* workspace */
  work            = (float *)         i_malloc (sz * sizeof (float));

  /* indices des pixels nuageux */
  indicesClearSky = (int *)           i_malloc (sz * sizeof (int));


  fprintf (stderr, "List of applied filters to each temporal signal:\n");
  fprintf (stderr, "Suppression of cloudy values\n");
  if (do_clean)
    fprintf (stderr, "Suppression of values outside mean +- 2stdev\n");
  if (do_atmo)
  {
    if (atmo_effect == 'u')
      fprintf (stderr, "morphological lowering of signal, 1/2 size=%d\n",
               win_atmo);
    else
      fprintf (stderr, "morphological raising of signal, 1/2 size=%d\n", win_atmo);
  }
  if (do_median)
    fprintf (stderr, "Applying median filter with total window size %d\n", 2*win_median + 1);
  if (do_localvar)
    fprintf (stderr, "Applying local variance filter with total window size %d, filtering values outside range [mean +- %lf*stddev]\n",
             2*win_localvar + 1, n_stddev_localvar);
  if (do_daily)
    fprintf (stderr, "interpolation of the signal on a daily basis\n");
  else
    fprintf (stderr, "interpolation of suppressed values\n");


  /* loop sur pixels de l'image  */
  for (py = 0; py < sy; py++)
  {
    fprintf (stderr, "Processing ... %d %%\r", 100 * (py + 1) / sy);
    
    if (!inrReadLine(fnameInput, py, lineInput))
    {
      printf("Error: could not read line %d from input file '%s'.\n", py, fnameInput);
      free(lineInput);
      free(lineCloudMask);
      free(lineOutput);
      free(signal);
      free(signalCloudMask);
      free(signalOutput);
      exit(-1);
    }
    if (!inrReadLine(fnameCloudMask, py, lineCloudMask))
    {
      printf("Error: could not read line %d from cloud mask '%s'.\n", py, fnameCloudMask);
      free(lineInput);
      free(lineCloudMask);
      free(lineOutput);
      free(signal);
      free(signalCloudMask);
      free(signalOutput);
      exit(-1);
    }
      
    for (px = 0; px < sx; px++)
    {
      /* si hors zone signal a 0 */
      if (!ROIMask[px + py * sx])
      {
        insert_null_signal (lineOutput, px, sx, sz_out);
        continue;
      }

      /* extraction sequence du pixel px,py et marquage des pixels ciel clair */
      extract_signal (signal, lineInput, px, sx, sz, typeInput);
      extract_signal_byte (signalCloudMask, lineCloudMask, px, sx, sz);

      nClearSky = mark_clouds (signalCloudMask, indicesClearSky, sz);
      if (nClearSky <= 1)
      {                        /* aucune valeur acceptable:0 */
        insert_null_signal (lineOutput, px, sx, sz_out);
        continue;
      }

      /* ici on doit traiter un pixel avec au moins deux dates non nuageuses */

      /* suppression systematique des donnees nuageuses */
      suppress_cloudy_dates (signal, signalCloudMask, sz);

      /* suppression optionnelle des valeurs abberrantes */
      if (do_clean)
      {
        signal_clean_extrema (signal, signalCloudMask, sz);
        nClearSky = mark_clouds (signalCloudMask, indicesClearSky, sz);
        if (nClearSky <= 1)
        {                /* aucune valeur acceptable:0 */
          insert_null_signal (lineOutput, px, sx, sz_out);
          continue;
        }
      }
      if (do_atmo)
      {
        pack_signal (signal, indicesClearSky, sz, nClearSky);

        /* filtrage atmosphere optionel */
        if (atmo_effect == 'u')        /* atmosphere tends to raise signal */
          fermeture (signal, signal, win_atmo, nClearSky);
        else                /* atmosphere tends to lower signal */
          ouverture (signal, signal, win_atmo, nClearSky);

        unpack_signal (signal, indicesClearSky, sz, nClearSky);
      }
      if (do_median)
      {
        /* median filtering */
        pack_signal (signal, indicesClearSky, sz, nClearSky);

        signal_filter_median (signal, signalOutput, win_median, nClearSky);
        copy_signal (signalOutput, signal, nClearSky);

        unpack_signal (signal, indicesClearSky, sz, nClearSky);
      }
      if (do_localvar)
      {
        /* local variance filtering */
        pack_signal (signal, indicesClearSky, sz, nClearSky);

        signal_filter_localvar (signal, signalOutput, win_localvar, n_stddev_localvar, nClearSky);
        copy_signal (signalOutput, signal, nClearSky);

        unpack_signal (signal, indicesClearSky, sz, nClearSky);
      }
      
      

      /* mise sur base journaliere */
      if (do_daily)
      {
        expand_signal_per_day (signal, indicesClearSky, signalOutput, nClearSky,
                               szdaily);
      }
      else
      {
        copy_signal (signal, signalOutput, sz);
      }

      /* interpolation des donnees manquantes */
      interpolate_cloudy_dates (signalOutput, indicesClearSky, nClearSky, sz_out);

      /* on met le tout dans l'image */
      insert_signal (signalOutput, lineOutput, px, sx, sz_out);
    }
    if (!inrWriteLine(fnameOutput, py, (unsigned char *)lineOutput))
    {
      printf("Error: could not write line %d to output file '%s'.\n", py, fnameOutput);
      free(lineInput);
      free(lineCloudMask);
      free(lineOutput);
      free(signal);
      free(signalCloudMask);
      free(signalOutput);
      exit(-1);
    }
  } 
  fprintf (stderr, "\n");

  {
    // transfers georeference info if available
    InrGeorefInfo georef;
    if (inrReadGeoref(fnameInput, &georef))
      inrWriteGeoref(fnameOutput, georef);
    else
      inrWriteEnviHdr(fnameOutput);
  }

  
  if (do_daily)
  {
    fprintf (stderr, "output sequence range is:\n");
    print_frame_date (0);
    print_frame_date (sz - 1);
  }


  if (do_daily)
    clean_date ();
    
  free(lineInput);
  free(lineCloudMask);
  free(lineOutput);
  free(signal);
  free(signalCloudMask);
  free(signalOutput);

  return 0;
}
