#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "linearalgebra.h"
#include "profile_features.h"

#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE       1024

#define DEG_DEFAULT     8
#define NODATA_DEFAULT  -9999



static char Ucmd[] = "-ifile inr_sequence [-roi roi_mask] -ofile fitted_sequence \
[-dates datefile] [-degree value] [-nodata value][-fail value]\n";

static char Udetail[] =
"Given a temporal sequence of data, fits a Nth degree polynomial function to\n\
the temporal profile of each point, and outputs the fitted sequence.\n\
\nInput:\n\
\t-ifile\tFile name of the input INRIMAGE sequence\n\
\tit must be a NDVI sequence\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output INRIMAGE file\n\
\nOptional:\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t-dates\tFile name of the list of dates that correspond to each layer in\n\
\t\tthe image sequence. Dates must be expressed in ddd format\n\
\t\tIf a datefile is not specified, fitting\n\
\t\twill be computed considering indices 0...nImages as the 'dates'\n\
\t\tof each layer in the sequence\n\
\t-degree\tDegree of the polynomial function to be used for fitting\n\
\t\tDefault value is 8\n\
\t-nodata\tnodata code in ndvi sequence (def -9999)\n\
\t-fail\tvalue to put when fit fails (def -9999)\n";


/*
 *  defineDates
 *
 *  Builds date vector from a datefile if it is specified
 *  If not, date vector will be [0, 1, .., nDates]
 */

static  int *defineDates(int hasFile, char *fnameDates, int nDates)
{
    int *dates = (int *)malloc(nDates*sizeof(int));
    int  date;
    int  iDate;
    FILE *fpDates;
    
    if (!hasFile)
    {
      // case 1: no dates file specified
      for (iDate=0; iDate<nDates; iDate++)
          dates[iDate] = iDate;
 
      return dates;
    }
    else
    {
        // case 2: dates file specified
        fpDates = fopen(fnameDates, "rt");
        if (!fpDates)
        {
            fprintf(stderr, "Error: could not open dates file '%s'.\n", fnameDates);
            free(dates);
            return NULL;
        }
        for (iDate=0; iDate<nDates; iDate++)
        {
            if (!fscanf(fpDates, "%d", &date))
            {
                fprintf(stderr, "Error: not enough dates (%d) in dates file '%s'.\n", iDate, fnameDates);
                free(dates);
                return NULL;
            }
            dates[iDate] = date;
        }
    }

    return dates;
}




int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameDates[FNAMESIZE];
  int  okDegree, okDates, okNoData,okFail,okRoi;
  InrType typeInput;
  InrType type;
  int     typesizeInput;
  InrGeorefInfo georef;
  
  unsigned char *dataInput;
  unsigned char *ROIMask;
  float         *dataOutput;
  // float *ndviInput ;

  int      degree;
  double   fail;
  double  *values,*maskedvalues;
  int     *dates,*maskeddates;
  double *coef ;
  float nodata ;

  int nfail=0 ;
  int nok=0 ;

  int size_x, size_y, size_z;
  int i, j, k;
  int x, y, z;
  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okRoi = igetopt1 ("-roi"  , "%s", fnameROIMask);
  okDates  = igetopt1 ("-dates" ,  "%s", fnameDates);
  okDegree = igetopt1 ("-degree",  "%d", &degree);
  okFail = igetopt1 ("-fail", "%lf", &fail);
  okNoData = igetopt1 ("-nodata", "%f", &nodata);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  if (!okFail)
    fail = -9999;
  if (!okNoData)
    nodata=-9999;
  if (!okDegree)
    degree = DEG_DEFAULT;
    

 

    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if ( typeInput != InrFloat  ) {
    fprintf(stderr,"Error: NDVI sequence must be coded as floats\n");
    return 0;
  }
  if (okRoi) {
    if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
      {
	fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
	return 0;
      }
    if (type != InrByte || k != 1)
      {
	fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
	return 0;
      }
    if (i != size_x || j != size_y)
      {
	fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
	return 0;
      }
    
    // reads the ROI mask image
    ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
    if (!ROIMask)
      {
	fprintf (stderr, "Error: could not read ROI mask data.\n");
	return 0;
      }
  }
  else { // roi is 1 everywhere
    int index;
    ROIMask = (unsigned char*)malloc(size_x*size_y);
    if (!ROIMask) {
      fprintf(stderr,"Error: memory allocation failed\n");
      return 0;
    }
    for (index=0;index<size_x*size_y;index++) ROIMask[index]=255;
  }

  // defines date vector (1,2..nDates if not specified)
  dates = defineDates(okDates, fnameDates, size_z);
  if (!dates)
    return 0;


  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  // ndviInput        = (float *)malloc( size_x*size_z *sizeof(float));
  dataOutput       = (float *)  malloc (size_x*size_y*size_z * sizeof(float));
  values           = (double *) malloc (size_z * sizeof(double));
  maskedvalues     = (double *) malloc (size_z * sizeof(double));
  maskeddates      = (int *) malloc (size_z * sizeof(int));
  int nval ;


  // writes output file header (output type is always float)
  if (!inrWriteHeader(fnameOutput, size_x, size_y, size_z, InrFloat))
  {
    printf("Error: could not write header information for output file '%s'.\n", fnameOutput);
    return 0;
  }
  // initializes output data line
  for (i=0; i<size_x*size_z; i++)
    dataOutput[i] = fail;
  
  coef = (double *) malloc ((degree+1) * sizeof(double));

  // processes the data
  for (y=0; y<size_y; y++)
  {
    fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);

    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    if (x == size_x)
    {
      // writes empty output data line
      if (!inrWriteLine(fnameOutput, y, (unsigned char *)dataOutput))
      {
        printf("Error: could not write line %d to output file '%s'.\n", y, fnameOutput);
        return 0;
      }
      continue;
    }

    // reads input data line y
    if (!inrReadLine(fnameInput, y,dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }

    // processes each point at a time
    for (x=0; x<size_x; x++)
      {
      // discards pixels outside of region of interest
      if (!ROIMask[x + y*size_x])
        continue;
        
      // reads values for each date of the data point
      nval=0;
     
      for (z=0; z<size_z; z++){
        values[z] = inrGetDataValue((unsigned char*)dataInput, InrFloat, x+z*size_x);
	// discard nodata points
	if ( values[z]!=nodata) {
	  maskedvalues[nval]=values[z];
	  maskeddates[nval]=dates[z];
	  nval++;
	}	
      }
      

      if (nval==0) continue;
      // check for constant signal (produces singular matrix)
      {
	float ndvimin,ndvimax;
	ndvimin=maskedvalues[0];
	ndvimax=maskedvalues[0];
	for (z=0;z<nval;z++) {
	  ndvimin=MIN(maskedvalues[z],ndvimin);
	  ndvimax=MAX(maskedvalues[z],ndvimax);
	}
	if (ndvimin==ndvimax) continue ;
      }


      // computes the polynomial for the sequence values
      if (!profFitPoly( maskedvalues, maskeddates, nval, coef, degree )){
	// echec du fit: raz
	for  (z=0; z<size_z; z++) values[z]=0.;
	nfail++;
	
     }
      else {
	nok ++ ;
	// calcul des valeurs du polynome
	for  (z=0; z<size_z; z++)
	  values[z]=profPolyValue( coef, degree, dates[z]) ;
       
      }
      


      for (z=0; z<size_z; z++)
	inrPutDataValue((unsigned char*)dataOutput,InrFloat,x+z*size_x,values[z]);
      }

    // writes output data line
    if (!inrWriteLine(fnameOutput, y, (unsigned char *)dataOutput))
    {
      printf("Error: could not write line %d to output file '%s'.\n", y, fnameOutput);
      return 0;
    }
    // re-initializes output data line
    for (i=0; i<size_x*size_z; i++)
      dataOutput[i] = fail;
  }

  // transfers georeference info if available
  if (inrReadGeoref(fnameInput, &georef))
    inrWriteGeoref(fnameOutput, georef);
  else
    inrWriteEnviHdr(fnameOutput);


    if (nfail || nok ) printf("\nfit ok on %d pixels and failed on %d pixels\n",nok,nfail);
  
    

  free(dataInput);
  free(dataOutput);
  free(ROIMask);
  free(values);
  free(dates);
  free(maskedvalues);
  free(maskeddates);
  
  return 0;
}
