#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE       1024


static char Ucmd[] = "-ifile inr_profiles -forest index -bare index \
-ofile inr_profiles_defor [-sig factor]\n";

static char Udetail[] =
"Given an input profile image, outputs a new profile image adding an estimated\n\
deforestation profile.\n\
\nDeforestation profile estimation is computed by interpolating between the\n\
and bare soil profiles, so that it corresponds to the average behavior of an\n\
area that starts as 100% forest and gradually turns into 100% bare soil\n\
\nInput:\n\
\t-ifile\tFile name of input INRIMAGE profile image\n\
\t-forest\tIndex that specifies the forest profile in the input file\n\
\t-bare\tIndex that specifies the bare soil profile in the input file\n\
\nOutput:\n\
\t-ofile\tFile name of the output INRIMAGE profile image\n\
\nOptional:\n\
\t-sig\tInterpolate using a sigmoid function with a given factor\n\
\t\tInterpolating function will be 1/(1+exp(factor*x))\n\
\t\tDefault is to use linear interpolation\n";



int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  InrType typeInput;
  int typesizeInput;
  int okSig;
  double sigFactor;
  
  unsigned char *dataInput;
  unsigned char *dataOutput;

  int iForest;
  int iBare;  
  
  int size_x, size_y, size_z;
  int x, y;
  
  double y_norm;
  
  double value;
  double valueForest;
  double valueBare;

  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile" , "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile" , "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-forest", "%d", &iForest))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-bare"  , "%d", &iBare))
    iusage_ (Ucmd, Udetail);
  okSig = igetopt1 ("-sig", "%lf", &sigFactor);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);
    

  // reads input profile
  dataInput = inrReadDataSequence(fnameInput, &size_x, &size_y, &size_z, &typeInput);
  if (!dataInput)
  {
    fprintf(stderr, "Error: could not read input file '%s'.\n", fnameInput);
    return 0;
  }
  if (size_z != 1)
  {
    fprintf (stderr, "Error: input profile image should contain only one layer/frame.\n");
    return 0;
  }
  if (iForest < 0 || iForest >= size_x)
  {
    fprintf (stderr, "Error: invalid forest profile index (valid indices for input file are 0..%d)\n", size_x-1);
    return 0;
  }
  if (iBare < 0 || iBare >= size_x)
  {
    fprintf (stderr, "Error: invalid bare soil profile index (valid indices for input file are 0..%d)\n", size_x-1);
    return 0;
  }
  
  typesizeInput = InrTypeSizes[typeInput]/8;
  dataOutput    = (unsigned char *) malloc ((size_x+1)*size_y * typesizeInput * sizeof(unsigned char));


  // copies input profiles to output data buffer
  for (y=0; y<size_y; y++)
  {
    for (x=0; x<size_x; x++)
    {
      value = inrGetDataValue(dataInput, typeInput, x + y*size_x);
      inrPutDataValue(dataOutput, typeInput, x + y*(size_x+1), value);
    }
  }

  // adds new deforestation profile through interpolation
  for (y=0; y<size_y; y++)
  {
    // reads forest and bare profile values for date y
    valueForest = inrGetDataValue(dataInput, typeInput, iForest + y*size_x);
    valueBare   = inrGetDataValue(dataInput, typeInput, iBare   + y*size_x);
    
    // interpolates values given date y
    y_norm = (double)y/(size_y-1);
    if (okSig)
    {
      // sigmoid interpolation
      double y_sig = 1./(1+exp(sigFactor * (y_norm-0.5)));
      value = y_sig*valueForest + (1-y_sig)*valueBare;
    }
    else
    {
      // linear interpolation
      value  = (1-y_norm)*valueForest + y_norm*valueBare;
    }
    
    // puts computed value into output data buffer
    inrPutDataValue(dataOutput, typeInput, size_x + y*(size_x+1), value);
  }
  
  // writes output profiles
  if (!inrWriteDataSequence(fnameOutput, size_x+1, size_y, 1, dataOutput, typeInput))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  
    
  fprintf(stderr, "Deforestation profile successfully computed.\n");

  free(dataInput);
  free(dataOutput);
  
  return 1;
}
