/***************************************************************************
                          sequence_profile_features.c  -  description
                             -------------------
    Modified             : June 2007
    By                   : (C) 2007 by Carlos de la Torre
    Email                : Carlos.de_la_Torre@inria.fr
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "linearalgebra.h"
#include "profile_features.h"

#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE       1024
#define DEG_DEFAULT     12
#define NODATA_DEFAULT  0



static char Ucmd[] = "-ifile inr_sequence -iprev inr_prev_sequence -roi roi_mask\
-ofile inr_features [-dates datefile] [-datesprev datefile] [-degree value]\
[-nodata value] [-ofit inr_fitted]\n";

static char Udetail[] =
"Given a temporal sequence of data, fits a Nth degree polynomial function to\n\
the temporal profile of each point, and outputs the fitted curves' features.\n\
\nInput:\n\
\t-ifile\tFile name of the input INRIMAGE sequence\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output INRIMAGE file containing one layer for\n\
\t\teach profile feature\n\
\nOptional:\n\
\t-iprev\tFile name of the INRIMAGE sequence for the previous year\n\
\t-dates\tFile name of the list of dates that correspond to each layer in\n\
\t\tthe image sequence. Dates must be expressed in integer format\n\
\t\t(e.g. day of the year). If a datefile is not specified, fitting\n\
\t\twill be computed considering indices 0...n as the 'dates'\n\
\t\tof each layer in the sequence\n\
\t-datesprev\tFile name of the list of dates that correspond to each\n\
\t\t\tlayer in the image sequence for the previous year\n\
\t\t\tCan only be specified if 'dates' parameter is also\n\
\t\t\tspecified\n\
\t-degree\tDegree of the polynomial function to be used for fitting\n\
\t\tDefault value is 12\n\
\t-nodata\tValue to be assigned when feature is not available\n\
\t\t(i.e., value for missing data)\n\
\t\tDefault value is 0\n\
\t-ofit\tFile name of an optional output image containing the fitted\n\
\t\tprofile values (not implemented)\n";


/*
 * GLOBAL CONTEXT
 */
typedef struct 
{
    char fnameInput[FNAMESIZE];
    char fnameInputPrev[FNAMESIZE];
    char fnameROIMask[FNAMESIZE];
    char fnameOutput[FNAMESIZE];
    char fnameDates[FNAMESIZE];
    char fnameDatesPrev[FNAMESIZE];
    int  okDegree, okDates, okDatesPrev, okNoData;

    InrType typeInput;
    InrType typeInputPrev;
    int     typesizeInput;
    int     typesizeInputPrev;

    InrGeorefInfo georef;

    unsigned char *dataInput;
    unsigned char *dataInputPrev;
    unsigned char *ROIMask;
    float         *dataOutput;

    int      degree;
    float    nodata;
    double  *values;
    double  *valuesPrev;
    int     *dates;
    int     *datesPrev;
    int      nFeatures;

    int size_x, size_y, size_z;
    int prev_size_z;
} ModisProfFeat_gc;
                                  


/*
 *  insertPointFeatures
 *
 *  Inserts feature information for one point into the output data line
 */
#define NUM_FEATURES  12
int insertPointFeatures(ModisProfFeat_gc *gc, int x, profFeatures *features)
{
    // wrapper or CFS
    gc->dataOutput[x + 0*gc->size_x] = (float)features->val[prof_mean];
    gc->dataOutput[x + 1*gc->size_x] = (float)features->val[prof_maxDate];
    gc->dataOutput[x + 2*gc->size_x] = (float)features->val[prof_delta45days];
    gc->dataOutput[x + 3*gc->size_x] = (float)features->val[prof_slopeAbsMean];
    gc->dataOutput[x + 4*gc->size_x] = (float)features->nModes;
  
    // CFS only
    gc->dataOutput[x + 5*gc->size_x] = (float)features->val[prof_min];
    gc->dataOutput[x + 6*gc->size_x] = (float)features->val[prof_deltaMaxMin];
    gc->dataOutput[x + 7*gc->size_x] = (float)features->val[prof_valueStddev];
    if (features->nModes > 0)
    {
        gc->dataOutput[x + 8*gc->size_x]  = (float)features->modes[0].datePeak;
        gc->dataOutput[x + 9*gc->size_x]  = (float)features->modes[0].dateWidthPeak;
        gc->dataOutput[x + 10*gc->size_x] = profModeHeight(features->modes[0]);
        gc->dataOutput[x + 11*gc->size_x] = (float)features->modes[0].slopeRight;
    }
    else
    {
        gc->dataOutput[x + 8*gc->size_x] = gc->nodata;
        gc->dataOutput[x + 9*gc->size_x] = gc->nodata;
        gc->dataOutput[x + 10*gc->size_x] = gc->nodata;
        gc->dataOutput[x + 11*gc->size_x] = gc->nodata;
    }
    //gc->dataOutput[x + 10*gc->size_x] = (float)features->val[prof_valueInterannualLoss];
    
    return 1;
}


/*
 *  defineDates
 *
 *  Builds date vector from a datefile if it is specified
 *  If not, date vector will be [0, 1, .., nDates]
 */
int *defineDates(int hasFile, char *fnameDates, int nDates)
{
    int *dates = (int *)malloc(nDates*sizeof(int));
    int  date;
    int  iDate;
    FILE *fpDates;
    
    if (!hasFile)
    {
      // case 1: no dates file specified
      for (iDate=0; iDate<nDates; iDate++)
          dates[iDate] = iDate;
 
      return dates;
    }
    else
    {
        // case 2: dates file specified
        fpDates = fopen(fnameDates, "rt");
        if (!fpDates)
        {
            fprintf(stderr, "Error: could not open dates file '%s'.\n", fnameDates);
            free(dates);
            return NULL;
        }
        for (iDate=0; iDate<nDates; iDate++)
        {
            if (!fscanf(fpDates, "%d", &date))
            {
                fprintf(stderr, "Error: not enough dates (%d) in dates file '%s'.\n", iDate, fnameDates);
                free(dates);
                return NULL;
            }
            dates[iDate] = date;
        }
    }

    return dates;
}



int main (int argc, char **argv)
{
    ModisProfFeat_gc gc_obj;           /* general context */
    ModisProfFeat_gc *gc = &gc_obj;
    InrBandNameInfo* bn;

    profFeatures features;
    InrType type;
    int i, j, k;
    int x, y, z;

    // reads arguments
    inr_init (argc, argv, "1.0", Ucmd, Udetail);
    if (!igetopt1 ("-ifile", "%s", gc->fnameInput))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-iprev", "%s", gc->fnameInputPrev))
        strcpy(gc->fnameInputPrev, gc->fnameInput);
    if (!igetopt1 ("-roi"  , "%s", gc->fnameROIMask))
        iusage_ (Ucmd, Udetail);
    if (!igetopt1 ("-ofile", "%s", gc->fnameOutput))
        iusage_ (Ucmd, Udetail);
    gc->okDates     = igetopt1 ("-dates"    , "%s", gc->fnameDates);
    gc->okDatesPrev = igetopt1 ("-datesprev", "%s", gc->fnameDatesPrev);
    gc->okDegree    = igetopt1 ("-degree"   , "%d", &gc->degree);
    gc->okNoData    = igetopt1 ("-nodata"   , "%f", &gc->nodata);

    // exits program if there are unused options still present
    if (tstopts ())
        iusage_ (Ucmd, Udetail);

    // invalid combination: 'dates' can be assumed for previous input, but not the other way around
    if (!gc->okDates && gc->okDatesPrev)
        iusage_ (Ucmd, Udetail);

    if (!gc->okNoData)
        gc->nodata = 0;
    if (!gc->okDegree)
        gc->degree = DEG_DEFAULT;
      
    gc->nFeatures = NUM_FEATURES;
      
    // reads header information for all input data and checks if they are appropriate
    if (!inrReadHeader(gc->fnameInput, &gc->size_x, &gc->size_y, &gc->size_z, &gc->typeInput))
    {
        fprintf(stderr, "Error: could not read header information for input file '%s'.\n", gc->fnameInput);
        return 0;
    }
    if (!inrReadHeader(gc->fnameInputPrev, &i, &j, &gc->prev_size_z, &gc->typeInputPrev))
    {
        fprintf(stderr, "Error: could not read header information for input file '%s'.\n", gc->fnameInputPrev);
        return 0;
    }
    if (gc->okDates && !gc->okDatesPrev && gc->prev_size_z != gc->size_z)
    {
        // if no previous date information was specified, z dimensions of input data must match
        fprintf (stderr, "Error: z dimensions of input data files do not match (should specify distinct date information?).\n");
        return 0;
    }
    if (i != gc->size_x || j != gc->size_y)
    {
        fprintf (stderr, "Error: dimensions of input data files do not match.\n");
        return 0;
    }
    if (!inrReadHeader(gc->fnameROIMask, &i, &j, &k, &type))
    {
        fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", gc->fnameROIMask);
        return 0;
    }
    if (type != InrByte || k != 1)
    {
        fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", gc->fnameROIMask);
        return 0;
    }
    if (i != gc->size_x || j != gc->size_y)
    {
        fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
        return 0;
    }
  
    // reads the ROI mask image
    gc->ROIMask = inrReadByteImage (gc->fnameROIMask, &i, &j);
    if (!gc->ROIMask)
    {
        fprintf (stderr, "Error: could not read ROI mask data.\n");
        return 0;
    }

    // defines date vector (1,2..nDates if not specified)
    gc->dates = defineDates(gc->okDates, gc->fnameDates, gc->size_z);
    if (!gc->dates)
        return 0;

    // defines previous date vector (1,2..nDates if not specified)
    // if no specific previous date information was specified, uses regular date specifications
    if (gc->okDatesPrev)
        gc->datesPrev = defineDates(gc->okDatesPrev, gc->fnameDatesPrev, gc->prev_size_z);
    else
        gc->datesPrev = defineDates(gc->okDates, gc->fnameDates, gc->prev_size_z);
    if (!gc->datesPrev)
        return 0;


    // allocates memory for the input and output data
    gc->typesizeInput     = InrTypeSizes[gc->typeInput]/8;
    gc->typesizeInputPrev = InrTypeSizes[gc->typeInputPrev]/8;
    gc->dataInput         = (unsigned char *) malloc (gc->size_x*gc->size_z * gc->typesizeInput * sizeof(unsigned char));
    gc->dataInputPrev     = (unsigned char *) malloc (gc->size_x*gc->prev_size_z * gc->typesizeInputPrev * sizeof(unsigned char));
    gc->dataOutput        = (float *)  malloc (gc->size_x*gc->nFeatures * sizeof(float));
    gc->values            = (double *) malloc (gc->size_z * sizeof(double));
    gc->valuesPrev        = (double *) malloc (gc->prev_size_z * sizeof(double));

    // writes output file header (output type is always float)
    if (!inrWriteHeader(gc->fnameOutput, gc->size_x, gc->size_y, gc->nFeatures, InrFloat))
    {
        printf("Error: could not write header information for output file '%s'.\n", gc->fnameOutput);
        return 0;
    }
    // initializes output data line
    for (i=0; i < gc->size_x * gc->nFeatures; i++)
        gc->dataOutput[i] = gc->nodata;
    

    // processes the data
    for (y=0; y < gc->size_y; y++)
    {
        fprintf (stderr, "Processing... %d%%\r", 100*y/gc->size_y);

        // checks if line y is within region of interest
        for (x=0; x < gc->size_x; x++)
            if (gc->ROIMask[x + y*gc->size_x])
                break;

        // if no unmasked point was detected, line is not within ROI and is skipped 
        if (x == gc->size_x)
        {
            // writes empty output data line
            if (!inrWriteLine(gc->fnameOutput, y, (unsigned char *)gc->dataOutput))
            {
              printf("Error: could not write line %d to output file '%s'.\n", y, gc->fnameOutput);
              return 0;
            }
            continue;
        }
  
        // reads input data line y
        if (!inrReadLine(gc->fnameInput, y, gc->dataInput))
        {
            fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, gc->fnameInput);
            return 0;
        }
        // reads previnput data line y
        if (!inrReadLine(gc->fnameInputPrev, y, gc->dataInputPrev))
        {
            fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, gc->fnameInputPrev);
            return 0;
        }
        
        // processes each point at a time
        for (x=0; x < gc->size_x; x++)
        {
            // discards pixels outside of region of interest
            if (!gc->ROIMask[x + y*gc->size_x])
                continue;
              
            // reads values for each date of the data point
            for (z=0; z < gc->size_z; z++)
                gc->values[z] = inrGetDataValue(gc->dataInput, gc->typeInput, x+z*gc->size_x);
            for (z=0; z < gc->prev_size_z; z++)
                gc->valuesPrev[z] = inrGetDataValue(gc->dataInputPrev, gc->typeInputPrev, x+z*gc->size_x);
        
            // ignores point if values are zero
            if (gc->values[0] == 0)
            {
                fprintf(stderr, "\nWarning: found value '0'\n");
                continue;        
            }

            // computes the features for the sequence values
            profComputeFeatures(gc->values, gc->dates, gc->size_z,
                                gc->valuesPrev, gc->datesPrev, gc->prev_size_z,
                                gc->degree, &features);

            // inserts feature information in the output data line
            insertPointFeatures(gc, x, &features);
        }
        // writes output data line
        if (!inrWriteLine(gc->fnameOutput, y, (unsigned char *)gc->dataOutput))
        {
            printf("Error: could not write line %d to output file '%s'.\n", y, gc->fnameOutput);
            return 0;
        }
        // re-initializes output data line
        for (i=0; i<gc->size_x*gc->nFeatures; i++)
            gc->dataOutput[i] = gc->nodata;
    }
  
    // transfers georeference info if available
    if (inrReadGeoref(gc->fnameInput, &gc->georef))
        inrWriteGeoref(gc->fnameOutput, gc->georef);
    else
        inrWriteEnviHdr(gc->fnameOutput);
    
    // append band name to header
    //    int nFeatures=12;
    bn = getBandNameCFS(gc->nFeatures) ;
    inrAppendBandName(gc->fnameOutput,bn);
        
    fprintf(stderr, "Profile features successfully computed based on polynomial functions of degree %d\n", gc->degree);
  
    free(gc->dataInput);
    free(gc->dataOutput);
    free(gc->ROIMask);
    free(gc->values);
    free(gc->dates);
    free(gc->datesPrev);
      
    return 0;
}
