#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "svd.h"


#define FNAMESIZE  1024


static char Ucmd[] = "-ifile inr_sequence -prop proportions_file -roi roi_mask \
-ofile profiles\n";

static char Udetail[] =
"Computes temporal reflectance profiles for individual classes using a linear\n\
mixture model. For that purpose, the algorithm uses:\n\
1. An input image sequence, defining the reflectances for a region over time\n\
2. A proportion image, defining the percentage of each class within each pixel\n\
\nInput:\n\
\t-ifile\tFile name of the input image sequence\n\
\t-prop\tFile name of the proportion image\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output file containing the mean temporal\n\
\t\treflectance profiles for each class\n";


int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameProp[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  int     typesizeInput;
  InrType typeInput;
  InrType type;
  
  unsigned char *dataInput;
  unsigned char *ROIMask;
  float         *prop;

  int sx, sy, sz;
  int i, j, k;
  int n_class;

  float *res, **test, **res2, **A, som, tot, som2, vr, *reflec;
  int t, taille_plan;
  int taille_ech, l ;
  float **V, *W;

  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-prop" , "%s", fnameProp))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"  , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &sx, &sy, &sz, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameProp, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for proportion image '%s'.\n", fnameProp);
    return 0;
  }
  if (type != InrFloat)
  {
    fprintf(stderr, "Error: proportion image '%s' has to be a image of type 'Float'.\n", fnameProp);
    return 0;
  }
  if (i != sx || j != sy)
  {
    fprintf (stderr, "Error: dimensions of input data and proportion image do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != sx || j != sy)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }

  // allocates memory for the input data
  typesizeInput = InrTypeSizes[typeInput]/8;
  dataInput     = (unsigned char *) malloc (sx*sz * typesizeInput * sizeof(unsigned char));

  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    free(dataInput);
    return 0;
  }

  prop = inrReadFloatSequence (fnameProp, &i, &j, &n_class);
  if (!prop)
  {
    fprintf (stderr, "Error: could not read proportion image data.\n");
    free(dataInput);
    free(ROIMask);
    return 0;
  }

    
  /* creation image des reflectances */
  reflec = (float *) i_calloc (n_class * sz, sizeof (float));
  inrWriteFloatImage (fnameOutput, n_class, sz, reflec);

  /* remplir mat_prop (la matrice des proportions) et  mat_ref  */
  taille_plan = sx * sy;
  taille_ech = compter_pixels (ROIMask, taille_plan);

  V = matrix (1, n_class, 1, n_class);
  W = vector (1, n_class);
  test = matrix (1, taille_ech + 1, 1, 1);
  res2 = matrix (1, n_class, 1, 1);
  A = matrix (1, taille_ech + 1, 1, n_class);
  mat_prop = matrix (1, taille_ech + 1, 1, n_class);
  mat_ref = vector (1, taille_ech + 1);
  l = 0;
  for (i = 0; i < taille_plan; i++)
  {
    if (ROIMask[i])
    {
      l++;
      for (j = 0; j < n_class; j++)
      {
        mat_prop[l][j + 1] = prop[i + taille_plan * j];
        A[l][j + 1] = prop[i + taille_plan * j];
      }
    }
  }
  
  conditionnement (mat_prop, taille_ech + 1, n_class, V, W);
  l = 0;
  res = vector (1, n_class);
  tot = 0;
  for (t = 0; t < sz; t++)
  {
    if (!inrReadFrame(fnameInput, t, dataInput))
    {
      printf("Error: could not read frame %d from input file '%s'.\n", t, fnameInput);
      free(dataInput);
      free(ROIMask);
      free(prop);
      free(reflec);
      return 0;
    }
   
    l = 0;
    fprintf (stderr, "date %d\r", t);
    for (i = 0; i < taille_plan; i++)
    {
      if (ROIMask[i])
      {
        l++;
        for (j = 0; j < n_class; j++)
          mat_prop[l][j + 1] = prop[i + taille_plan * j];
      }
    }
    l = 0;
    for (i = 0; i < taille_plan; i++)
    {
      if (ROIMask[i])
      {
        l++;
        //mat_ref[l] = (float) dataInput[i + t * taille_plan];
        mat_ref[l] = (float) inrGetDataValue(dataInput, typeInput, i);
      }
    }
    svdcmp (mat_prop, taille_ech + 1, n_class, W, V);
    svbksb (mat_prop, W, V, taille_ech + 1, n_class, mat_ref, res);

    for (j = 0; j < n_class; j++)
      reflec[j + n_class * t] = (float) res[j + 1];
    for (j = 0; j < n_class; j++)
      res2[j + 1][1] = res[j + 1];

    mul_2matrice (A, taille_ech + 1, n_class, res2, n_class, 1, test);
    som = som2 = 0.0;
    for (i = 1; i <= taille_ech; i++)
    {
      som += fabs (test[i][1] - mat_ref[i]);
      som2 += (test[i][1] - mat_ref[i]) * (test[i][1] - mat_ref[i]);
    }

    tot += som /= taille_ech;
    som2 /= taille_ech;
    vr = som2 - (som * som);
  }                                /* end for t */

  printf ("\n");
  tot /= sz;
  printf (" err= %f\n", tot);

  /* ecriture resultat */
  inrWriteFloatImage (fnameOutput, n_class, sz, reflec);

  free_matrix (V, 1, n_class, 1, n_class);
  free_vector (W, 1, n_class);
  i_Free ((void *) &dataInput);
  i_Free ((void *) &ROIMask);
  i_Free ((void *) &reflec);
  free_matrix (mat_prop, 1, taille_ech + 1, 1, n_class);
  free_matrix (A, 1, taille_ech + 1, 1, n_class);
  free_matrix (test, 1, taille_ech + 1, 1, 1);
  free_matrix (res2, 1, n_class, 1, 1);
  free_vector (mat_ref, 1, taille_ech + 1);
  free_vector (res, 1, n_class);

  return 0;
}
