#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024
#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)



static char Ucmd[] = "-ifile inr_sequence -prop proportions_file \
-class classification_file -roi roi_mask -ofile profiles\n";

static char Udetail[] =
"Computes temporal reflectance profiles for individual classes using only pure\n\
pixels. For that purpose, the algorithm uses:\n\
1. An input image sequence, defining the reflectances for a region over time\n\
2. A proportion image, defining the percentage of each class within each pixel\n\
   OR\n\
   A classification image, defining the classes for each pure-pixel\n\
\nInput:\n\
\t-ifile\tFile name of the input image sequence\n\
\t-prop\tFile name of a proportion image\n\
\t-class\tFile name of a classification image\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output file containing the mean temporal\n\
\t\treflectance profiles for each class\n\
\nOutput(optional):\n\
\t-otext\tBase name of the output files containing the temporal\n\
\t\treflectance profiles for each class\n";


/*
 *  outputValuesInit
 *
 *  Initializes output files for outputValues
 */
int outputValuesInit(char *fnameOutputBase, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;

  for (iClass=0; iClass<nClass; iClass++)
  {
    // defines output values file for the given class and opens it
    sprintf(fnameOutput, "%s.class%02d.values.txt", fnameOutputBase, iClass+1);
    fpOutput = fopen(fnameOutput, "wt");
    if (!fpOutput)
    {
      fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
      return 0;
    }
    fclose(fpOutput);
  }
  return 1;
}


/*
 *  outputValues
 *
 *  Writes values for individual pixels
 */
int outputValues(char *fnameOutputBase, float *values, int nValues, int class_id)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   i;

  // defines output values file for the given class and opens it for appending
  sprintf(fnameOutput, "%s.class%02d.values.txt", fnameOutputBase, class_id+1);
  fpOutput = fopen(fnameOutput, "at");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
    return 0;
  }

  // writes output values for current frame and class
  for (i=0; i<nValues; i++)
    fprintf(fpOutput, "%lf\t", values[i]);
  fprintf(fpOutput, "\n");

  fclose(fpOutput);
  return 1;
}


/*
 *  outputCoordinates
 *
 *  Writes coordinate values for individual pixels
 */
int outputCoordinates(char *fnameOutputBase, int *indices, int nIndices, int class_id, int size_x, int size_y)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   i;
  int   x, y;

  // defines output coordinates file for the given class and opens it
  sprintf(fnameOutput, "%s.class%02d.coordinates.txt", fnameOutputBase, class_id+1);
  fpOutput = fopen(fnameOutput, "wt");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
    return 0;
  }

  // writes output coordinates for current frame and class
  for (i=0; i<nIndices; i++)
  {
    x = indices[i]%size_x;
    y = indices[i]/size_x;
    fprintf(fpOutput, "%d\t%d\n", x, y);
  }

  fclose(fpOutput);
  return 1;
}




int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameProportions[FNAMESIZE];
  char fnameClassif[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameTextBase[FNAMESIZE];
  char fnameTextOut[FNAMESIZE];
  char fnameStatsOut[FNAMESIZE];
  int  okProportions, okClassif, okTextOut;
  InrType typeInput;
  InrType type;
  int     typesizeInput;
  
  unsigned char *dataInput;
  unsigned char *ROIMask;
  unsigned char *classif;
  float         *proportions;
  float         *values;
  int           *indices;
  float         *profiles_mean;
  float         *profiles_var;
  int           *profiles_numsamples;
  float         *stats_min;
  float         *stats_max;
  int           *stats_class_id;

  int size_x, size_y, size_z;
  int i, j, k;
  int z;
  int class;
  int nClass;
  
  int    nValues;
  int    iValues;
  float  value;
  double mean;
  double var;


  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile"  , "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"    , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile"  , "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);

  okProportions = igetopt1 ("-prop" , "%s", fnameProportions);
  okClassif     = igetopt1 ("-class", "%s", fnameClassif);
  okTextOut     = igetopt1 ("-otext", "%s", fnameTextBase);

  // either a proportions image or a classification image must be supplied, but not both
  if ((!okProportions && !okClassif) || (okProportions && okClassif))
    iusage_ (Ucmd, Udetail);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (okProportions)
  {
    if (!inrReadHeader(fnameProportions, &i, &j, &k, &type))
    {
      fprintf(stderr, "Error: could not read header information for proportion image '%s'.\n", fnameProportions);
      return 0;
    }
    if (type != InrFloat)
    {
      fprintf(stderr, "Error: proportion image '%s' has to be a image of type 'Float'.\n", fnameProportions);
      return 0;
    }
    if (i != size_x || j != size_y)
    {
      fprintf (stderr, "Error: dimensions of input data and proportion image do not match.\n");
      return 0;
    }
  }
  if (okClassif)
  {
    if (!inrReadHeader(fnameClassif, &i, &j, &k, &type))
    {
      fprintf(stderr, "Error: could not read header information for classification image '%s'.\n", fnameClassif);
      return 0;
    }
    if (type!=InrByte || k != 1)
    {
      fprintf(stderr, "Error: input classification image '%s' has to be a single image of type 'Byte'.\n", fnameInput);
      return 0;
    }
    if (i != size_x || j != size_y)
    {
      fprintf (stderr, "Error: dimensions of input data and classification image do not match.\n");
      return 0;
    }
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
 
  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }

  // reads the proportions/classification image
  proportions = NULL;
  classif     = NULL;
  if (okProportions)
    proportions = inrReadFloatSequence (fnameProportions, &i, &j, &nClass);
  if (okClassif)
  {
    classif = inrReadByteImage (fnameClassif, &i, &j);
    nClass = 0;
    for (k=0; k<i*j; k++)
      nClass = MAX(nClass, classif[k]);
  }
  if (!proportions && !classif)
  {
    fprintf (stderr, "Error: could not read proportion/classification image data.\n");
    free(ROIMask);
    return 0;
  }

  // allocates memory for the input and output data
  typesizeInput       = InrTypeSizes[typeInput]/8;
  dataInput           = (unsigned char *) malloc (size_x*size_y * typesizeInput * sizeof(unsigned char));
  values              = (float *) malloc (size_x*size_y * sizeof(float));
  indices             = (int *)   malloc (size_x*size_y * sizeof(int));
  profiles_mean       = (float *) malloc (nClass*size_z * sizeof(float));
  profiles_var        = (float *) malloc (nClass*size_z * sizeof(float));
  profiles_numsamples = (int *)   malloc (nClass*size_z * sizeof(int));
  stats_min           = (float *) malloc (size_x*size_y * sizeof(float));
  stats_max           = (float *) malloc (size_x*size_y * sizeof(float));
  stats_class_id      = (int *)   malloc (size_x*size_y * sizeof(int));
  
  outputValuesInit(fnameTextBase, nClass);

  // processes the data
  for (z=0; z<size_z; z++)
  {
    fprintf (stderr, "Processing frame %d...\r", z+1);

    // reads input data frame for date z
    if (!inrReadFrame(fnameInput, z, dataInput))
    {
      fprintf(stderr, "Error: could not read frame %d from input file '%s'.\n", z, fnameInput);
      free(dataInput);
      free(ROIMask);
      if (okProportions)
        free(proportions);
      if (okClassif)
        free(classif);
      free(values);
      free(indices);
      free(profiles_mean);
      free(profiles_var);
      free(profiles_numsamples);
      return 0;
    }
    
    // processes each class at a time
    for (class=0; class<nClass; class++)
    {
      nValues = 0;
      mean = 0;
      var  = 0;
      for (i=0; i<size_x*size_y; i++)
      {
        // discards pixels outside of region of interest
        if (!ROIMask[i])
          continue;
        
        // discards non-pure pixels for current class
        if (okProportions && proportions[i + class*size_x*size_y] != 1)
          continue;
        if (okClassif && classif[i] != class+1)
          continue;
          
        // reads value
        value = (float)inrGetDataValue(dataInput, typeInput, i);
        
        // ignores zero values
        if (value == 0)
          continue;
          
        indices[nValues] = i;
        values[nValues]  = value;
        mean            += value;
        nValues++;
        
        // updates stats
        if (z == 0)
        {
          stats_min[i] = value;
          stats_max[i] = value;
          stats_class_id[i] = class+1;
        }
        else
        {
          stats_min[i] = MIN(stats_min[i], value);
          stats_max[i] = MAX(stats_max[i], value);
        }
      }
      // computes mean and variance
      if (nValues > 2)
      {
        mean /= nValues;
        for (iValues=0; iValues<nValues; iValues++)
          var += (values[iValues] - mean)*(values[iValues] - mean);
        var  /= nValues - 1;

        profiles_mean[class + z*nClass]       = (float)mean;
        profiles_var[class + z*nClass]        = (float)var;
        profiles_numsamples[class + z*nClass] = nValues;
      }
      else  
      {
        // invalid mean and variance
        profiles_mean[class + z*nClass]       = -1;
        profiles_var[class + z*nClass]        = 0;
        profiles_numsamples[class + z*nClass] = nValues;
      }
      
      // if specified, outputs values to a text file
      if (okTextOut)
      {
        outputValues(fnameTextBase, values, nValues, class);
        if (z == 0)
          outputCoordinates(fnameTextBase, indices, nValues, class, size_x, size_y);
      }
    }
  }
  
  // writes output profile data
  inrWriteFloatImage (fnameOutput, nClass, size_z, profiles_mean);
  
  // writes output text files, if specified
  if (okTextOut)
  {
    // writes base text file with profile means, standard deviations, etc.
    for (class=0; class<nClass; class++)
    {
      // defines output text file for current class and opens it
      sprintf(fnameTextOut, "%s.class%02d.txt", fnameTextBase, class+1);
	    FILE *fpTextOut = fopen(fnameTextOut, "wt");
	    if (!fpTextOut)
      {
	      fprintf(stderr, "Error: could not open output text file '%s'.\n", fnameTextOut);
        continue;
      }
      
      fprintf(fpTextOut, "! MEAN     \tMEAN-STDDEV\tMEAN+STDDEV\tSTANDARD_DEVIATION\tNUM_SAMPLES\n");
      for (z=0; z<size_z; z++)
	    {
	      fprintf(fpTextOut, "%12.5f\t%12.5f\t%12.5f\t%12.5f\t%d\n", 
                profiles_mean[class + z*nClass],
                profiles_mean[class + z*nClass] - sqrt(profiles_var[class + z*nClass]),
                profiles_mean[class + z*nClass] + sqrt(profiles_var[class + z*nClass]),
                sqrt(profiles_var[class + z*nClass]),
                profiles_numsamples[class + z*nClass]);
	    }
      fclose(fpTextOut);
    }

    // writes statistics file with information for each pixel profile
    for (class=0; class<nClass; class++)
    {
      // defines output stats file for current class and opens it
      sprintf(fnameStatsOut, "%s.class%02d.stats.txt", fnameTextBase, class+1);
      FILE *fpStatsOut = fopen(fnameStatsOut, "wt");
      if (!fpStatsOut)
      {
        fprintf(stderr, "Error: could not open output Stats file '%s'.\n", fnameStatsOut);
        continue;
      }
      
      fprintf(fpStatsOut, "! MIN     \tMAX\n");
      for (i=0; i<size_x*size_y; i++)
      {
        if (stats_class_id[i] != class+1)
          continue;
          
        fprintf(fpStatsOut, "%12.5f\t%12.5f\n", stats_min[i], stats_max[i]);
      }
      fclose(fpStatsOut);
    }
    
  }
  
  fprintf(stderr, "Profile image successfully computed for %d classes and %d dates.\n", nClass, size_z);

  free(dataInput);
  free(ROIMask);
  if (okProportions)
    free(proportions);
  if (okClassif)
    free(classif);
  free(values);
  free(indices);
  free(profiles_mean);
  free(profiles_var);
  free(profiles_numsamples);
  
  return 0;
}
