#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "profile_features.h"


#define FNAMESIZE   1024
#define MIN(x,y)    ((x<y)?x:y)
#define MAX(x,y)    ((x>y)?x:y)
#define DEG_DEFAULT 8


static char Ucmd[] = "-ifile inr_sequence -prop proportions_file \
-class classification_file -roi roi_mask -oname profiles\n";

static char Udetail[] =
"Computes temporal reflectance profiles for individual classes using only pure\n\
pixels. For that purpose, the algorithm uses:\n\
1. An input image sequence, defining the reflectances for a region over time\n\
2. A proportion image, defining the percentage of each class within each pixel\n\
   OR\n\
   A classification image, defining the classes for each pure-pixel\n\
\nInput:\n\
\t-ifile\tFile name of the input image sequence\n\
\t-prop\tFile name of a proportion image\n\
\t-class\tFile name of a classification image\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-oname\tBase name of the output files containing the computed\n\
\t\ttemporal reflectance profiles for each class\n\
\nOptional:\n\
\t-dates\tFile name of the list of dates that correspond to each layer in\n\
\t\tthe image sequence. Dates must be expressed in integer format\n\
\t\t(e.g. day of the year). If a datefile is not specified, fitting\n\
\t\twill be computed considering indices 0...nImages as the 'dates'\n\
\t\tof each layer in the sequence\n\
\t-degree\tDegree of the polynomial function to be used for fitting\n\
\t\tDefault value is 8\n";


/*
 *  outputInrProfilesInit
 *
 *  Initializes output file for outputInrProfiles
 */
int outputInrProfilesInit(char *fnameOutputBase, char *fnameInput)
{
  char fnameOutput[FNAMESIZE];
  int  size_x, size_y, size_z;
  InrType       type;
  InrGeorefInfo georef;

  // reads parameters from input file  
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &type))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }

  // defines output file and opens it 
  sprintf(fnameOutput, "%s.profiles.inr", fnameOutputBase);
  if (!inrWriteHeader(fnameOutput, size_x, size_y, size_z, InrFloat))
  {
    fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
    return 0;
  }
  
  // initializes all values to zero
  inrResetData(fnameOutput, 0);
  
  // writes envi header with georeference information, if available  
  if (inrReadGeoref(fnameInput, &georef))
    inrWriteGeoref(fnameOutput, georef);
  else
    inrWriteEnviHdr(fnameOutput);
   
  return 1;
}

/*
 *  outputInrProfiles
 *
 *  Writes profiles Initializes output file for outputInrProfiles
 */
int outputInrProfiles(char *fnameOutputBase, double *coef, int degree,
                      float *values, int *dates, int nDates, int x, int y)
{
  char fnameOutput[FNAMESIZE];
  int  i, iDate;
  
  // computes profile
  for (iDate=0; iDate<nDates; iDate++)
  {
    double value = 0;
    for (i=0; i<(degree+1); i++)
      value += coef[i]*pow(dates[iDate],i);
    values[iDate] = (float)value;
  }

  // defines output file and writes point to it
  sprintf(fnameOutput, "%s.profiles.inr", fnameOutputBase);
  if (!inrWritePoint(fnameOutput, x, y, (unsigned char *) values))
  {
    fprintf(stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  return 1;
}


/*
 *  outputCoefValuesInit
 *
 *  Initializes output files for outputCoefValues
 */
int outputCoefValuesInit(char *fnameOutputBase, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;

  for (iClass=0; iClass<nClass; iClass++)
  {
    // defines output file for the given class and opens it in
    sprintf(fnameOutput, "%s.class%02d.coef.txt", fnameOutputBase, iClass+1);
    fpOutput = fopen(fnameOutput, "wt");
    if (!fpOutput)
    {
      fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
      return 0;
    }
    fclose(fpOutput);
  }
  return 1;
}

/*
 *  outputCoefValues
 *
 *  Writes coefficient values for individual pixels
 */
int outputCoefValues(char *fnameOutputBase, double *coef, int degree, int class_id)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   i;

  // defines output file for the given class and opens it in for appending
  sprintf(fnameOutput, "%s.class%02d.coef.txt", fnameOutputBase, class_id+1);
  fpOutput = fopen(fnameOutput, "at");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
    return 0;
  }

  // writes coefficient values
  for (i=0; i<(degree+1); i++)
    fprintf(fpOutput, "%10g\t", coef[i]);
  fprintf(fpOutput, "\n");

  fclose(fpOutput);
  return 1;
}


/*
 *  outputProfileValuesInit
 *
 *  Initializes output files for outputProfileValues
 */
int outputProfileValuesInit(char *fnameOutputBase, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;

  for (iClass=0; iClass<nClass; iClass++)
  {
    // defines output file for the given class and opens it in
    sprintf(fnameOutput, "%s.class%02d.profiles.txt", fnameOutputBase, iClass+1);
    fpOutput = fopen(fnameOutput, "wt");
    if (!fpOutput)
    {
      fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
      return 0;
    }
    fclose(fpOutput);
  }
  return 1;
}

/*
 *  outputProfileValues
 *
 *  Writes profile values for individual pixels
 */
int outputProfileValues(char *fnameOutputBase, double *coef, int degree,
                        int *dates, int nDates, int class_id)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iDate;
  int   i;

  // defines output file for the given class and opens it in for appending
  sprintf(fnameOutput, "%s.class%02d.profiles.txt", fnameOutputBase, class_id+1);
  fpOutput = fopen(fnameOutput, "at");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
    return 0;
  }

  // writes computed profile
  for (iDate=0; iDate<nDates; iDate++)
  {
    double value = 0;
    for (i=0; i<(degree+1); i++)
      value += coef[i]*pow(dates[iDate],i);
    fprintf(fpOutput, "%12.5lf\t", value);
  }
  fprintf(fpOutput, "\n");
  fclose(fpOutput);
  return 1;
}



/*
 *  outputFeatureValuesInit
 *
 *  Initializes output files for outputFeatureValues
 */
int outputFeatureValuesInit(char *fnameOutputBase, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;
  int   iMode;
  int   i;

  for (iClass=0; iClass<nClass; iClass++)
  {
    // defines output file for the given class and opens it in
    sprintf(fnameOutput, "%s.class%02d.features.txt", fnameOutputBase, iClass+1);
/*
    sprintf(fnameOutput, "%s.features.allpoints.txt", fnameOutputBase);
*/
    fpOutput = fopen(fnameOutput, "wt");
    if (!fpOutput)
    {
      fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
      return 0;
    }
    
    fprintf(fpOutput, "%12s\t", "x");
    fprintf(fpOutput, "%12s\t", "y");
    fprintf(fpOutput, "%12s\t", "CLASS");
    
    for (i=0; i<PROF_NUM_FEATURES; i++)
      fprintf(fpOutput, "%12s\t", profFeaturesNames[i]);
    
    fprintf(fpOutput, "%12s\t", "nModes");
//*** DEBUG - MAXIMUM NUMBER OF MODES==3 ***
//    for (iMode=0; iMode<PROF_MAX_MODES; iMode++)
    for (iMode=0; iMode<3; iMode++)
    {
      fprintf(fpOutput, "%12s%d\t", "datePeak", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "dateWidthPeak", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "dateWidthLeft", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "dateWidthRight", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "valuePeak", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "valueDeltaLeft", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "valueDeltaRight", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "integral", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "slopeLeft", iMode+1);
      fprintf(fpOutput, "%12s%d\t", "slopeRight", iMode+1);
    }
    fprintf(fpOutput, "\n");
    fclose(fpOutput);
  }
/*
*/
  return 1;
}

/*
 *  outputFeatureValues
 *
 *  Writes feature values for individual pixels
 */
int outputFeatureValues(char *fnameOutputBase, int x, int y, profFeatures features, int class_id)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iMode;
  int   i;

  // defines output file for the given class and opens it in for appending
  sprintf(fnameOutput, "%s.class%02d.features.txt", fnameOutputBase, class_id+1);
/*
  sprintf(fnameOutput, "%s.features.allpoints.txt", fnameOutputBase);
*/
  fpOutput = fopen(fnameOutput, "at");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
    return 0;
  }

  // writes feature values
  
  fprintf(fpOutput, "%12d\t", x);
  fprintf(fpOutput, "%12d\t", y);
  fprintf(fpOutput, "%12d\t", class_id+1);

  for (i=0; i<PROF_NUM_FEATURES; i++)
    fprintf(fpOutput, "%12.5lf\t", features.val[i]);

  fprintf(fpOutput, "%12d\t"   , features.nModes);

//*** DEBUG ***
  for (iMode=0; iMode<features.nModes; iMode++)
//  for (iMode=0; iMode<PROF_MAX_MODES; iMode++)
//  for (iMode=0; iMode<3; iMode++)
  {
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].datePeak);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].dateWidthPeak);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].dateWidthLeft);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].dateWidthRight);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].valuePeak);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].valueDeltaLeft);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].valueDeltaRight);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].integral);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].slopeLeft);
    fprintf(fpOutput, "%12.5lf\t", features.modes[iMode].slopeRight);
  }

  fprintf(fpOutput, "\n");
  fclose(fpOutput);
  return 1;
}



/*
 *  outputCoef
 *
 *  Writes base text file with profile means, standard deviations, etc.
 */
int outputCoef(char *fnameOutputBase, double *coef_mean, double *coef_stddev,
               int degree, int *class_numsamples, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;
  int   i;

  sprintf(fnameOutput, "%s.coef.txt", fnameOutputBase);
  fpOutput = fopen(fnameOutput, "wt");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output text file '%s'.\n", fnameOutput);
    return 0;
  }
  for (iClass=0; iClass<nClass; iClass++)
  {
    fprintf(fpOutput, "CLASS %02d\n", iClass+1);
    
    fprintf(fpOutput, "  coefficients:\t");
    for (i=0; i<(degree+1); i++)
      fprintf(fpOutput, "%10g\t", coef_mean[i + iClass*(degree+1)]);
    fprintf(fpOutput, "\n");
    
    fprintf(fpOutput, "  stddev:\t");
    for (i=0; i<(degree+1); i++)
      fprintf(fpOutput, "%10g\t", coef_stddev[i + iClass*(degree+1)]);
    fprintf(fpOutput, "\n");

    fprintf(fpOutput, "  num_samples:\t%d", class_numsamples[iClass]);
    fprintf(fpOutput, "\n\n");
  }
  fclose(fpOutput);
  
  return 1;
}


/*
 *  outputProfiles
 *
 *  Writes temporal profiles for each class based on coefficients
 */
int outputProfiles(char *fnameOutputBase, double *coef_mean, int degree,
                   int *dates, int nDates, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;
  int   iDate;
  int   i;

  // defines output profiles file and opens it
  sprintf(fnameOutput, "%s.profiles.txt", fnameOutputBase);
  fpOutput = fopen(fnameOutput, "wt");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output profiles file '%s'.\n", fnameOutput);
    return 0;
  }

  // writes computed profile for each class 
  for (iDate=0; iDate<nDates; iDate++)
  {
    fprintf(fpOutput, "%d\t", dates[iDate]);
    for (iClass=0; iClass<nClass; iClass++)
    {
      double value = 0;
      for (i=0; i<(degree+1); i++)
        value += coef_mean[i + iClass*(degree+1)]*pow(dates[iDate],i);
      fprintf(fpOutput, "%12.5lf\t", value);
    }
    fprintf(fpOutput, "\n");
  }
  fclose(fpOutput);
  return 1;
}



/*
 *  outputFeatures
 *
 *  Writes features computed for each class based on their temporal profiles
 */
int outputFeatures(char *fnameOutputBase, profFeatures *features_mean,
                   profFeatures *features_stddev, int *class_numsamples, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   iClass;
  int   iMode;
  int   i;
  char  feat[50];

  sprintf(fnameOutput, "%s.features.txt", fnameOutputBase);
  fpOutput = fopen(fnameOutput, "wt");
  if (!fpOutput)
  {
    fprintf(stderr, "Error: could not open output text file '%s'.\n", fnameOutput);
    return 0;
  }
  for (iClass=0; iClass<nClass; iClass++)
  {
    fprintf(fpOutput, "CLASS %02d\n", iClass+1);
    
    fprintf(fpOutput, "  samples: %d\n", class_numsamples[iClass]);

    for (i=0; i<PROF_NUM_FEATURES; i++)
    {
      sprintf(feat, "%s:", profFeaturesNames[i]);
      fprintf(fpOutput, "  %-20s %12.5lf      (stddev: %12.5lf)\n", feat, features_mean[iClass].val[i], features_stddev[iClass].val[i]);
    }

    fprintf(fpOutput, "  %-20s %12.5lf      (stddev: %12.5lf)\n", "nModes:", features_mean[iClass].nModesStats  , features_stddev[iClass].nModesStats);

    for (iMode=0; iMode<features_mean[iClass].nModes; iMode++)
    {
      fprintf(fpOutput, "  mode %d:\n", iMode+1);
      fprintf(fpOutput, "    frequency: %5.2f%%\n", (double)features_mean[iClass].modes[iMode].nSamples*100/class_numsamples[iClass]);
      fprintf(fpOutput, "    datePeak:        %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].datePeak       , features_stddev[iClass].modes[iMode].datePeak);
      fprintf(fpOutput, "    dateWidthPeak:   %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].dateWidthPeak  , features_stddev[iClass].modes[iMode].dateWidthPeak);
      fprintf(fpOutput, "    dateWidthLeft:   %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].dateWidthLeft  , features_stddev[iClass].modes[iMode].dateWidthLeft);
      fprintf(fpOutput, "    dateWidthRight:  %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].dateWidthRight , features_stddev[iClass].modes[iMode].dateWidthRight);
      fprintf(fpOutput, "    valuePeak:       %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].valuePeak      , features_stddev[iClass].modes[iMode].valuePeak);
      fprintf(fpOutput, "    valueDeltaLeft:  %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].valueDeltaLeft , features_stddev[iClass].modes[iMode].valueDeltaLeft);
      fprintf(fpOutput, "    valueDeltaRight: %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].valueDeltaRight, features_stddev[iClass].modes[iMode].valueDeltaRight);
      fprintf(fpOutput, "    integral:        %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].integral       , features_stddev[iClass].modes[iMode].integral);
      fprintf(fpOutput, "    slopeLeft:       %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].slopeLeft      , features_stddev[iClass].modes[iMode].slopeLeft);
      fprintf(fpOutput, "    slopeRight:      %12.5lf      (stddev: %12.5lf)\n", features_mean[iClass].modes[iMode].slopeRight     , features_stddev[iClass].modes[iMode].slopeRight);
    }
    fprintf(fpOutput, "\n\n");
  }
  fclose(fpOutput);
  
  return 1;
}



/*
 *  outputCoordinates
 *
 *  Writes coordinate values for individual pixels
 */
int outputCoordinates(char *fnameOutputBase, unsigned char *ROIMask, unsigned char *classif,
                      int size_x, int size_y, int nClass)
{
  char  fnameOutput[FNAMESIZE];
  FILE *fpOutput;
  int   pos;
  int   x, y;
  int   iClass;
  
  for (iClass=0; iClass<nClass; iClass++)
  {
    // defines output coordinates file for the given class and opens it
    sprintf(fnameOutput, "%s.class%02d.coordinates.txt", fnameOutputBase, iClass+1);
    fpOutput = fopen(fnameOutput, "wt");
    if (!fpOutput)
    {
      fprintf(stderr, "Error: could not open output file '%s'.\n", fnameOutput);
      return 0;
    }
  
    // writes output coordinates for current frame and class
    for (y=0; y<size_y; y++)
    {
      for (x=0; x<size_x; x++)
      {
        pos = x + y*size_x;
        if (!ROIMask[pos])
          continue; 

        if (classif[pos] == iClass+1)
          fprintf(fpOutput, "%d\t%d\n", x, y);
      }
    }
  
    fclose(fpOutput);
  }
  return 1;
}



/*
 *  defineDates
 *
 *  Builds date vector from a datefile if it is specified
 *  If not, date vector will be [0, 1, .., nDates]
 */
int *defineDates(int hasFile, char *fnameDates, int nDates)
{
  int *dates = (int *)malloc(nDates*sizeof(int));
  int  date;
  int  iDate;
  FILE *fpDates;
  
  if (!hasFile)
  {
    // case 1: no dates file specified
    for (iDate=0; iDate<nDates; iDate++)
      dates[iDate] = iDate;
    return dates;
  }
  else
  {
    // case 2: dates file specified
    fpDates = fopen(fnameDates, "rt");
    if (!fpDates)
    {
      fprintf(stderr, "Error: could not open dates file '%s'.\n", fnameDates);
      free(dates);
      return NULL;
    }
    for (iDate=0; iDate<nDates; iDate++)
    {
      if (!fscanf(fpDates, "%d", &date))
      {
        fprintf(stderr, "Error: not enough dates (%d) in dates file '%s'.\n", iDate, fnameDates);
        free(dates);
        return NULL;
      }
      dates[iDate] = date;
    }
  }
  return dates;
}



int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameProportions[FNAMESIZE];
  char fnameClassif[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameOutputBase[FNAMESIZE];
  char fnameDates[FNAMESIZE];
  int  okProportions, okClassif;
  int  okDegree, okDates;
  InrType typeInput;
  InrType type;
  int     typesizeInput;
  
  unsigned char *dataInput;
  unsigned char *ROIMask;
  float         *proportions;
  unsigned char *classif;
  
  int      degree;
  double  *values;
  float   *values_fit;
  int     *dates;
  double  *coef;
  double  *coef_sum;
  double  *coef_sum2;
  double  *coef_mean;
  double  *coef_stddev;
  int     *class_numsamples;
  int      class_id;
  
  
  profFeatures  features;
  profFeatures *features_sum;
  profFeatures *features_sum2;
  profFeatures *features_mean;
  profFeatures *features_stddev;

  int size_x, size_y, size_z;
  int i, j, k;
  int x, y, z;
  int iClass;
  int nClass;
  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"  , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-oname", "%s", fnameOutputBase))
    iusage_ (Ucmd, Udetail);
  okProportions = igetopt1 ("-prop"  , "%s", fnameProportions);
  okClassif     = igetopt1 ("-class" , "%s", fnameClassif);
  okDegree      = igetopt1 ("-degree", "%d", &degree);
  okDates       = igetopt1 ("-dates" , "%s", fnameDates);

  // either a proportions image or a classification image must be supplied, but not both
  if ((!okProportions && !okClassif) || (okProportions && okClassif))
    iusage_ (Ucmd, Udetail);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okDegree)
    degree = DEG_DEFAULT;
    
    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (okProportions)
  {
    if (!inrReadHeader(fnameProportions, &i, &j, &k, &type))
    {
      fprintf(stderr, "Error: could not read header information for proportion image '%s'.\n", fnameProportions);
      return 0;
    }
    if (type != InrFloat)
    {
      fprintf(stderr, "Error: proportion image '%s' has to be a image of type 'Float'.\n", fnameProportions);
      return 0;
    }
    if (i != size_x || j != size_y)
    {
      fprintf (stderr, "Error: dimensions of input data and proportion image do not match.\n");
      return 0;
    }
  }
  if (okClassif)
  {
    if (!inrReadHeader(fnameClassif, &i, &j, &k, &type))
    {
      fprintf(stderr, "Error: could not read header information for classification image '%s'.\n", fnameClassif);
      return 0;
    }
    if (type!=InrByte || k != 1)
    {
      fprintf(stderr, "Error: input classification image '%s' has to be a single image of type 'Byte'.\n", fnameClassif);
      return 0;
    }
    if (i != size_x || j != size_y)
    {
      fprintf (stderr, "Error: dimensions of input data and classification image do not match.\n");
      return 0;
    }
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
 
  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }

  // reads the proportions/classification image
  proportions = NULL;
  classif     = NULL;
  if (okProportions)
    proportions = inrReadFloatSequence (fnameProportions, &i, &j, &nClass);
  if (okClassif)
  {
    classif = inrReadByteImage (fnameClassif, &i, &j);
    nClass = 0;
    for (k=0; k<i*j; k++)
      nClass = MAX(nClass, classif[k]);
  }
  if (!proportions && !classif)
  {
    fprintf (stderr, "Error: could not read proportion/classification image data.\n");
    free(ROIMask);
    return 0;
  }
  
  // defines date vector (1,2..nDates if not specified)
  dates = defineDates(okDates, fnameDates, size_z);
  if (!dates)
    return 0;

  // outputs coordinate values for individual pixels of each class
  outputCoordinates(fnameOutputBase, ROIMask, classif, size_x, size_y, nClass);

  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  values           = (double *) malloc (size_z * sizeof(double));
  values_fit       = (float  *) malloc (size_z * sizeof(float));
  coef             = (double *) malloc ((degree+1) * sizeof(double));
  coef_sum         = (double *) malloc (nClass*(degree+1) * sizeof(double));
  coef_sum2        = (double *) malloc (nClass*(degree+1) * sizeof(double));
  coef_mean        = (double *) malloc (nClass*(degree+1) * sizeof(double));
  coef_stddev      = (double *) malloc (nClass*(degree+1) * sizeof(double));
  class_numsamples = (int *)    malloc (nClass * sizeof(int));
  features_sum     = (profFeatures *) malloc (nClass * sizeof(profFeatures));
  features_sum2    = (profFeatures *) malloc (nClass * sizeof(profFeatures));
  features_mean    = (profFeatures *) malloc (nClass * sizeof(profFeatures));
  features_stddev  = (profFeatures *) malloc (nClass * sizeof(profFeatures));
  
  for (iClass=0; iClass<nClass; iClass++)
  {
    class_numsamples[iClass] = 0;
    for (i=0; i<(degree+1); i++)
    {
      coef[i] = 0;
      coef_sum[i + iClass*(degree+1)] = 0;
      coef_sum2[i + iClass*(degree+1)] = 0;
    }
    profInitStruct(&features_sum[iClass]);
    profInitStruct(&features_sum2[iClass]);
  }
  outputInrProfilesInit  (fnameOutputBase, fnameInput);
  outputCoefValuesInit   (fnameOutputBase, nClass);
  outputProfileValuesInit(fnameOutputBase, nClass);
  outputFeatureValuesInit(fnameOutputBase, nClass);


  
  // PROCESSES THE DATA
  
  for (y=0; y<size_y; y++)
  {
    fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);

    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    if (x == size_x)
      continue;

    // reads input data line y
    if (!inrReadLine(fnameInput, y, dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }
    
    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      // discards pixels outside of region of interest
      if (!ROIMask[x + y*size_x])
        continue;
        
      // checks which class current pixel belongs to (considers only pure pixels)
      class_id = -1;

      if (okProportions)
        for (iClass=0; iClass<nClass; iClass++)
          if (proportions[x + y*size_x + iClass*size_x*size_y] == 1)
            class_id = iClass;
      if (okClassif)
        class_id = classif[x + y*size_x] - 1;

      if (class_id == -1)
        continue;    
        
      // reads values for each date of the data point
      for (z=0; z<size_z; z++)
        values[z] = inrGetDataValue(dataInput, typeInput, x+z*size_x);

      // ignores point if values are zero
      if (values[0] == 0)
        continue;        
      
      
      // computes the coefficients that best fit the profile values
      if (!profFitPoly(values, dates, size_z, coef, degree))
        continue;


      // initializes feature structure
      profInitStruct(&features);
      // computes profile features based on the polynomial coefficients
      profComputePolyFeatures(coef, degree, dates[0], dates[size_z-1], 1, &features);
      // computes profile features based on the raw values
      profComputeValueFeatures(values, dates, size_z, NULL, NULL, 0, &features);



      // updates coefficient and feature sum and sum^2 values (for mean and stddev computation)
      for (i=0; i<(degree+1); i++)
      {
        coef_sum[i + class_id*(degree+1)]  += coef[i];
        coef_sum2[i + class_id*(degree+1)] += coef[i]*coef[i];
      }
      profFeaturesSum(&features_sum[class_id], &features_sum2[class_id], features);
      
      class_numsamples[class_id]++;
      

      
      // outputs values
      if (!outputInrProfiles  (fnameOutputBase, coef, degree, values_fit, dates, size_z, x, y))
        return 0;
      if (!outputCoefValues   (fnameOutputBase, coef, degree, class_id))
        return 0;
      if (!outputProfileValues(fnameOutputBase, coef, degree, dates, size_z, class_id))
        return 0;
      if (!outputFeatureValues(fnameOutputBase, x, y, features, class_id))
        return 0;
    }
  }
  
  // computes mean and stddev of all coefficients and features
  for (iClass=0; iClass<nClass; iClass++)
  {
    int n = class_numsamples[iClass];
    
    for (i=0; i<(degree+1); i++)
    {
      double sum  = coef_sum[i + iClass*(degree+1)];
      double sum2 = coef_sum2[i + iClass*(degree+1)];
      
      if (n < 2)
      {
        coef_mean[i + iClass*(degree+1)] = 0;
        coef_stddev[i + iClass*(degree+1)] = -1;
      }
      else
      {
        coef_mean[i + iClass*(degree+1)] = sum/n;
        coef_stddev[i + iClass*(degree+1)] = sqrt((sum2 - sum*sum/n)/(n-1));
      }
    }
    profInitStruct(&features_mean[iClass]);
    profInitStruct(&features_stddev[iClass]);
    if (n > 1)
    {
      profFeaturesMean(features_sum[iClass], n, &features_mean[iClass]);
      profFeaturesStddev(features_sum[iClass], features_sum2[iClass], n, &features_stddev[iClass]);
    }
    
  }
  
  // outputs results
  outputCoef(fnameOutputBase, coef_mean, coef_stddev, degree, class_numsamples, nClass); 
  outputProfiles(fnameOutputBase, coef_mean, degree, dates, size_z, nClass);
  outputFeatures(fnameOutputBase, features_mean, features_stddev, class_numsamples, nClass); 
  
    
  fprintf(stderr, "Profiles successfully computed for %d classes and fitted to a polynomial function of degree %d\n", nClass, degree);

  free(dataInput);
  free(ROIMask);
  if (okProportions)
    free(proportions);
  if (okClassif)
    free(classif);
  free(values);
  free(values_fit);
  free(dates);
  free(coef);
  free(coef_sum);
  free(coef_sum2);
  free(coef_mean);
  free(coef_stddev);
  free(class_numsamples);
  free(features_sum);
  free(features_sum2);
  free(features_mean);
  free(features_stddev);
  
  return 0;
}
