#include <stdio.h>
#include <stdlib.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "svd.h"
#include "sequence_proportions_estimation.h"

#define FNAMESIZE  1024


static char Ucmd[] = "-c1 modis_ch1 -c2 modis_ch2 \n\
\t\t-prof1 profiles_ch1 -prof2 prof_ch2\n\
\t\t-roi roi_mask -ofile proportions [-chan1] [-chan2]\n";

static char Udetail[] =
"Estimates proportions for each land cover within MODIS pixels.\n\
\nThe observed reflectance of each MODIS pixel is considered to be a linear\n\
combination of the pure reflectances of each class, weighted by their\n\
proportions within the pixel.\n\
\nInput:\n\
\t-c1\tFilename of the image sequence containing observed MODIS\n\
\t\treflectances for channel 1\n\
\t-c2\tFilename of the image sequence containing observed MODIS\n\
\t\treflectances for channel 2\n\
\t-prof1\tFile name of the image sequence that contains the reflectance\n\
\t\tprofiles of each class for channel 1\n\
\t-prof2\tFile name of the image sequence that contains the reflectance\n\
\t\tprofiles of each class for channel 2\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output proportions image\n\
\nOptional:\n\
\t-chan1\tUse channel 1\n\
\t-chan2\tUse channel 2\n\
\t\t(default is to use both channels)\n";



int main (int argc, char **argv)
{
  char fnameChannel1[FNAMESIZE];
  char fnameChannel2[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameProfiles1[FNAMESIZE];
  char fnameProfiles2[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  int  okChannel1, okChannel2;

  
  unsigned char *dataChannel1;
  unsigned char *dataChannel2;
  unsigned char *dataProfiles1;
  unsigned char *dataProfiles2;
  unsigned char *ROIMask;
  float         *dataOutput;
  
  InrGeorefInfo georef;
  InrType type;
  InrType typeChannel1, typeChannel2;
  InrType typeProfiles1, typeProfiles2;
  int typesizeChannel1, typesizeChannel2;
  int size_x, size_y;
  int size_image;
  int nDates, nDates1, nDates2;
  int date;
  int nClass;
  int class;
  int x, y;
  int i, j, k;

  double       *proportions;
  dataPropEstim pixelPropEstim;


  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  if (!igetopt1 ("-c1", "%s", fnameChannel1))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-c2", "%s", fnameChannel2))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-prof1", "%s", fnameProfiles1))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-prof2", "%s", fnameProfiles2))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi", "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  okChannel1 = igetopt0 ("-chan1");
  okChannel2 = igetopt0 ("-chan2");

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  /* if neither chan1 nor chan2 is specified, both channels are processed */
  if (!okChannel1 && !okChannel2)
    okChannel1 = okChannel2 = 1;


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameChannel1, &size_x, &size_y, &nDates1, &typeChannel1))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameChannel1);
    return 0;
  }
  if (!inrReadHeader(fnameChannel2, &i, &j, &nDates2, &typeChannel2))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameChannel2);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data for channels 1 and 2 do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameProfiles1, &nClass, &j, &k, &typeProfiles1))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameProfiles1);
    return 0;
  }
  if (j != nDates1)
  {
    fprintf(stderr, "Error: dimensions of profile image '%s' don't match those of input file %s.\n", fnameProfiles1, fnameChannel1);
    return 0;
  }
  if (k != 1)
  {
    fprintf(stderr, "Error: profile '%s' has to be a single image.\n", fnameProfiles1);
    return 0;
  }
  if (!inrReadHeader(fnameProfiles2, &i, &j, &k, &typeProfiles2))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameProfiles2);
    return 0;
  }
  if (i != nClass)
  {
    fprintf(stderr, "Error: number of classes in profile images '%s' and '%s' don't match.\n", fnameProfiles1, fnameProfiles2);
    return 0;
  }
  if (j != nDates2)
  {
    fprintf(stderr, "Error: dimensions of profile image '%s' don't match those of input file %s.\n", fnameProfiles2, fnameChannel2);
    return 0;
  }
  if (k != 1)
  {
    fprintf(stderr, "Error: profile '%s' has to be a single image.\n", fnameProfiles2);
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  
  // defines auxiliary variables
  size_image = size_x * size_y;
  typesizeChannel1 = InrTypeSizes[typeChannel1]/8;
  typesizeChannel2 = InrTypeSizes[typeChannel2]/8;
  nDates = 0;
  if (okChannel1)
    nDates = nDates1;
  if (okChannel2)
    nDates = nDates2;
  if (okChannel1 && okChannel2)
    nDates = nDates1 + nDates2;


  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }

  // reads the reflectance profile data
  dataProfiles1 = inrReadDataSequence (fnameProfiles1, &i, &j, &k, &type);
  dataProfiles2 = inrReadDataSequence (fnameProfiles2, &i, &j, &k, &type);


  // allocates memory for the input and output data
  dataChannel1 = (unsigned char *) malloc (size_x*nDates * typesizeChannel1 * sizeof(unsigned char));
  dataChannel2 = (unsigned char *) malloc (size_x*nDates * typesizeChannel2 * sizeof(unsigned char));
  dataOutput   = (float *)  malloc (nClass * size_image * sizeof (float));
  proportions  = (double *) malloc (nClass * sizeof (double));


  // initializes data structure for estimating proportions within a pixel
  pixelPropEstim.nDates = nDates;
  pixelPropEstim.nClass = nClass;
  pixelPropEstim.ref_pixel = vector (1, nDates);
  pixelPropEstim.ref_class = matrix (1, nDates, 1, nClass);
  if (okChannel1)
  {
    // fills structure with each class's reflectance profile for channel 1
    for (date = 0; date < nDates1; date++)
    {
      for (class = 0; class < nClass; class++)
      {
        int    pos   = class + nClass*date;
        double value = inrGetDataValue(dataProfiles1, typeProfiles1, pos);
        pixelPropEstim.ref_class[date + 1][class + 1] = (float) value;
      }
    }
  }
  if (okChannel2)
  {
    // fills structure with each class's reflectance profile for channel 1
    if (!okChannel1)
      nDates1 = 1;
    for (date = 0; date < nDates2; date++)
    {
      for (class = 0; class < nClass; class++)
      {
        int    pos   = class + nClass*date;
        double value = inrGetDataValue(dataProfiles2, typeProfiles2, pos);
        pixelPropEstim.ref_class[nDates1 + date][class + 1] = (float) value;
      }
    }
  }

  // initializes output proportions to zero
  for (i = 0; i < size_image; i++)
    for (class = 0; class < nClass; class++)
      dataOutput[i + size_image * class] = 0.0;


  // processes the data
  for (y=0; y<size_y; y++)
  {
    fprintf (stderr, "Processing ... %d %%\r", 100 * (y + 1) / size_y);
    
    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    if (x == size_x)
      continue;

    // reads data for current line
    if (okChannel1 && !inrReadLine(fnameChannel1, y, dataChannel1))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameChannel1);
      return 0;
    }
    if (okChannel2 && !inrReadLine(fnameChannel2, y, dataChannel2))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameChannel2);
      return 0;
    }

    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      // discards pixels outside of region of interest
      if (!ROIMask[x + y*size_x])
        continue;

      if (okChannel1)
      {
        // fills structure with the pixel's observed reflectances for channel 1
        for (date = 0; date < nDates1; date++)
        {
          int    pos   = x + date*size_x;
          double value = inrGetDataValue(dataChannel1, typeChannel1, pos);
          pixelPropEstim.ref_pixel[date + 1] = (float) value;
        }
      }
      if (okChannel2)
      {
        // fills structure with the pixel's observed reflectances for channel 2
        if (!okChannel1)
          nDates1 = 1;
        for (date = 0; date < nDates2; date++)
        {
          int    pos   = x + date*size_x;
          double value = inrGetDataValue(dataChannel2, typeChannel2, pos);
          pixelPropEstim.ref_pixel[nDates1 + date] = (float) value;
        }
      }


      // estimates proportions within the pixel
      estim_prop_pix (proportions, nClass, &pixelPropEstim);


      for (class = 0; class < nClass; class++)
      {
        int pos = x + y*size_x + class*size_image;
        if (proportions[class] < 0)
          dataOutput[pos] = 0.0;
        else
          dataOutput[pos] = (float) proportions[class];
      }
    }
  }

  // writes output translated classification
  if (!inrWriteFloatSequence(fnameOutput, size_x, size_y, nClass, dataOutput))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  // copies georreference info if present
  if (inrReadGeoref(fnameROIMask, &georef))
    inrWriteGeoref(fnameOutput, georef);



  free (dataChannel1);
  free (dataChannel2);
  free (dataProfiles1);
  free (dataProfiles2);
  free (ROIMask);
  free (dataOutput);
  free_matrix (pixelPropEstim.ref_class, 1, nDates, 1, nClass);
  free_vector (pixelPropEstim.ref_pixel, 1, nDates);
  free (proportions);

  fprintf(stderr, "Proportion image successfully computed for %d classes.\n", nClass);
  return 1;
}
