#include <stdio.h>
#include <stdlib.h>
#include <inrimage/image.h>
#include "inr_io.h"
#include "svd.h"
#include "sequence_proportions_estimation.h"

#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE  1024

#define WINSIZE_DEFAULT 3


static char Ucmd[] = "-c1 modis_ch1 -c2 modis_ch2 \n\
\t\t-prof1 profiles_ch1 -prof2 prof_ch2\n\
\t\t-class inr_classification -roi roi_mask \n\
\t\t-ofile inr_proportions \n\
\t\t[-win size] [-localprof] [-chan1] [-chan2]\n";

static char Udetail[] =
"Estimates proportions within unclassified MODIS pixels based on neighbouring\n\
classified pixels.\n\
\nThe observed reflectance of each MODIS pixel is considered to be a linear\n\
combination of the pure reflectances of each class, weighted by their\n\
proportions within the pixel.\n\
A local window centered on the unclassified pixel is used so as to define a\n\
neighbourhood of that pixel. Classes to be considered for proportions estimation\n\
are limited to those that are present within the neighbourhood (i.e., there\n\
must be at least one classified pixel with that class label within the local\n\
window). Class reflectance profiles may be specified globally (-prof1, -prof2)\n\
or estimated locally using the mean reflectance profiles of classified pixels\n\
within the local window.\n\
\nInput:\n\
\t-c1\tFile name of the image sequence containing observed MODIS\n\
\t\treflectances for channel 1\n\
\t-c2\tFile name of the image sequence containing observed MODIS\n\
\t\treflectances for channel 2\n\
\t-prof1\tFile name of the image sequence that contains the reflectance\n\
\t\tprofiles of each class for channel 1\n\
\t-prof2\tFile name of the image sequence that contains the reflectance\n\
\t\tprofiles of each class for channel 2\n\
\t-class\tFile name of a classification image\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output proportions image\n\
\nOptional:\n\
\t-win\tValue defining size of the window used for local proportions\n\
\t\testimation. Window will have UL and LR corners at points\n\
\t\t(x-size,y-size), (x+size,y+size)\n\
\t\tDefault value is 3 (about 1km between UL and center)\n\
\t-localprof\tEstimate class reflectance profiles from classified pixels\n\
\t\t\twithin the local window\n\
\t\t\tDefault is to always use reflectance profiles provided\n\
\t-chan1\tUse channel 1\n\
\t-chan2\tUse channel 2\n\
\t\t(default is to use both channels)\n";


int fillPropEstim_GlobalProf(char *fnameChannel1, unsigned char *dataChannel1, InrType typeChannel1,
                             char *fnameChannel2, unsigned char *dataChannel2, InrType typeChannel2,
                             unsigned char *dataClassif, int winSize,
                             int okChannel1, int okChannel2,
                             int x, int y, int size_x, int size_y,
                             int nDates1, int nDates2,
                             dataPropEstim *propEstim)
{
  double value;
  float  total;
  int    class, iClass;
  int    iDate, iBegin;
  int    x_begin, x_end;
  int    y_begin, y_end;
  int    x_win, y_win;
  
  // defines local window
  x_begin = MAX(x - winSize, 0);
  x_end   = MIN(x + winSize, size_x-1);
  y_begin = MAX(y - winSize, 0);
  y_end   = MIN(y + winSize, size_y-1);
  
  // resets profiles, prior probabilities, etc (but not global class profiles)
  propEstimReset(propEstim, 0);

  // processes classified pixels within the local window
  for (y_win = y_begin; y_win <= y_end ; y_win++)   
  {
    for (x_win = x_begin; x_win <= x_end ; x_win++)   
    {
      class = dataClassif[x_win + y_win*size_x];
      if (!class)
        continue;
        
      // increments class count
      propEstim->prior_prob[class]++;
    }
  }
  
  // finalizes prior probability computation: divides each class count by total count
  total = 0;
  for (iClass=0; iClass<propEstim->nClass; iClass++)
    total += propEstim->prior_prob[iClass + 1];
  if (total > 0)
    for (iClass=0; iClass<propEstim->nClass; iClass++)
      propEstim->prior_prob[iClass + 1] /= total;
    
  // records classes found within local window
  propEstim->nClassLocal = 0;
  for (iClass=0; iClass<propEstim->nClass; iClass++)
  {
    if (propEstim->prior_prob[iClass + 1] > 0)
    {
      propEstim->local_class[propEstim->nClassLocal] = iClass+1;
      propEstim->nClassLocal++;
    }
  }
  
  // fills up pixel reflectance profile with the appropriate channels
  iBegin = 1;
  if (okChannel1)
  {
    // reads channel 1 data for current point
    if (!inrReadPoint(fnameChannel1, x, y, dataChannel1))
    {
      fprintf(stderr, "Error: could not read point (%d,%d) from input file '%s'.\n", x, y, fnameChannel1);
      return 0;
    }
    // fills profile with reflectances for channel 1
    for (iDate=0; iDate<nDates1; iDate++)
    {
      value = inrGetDataValue(dataChannel1, typeChannel1, iDate);
      propEstim->ref_pixel[1 + iDate] = (float) value;
    }
    iBegin += nDates1;
  }
  if (okChannel2)
  {
    // reads channel 2 data for current point
    if (!inrReadPoint(fnameChannel2, x, y, dataChannel2))
    {
      fprintf(stderr, "Error: could not read point (%d,%d) from input file '%s'.\n", x, y, fnameChannel2);
      return 0;
    }
    // fills profile with reflectances for channel 2
    for (iDate=0; iDate<nDates2; iDate++)
    {
      value = inrGetDataValue(dataChannel2, typeChannel2, iDate);
      propEstim->ref_pixel[iBegin + iDate] = (float) value;
    }
  }
  
  return propEstim->nClassLocal;
}


int fillPropEstim_LocalProf(char *fnameChannel1, unsigned char *dataChannel1, InrType typeChannel1,
                            char *fnameChannel2, unsigned char *dataChannel2, InrType typeChannel2,
                            unsigned char *dataClassif, int winSize,
                            int okChannel1, int okChannel2,
                            int x, int y, int size_x, int size_y,
                            int nDates1, int nDates2,
                            dataPropEstim *propEstim)
{
  double value;
  float  total;
  int    class, iClass;
  int    iDate, iBegin;
  int    x_begin, x_end;
  int    y_begin, y_end;
  int    x_win, y_win;
  
  // defines local window
  x_begin = MAX(x - winSize, 0);
  x_end   = MIN(x + winSize, size_x-1);
  y_begin = MAX(y - winSize, 0);
  y_end   = MIN(y + winSize, size_y-1);
  
  // resets all profiles, prior probabilities, etc
  propEstimReset(propEstim, 1);

  // processes classified pixels within the local window
  for (y_win = y_begin; y_win <= y_end ; y_win++)   
  {
    for (x_win = x_begin; x_win <= x_end ; x_win++)   
    {
      class = dataClassif[x_win + y_win*size_x];
      if (!class)
        continue;
        
      // increments class count
      propEstim->prior_prob[class]++;
      
        
      // adds profile to the class's profile
      iBegin = 1;
      if (okChannel1)
      {
        // reads channel 1 data for window point
        if (!inrReadPoint(fnameChannel1, x_win, y_win, dataChannel1))
        {
          fprintf(stderr, "Error: could not read point (%d,%d) from input file '%s'.\n", x_win, y_win, fnameChannel1);
          return 0;
        }
        // adds reflectance profile for channel 1
        for (iDate=0; iDate<nDates1; iDate++)
        {
          value = inrGetDataValue(dataChannel1, typeChannel1, iDate);
          propEstim->ref_class[iBegin + iDate][class] += (float) value;
        }
        iBegin += nDates1;
      }
      if (okChannel2)
      {
        // reads channel 2 data for window point
        if (!inrReadPoint(fnameChannel2, x_win, y_win, dataChannel2))
        {
          fprintf(stderr, "Error: could not read point (%d,%d) from input file '%s'.\n", x_win, y_win, fnameChannel2);
          return 0;
        }
        // adds reflectance profile for channel 2
        for (iDate=0; iDate<nDates2; iDate++)
        {
          value = inrGetDataValue(dataChannel2, typeChannel2, iDate);
          propEstim->ref_class[iBegin + iDate][class] += (float) value;
        }
      }
    }
  }
  
  // finalizes class profile computation: divides sum by class count to obtain mean
  for (iClass=0; iClass<propEstim->nClass; iClass++)
  {
    if (propEstim->prior_prob[iClass + 1] == 0)
      continue;
    for (iDate=0; iDate<propEstim->nDates; iDate++)
      propEstim->ref_class[iDate + 1][iClass + 1] /= propEstim->prior_prob[iClass + 1];
  }

  // finalizes prior probability computation: divides each class count by total count
  total = 0;
  for (iClass=0; iClass<propEstim->nClass; iClass++)
    total += propEstim->prior_prob[iClass + 1];
  if (total > 0)
    for (iClass=0; iClass<propEstim->nClass; iClass++)
      propEstim->prior_prob[iClass + 1] /= total;
    
  // records classes found within local window
  propEstim->nClassLocal = 0;
  for (iClass=0; iClass<propEstim->nClass; iClass++)
  {
    if (propEstim->prior_prob[iClass + 1] > 0)
    {
      propEstim->local_class[propEstim->nClassLocal] = iClass+1;
      propEstim->nClassLocal++;
    }
  }
  
  // fills up pixel reflectance profile with the appropriate channels
  iBegin = 1;
  if (okChannel1)
  {
    // reads channel 1 data for current point
    if (!inrReadPoint(fnameChannel1, x, y, dataChannel1))
    {
      fprintf(stderr, "Error: could not read point (%d,%d) from input file '%s'.\n", x, y, fnameChannel1);
      return 0;
    }
    // fills profile with reflectances for channel 1
    for (iDate=0; iDate<nDates1; iDate++)
    {
      value = inrGetDataValue(dataChannel1, typeChannel1, iDate);
      propEstim->ref_pixel[1 + iDate] = (float) value;
    }
    iBegin += nDates1;
  }
  if (okChannel2)
  {
    // reads channel 2 data for current point
    if (!inrReadPoint(fnameChannel2, x, y, dataChannel2))
    {
      fprintf(stderr, "Error: could not read point (%d,%d) from input file '%s'.\n", x, y, fnameChannel2);
      return 0;
    }
    // fills profile with reflectances for channel 2
    for (iDate=0; iDate<nDates2; iDate++)
    {
      value = inrGetDataValue(dataChannel2, typeChannel2, iDate);
      propEstim->ref_pixel[iBegin + iDate] = (float) value;
    }
  }
  
  return propEstim->nClassLocal;
}


int main (int argc, char **argv)
{
  char fnameChannel1[FNAMESIZE];
  char fnameChannel2[FNAMESIZE];
  char fnameProfiles1[FNAMESIZE];
  char fnameProfiles2[FNAMESIZE];
  char fnameClassif[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  int  okChannel1, okChannel2;
  int  okWinSize, okLocalProf;
  int  winSize;
  
  unsigned char *dataChannel1;
  unsigned char *dataChannel2;
  unsigned char *dataProfiles1;
  unsigned char *dataProfiles2;
  unsigned char *dataClassif;
  unsigned char *ROIMask;
  float         *dataOutput;
  
  InrGeorefInfo georef;
  InrType type;
  InrType typeChannel1, typeChannel2;
  InrType typeProfiles1, typeProfiles2;
  int typesizeChannel1, typesizeChannel2;
  int size_x, size_y;
  int size_image;
  int nDates, nDates1, nDates2;
  int iDate, iBegin;
  int nClass;
  int iClass, class;
  int x, y;
  int i, j, k;

  int    pos;
  double value;
  
  double       *proportions;
  dataPropEstim pixelPropEstim;


  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);

  if (!igetopt1 ("-c1"   , "%s", fnameChannel1))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-c2"   , "%s", fnameChannel2))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-prof1", "%s", fnameProfiles1))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-prof2", "%s", fnameProfiles2))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-class", "%s", fnameClassif))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi", "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  okWinSize   = igetopt1 ("-win", "%d", &winSize);
  okLocalProf = igetopt0 ("-localprof");
  okChannel1  = igetopt0 ("-chan1");
  okChannel2  = igetopt0 ("-chan2");

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  /* if neither chan1 nor chan2 is specified, both channels are processed */
  if (!okChannel1 && !okChannel2)
    okChannel1 = okChannel2 = 1;

  // defines value for window size
  if (!okWinSize)
    winSize = WINSIZE_DEFAULT;

  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameChannel1, &size_x, &size_y, &nDates1, &typeChannel1))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameChannel1);
    return 0;
  }
  if (!inrReadHeader(fnameChannel2, &i, &j, &nDates2, &typeChannel2))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameChannel2);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data for channels 1 and 2 do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameClassif, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for classification image '%s'.\n", fnameClassif);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: Classification image '%s' has to be a single image of type 'Byte'.\n", fnameClassif);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and classification image do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameProfiles1, &nClass, &j, &k, &typeProfiles1))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameProfiles1);
    return 0;
  }
  if (j != nDates1)
  {
    fprintf(stderr, "Error: dimensions of profile image '%s' don't match those of input file %s.\n", fnameProfiles1, fnameChannel1);
    return 0;
  }
  if (k != 1)
  {
    fprintf(stderr, "Error: profile '%s' has to be a single image.\n", fnameProfiles1);
    return 0;
  }
  if (!inrReadHeader(fnameProfiles2, &i, &j, &k, &typeProfiles2))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameProfiles2);
    return 0;
  }
  if (i != nClass)
  {
    fprintf(stderr, "Error: number of classes in profile images '%s' and '%s' don't match.\n", fnameProfiles1, fnameProfiles2);
    return 0;
  }
  if (j != nDates2)
  {
    fprintf(stderr, "Error: dimensions of profile image '%s' don't match those of input file %s.\n", fnameProfiles2, fnameChannel2);
    return 0;
  }
  if (k != 1)
  {
    fprintf(stderr, "Error: profile '%s' has to be a single image.\n", fnameProfiles2);
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
  
  // defines auxiliary variables
  size_image = size_x * size_y;
  typesizeChannel1 = InrTypeSizes[typeChannel1]/8;
  typesizeChannel2 = InrTypeSizes[typeChannel2]/8;
  nDates = 0;
  if (okChannel1)
    nDates = nDates1;
  if (okChannel2)
    nDates = nDates2;
  if (okChannel1 && okChannel2)
    nDates = nDates1 + nDates2;


  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }

  // reads the classification data
  dataClassif = inrReadByteImage (fnameClassif, &i, &j);
  if (!dataClassif)
  {
    fprintf (stderr, "Error: could not read classification data.\n");
    return 0;
  }
  // verifies total number of classes in the classification image
  nClass = 0;
  for (i=0; i<size_x*size_y; i++)
    nClass = MAX(nClass, dataClassif[i]);


  // reads the reflectance profile data
  dataProfiles1 = inrReadDataSequence (fnameProfiles1, &i, &j, &k, &type);
  dataProfiles2 = inrReadDataSequence (fnameProfiles2, &i, &j, &k, &type);

  // allocates memory for the input and output data
  dataChannel1  = (unsigned char *) malloc (nDates1 * typesizeChannel1*sizeof(unsigned char));
  dataChannel2  = (unsigned char *) malloc (nDates2 * typesizeChannel2 * sizeof(unsigned char));
  dataOutput    = (float *)  malloc (nClass * size_image * sizeof (float));
  proportions   = (double *) malloc (nClass * sizeof (double));


  // initializes data structure for estimating proportions within a pixel
  pixelPropEstim.nDates = nDates;
  pixelPropEstim.nClass = nClass;
  pixelPropEstim.ref_pixel   = vector (1, nDates);
  pixelPropEstim.ref_class   = matrix (1, nDates, 1, nClass);
  pixelPropEstim.prior_prob  = vector (1, nClass);
  pixelPropEstim.local_class = (int *)malloc(nClass*sizeof(int));
  // fills up class profiles with the appropriate channels
  iBegin = 1;
  if (okChannel1)
  {
    // fills structure with each class's reflectance profile for channel 1
    for (iDate=0; iDate<nDates1; iDate++)
    {
      for (iClass = 0; iClass < nClass; iClass++)
      {
        pos   = iClass + nClass*iDate;
        value = inrGetDataValue(dataProfiles1, typeProfiles1, pos);
        pixelPropEstim.ref_class[iBegin + iDate][1 + iClass] = (float) value;
      }
    }
    iBegin += nDates1;
  }
  if (okChannel2)
  {
    // fills structure with each class's reflectance profile for channel 1
    for (iDate=0; iDate<nDates2; iDate++)
    {
      for (iClass = 0; iClass < nClass; iClass++)
      {
        pos   = iClass + nClass*iDate;
        value = inrGetDataValue(dataProfiles2, typeProfiles2, pos);
        pixelPropEstim.ref_class[iBegin + iDate][1 + iClass] = (float) value;
      }
    }
  }

  // initializes output proportions to zero
  for (i = 0; i < size_image; i++)
    for (class = 0; class < nClass; class++)
      dataOutput[i + size_image * class] = 0.0;


  // processes the data
  //*** DEBUG ***
  for (y=0; y<size_y; y++)
//  for (y=0; y<690; y++)
  {
    fprintf (stderr, "Processing ... %d %%\r", 100 * (y + 1) / size_y);

    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      // discards pixels outside of region of interest
      if (!ROIMask[x + y*size_x])
        continue;

      // if pixel is classified, sets proportion to 100% for the class and moves on
      class = dataClassif[x + y*size_x];
      if (class > 0)
      {
        pos = x + y*size_x + (class-1)*size_image;
        dataOutput[pos] = 1;
        continue;
      }

      if (okLocalProf)
        // fills reflectance profile structure appropriately (using local profiles)
        fillPropEstim_LocalProf(fnameChannel1, dataChannel1, typeChannel1,
                                fnameChannel2, dataChannel2, typeChannel2,
                                dataClassif, winSize, okChannel1, okChannel2,
                                x, y, size_x, size_y, nDates1, nDates2,
                                &pixelPropEstim);
      else
        // fills reflectance profile structure appropriately (using global profiles)
        fillPropEstim_GlobalProf(fnameChannel1, dataChannel1, typeChannel1,
                                fnameChannel2, dataChannel2, typeChannel2,
                                dataClassif, winSize, okChannel1, okChannel2,
                                x, y, size_x, size_y, nDates1, nDates2,
                                &pixelPropEstim);

      // no estimation is made if local window contains no classified points
      if (pixelPropEstim.nClassLocal == 0)
        continue;

      // if only a single class is present in the window, then this class is
      // assigned to the pixel and no estimation is necessary
      if (pixelPropEstim.nClassLocal == 1)
      {
        // obtains which is the only class present within the window
        class = pixelPropEstim.local_class[0];

        // sets proportion to 100% of the class
        pos = x + y*size_x + (class-1)*size_image;
        dataOutput[pos] = 1;
        continue;
      }


      // estimates proportions within the pixel
      estim_prop_pix (proportions, pixelPropEstim.nClassLocal, &pixelPropEstim);


      // transfers estimated proportions to output buffer
      for (iClass = 0; iClass < pixelPropEstim.nClassLocal; iClass++)
      {
        class = pixelPropEstim.local_class[iClass];
        pos   = x + y*size_x + (class-1)*size_image;
        if (proportions[iClass] < 0)
          dataOutput[pos] = 0.0;
        else
          dataOutput[pos] = (float) proportions[iClass];
      }
    }
  }

  // writes output translated classification
  if (!inrWriteFloatSequence(fnameOutput, size_x, size_y, nClass, dataOutput))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  // copies georreference info if present
  if (inrReadGeoref(fnameROIMask, &georef))
    inrWriteGeoref(fnameOutput, georef);



  free (dataChannel1);
  free (dataChannel2);
  free (dataClassif);
  free (dataProfiles1);
  free (dataProfiles2);
  free (ROIMask);
  free (dataOutput);
  free_matrix (pixelPropEstim.ref_class, 1, nDates, 1, nClass);
  free_vector (pixelPropEstim.ref_pixel, 1, nDates);
  free_vector (pixelPropEstim.prior_prob, 1, nClass);
  free (pixelPropEstim.local_class);
  free (proportions);

  fprintf(stderr, "Proportion image successfully computed for %d classes.\n", nClass);
  return 1;
}
