#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define FNAMESIZE  1024
#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)



static char Ucmd[] = "-ifile inr_proportions -class inr_classification -ofile inr_merged_proportions\n";

static char Udetail[] =
"Given an INRIMAGE proportions image and a classification image, merges the two\n\
into a new proportions image where all classified points have proportion = 1\n\
for the specified class and 0 otherwise. Unclassified points are given the\n\
proportions specified in the original proportions image.\n\
\nInput:\n\
\t-ifile\tFile name of the input proportions image\n\
\t-class\tFile name of a classification image\n\
\nOutput:\n\
\t-ofile\tFile name of the output new proportions image\n";



int main (int argc, char **argv)
{
  char fnameClassif[FNAMESIZE];
  char fnameProp[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  InrGeorefInfo georef;
  InrType       type;
  InrType       typeProp;
  int           typesizeProp;


  unsigned char *dataClassif;
  unsigned char *dataProp;
  
  int size_x, size_y, size_z;
  int i, j, k, z;
  int class;
  int nClass;
  

  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameProp))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-class", "%s", fnameClassif))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);


  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameProp, &size_x, &size_y, &size_z, &typeProp))
  {
    fprintf(stderr, "Error: could not read header information for proportions image '%s'.\n", fnameProp);
    return 0;
  }
  if (!inrReadHeader(fnameClassif, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for classification image '%s'.\n", fnameClassif);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: Classification image '%s' has to be a single image of type 'Byte'.\n", fnameClassif);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of proportions image and classification image do not match.\n");
    return 0;
  }
  typesizeProp = InrTypeSizes[typeProp]/8;

 
  // reads the proportions data
  dataProp = inrReadDataSequence(fnameProp, &i, &j, &k, &type);
  if (!dataProp)
  {
    fprintf (stderr, "Error: could not read proportions data.\n");
    return 0;
  }
  // reads the classification data
  dataClassif = inrReadByteImage (fnameClassif, &i, &j);
  if (!dataClassif)
  {
    fprintf (stderr, "Error: could not read classification data.\n");
    return 0;
  }
  // verifies total number of classes in the classification image
  nClass = 0;
  for (i=0; i<size_x*size_y; i++)
    nClass = MAX(nClass, dataClassif[i]);
    

  // extends proportions image if nClass is beyond size_z
  if (nClass > size_z)
  {
    // reallocates memory for 'dataProp'
    dataProp = realloc(dataProp, size_x*size_y*nClass * typesizeProp * sizeof(unsigned char));
    if (!dataProp)
    {
      fprintf (stderr, "Error: could not reallocate memory to incorporate %d classes.\n", nClass);
      return 0;
    }
    // initializes all new values to zero
    for (i = size_x*size_y*size_z; i < size_x*size_y*nClass; i++)
      inrPutDataValue(dataProp, typeProp, i, 0);
    
    size_z = nClass;
  }
  


  // PROCESSES THE DATA

  for (i=0; i<size_x*size_y; i++)
  {
    class = dataClassif[i];
    
    if (!class)
      continue;

    // if position has a classification, proportions for the class is 1, and all
    // other proportions are 0
    for (z=0; z<size_z; z++)
      inrPutDataValue(dataProp, typeProp, i + z*size_x*size_y, 0);

    inrPutDataValue(dataProp, typeProp, i + (class-1)*size_x*size_y, 1);
  }  
  
  
  
  // writes output merged proportions file
  if (!inrWriteDataSequence(fnameOutput, size_x, size_y, size_z, dataProp, typeProp))
  {
    fprintf (stderr, "Error: could not write to output file '%s'.\n", fnameOutput);
    return 0;
  }
  
  // copies georreference info if present
  if (inrReadGeoref(fnameProp, &georef))
    inrWriteGeoref(fnameOutput, georef);
  
  fprintf(stderr, "Classification/proportions merged successfully.\n");

  free(dataProp);
  free(dataClassif);
  return 1;
}
