#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "inr_io.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define FNAMESIZE         1024
#define TOLERANCE_DEFAULT 0.2



static char Ucmd[] = "-ifile inr_proportions -truth inr_proportions_ground_truth \
roi roi_mask [-tol value]\n";

static char Udetail[] =
"Given a proportions image, compares it to a specified ground truth proportions\n\
image so as to evaluate its accuracy.\n\
\nStatistics computed for comparison are, for each class:\n\
  . RMSE of the proportions estimation\n\
  . Percentage of proportions that have true value within range:\n\
      [proportion-tol/2 , proportion+tol/2]\n\
  . Percentage of proportions that fall within the same range as the\n\
    true values. Ranges used are:\n\
      { [0, tol) ; [tol, 2*tol) ; ... ; [k*tol, 1] }\n\
All input files must be in INRIMAGE format.\n\
Output is given in stdout.\n\
\nInput:\n\
\t-ifile\tFile name of the input classification image\n\
\t-truth\tFile name of a classification image to be used as ground truth\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOptional:\n\
\t-tol\tTolerance value for considering a proportion as correct\n\
\t\tDefault value is 0.2\n";



void outputStats(double *stats, int nClass)
{
  int i;
  double sum = 0;
  
  for (i=0; i<nClass; i++)
    printf("\t%d", i+1);
  printf("\n");
  for (i=0; i<nClass; i++)
  {
    printf("\t%.2lf", stats[i]);
    sum += stats[i];
  }
  printf("\n\n");
  printf("\tOVERALL: %.2lf\n\n", sum/nClass);
}

void outputStatsPerc(double *stats, int nClass)
{
  int i;
  double sum = 0;
  
  for (i=0; i<nClass; i++)
    printf("\t%d", i+1);
  printf("\n");
  for (i=0; i<nClass; i++)
  {
    printf("\t%.2lf%%", stats[i]*100);
    sum += stats[i]*100;
  }
  printf("\n\n");
  printf("\tOVERALL: %.2lf%%\n\n", sum/nClass);
}

void outputRanges(double tolerance, int nRanges)
{
  int i;
  double beginRange, endRange;
  
  printf("Ranges:\t");
  for (i=0; i<nRanges; i++)
  {
    beginRange = i*tolerance;
    endRange   = MIN((i+1)*tolerance,1);
    printf("[%.2lf,%.2lf", beginRange, endRange);
    if (i != nRanges-1)
      printf(") ; ");
    else
      printf("]");
  }
  printf("\n\n");
}




int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameTruth[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  InrType typeInput;
  InrType typeTruth;
  InrType type;
  
  unsigned char *dataInput;
  unsigned char *dataTruth;
  unsigned char *ROIMask;

  int    okTolerance;
  double tolerance;
  
  int    nClass;
  int    class;
  int    pos;
  double valueInput;
  double valueTruth;
  
  int size_x, size_y, size_z;
  int i, j, k;
  
  // statistics
  double *RMSE;
  double *perc_within_tol;
  double *perc_same_range;

  // auxiliary variables for computing statistics
  int rangeInput;
  int rangeTruth;
  int nRanges;
  int nPoints;

  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-truth", "%s", fnameTruth))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"  , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  okTolerance = igetopt1 ("-tol", "%lf", &tolerance);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okTolerance)
    tolerance = TOLERANCE_DEFAULT;
  nRanges = (int)ceil(1./tolerance - 1e-10);
    
  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (typeInput != InrFloat && typeInput != InrDouble)
  {
    fprintf(stderr, "Error: Input proportions image '%s' has to be of type 'Float' or 'Double'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameTruth, &i, &j, &k, &typeTruth))
  {
    fprintf(stderr, "Error: could not read header information for ground truth image '%s'.\n", fnameTruth);
    return 0;
  }
  if (typeTruth != InrFloat && typeInput != InrDouble)
  {
    fprintf(stderr, "Error: Ground truth proportions image  '%s' has to be of type 'Float' or 'Double'.\n", fnameTruth);
    return 0;
  }
  if (i != size_x || j != size_y || k != size_z)
  {
    fprintf (stderr, "Error: dimensions of input and ground truth images do not match.\n");
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }
 
  // reads the input proportions data
  dataInput = inrReadDataSequence(fnameInput, &i, &j, &k, &type);
  if (!dataInput)
  {
    fprintf (stderr, "Error: could not read input proportions data.\n");
    return 0;
  }
  // reads the ground truth proportions data
  dataTruth = inrReadDataSequence(fnameTruth, &i, &j, &k, &type);
  if (!dataTruth)
  {
    fprintf (stderr, "Error: could not read ground truth proportions data.\n");
    return 0;
  }
  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }


  // verifies total number of classes 
  nClass = size_z;


  // allocates memory 
  RMSE                  = (double *) malloc (nClass * sizeof(double));
  perc_within_tol       = (double *) malloc (nClass * sizeof(double));
  perc_same_range       = (double *) malloc (nClass * sizeof(double));
  
  // initializes variables
  nPoints = 0;
  for (i=0; i<nClass; i++)
  {
    RMSE[i] = 0;
    perc_within_tol[i] = 0;
    perc_same_range[i] = 0;
  }


  // PROCESSES THE DATA

  for (i=0; i<size_x*size_y; i++)
  {
    // skips pixels outside of the region of interest
    if (!ROIMask[i])
      continue;
        
    for (class=0; class<nClass; class++)
    {
      // gets input and truth values
      pos = i + class*size_x*size_y;
      valueInput = inrGetDataValue(dataInput, typeInput, pos);
      valueTruth = inrGetDataValue(dataTruth, typeTruth, pos);
        
      // updates RMSE
      RMSE[class] += (valueInput-valueTruth)*(valueInput-valueTruth);
  
      // updates perc_within_tol
      if (fabs(valueInput-valueTruth) <= tolerance/2)
        perc_within_tol[class]++;
        
      // updates perc_same_range
      rangeInput = (int)MIN(floor(valueInput / tolerance), nRanges-1);
      rangeTruth = (int)MIN(floor(valueTruth / tolerance), nRanges-1);
      if (rangeInput == rangeTruth)
        perc_same_range[class]++;
    }
    nPoints++;
  }
  // finishes up computing the statistics
  for (class=0; class<nClass; class++)
  {
    RMSE[class] = sqrt(RMSE[class]/nPoints);
    perc_within_tol[class] /= nPoints;
    perc_same_range[class] /= nPoints;
  }


  // OUTPUTS STATISTICS

  printf("\n");
  printf("RMSE - Root Mean Square Error\n");
  printf("-----------------------------\n");
  outputStats(RMSE, nClass);

  printf("\n");
  printf("Percentage of Proportions Within Tolerance Range\n");
  printf("------------------------------------------------\n");
  outputStatsPerc(perc_within_tol, nClass);

  printf("\n");
  printf("Percentage of Proportions In Same Range\n");
  printf("---------------------------------------\n");
  outputRanges(tolerance, nRanges);
  outputStatsPerc(perc_same_range, nClass);

    
  free(dataInput);
  free(dataTruth);
  free(RMSE);
  free(perc_within_tol);
  free(perc_same_range);
  
  return 0;
}
