/***************************************************************************
                          fuzzyclassdef.h  -  description
                             -------------------
    Created              : 7 August, 2007
    By                   : (C) 2007 by Carlos de la Torre
    Email                : Carlos.de_la_Torre@inria.fr
 ***************************************************************************/

#ifndef __FUZZYCLASSDEF_H__
#define __FUZZYCLASSDEF_H__

#include "linearalgebra.h"
#include "inr_io.h"

/*
 *
 * The classes can be defined in two ways:
 * - The maha strategy requires a means vector and covariance matrix
 *   for the values of the features in every class.
 * - The fusion strategies requiere the mean and standard deviation 
 *   of every class/feature pair.
 *
 */

typedef enum
{
  Maha,
  FusionWA,
  FusionDR,
} strategy;
 
struct classdef_t {
  strategy  strat;
  int       nClasses;
  int       nFeatures;

  //For the maha strategy
  int**     features_index; //and DS 
  matrice** mean; // for each class, a line with the mean of each feature.
  matrice** cov_matrix;
  
  //For both fusion strategies
  float**   cfMean;
  float**   cfStdDev;
  //For the FusionWA strategy
  float**   weight;

  //For the FusionDS strategy
  int* features_number;

};

typedef struct classdef_t classdef;

/*
 * Initialize and destroy:
 *
 * defCreate          : Creates an empty classdef.
 * defCreateFromFile  : Creates a classdef and reads the means and covarainces from a file
 * defDestroy         : Guess what...
 *
 */
classdef *defCreate          (int nClass, int nFeatures, strategy s);
classdef *defCreateFromFile  (char *filename);
classdef *defDestroy         (classdef *def);

/*
 * defGetClassPoint
 * Given an array 'values' with the values for all the features,
 * this method returns a matrice (only one line) with only the values
 * of the featuresselected for class 'iClass'.
 *
 */
int defGetClassPoint(classdef *def, int iClass, float *values, matrice *point);

/*
 * Accesors: incomplete. The rest should be created... and used!!!
 *
 * defGetClassFeatMean: returns the mean of feature iFeat in class iClass
 * defGetClassFeatStdDev: returns the standard deviation of feature iFeat in class iClass
 */
matrice  *defGetCovariance     (classdef* def, int iClass);
int       defSetCovariance     (classdef* def, int iClass, matrice *cov);
int       defGetClassNfeatures (classdef* def, int iClass);
float     defGetClassFeatMean  (classdef* def, int iClass, int iFeat);
float     defGetClassFeatStdDev(classdef* def, int iClass, int iFeat);


/*
 * defInvertCov
 * It inverts the covariance matrix of every class (useful to
 * compute the Mahalanobis distance)
 * 
 */
int defInvertCov(classdef *def);

                            
//For the training

// Config the Maha strategy: Which features interest which classes
int defConfigFromFile            (classdef *def, char *filename);
int defToTextFile                (classdef *def, FILE *fp);
int defComputeDefinitions        (classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y, int z);

//Auxiliars
int defCreateMahaFromFile        (FILE* fp, classdef* def);
int defCreateFusionWAFromFile    (FILE* fp, classdef* def);
int defCreateFusionDRFromFile    (FILE* fp, classdef* def);

int defConfigMahaFromFile        (classdef *def, FILE *fp);
int defConfigFusionWAFromFile    (classdef *def, FILE *fp);
int defConfigDSFromFile          (classdef *def, FILE *fp);

int defMahaToTextFile            (classdef *def, FILE *fpOutput);
int defFusionDRToTextFile        (classdef *def, FILE *fpOutput);
int defFusionWAToTextFile        (classdef *def, FILE *fpOutput);

int defComputeFusionWADefinitions(classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y);
int defComputeMahaDefinitions    (classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y, int z);
int defComputeFusionDRDefinitions(classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y);


#endif //__FUZZYCLASSDEF_H__

