/***************************************************************************
                          fuzzyclassdef.c  -  description
                             -------------------
    Created              : 7 August, 2007
    By                   : (C) 2007 by Carlos de la Torre
    Email                : Carlos.de_la_Torre@inria.fr
 ***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "fuzzyclassdef.h"
#include "linearalgebra.h"
#include "inr_io.h"

#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)


/*
 * defCreate
 * 
 * Creates a new class definition from scratch
 * All features are initially considered to be continuous
 */
classdef *defCreate (int nClass, int nFeatures, strategy s)
{
  int iClass;

  classdef *def = (classdef *)malloc(sizeof(struct classdef_t));

  def->nClasses          = nClass;
  def->nFeatures         = nFeatures;
  def->strat             = s;
  if(s == Maha)
  {
    def->cov_matrix        = (matrice**) malloc(nClass*sizeof(matrice*));
    def->mean              = (matrice**) malloc(nClass*sizeof(matrice*));
    def->features_index    = (int**) malloc(nClass*sizeof(int));
    def->cfMean            = NULL;
    def->cfStdDev          = NULL;
    def->weight            = NULL;
  }
  else if(s == FusionWA)
  {
    def->cov_matrix        = NULL;
    def->mean              = NULL;
    def->features_index    = NULL;
    def->cfMean            = (float**) malloc(nClass*sizeof(float*));
    def->cfStdDev          = (float**) malloc(nClass*sizeof(float*));
    def->weight            = (float**) malloc(nClass*sizeof(float*));
    for (iClass=0; iClass<nClass; iClass++)
    {
      def->cfMean[iClass]   = (float*) calloc(nFeatures,sizeof(float));
      def->cfStdDev[iClass] = (float*) calloc(nFeatures,sizeof(float));
      def->weight[iClass]   = (float*) calloc(nFeatures,sizeof(float)); 
    }
  }
  else if(s == FusionDR)
  {
    def->cov_matrix        = NULL;
    def->mean              = NULL;
    def->features_index    = NULL;
    def->cfMean            = (float**) malloc(nClass*sizeof(float*));
    def->cfStdDev          = (float**) malloc(nClass*sizeof(float*));
    def->weight            = NULL;
    for (iClass=0; iClass<nClass; iClass++)
    {
      def->cfMean[iClass]   = (float*) calloc(nFeatures,sizeof(float));
      def->cfStdDev[iClass] = (float*) calloc(nFeatures,sizeof(float));
    }
  }
  return def;
}



/*
 * defCreateFromFile
 * 
 * Creates a new class definition from a class definition file.
 * The first line of the file must state the strategy, which implies
 * the format of the file.
 */
classdef *defCreateFromFile(char *filename)
{
  int nClass;
  int nFeatures;
  int ret = 0;
  
  classdef* def;
  FILE *fp;
  int strat; //strategy

  fp = fopen(filename, "rt");
  if (!fp)
  {
    fprintf (stderr, "Error: could not open class definition file.\n");
    return NULL;
  }
  // reads the strategy (represented by a number... check the enum in the header file)
  if (fscanf(fp, "STRATEGY:%d\n", &strat) != 1)
  {
    fprintf (stderr, "Error: could not read the strategy.\n");
    fclose(fp);
    return NULL;
  }
  // reads number of classes and of features
  if (fscanf(fp, "CLASSES:%d\n", &nClass) != 1)
  {
    fprintf (stderr, "Error: could not read number of classes from class definition file.\n");
    return NULL;
  }
  if (fscanf(fp, "FEATURES:%d\n", &nFeatures) != 1)
  {
    fprintf (stderr, "Error: could not read number of features from class definition file.\n");
    return NULL;
  }

  // creates a new class definition structure  
  def = defCreate(nClass, nFeatures, strat);

  switch(strat)
  {
    case Maha:
      ret = defCreateMahaFromFile(fp, def);
      break;
    case FusionWA:
      ret = defCreateFusionWAFromFile(fp, def);
      break;
    case FusionDR:
      ret = defCreateFusionDRFromFile(fp, def);
      break;
    default:
      fprintf(stderr,"Error: Unknown strategy.\n");
      ret = 0;
  }

  if(!ret)
  {
    def = defDestroy(def);
  }

  return def;
}


/*
 * defGetClassFeatMean
 * 
 * Accesor: returns the mean of feature iFeat in class iClass
 */
float defGetClassFeatMean(classdef* def, int iClass, int iFeat)
{
  return def->cfMean[iClass-1][iFeat-1];
}


/*
 * defGetClassFeatStdDev
 * 
* Accesor: returns the standard deviation of feature iFeat in class iClass
 */
float defGetClassFeatStdDev(classdef* def, int iClass, int iFeat)
{
  return def->cfStdDev[iClass-1][iFeat-1];
}

/*
 * defCreateMahaFromFile
 * 
 * Creates a new class definition from a class definition file,
 * given the maha strategy.
 */
int defCreateMahaFromFile(FILE* fp, classdef* def)
{
  int iClass, ncFeatures;
  int i;

  // reads the classses mean and covariance matrices
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    if(fscanf(fp, "class %*d num. Features: %d\n",&ncFeatures) != 1)
    {
      fprintf (stderr, "Error: problems reading number of features for class %d.\n",iClass+1);
      return 0;
    }
    // Read the features index
    fscanf(fp, "class %*d features:\n");
    def->features_index[iClass] = (int*) calloc(ncFeatures,sizeof(int));
    for (i=0; i<ncFeatures; i++)
    {
      if (fscanf(fp, "%d ", &(def->features_index[iClass][i])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d mean.\n",iClass);
        return 0;
      }    
    }
    fscanf(fp, "\n");
    // Read the vector of means
    fscanf(fp, "class %*d means:\n");
    def->mean[iClass] = Matrice_Nulle(1,ncFeatures);
    for (i=0; i<ncFeatures; i++)
    {
      if (fscanf(fp, "%lf ", &(def->mean[iClass]->val[i])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d mean.\n",iClass);
        return 0;
      }    
    }
    fscanf(fp, "\n");
    // Read the covariance matrix
    fscanf(fp, "class %*d covariance matrix:\n");
    def->cov_matrix[iClass] = Matrice_Nulle(ncFeatures, ncFeatures);
    for (i=0; i<ncFeatures*ncFeatures; i++)
    {
      if (fscanf(fp, "%lf ", &(def->cov_matrix[iClass]->val[i])) != 1)
      {
        fprintf (stderr, "Error: problems reading the covariance matrix for class %d.\n",iClass);
        return 0;
      }
    }
  }
  
  return 1;
}



/*
 * defCreateFusionWAFromFile
 * 
 * Creates a new class definition from a class definition file,
 * given the FusionWA strategy.
 * It takes a FILE* already opened and advanced into the file.
 * The begining is exactly the same as the FusionDR strategy file.
 */
int defCreateFusionWAFromFile(FILE* fp, classdef* def)
{
  int iClass, iFeat;

  if (!defCreateFusionDRFromFile(fp, def)) return 0;

  // Above, the FILE* was advanced up to the weights matrix.
  // Read the weigths matrix
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      if (fscanf(fp, "%f ", &(def->weight[iClass][iFeat])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d, feature %d weigth.\n",iClass+1, iFeat+1);
        return 0;
      }    
    }
    fscanf(fp, "\n");
  }

  return 1;
}

int defCreateFusionDRFromFile(FILE* fp, classdef* def)
{
  int iClass, iFeat;

  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    fscanf(fp, "class %*d:\n");
    // Read the mean of every feature in this class
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      if (fscanf(fp, "%f ", &(def->cfMean[iClass][iFeat])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d, feature %d mean.\n",iClass+1, iFeat+1);
        return 0;
      }    
    }
    fscanf(fp, "\n");
    // Read the standar deviation of every feature in this class
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      if (fscanf(fp, "%f ", &(def->cfStdDev[iClass][iFeat])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d, feature %d standard deviation.\n",iClass+1, iFeat+1);
        return 0;
      }    
    }
    fscanf(fp, "\n");
  }

  return 1;
}

/*
 * defDestroy
 * 
 */
classdef* defDestroy(classdef *def)
{
  int i;
  if(def){
    if(def->features_index)
    {
      for (i=0; i<def->nClasses; i++)
      {
        if(def->features_index[i]) free(def->features_index[i]);
      }
      free(def->features_index);
    }
    if(def->mean)
    {
      for (i=0; i<def->nClasses; i++)
      {
        if(def->mean[i]) free_matrice(&def->mean[i]);
      }
      free(def->mean);
    }
    if(def->cov_matrix)
    {
      for (i=0; i<def->nClasses; i++)
      {
        if(def->cov_matrix[i]) free_matrice(&def->cov_matrix[i]);
      }
      free(def->cov_matrix);
    }
    if(def->cfMean)
    {
      for (i=0; i<def->nClasses; i++)
      {
        if(def->cfMean[i]) free(def->cfMean[i]);
      }
      free(def->cfMean);
    }
    if(def->cfStdDev)
    {
      for (i=0; i<def->nClasses; i++)
      {
        if(def->cfStdDev[i]) free(def->cfStdDev[i]);
      }
      free(def->cfStdDev);
    }
    if(def->weight)
    {
      for (i=0; i<def->nClasses; i++)
      {
        if(def->weight[i]) free(def->weight[i]);
      }
      free(def->weight);
    }
    free(def);
  }
  return NULL;
}


/*
 * For the Maha strategy.
 * defInvertCov
 * It inverts the covariance matrix of every class (useful to
 * compute the Mahalanobis distance)
 * 
 */
int defInvertCov(classdef *def)
{
  matrice *cov=NULL;
  matrice *inv=NULL;
  int iClass; 

  // inverts covariance matrices
  for (iClass=1; iClass<=def->nClasses; iClass++)
  {
    cov = defGetCovariance(def, iClass);
    inv = Inverser_Matrice(cov);
    defSetCovariance(def, iClass, inv); // frees the other one itself
  }

  return 1;
}


/*
 * For the Maha strategy
 */
matrice *defGetCovariance(classdef *def, int iClass)
{
  if (iClass<=0 || iClass>def->nClasses)
  {
    return NULL;
  }
  iClass--;
  return def->cov_matrix[iClass];
}

/*
 * For the Maha strategy
 */
int defSetCovariance(classdef *def, int iClass, matrice *cov)
{
  if (iClass<=0 || iClass>def->nClasses)
    return 0;
  iClass--;
  free_matrice(&def->cov_matrix[iClass]);
  def->cov_matrix[iClass] = cov;

  return 1;
}

/*
 * For the Maha strategy.
 * Returns the number of interesting features for class 'iClass'.
 */
int defGetClassNfeatures(classdef *def, int iClass)
{
  if(!def) return 0;
  
  iClass--;
  return def->mean[iClass]->nbColonnes;
}


/*
 * For the maha strategy.
 * Given the array with the 'values' of all the features,
 * return a vector, 'point', with the values of the features interesting for class 'iClass'.
 */
int defGetClassPoint(classdef *def, int iClass, float *values, matrice *point)
{
  int i=0;
  int* index = NULL;
  index = def->features_index[iClass-1]; // shortcut to the Index of class 'iClass' in 'def'
  for(i=0; i<point->nbLignes; i++)
  {
    point->val[i] = (double)values[index[i]-1];
  }

  return 1;
}


/*
 *
 */
int defConfigFromFile(classdef *def, char *filename)
{
  FILE *fp;
  int ret=0;

  fp = fopen(filename, "rt");
  if (!fp)
  {
    fprintf (stderr, "Error: could not open configuration file.\n");
    return 0;
  }

  switch(def->strat)
  {
    case Maha:
      ret = defConfigMahaFromFile(def, fp);
      break;
    case FusionWA:
      ret = defConfigFusionWAFromFile(def, fp);
      break;
    case FusionDR:
      // No configuration needed for this strategy, yet.
      ret = 1;
      break;
    default:
      fprintf (stderr, "Error: Unknown strategy.\n");
      ret = 0;
  }
  fclose(fp);
  return ret;
}


/*
 *
 */
int defConfigMahaFromFile(classdef *def, FILE *fp)
{
  int iClass, ncFeatures;
  int i;

  // reads the features for each class to create the index
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    if(fscanf(fp, "class %*d num. Features: %d\n",&ncFeatures) != 1)
    {
      fprintf (stderr, "Error: problems reading number of features for class %d.\n",iClass+1);
      return 0;
    }
    if(ncFeatures > def->nFeatures || ncFeatures <= 0)
    {
      fprintf (stderr, "Error: Incorrect number of features for class %d.\n",iClass+1);
      return 0;
    }
    // Read the features index
    fscanf(fp, "class %*d features:\n");
    def->features_index[iClass] = (int*) calloc(ncFeatures,sizeof(int));
    def->mean[iClass]           = Matrice_Nulle(1,ncFeatures);
    def->cov_matrix[iClass]     = Matrice_Nulle(ncFeatures,ncFeatures);
    for (i=0; i<ncFeatures; i++)
    {
      if (fscanf(fp, "%d ", &(def->features_index[iClass][i])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d mean.\n",iClass+1);
        return 0;
      }
      if(  def->features_index[iClass][i] > def->nFeatures || def->features_index[iClass][i] <= 0)
      {
        fprintf (stderr, "Error: Incorrect feature number for class %d.\n",iClass+1);
        return 0;
      }
    }
    fscanf(fp, "\n");
  }
  
  return 1;
}


/*
 *
 */
int defConfigFusionWAFromFile(classdef *def, FILE *fp)
{
  int iClass, iFeat;

  // reads the weigths matrix
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      if (fscanf(fp, "%f ", &(def->weight[iClass][iFeat])) != 1)
      {
        fprintf (stderr, "Error: problems reading class %d, feature %d weigth.\n",iClass+1, iFeat+1);
        return 0;
      }
    }
    fscanf(fp, "\n");
  }
  
  return 1;
}


/*
 * defToTextFile
 * Outputs the results into a text file.
 * The format depends on the strategy.
 */
int defToTextFile(classdef *def, FILE *fp)
{
  // outputs the strategy, number of classes and of features
  fprintf(fp, "STRATEGY: %d\n", (int)def->strat);
  fprintf(fp, "CLASSES: %d\n", def->nClasses);
  fprintf(fp, "FEATURES: %d\n", def->nFeatures);
  switch(def->strat)
  {
    case Maha:
      defMahaToTextFile(def, fp);
      break;
    case FusionWA:
      defFusionWAToTextFile(def, fp);
      break;
    case FusionDR:
      defFusionDRToTextFile(def, fp);
      break;
    default:
      ;
  }

  return 1;
}

/*
 * defMahaToTextFile
 * Completes the classes definitions file according to the Maha strategy
 */
int defMahaToTextFile(classdef *def, FILE *fpOutput)
{
  int class, nClass;
  int ncFeatures, i, j;

  // outputs mean values for each class, conditioned to the discrete variables
  for (class=0; class<def->nClasses; class++)
  {
    nClass = class+1; // they refer to the same class, actually
    ncFeatures = defGetClassNfeatures(def,nClass);
    fprintf(fpOutput, "class %d num. Features: %d\n", nClass, ncFeatures);
    fprintf(fpOutput, "class %d features:\n", nClass);
    for (i=0; i<ncFeatures; i++)
      fprintf(fpOutput, "%d ", def->features_index[class][i]);

    fprintf(fpOutput, "\n");
    fprintf(fpOutput, "class %d means:\n", nClass);
    for (i=0; i<ncFeatures; i++)
      fprintf(fpOutput, "%lf ", def->mean[class]->val[i]);

    fprintf(fpOutput, "\n");
    fprintf(fpOutput, "class %d covariance matrix:\n", nClass);
    for (i=0; i<ncFeatures; i++)
    {
      for (j=0; j<ncFeatures; j++)
      {
        fprintf(fpOutput, "%lf ", def->cov_matrix[class]->val[j+i*ncFeatures]);
      }
      fprintf(fpOutput, "\n");
    }
  }

  return 1;
}


/*
 * defFusionDRToTextFile
 * Completes the classes definitions file according to the FusionDR strategy
 */
int defFusionDRToTextFile(classdef *def, FILE *fpOutput)
{
  int class, nClass, i;

  // outputs mean values for each class, conditioned to the discrete variables
  for (class=0; class<def->nClasses; class++)
  {
    nClass = class+1; // they refer to the same class, actually
    fprintf(fpOutput, "class %d:\n", nClass);
    for (i=0; i<def->nFeatures; i++)
      fprintf(fpOutput, "%lf ", def->cfMean[class][i]);

    fprintf(fpOutput, "\n");
    for (i=0; i<def->nFeatures; i++)
      fprintf(fpOutput, "%lf ", def->cfStdDev[class][i]);

    fprintf(fpOutput, "\n");
  }

  return 1;
}

/*
 * defFusionWAToTextFile
 * Completes the classes definitions file according to the FusionWA strategy
 * For the moment, the automatic part of this two strategies is the same.
 * In the resulting file from this method should be added the weights matrix.
 * Actually, the weigths should be computed automatically
 */
int defFusionWAToTextFile(classdef *def, FILE *fpOutput)
{
  int iClass, iFeat;
  
  if(!defFusionDRToTextFile(def, fpOutput)) return 0;
  fprintf(fpOutput, "\n");

  // outputs mean values for each class, conditioned to the discrete variables
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
      fprintf(fpOutput, "%lf ", def->weight[iClass][iFeat]);

    fprintf(fpOutput, "\n");
  }
  
  return 1;
}


/*
 * defComputeDefinitions
 *
 */
int defComputeDefinitions(classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y, int z)
{
  int ret=0;

  switch(def->strat)
  {
    case Maha:
      ret = defComputeMahaDefinitions(def, data, dataType, classif, x, y, z);
      break;
    case FusionWA:
      ret = defComputeFusionWADefinitions(def, data, dataType, classif, x, y);
      break;
    case FusionDR:
      ret = defComputeFusionDRDefinitions(def, data, dataType, classif, x, y);
      break;
    default:
      fprintf (stderr, "Error: Unknown strategy.\n");
      ret = 0;
  }

  return ret;
}


/*
 *
 * defComputeFusionDRDefinitions
 * Completes the classes definitions file according to the FusionDR strategy
 */
int defComputeFusionDRDefinitions(classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y)
{
  return defComputeFusionWADefinitions(def, data, dataType, classif, x, y);
}


/*
 * defComputeFusionWADefinitions
 * Completes the classes definitions file according to the FusionWA strategy
 * For the moment, the automatic part of this two strategies is the same.
 * In the resulting file from this method should be added the weights matrix.
 * Actually, the weigths should be computed automatically
 */
int defComputeFusionWADefinitions(classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y)
{
  int   size[def->nClasses];
  int   pos=0, iClass=0, iFeat=0;
  float fValue=0.0, delta=0.0;

  // Initialize arrays
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    size[iClass]=0;
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      def->cfMean[iClass][iFeat] = 0.0;
      def->cfStdDev[iClass][iFeat] = 0.0;
    }
  }
  
  // PROCESSES THE DATA
  //
  iClass = 0;
  for (pos=0; pos < x*y; pos++)
  {
    if (pos%x == 0)
      fprintf (stderr, "Computing mean values... %d%%\r", 100*pos/(x*y));

    // read class of the point
    iClass = (int) classif[pos];

    // discards non-classified points
    if (!iClass)
      continue;

    iClass--;
    // Recursive calculus of mean and stddev.
    // Algorithm III in http://en.wikipedia.org/wiki/Algorithms_for_calculating_variance.
    //  foreach x in data:
    //    n = n + 1
    //    delta = x - mean
    //    mean = mean + delta/n
    //    S = S + delta*(x - mean)      // This expression uses the new value of mean
    //  end for
    //  variance = S/(n - 1)
    size[iClass]++;
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      fValue = inrGetDataValue(data, dataType, pos + x*y*iFeat);
      delta = fValue - def->cfMean[iClass][iFeat];
      def->cfMean[iClass][iFeat] += delta/(float)size[iClass];
      def->cfStdDev[iClass][iFeat] += delta*(fValue-def->cfMean[iClass][iFeat]);
    }
  }

  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    for (iFeat=0; iFeat<def->nFeatures; iFeat++)
    {
      def->cfStdDev[iClass][iFeat] = sqrt(def->cfStdDev[iClass][iFeat]/(size[iClass] - 1.0));
    }
  }
  
  return 1;
}

int defComputeMahaDefinitions(classdef* def, unsigned char *data, InrType dataType, unsigned char *classif, int x, int y, int z)
{
  int iClass,ncFeatures,i,j;
  int *cSize, pos;
  float* values;
  matrice *point, *mean, *cov;
  
  
  // allocates memory for step 2 and 3
  cSize      = (int *) malloc (def->nClasses * sizeof(int));
  for (iClass=0; iClass<def->nClasses; iClass++)
    cSize[iClass] = 0;

  values = (float*) calloc(z,sizeof(double));

  // 2. computes means
  for (pos=0; pos<x*y; pos++)
  {
    if (pos%x == 0)
      fprintf (stderr, "Computing mean values... %d%%\r", 100*pos/(x*y));

    // reads classification of the point
    iClass = (int) classif[pos];

    // discards non-classified points
    if (!iClass)
      continue;

    ncFeatures = defGetClassNfeatures(def, iClass);

    // gets features values for the point in pos
    for(i=0; i<z; i++)
      values[i] = inrGetDataValue(data, dataType, pos + x*y*i);
    
    // Keeps the fetures interesting for this class
    point = Matrice_Nulle(ncFeatures, 1);
    if (!defGetClassPoint(def, iClass, values, point))
      continue;

//     defGetMean(def, iClass,  mean);
    for (i=0; i<ncFeatures; i++)
      def->mean[iClass-1]->val[i] += point->val[i];

    // increments class size
    cSize[iClass-1]++;
    free_matrice(&point);
  }

  // finalizes mean computation by dividing total sum by cSize
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    int n = cSize[iClass];
//     defGetMean(def, iClass+1, mean);
    mean = def->mean[iClass];
    if (n > 0)
      Multiplication_Scalaire_Matrice(1./n, mean);
  }

  // 3. computes covariance matrices for all classes
  for (pos=0; pos<x*y; pos++)
  {
    if (pos%x == 0)
      fprintf (stderr, "Computing covariance values... %d%%\r", 100*pos/(x*y));

    // reads classification of the point
    iClass = (int) classif[pos];

    // discards non-classified points
    if (!iClass)
      continue;

    ncFeatures = defGetClassNfeatures(def,iClass);
    // gets point values for the position pos
    for(i=0; i<z; i++)
      values[i] = inrGetDataValue(data, dataType, pos + x*y*i);
    
    point = Matrice_Nulle(ncFeatures, 1);
    if (!defGetClassPoint(def, iClass, values, point))
      continue;

    // subtracts the mean for the continuous values: (x-m)
    for (i=0; i<ncFeatures; i++)
      point->val[i] -= def->mean[iClass-1]->val[i];

    // updates covariance matrix: sigma += (x-m) * t(x-m)
    cov = defGetCovariance(def, iClass);
    for (i=0; i<ncFeatures; i++)
      for (j=0; j<ncFeatures; j++)
        cov->val[i+ ncFeatures*j ] += point->val[i]*point->val[j];
    
    free_matrice(&point);
  }
  // finalizes covariances computation by dividing total sum by cSize
  for (iClass=0; iClass<def->nClasses; iClass++)
  {
    int n = cSize[iClass];
    cov = defGetCovariance(def, iClass+1);
    if (n > 0)
      Multiplication_Scalaire_Matrice(1./n, cov);
  }

  return 1;

}
