#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inrimage/image.h>
#include "fuzzyclassdef.h"
#include "inr_io.h"
#include "file_tools.h"
#include "linearalgebra.h"


#define MIN(x,y) ((x<y)?x:y)
#define MAX(x,y) ((x>y)?x:y)

#define PI  3.14159265358979323846

#define FNAMESIZE         1024
#define TOLERANCE_DEFAULT 0.0



static char Ucmd[] = "-ifile inr_memberships -roi roi_mask -ofile inr_classification [-tol min_membership]\n";

static char Udetail[] =
"Assigns a pixel to the class with maximum membership\n\
Input and output files are in INRIMAGE format.\n\
\nInput:\n\
\t-ifile\tFile name of the input memberships file\n\
\t-roi\tFile name of a mask image that defines the region of interest\n\
\t\tMask should be zero for pixels to be masked out and non-zero\n\
\t\totherwise\n\
\nOutput:\n\
\t-ofile\tFile name of the output classificantion image\n\
\nOptional:\n\
\t-tol\tMinimum membership tolerance value for assigning a pixel to a class.\n\
\t\tIf, for one pixel , the class with maximum membership does\n\
\t\tnot exceed this value, the pixel is left unclassified (class 0)\n\
\t\tDefault value is 0 (all pixels within the ROI are labeled)\n";



int main (int argc, char **argv)
{
  char fnameInput[FNAMESIZE];
  char fnameROIMask[FNAMESIZE];
  char fnameOutput[FNAMESIZE];
  InrType typeInput;
  InrType type;
  int     typesizeInput;
  InrGeorefInfo georef;
  
  int    okTolerance;
  double tolerance=0.0;
  
  unsigned char *dataInput;
  unsigned char *ROIMask;
  unsigned char *dataClassif;
  
  
  int newClass, iClass;
  float max_membership=0.0, cMembership=0.0, sum_membership;
  
  int size_x, size_y, size_z;
  int i, j, k;
  int x, y;
  
  /* lecture options */
  inr_init (argc, argv, "1.0", Ucmd, Udetail);
  if (!igetopt1 ("-ifile", "%s", fnameInput))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-roi"  , "%s", fnameROIMask))
    iusage_ (Ucmd, Udetail);
  if (!igetopt1 ("-ofile", "%s", fnameOutput))
    iusage_ (Ucmd, Udetail);
  okTolerance  = igetopt1 ("-tol", "%lf", &tolerance);

  if (tstopts ())                /* s'il reste des options : exit avec message */
    iusage_ (Ucmd, Udetail);

  if (!okTolerance)
    tolerance = TOLERANCE_DEFAULT;

  // reads header information for all input data and checks if they are appropriate
  if (!inrReadHeader(fnameInput, &size_x, &size_y, &size_z, &typeInput))
  {
    fprintf(stderr, "Error: could not read header information for input file '%s'.\n", fnameInput);
    return 0;
  }
  if (!inrReadHeader(fnameROIMask, &i, &j, &k, &type))
  {
    fprintf(stderr, "Error: could not read header information for ROI mask '%s'.\n", fnameROIMask);
    return 0;
  }
  if (type != InrByte || k != 1)
  {
    fprintf(stderr, "Error: ROI mask '%s' has to be a single image of type 'Byte'.\n", fnameROIMask);
    return 0;
  }
  if (i != size_x || j != size_y)
  {
    fprintf (stderr, "Error: dimensions of input data and ROI mask do not match.\n");
    return 0;
  }

  // reads the ROI mask image
  ROIMask = inrReadByteImage (fnameROIMask, &i, &j);
  if (!ROIMask)
  {
    fprintf (stderr, "Error: could not read ROI mask data.\n");
    return 0;
  }
  
  // allocates memory for the input and output data
  typesizeInput    = InrTypeSizes[typeInput]/8;
  dataInput        = (unsigned char *) malloc (size_x*size_z * typesizeInput * sizeof(unsigned char));
  dataClassif      = (unsigned char *) malloc (size_x*size_y * sizeof(unsigned char));
  
  // PROCESSES THE DATA
  for (y=0; y<size_y; y++)
  {
    fprintf (stderr, "Processing... %d%%\r", 100*y/size_y);

    // checks if line y is within region of interest
    for (x=0; x<size_x; x++)
      if (ROIMask[x + y*size_x])
        break;
    // if no unmasked point was detected, line is not within ROI and is skipped 
    if (x == size_x)
    {
      // all points in the line are rated unclassified
      for (x=0; x<size_x; x++)
        dataClassif[x + y*size_x] = 0;

      continue;
    }

    // reads input data line y
    if (!inrReadLine(fnameInput, y, dataInput))
    {
      fprintf(stderr, "Error: could not read line %d from input file '%s'.\n", y, fnameInput);
      return 0;
    }
    
    // processes each point at a time
    for (x=0; x<size_x; x++)
    {
      // pixels outside of region of interest are rated unclassified
      if (!ROIMask[x + y*size_x])
      {
        dataClassif[x + y*size_x] = 0;
        continue;
      }

      // compute sum of all membership
      sum_membership=0.0;
      for(iClass=0; iClass<size_z; iClass++)
      {
	sum_membership += inrGetDataValue(dataInput, typeInput, x + size_x*iClass);
      }
      // gets class memberships for point x of the input line
      newClass = 0;
      max_membership = 0.0;
      for(iClass=0; iClass<size_z; iClass++)
      {
        cMembership = inrGetDataValue(dataInput, typeInput, x + size_x*iClass)/sum_membership;
        if(cMembership > max_membership)
        {
          max_membership = cMembership;
          if(max_membership >= tolerance)
            newClass = iClass+1;
        }
      }
      // Assign to pixel x label of class 
      // with maximum membership (if higher than tolerance).
      dataClassif[x + y*size_x] = newClass;
    }
  }

  // outputs results
  inrWriteByteImage(fnameOutput, size_x, size_y, dataClassif);

  // transfers georeference info if available
  if (inrReadGeoref(fnameInput, &georef)) inrWriteGeoref(fnameOutput, georef);

  fprintf(stderr, "Image classification successfully completed.\n");

  free(dataInput);
  free(dataClassif);
  
  return 0;
}
